

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORELINEDISTFILTER_H
#define ADORE_CORELINEDISTFILTER_H

#include <limits>

namespace adore
{

	namespace core
	{

		/**
		 * A class which contains methods computing the distance of points to a line. The line is given as a 3-vector.
		 * @tparam Real the real type, float or double
		 *  
		*/
		template <typename Real>
		class LineDistanceT
		{
			public:
				/**
				 * Constructor, takes a line of the form a*x + b*y + c = 0
				 * @param a the x coefficient
				 * @param b the y coefficient
				 * @param c the constant
				 */
				LineDistanceT( Real a = 0, Real b = 0, Real c = 0 ):okay( true )
				{
					setCoeffs( a, b, c );
				}

				/**
				 * take the coefficients of an implicit line (a*x + b*y + c = 0) and saves them
				 * after normalizing the normal (_a, _b)^T to 1.
				 * @param _a the x coefficient
				 * @param _b the y coefficient
				 * @param _c the constant
				 */
				bool setCoeffs( Real _a, Real _b, Real _c )
				{
					using std::numeric_limits;
					Real a2 = _a * _a;
					Real b2 = _b * _b;
					if ( a2 <= numeric_limits<Real>::epsilon() && b2 <= numeric_limits<Real>::epsilon() )
					{
						okay = false;
						return false;
					}
					Real div = sqrt( a2 + b2 );
					a = _a / div;
					b = _b / div;
					c = _c / div;
					return true;
				}

				/**
				 * returns the distance of the point to the line
				 * @param x the x co-ordinate of the point
				 * @param y the y co-ordinate of the point
				 * @return the distance of the point to the line
				 */
				Real getDistance( Real x, Real y ) const
				{
					return a * x + b * y + c;
				}

				/**
				 * @return false, if the coefficient were not sane
				 */
				bool isOkay()
				{
					return okay;
				}

			private:
				Real a, b, c;	///the line
				bool okay;	///sane coeffs?
		};

		/**
		 * line distances computed as floats
		 */
		typedef LineDistanceT<float> LineDistanceF;

		/**
		 * line distances computed as doubles
		 */
		typedef LineDistanceT<double> LineDistanceD;
		
		/**
		 * line distances computed as floats are the default
		 */
		typedef LineDistanceF LineDistance;
	}
}

#endif
