

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGCANNY_H
#define ADORE_IMGPROCESSINGCANNY_H

#include <stack>
#include <adrcore/util/stlutil.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/hysteresisthresholder.h>
#include <adrcore/imgprocessing/borderfiller.h>

namespace adore
{

	namespace core
	{
		/**
		 * Canny edge detector
		 * @tparam img_in_typeT type of magnitude image
		 * @tparam img_out_typeT type of image to hold the edges
		 *  
		 */
		template<class img_in_typeT, class img_out_typeT>
		class Canny
		{
			public:
				typedef img_in_typeT img_in_type;
				typedef img_out_typeT img_out_type;
				typedef uint8_t secT;///the image containing the section info needs to be uint8_t
			private:
				stack<int> maxima;	///holds the maxima
				HysteresisThresholder<img_in_type, img_out_type> ht; ///a hysteresis thresholder
				BorderFiller<img_in_type> bf; ///to zero the border to avoid border-maxima
				img_in_type min, max;	///hysteresis parameter

			public:

				/**
				 * Constructor
				 */
				Canny() :maxima(),ht(),bf ( 1 ), min ( img_in_type ( 0.2 ) ), max ( img_in_type ( 0.6 ) )
				{
					;
				}

				/**
				 * Destructor
				 */
				~Canny()
				{
					;
				}

				/**
				 * computes canny edges from two images, one containing gradient magnitudes, the other the gradient directions.
				 * @param mag an image containing gradient magnitudes
				 * @param sec the gradient direction from the interval [0,7]
				 * @param dst the image which shall contain the result
				 */
				void process ( Channel<img_in_type> *mag, Channel<secT> *sec, Channel<img_out_type> *dst );

				/**
				 * set the value where to stop the hysteresis thresholding
				 * in Matlab: low = 0.5 * mean(norm (<-not squared norm!))
				 * @param theValue hysteresis stop threshold
				 */
				void setMin ( const img_in_type& theValue )
				{
					min = theValue;
				}

				/**
				 * @return the hysteresis stop threshold
				 */
				img_in_type getMin() const
				{
					return min;
				}

				/**
				 * set the value where to start the hysteresis thresholding
				 * in Matlab: high = 2.5 * low
				 * @param theValue hysteresis start threshold
				 */
				void setMax ( const img_in_type& theValue )
				{
					max = theValue;
				}

				/**
				 * @return the hysteresis start threshold
				 */
				img_in_type getMax() const
				{
					return max;
				}

			protected:

				/**
				 * suppresses all pixels that are not maxima in gradient direction or below a threshold
				 * modifies the mag image by negating the values to be suppressed
				 * @param mag an image of gradient magnitudes
				 * @param sec an image of gradient directions
				 * @param maxima the found maxima as a stack of offsets
				 * @param min the gradient strength threshold
				 */
				void nonMaximaSuppression ( Channel<img_in_type> *mag,
				                            Channel<secT> *sec,
				                            std::stack<int> &maxima, const double &min ) const;

				/**
				 * undo of the nonMaximaSuppression, done by negating all negative values
				 * @param mag the magnitude image, will be modified (i.e. restored)
				 */
				void undoNonMaximaSuppression ( Channel<img_in_type> *mag ) const;

		};

		template<class img_in_type, class img_out_type>
		void Canny<img_in_type, img_out_type>::process ( Channel<img_in_type> *mag, Channel<secT> *sec, Channel<img_out_type> *dst )
		{
			maxima.empty();

			bf.process ( mag, mag );

			UsingROI ur1 ( mag );
			UsingROI ur2 ( sec );

			mag->minROIDistance ( 1 );
			sec->minROIDistance ( 1 );

			nonMaximaSuppression ( mag, sec, maxima, max );

			ht.setMaxima ( &maxima );
			ht.setThreshold ( min );
			ht.process ( mag, dst );

			undoNonMaximaSuppression ( mag );
		}

		template<class img_in_type, class img_out_type>
		void Canny<img_in_type, img_out_type>::nonMaximaSuppression ( Channel<img_in_type> *mag,
		                Channel<secT> *sec,
		                stack<int> &maxima, const double &min ) const
		{
			//alle maxima auf nen Stack pushen
			//maximum wird in sector-Richtung geprft

			typename Channel<img_in_type>::Iterator2D magIt = mag->getIteratorBegin();
			typename Channel<secT>::Iterator2D secIt = sec->getIteratorBegin();

			for ( ; magIt != mag->getIteratorEnd(); ++magIt, ++secIt )
			{
				img_in_type absMag = tabs<img_in_type> ( *magIt );
				if ( absMag > min )
				{
//					int mx, my, sx, sy;
//					mag->getCoors( magIt.getPointer(), mx, my );
//					sec->getCoors( secIt.getPointer(), sx, sy );
//					printf( "(%d, %d), (%d, %d)\n", mx, my, sx, sy );

					int beforeIdx = SECTOR_CORONA[*secIt][0];
					int afterIdx = SECTOR_CORONA[*secIt][1];

//					printf( "sec:%d, %d->%d\n", *secIt, beforeIdx, afterIdx );

					img_in_type *before = magIt.getPointer ( beforeIdx );
					img_in_type *after = magIt.getPointer ( afterIdx );

//					mag->getCoors( before, mx, my );
//					mag->getCoors( after, sx, sy );
//					printf( "(%d, %d), (%d, %d)\n", mx, my, sx, sy );

					img_in_type absBefore = tabs<img_in_type> ( *before );
					img_in_type absAfter = tabs<img_in_type> ( *after );
//					printf( "%f -> %f -> %f\n", absBefore, *magIt, absAfter );
					if ( absMag >= absBefore && absMag >= absAfter )
					{
						maxima.push ( mag->getOffset ( magIt.getPointer() ) );
					}
					else
					{
						*magIt = -absMag;
					}

//					pressKey();
				}
				else
				{
					*magIt = -absMag;
				}
			}
//			printf( "found %d maxima\n", maxima.size() );
		}

		template<class img_in_type, class img_out_type>
		void Canny<img_in_type, img_out_type>::undoNonMaximaSuppression ( Channel<img_in_type> *mag ) const
		{
			typename Channel<img_in_type>::Iterator2D magIt = mag->getIteratorBegin();
			for ( ; magIt != mag->getIteratorEnd(); ++magIt )
			{
				*magIt = tabs<img_in_type> ( *magIt );
			}
		}


	}

}

#endif
