

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGCHANNELFUNCTIONS_H
#define ADORE_IMGPROCESSINGCHANNELFUNCTIONS_H

#include <adrcore/types/aarray.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/calctype.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/bilinearinterpolator.h>

namespace adore
{

	namespace core
	{
		/**
		 * fill the target buffer that is addressed by IteratorT it through affinely
		 * warping the coordinates (-xHood:xHood)(-yHood:yHood) and taking the pixelvalue
		 * from the source image at position (srcX, srcY)+(warpedX, warpedY).
		 * The buffer behind IteratorT it has to be big enough to hold
		 * (2 * xHood + 1) * (2 * yHood + 1) values.
		 * @tparam typeT the image type
		 * @tparam T_numType the matrix type
		 * @tparam IteratorT the iterator type
		 * @param src the source image to warp
		 * @param srcX x-coordinate of the center of the region to be warped
		 * @param srcY y-coordinate of the center of the region to be warped
		 * @param A 2x2 matrix that describes the warping
		 * @param it iterator over the collection which shall hold the result
		 * @param width width of the region to be warped
		 * @param height height of the region to be warped
		 */
		template<class typeT, class T_numType, class IteratorT>
		void warpAffineRegion( const Channel<typeT> &src, float srcX, float srcY,
		                       const AArray<T_numType, 2> &A,
		                       IteratorT it, int width, int height )
		{
			int addX = 0, addY = 0;
			if ( isOdd( width ) ) addX = 1;
			if ( isOdd( height ) ) addY = 1;

			int xHood = width >> 1;
			int yHood = height >> 1;

			BilinearInterpolator<typeT> sampler;
			for ( int y = -yHood; y < yHood + addY; ++y )
			{
				for ( int x = -xHood; x < xHood + addX; ++x, ++it )
				{
					float newX = srcX + A( 0, 0 ) * x + A( 0, 1 ) * y;
					float newY = srcY + A( 1, 0 ) * x + A( 1, 1 ) * y;


					if ( newX >= 0 && newX <= src.getWidth() - 1 && newY >= 0 && newY <= src.getHeight() - 1 )
					{
						sampler.init( newX, newY );
						*it = sampler.valueOf( src );
					}
					else
					{
						*it = 0;
					}
				}
			}
		}

		/**
		 * fill the target buffer that is addressed by IteratorT it through affinely
		 * warping the coordinates (-width/2:width/2)..(-height/2:height/2) and taking the pixelvalue
		 * from the source image at position (centerX, centerY)+(warpedX, warpedY).
		 * The buffer behind IteratorT it has to be big enough to hold width * height values.
		 * @tparam typeT the image type
		 * @tparam realT the matrix type
		 * @tparam IteratorT the iterator type
		 * @param src the source image to warp
		 * @param centerX x-coordinate of the center of the region to be warped
		 * @param centerY y-coordinate of the center of the region to be warped
		 * @param A 2x2 matrix that describes the warping
		 * @param it iterator over the collection which shall hold the result
		 * @param width width of the region to be warped
		 * @param height height of the region to be warped
		 */
		template<class imgT, class realT, class IteratorT>
		void warpAffine2( const Channel<imgT> & src, realT centerX, realT centerY,
		                  const AArray<realT, 2> & A, IteratorT it, int width, int height )
		{
			realT xHood = realT( width ) / 2; if(width % 2 == 0) xHood -= 0.5;
			realT yHood = realT( height ) / 2; if(height % 2 == 0) yHood -= 0.5;
			realT srcX = centerX - A( 0, 0 ) * xHood - A( 0, 1 ) * yHood;
			realT srcY = centerY - A( 1, 0 ) * xHood - A( 1, 1 ) * yHood;
			BilinearInterpolator<imgT> sampler;
			for ( int i = 0; i < height; ++i, srcX += A( 0, 1 ), srcY += A( 1, 1 ) )
			{
				realT x = srcX;
				realT y = srcY;
				for ( int j = 0; j < width; ++j, ++it, x += A( 0, 0 ), y += A( 1, 0 ) )
				{
					if ( x >= 0 && x <= src.getWidth() - 1 && y >= 0 && y <= src.getHeight() - 1 )
					{
						sampler.init( x, y );
						*it = sampler.valueOf( src );
					}
					else
						*it = 0;
				}
			}
		}


		/**
		 * fill the target buffer that is addressed by IteratorT it through affinely
		 * warping the coordinates (-xHood:xHood)(-yHood:yHood) and taking the pixelvalue
		 * from the source image at position (srcX, srcY)+(warpedX, warpedY).
		 * The buffer behind IteratorT it has to be big enough to hold
		 * (2 * xHood + 1) * (2 * yHood + 1) values.
		 * @tparam typeT the image type
		 * @tparam T_numType the matrix type
		 * @tparam IteratorT the iterator type
		 * @param src the source image to warp
		 * @param srcX x-coordinate of the center of the region to be warped
		 * @param srcY y-coordinate of the center of the region to be warped
		 * @param A 2x2 matrix that describes the warping
		 * @param it iterator over the collection which shall hold the result
		 * @param width width of the region to be warped
		 * @param height height of the region to be warped
		*/
		template<class typeT, class T_numType, class IteratorT>
		void warpAffine( const Channel<typeT> &src, float srcX, float srcY,
		                 const AArray<T_numType, 2> &A,
		                 IteratorT it, int xHood, int yHood )
		{
//			printf( "warp\n" );
			BilinearInterpolator<typeT> sampler;
			for ( int y = -yHood; y <= yHood; ++y )
			{
				for ( int x = -xHood; x <= xHood; ++x, ++it )
				{
					float newX = srcX + A( 0, 0 ) * x + A( 0, 1 ) * y;
					float newY = srcY + A( 1, 0 ) * x + A( 1, 1 ) * y;

//					printf( "%f, %f", newX, newY );

					if ( newX >= 0 && newX <= src.getWidth() - 1 && newY >= 0 && newY <= src.getHeight() - 1 )
					{
						sampler.init( newX, newY );
						*it = sampler.valueOf( src );
//						printf( "_%f", *it );
					}
					else
					{
						*it = 0;
//						printf( "#%f", *it );
					}
				}
//				printf( "\n" );
			}
		}

		/**
		 * copies a part of an image into a buffer, bilinearly interpolated.
		 * @tparam typeT the image type
		 * @tparam IteratorT the iterator type
		 * @param chan the source image to copy from
		 * @param srcX x-coordinate of the center of the region to copy
		 * @param srcY y-coordinate of the center of the region to copy
		 * @param it iterator over the collection which shall hold the result
		 * @param xHood 2*xHood + 1 = width of the region to copy
		 * @param yHood 2*yHood + 1 = height of the region to copy
		 */
		template<class typeT, class IteratorT>
		void getPatch( const Channel<typeT> &chan, float srcX, float srcY, IteratorT it, int xHood, int yHood )
		{
			BilinearInterpolator<typeT> sampler;
			for ( int y = -yHood; y <= yHood; ++y )
			{
				for ( int x = -xHood; x <= xHood; ++x, ++it )
				{
					float newX = srcX + x;
					float newY = srcY + y;
					if ( newX >= 0 && newX <= chan.getWidth() - 1 && newY >= 0 && newY <= chan.getHeight() - 1 )
					{
						sampler.init( newX, newY );
						*it = sampler.valueOf( chan );
					}
				}
			}
		}

		/**
		 * copies a part of an image into a buffer, bilinearly interpolated.
		 * unlike getPatch, this function assumes a buffer of even width and height
		 * (srcX, srcY) lies on the common corner of the four center pixels
		 * @tparam typeT the image type
		 * @tparam IteratorT the iterator type
		 * @param chan the source image to copy from
		 * @param srcX x-coordinate of the center of the region to copy
		 * @param srcY y-coordinate of the center of the region to copy
		 * @param it iterator over the collection which shall hold the result
		 * @param xHood 2*xHood + 1 = width of the region to copy
		 * @param yHood 2*yHood + 1 = height of the region to copy
		 */
		template<class typeT, class IteratorT>
		void getPatch2( const Channel<typeT> &chan, float srcX, float srcY, IteratorT it, int xHood, int yHood )
		{
			BilinearInterpolator<typeT> sampler;
			for ( int y = -yHood; y < yHood; ++y )
			{
				for ( int x = -xHood; x < xHood; ++x, ++it )
				{
					float newX = srcX + x + 0.5F;
					float newY = srcY + y + 0.5F;
					if ( newX >= 0 && newX <= chan.getWidth() - 1 && newY >= 0 && newY <= chan.getHeight() - 1 )
					{
						sampler.init( newX, newY );
						*it = sampler.valueOf( chan );
					}
				}
			}
		}

		/**
		 * Finds the pixels with the minimum and maximum values.
		 * @tparam typeT the image type
		 * @param chan the image in which the extrema shall be found.
		 * @param min will hold the minimum value
		 * @param max will hold the maximum value
		 * @param minL will point to the minimum value
		 * @param maxL will point to the maximum value
		 */
		template<class typeT>
		void minMaxLoc( const Channel<typeT> &chan, typeT &min, typeT &max, typeT* &minL, typeT* &maxL )
		{
			typename Channel<typeT>::Iterator2D begin = chan.getIteratorBegin();
			typename Channel<typeT>::Iterator2D end = chan.getIteratorEnd();

			min = *begin;
			max = *begin;
			++begin;

			for ( ; begin != end; ++begin )
			{
				typeT val = *begin;

				if ( val < min )
				{
					min = val;
					minL = begin.getPointer();
				}

				if ( val > max )
				{
					max = val;
					maxL = begin.getPointer();
				}
			}
		}

		/**
		 * calculates the center of gravity for a number of pixel coordinates given by offsets or pointers
		 * @tparam typeT the image type
		 * @tparam IteratorT the iterator type. Will be dereferenced to offsets or pointers
		 * @param chan the channel to which the offsets or pointers belong, chan.getCoors is used to calculate the (x,y)-coordinates
		 * @param start the begin of the collection containing the offsets or pointers
		 * @param stop the end of the collection containing the offsets or pointers
		 * @param outX will contain the x-coordinate of the center of gravity
		 * @param outY will contain the y-coordinate of the center of gravity
		 */
		template<class typeT, class IteratorT>
		void getCenterOfGravity( const Channel<typeT> &chan,
		                         IteratorT start,
		                         IteratorT stop,
		                         float &outX, float &outY )
		{
			outX = outY = 0;
			if ( start == stop ) return;

			int x = 0, y = 0;

			int count = 0;
			for ( ; start != stop; ++start, ++count )
			{
				chan.getCoors( *start, x, y );
				outX += x;
				outY += y;
			}
			outX /= count;
			outY /= count;
		}

		/**
		 * sets the ROI of a channel while respecting certain distances to the image borders. The ROI is shrinked to fit.
		 * @tparam typeT the image type
		 * @param _x the x-coordinate of the top left ROI point
		 * @param _y the y-coordinate of the top left ROI point
		 * @param width the width of the ROI
		 * @param height the height of the ROI
		 * @param topDist required distance to the top of the image
		 * @param bottomDist required distance to the bottom of the image
		 * @param leftDist required distance to the left of the image
		 * @param rightDist required distance to the bottom of the image
		 */
		template<class typeT>
		void setROISavely( Channel<typeT> &chan, int _x, int _y, uint width, uint height,
		                   uint topDist, uint bottomDist, uint leftDist, uint rightDist )
		{
			assert( width > 0 );
			assert( height > 0 );

			uint x = _x;
			uint y = _y;
			if ( _x < int( leftDist ) )
			{
				x = leftDist;
				width -= ( x - _x );
			}
			else if ( x >=  chan.getWidth() - rightDist )
			{
				x =  chan.getWidth() - rightDist - 1;
				width -= ( _x - x );
			}

			if ( _y < int( topDist ) )
			{
				y = topDist;
				height -= ( y - _y );
			}
			else if ( y >= chan.getHeight() - bottomDist )
			{
				y = chan.getHeight() - bottomDist;
				height -= ( _y - y );
			}

			chan.setROI( x, y, width, height );
		}

		/**
		 * Sets the ROI of a channel while respecting certain distances to the image borders. The ROI is shrinked to fit.
		 * This is a convenience function which allows only one paramter to set the distance.
		 * @tparam typeT the image type
		 * @param _x the x-coordinate of the top left ROI point
		 * @param _y the y-coordinate of the top left ROI point
		 * @param width the width of the ROI
		 * @param height the height of the ROI
		 * @param dist required distance to each image border
		 */
		template<class typeT>
		void setROISavely( Channel<typeT> &chan, int x, int y, uint width, uint height, uint dist = 0 )
		{
			setROISavely( chan, x, y, width, height, dist, dist, dist, dist );
		}

		/**
		 * Point-reflects an image at its center.
		 * Takes the ROI into account. Works in-place.
		 * @tparam imgT the image type
		 * @param chan the image to point reflect.
		 */
		template <class imgT>
		void centerReflection( Channel<imgT> *chan )
		{
			assert( chan );
			Channel<imgT> &ip = *chan;
			int roiX = chan->getROI().getX();
			int roiY = chan->getROI().getY();
			int roiW = chan->getROI().getWidth();
			int roiH = chan->getROI().getHeight();

			int halfH = roiH >> 1;
			for ( int y = 0; y < halfH; ++y )
			{
//				printf( "row: %d\n", y );
				int s1y = roiY + y;
				int s2y = roiY + roiH - y - 1;
				for ( int x = 0; x < roiW; ++x )
				{
					int s1x = roiX + x;
					int s2x = roiX + roiW - x - 1;
					//swap:
					float temp = ip( s1x, s1y );
					ip( s1x, s1y ) = ip( s2x, s2y );
					ip( s2x, s2y ) = temp;
				}
			}
			if ( isOdd( roiH ) )
			{
				int y = roiY + ( roiH >> 1 );
//				printf( "centerReflection: oddrow is %d\n", y );
				for ( int x = 0; x < roiW >> 1; ++x )
				{
					int s1x = roiX + x;
					int s2x = roiX + roiW - x - 1;
//					printf( "swapping %d, %d\n", s1x, s2x );
					float temp = ip( s1x, y );
					ip( s1x, y ) = ip( s2x, y );
					ip( s2x, y ) = temp;
				}
			}
		}

	}
}

#endif
