

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGCHANNELPROCESSOR_H
#define ADORE_IMGPROCESSINGCHANNELPROCESSOR_H

#include <adrcore/imgprocessing/imageplane.h>

namespace adore
{

	namespace core
	{


		/**
		 * A template-free base class of ChannelProcessor.
		 *  
		 */
		class ChannelProcessorBase
		{
			public:
				/**
				 * Virtual Destructor.
				 */
				virtual ~ChannelProcessorBase()
				{
					;
				}

				/**
				 * In case a ChannelProcessor does not process the whole image, it may overload this function to
				 * annouce this fact. ConvolverSeparable2D uses this function to determin the size of temporary space.
				 * @param width the width of the whole ROI.
				 * @param height the height of the whole ROI.
				 * @param resultWidth the width of the result.
				 */
				virtual void getProcessedResolution( int width, int height, int &resultWidth, int &resultHeight )
				{
					resultWidth = width;
					resultHeight = height;
				}
		};

		/**
		 * A ChannelProcessor takes a number of input channels and produces a number of output channels.
		 * @tparam numIn the number of input channels.
		 * @tparam numOut the number of output channels.
		 * @tparam typeIn the numerical type of the input channels.
		 * @tparam typeOut the numerical type of the output channels.
		 */
		template<int numIn, int numOut, class typeIn, class typeOut>
		class ChannelProcessor : public ChannelProcessorBase
		{
			public:
				/**
				 * virtual Destructor
				 */
				virtual ~ChannelProcessor()
				{
					;
				}

				/**
				 * the main interface of a Channel processor
				 * @param in the input; a set of channels
				 * @param out the output; a set of channels
				 */
				virtual void process( typename Channel<typeIn>::Set &in, typename Channel<typeOut>::Set &out ) = 0;
		};

		/**
		 * A specialization of ChannelProcessor. The number of input and output channels is fixed to 1.
		 * This does not mean, that only single channels can be processed, but that one channel is processed at a time, resulting in exactly one channel
		 * Hence, if sets of channels are passed to this ChannelProcessor, these channels are batch-processed.
		 * @tparam typeIn the numerical type of the input channels.
		 * @tparam typeOut the numerical type of the output channels.
		 */
		template<class typeIn, class typeOut>
		class ChannelProcessor<1, 1, typeIn, typeOut> : public ChannelProcessorBase
		{
			public:
				/**
				 * virtual Destructor.
				 */
				virtual ~ChannelProcessor()
				{
					;
				}

				/**
				 * Processes many channels and genereates an equal number of output channels.
				 * @param in a set of input channels.
				 * @param out a set of output channels.
				 */
				void process( typename Channel<typeIn>::Set &in, typename Channel<typeOut>::Set &out )
				{
					assert( in.getNumChans() == out.getNumChans() );
					for ( int c = in.getNumChans(); c >= 0; --c ) process( in[c], out[c] );
				}

				/**
				 * Be careful since imageplane is derived from channel.
				 * Therefore the different name of this function to make
				 * is less easily confused with the one below.
				 * @param in an image plane
				 * @param out an image plane
				 */
				void processImage( ImagePlane<typeIn> &in, ImagePlane<typeOut> &out )
				{
					assert( in.getNumChans() == out.getNumChans() );
					int cmI = in.getChannelMode();
					int cmO = out.getChannelMode();
					for ( int c = in.getNumChans() - 1; c >= 0; --c )
					{
						in.setChannelMode( c );
						out.setChannelMode( c );
						process( &in, &out );
					}
					in.setChannelMode( cmI );
					out.setChannelMode( cmO );
				}

				/**
				 * Processes exactly one channel.
				 * @param in a channel.
				 * @param out a channel.
				 */
				virtual void process( Channel<typeIn> *in, Channel<typeOut> *out ) = 0;

				/**
				 * In-place processing.
				 * @param in the channel to process.
				 */
				inline void process( Channel<typeIn> *in )
				{
					process( in, in );
				}
		};

		/**
		 * A ChannelProcessor which takes a number of input channels and creates exactly one output channel
		 * @tparam numIn the number of input channels.
		 * @tparam typeIn the numerical type of the input channels.
		 * @tparam typeOut the numerical type of the output channels.
		 */
		template<int numIn, class typeIn, class typeOut>
		class ChannelProcessor<numIn, 1, typeIn, typeOut> : public ChannelProcessorBase
		{
			public:
				/**
				 * virtual Destructor
				 */
				virtual ~ChannelProcessor()
				{
					;
				}

				/**
				 * Processes a set of channels and puts the result into the first channel of another set of channels
				 * @param in a set of input channels.
				 * @param out a set of output channels.
				 */
				void process( typename Channel<typeIn>::Set &in, typename Channel<typeOut>::Set &out )
				{
					process( in, out[0] );
				}

				/**
				 * Processes a set of channels and puts the result into another channel
				 * @param in a set of input channels.
				 * @param out a channels.
				 */
				virtual void process( typename Channel<typeIn>::Set &in, Channel<typeOut> *out ) = 0;

				/**
				 * Processes an image with potentially many channels and puts the result into another channel
				 * @param img an image.
				 * @param out a channels.
				 */
				void process( ImagePlane<typeIn> *img, Channel<typeOut> *out )
				{
					ChannelSet<typeIn> in ( true );
					img->createChannels ( in );
					this->process( in, out );
				}
		};

		/**
		 * A ChannelProcessor which takes exactly one input channel and creates a number of output channels
		 * @tparam numOut the number of output channels.
		 * @tparam typeIn the numerical type of the input channels.
		 * @tparam typeOut the numerical type of the output channels.
		 */
		template<int numOut, class typeIn, class typeOut>
		class ChannelProcessor<1, numOut, typeIn, typeOut> : public ChannelProcessorBase
		{
			public:
				/**
				 * virtual Destructor
				 */
				virtual ~ChannelProcessor()
				{
					;
				}
				
				/**
				 * Processes the first channel of a set of channels and puts the resulting channels into another set of channels
				 * @param in a set of input channels.
				 * @param out a set of output channels.
				 */
				void process( typename Channel<typeIn>::Set &in, typename Channel<typeOut>::Set &out )
				{
					process( in[0], out );
				}

				/**
				 * Processes a channel and puts the resulting channels into another set of channels
				 * @param in a channel.
				 * @param out a set of output channels.
				 */
				virtual void process( Channel<typeIn> &in, typename Channel<typeOut>::Set &out ) = 0;
		};


	}

}

#endif
