

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORECONVOLVER2D_H
#define ADORE_CORECONVOLVER2D_H

#include <adrcore/imgprocessing/channelprocessor.h>
#include <adrcore/imgprocessing/iteratorgetter.h>
#include <adrcore/imgprocessing/ringbuffer2d.h>
#include <adrcore/imgprocessing/imgconstants.h>
#include <adrcore/types/ringbuffer.h>

#define DBGCONVOLVE DBGALL

namespace adore
{

	namespace core
	{

		/**
		 * Applys a kernel to an image
			 
		*/
		template<class ConvKernT>
		class Convolver2D : public ChannelProcessor<1, 1, typename ConvKernT::img_in_type, typename ConvKernT::img_out_type>
		{
			private:
				enum
				{
					DIRECT = 0,
					BUFFER_IN = 1,
					BUFFER_OUT = 2,
					BUFFER_IN_OUT = 3
				};

			public:

				/*
				    Define in and out types
				*/
				typedef typename ConvKernT::img_in_type img_in_type;
				typedef typename ConvKernT::img_out_type img_out_type;

			public:
				Convolver2D( BORDER_MODE mode = BORDER_COPY ):ChannelProcessor<1, 1, img_in_type, img_out_type>()
				{
					rbIn.setMode( mode );
				}

				void process( Channel<img_in_type> *in, Channel<img_out_type> *out )
				{
//					printf( "Convolver2D-mode is %s\n", rb.getMode() == BORDER_COPY ? "copy" : "zero" );
					uint kthood = kernel.getTopHood();
					uint kbhood = kernel.getBottomHood();
					uint klhood = kernel.getLeftHood();
					uint krhood = kernel.getRightHood();

					//determin the mode of operation
					int mode = DIRECT;
					//in place or not
					if ( in == out ) mode |= BUFFER_OUT;

					//border needs special treatment or not
					ROI roi = in->getROI();
					if ( roi.getX() >= klhood &&
					                roi.getY() >= kthood &&
					                in->getWidth() - roi.getWidth() - roi.getX() >=krhood &&
					                in->getHeight() - roi.getHeight() - roi.getY() >= kbhood )
					{
						//values can be taken directly from the image
						DBGOUT( DBGCONVOLVE, "Convolver 2D does not need to buffer in chan\n" );
					}
					else
					{
						DBGOUT( DBGCONVOLVE, "Convolver 2D buffers in chan\n" );
						mode |= BUFFER_IN;
					}

					switch ( mode )
					{
						case DIRECT:
							DBGOUT( DBGCONVOLVE, "Convolver2D: direct\n" );
							processDirect( in, out );
							break;
						case BUFFER_IN:
							DBGOUT( DBGCONVOLVE, "Convolver2D: in gets buffered\n" );
							processBufferIn( in, out );
							break;
						case BUFFER_OUT:
							DBGOUT( DBGCONVOLVE, "Convolver2D: out gets buffered\n" );
							processBufferOut( in ); //inPlace
							break;
						case BUFFER_IN_OUT:
							DBGOUT( DBGCONVOLVE, "Convolver2D: in and out get buffered\n" );
							processBufferInOut( in );  //inPlace
							break;
						default:
							assert( false );
							break;
					}
				}

				ConvKernT &getKernel()
				{
					return kernel;
				}

				void setMode( const BORDER_MODE theValue )
				{
					rbIn.setMode( theValue );
				}

				BORDER_MODE getMode() const
				{
					return rbIn.getMode();
				}

			private:

				ConvKernT kernel;
				Ringbuffer2D<img_in_type> rbIn;
				Ringbuffer2D<img_in_type> rbOut;

				template<class IteratorT>
				void sampleBuffer( IteratorT outItBegin, IteratorT outItEnd )
				{
					vector<typename Ringbuffer2D<img_in_type>::lineIterator> lits;
					rbIn.beginOfLines( lits );

//					typename Channel<img_out_type>::Iterator1D oit  = out->getXIteratorBegin( dstRow );
//					DBGOUT( DBGALL, "entering for loop\n" );
//					for ( ; oit != out->getXIteratorEnd( dstRow ); ++oit )
						for ( ; outItBegin != outItEnd; ++outItBegin )
						{
							img_out_type res = img_out_type( kernel.convolve( lits.begin() ) );
//							printf( "%d ", res );
							*outItBegin = res;

							//increment the iterator of each line
							for ( typename vector<typename Ringbuffer2D<img_in_type>::lineIterator>::iterator vit = lits.begin();
							                vit != lits.end(); ++vit )
							{
								++( *vit );
							}
						}
//					printf( "\n" );
				}

				void processDirect( Channel<img_in_type> *in, Channel<img_out_type> *out )
				{
					uint kthood = kernel.getTopHood();
					uint kbhood = kernel.getBottomHood();
					uint klhood = kernel.getLeftHood();
					uint srcRow = in->getROI().getY() - kthood;
					uint dstRow = out->getROI().getY();

					typedef typename Channel<img_in_type>::Iterator1D InItT;
//					InItT iit  = in->getXIteratorBegin( dstRow );
					RingBuffer<InItT> inIts( kthood + kbhood + 1 );
					for ( ; srcRow <= in->getROI().getY() + kbhood; ++srcRow )
					{
						inIts.insert( in->getXIterator( srcRow, in->getROI().getX() - klhood ) );
					}

//					DBGOUT( DBGALL, "processing line:\n" );
					for ( ; srcRow < in->getROI().getY() + in->getROI().getHeight() + kbhood; ++srcRow, ++dstRow )
					{
						typename Channel<img_out_type>::Iterator1D oit      = out->getXIteratorBegin( dstRow );
						typename Channel<img_out_type>::Iterator1D oitStop  = out->getXIteratorEnd( dstRow );
						for ( ; oit != oitStop; ++oit )
						{
							img_out_type res = img_out_type( kernel.convolve( inIts.begin() ) );
//							printf( "%d ", res );
							*oit = res;

							for ( typename RingBuffer<InItT>::iterator vit = inIts.begin();
							                vit != inIts.end(); ++vit )
							{
								++( *vit );
							}
						}
						inIts.insert( in->getXIterator( srcRow, in->getROI().getX() - klhood ) );
					}
				}

				void processBufferIn( Channel<img_in_type> *in, Channel<img_out_type> *out )
				{
					uint srcRow = in->getROI().getY();
					uint dstRow = out->getROI().getY();
					uint kthood = kernel.getTopHood();
					uint kbhood = kernel.getBottomHood();
					uint klhood = kernel.getLeftHood();
					uint krhood = kernel.getRightHood();
					ROI inROI = in->getROI();
					//initialize buffer
//					DBGOUT( DBGALL, "initializing buffer for kernel(l, r, t, b) %d, %d, %d, %d\n", klhood, krhood, kthood, kbhood );
					rbIn.init( inROI.getWidth() + klhood + krhood, kthood + kbhood + 1 );

					//preprocessing
					uint i;
					typename Channel<img_in_type>::Iterator1D iit = in->getXIterator( inROI.getY(), inROI.getX() );
					switch ( rbIn.getMode() )
					{
						case BORDER_COPY:
							for ( i = 0; i < kthood; ++i )
							{
								rbIn.insert( iit, klhood, inROI.getWidth() );
							}
							break;
						case BORDER_ZERO:
							for ( i = 0; i < kthood; ++i )
							{
								rbIn.insertZeros();
							}
							break;
						default:
							break;
					}

					//load first lines into ringbuffer
//					DBGOUT( DBGALL, "filling buffer\n" );
					for ( ; srcRow < inROI.getY() + kbhood; ++srcRow )
					{
						rbIn.insert( in->getXIterator( srcRow, inROI.getX() ), klhood, inROI.getWidth() );
					}

//					DBGOUT( DBGALL, "processing line:\n" );
					for ( ; srcRow < inROI.getY() + inROI.getHeight(); ++srcRow, ++dstRow )
					{
//						DBGOUT( DBGALL, "%d of %d\n", srcRow, in->getHeight() );
						rbIn.insert( in->getXIterator( srcRow, inROI.getX() ), klhood, inROI.getWidth() );
//						rb.print(); pressKey();
						sampleBuffer( out->getXIteratorBegin( dstRow ), out->getXIteratorEnd( dstRow ) );
					}
					--srcRow;
					for ( ; dstRow < out->getROI().getY() + out->getROI().getHeight(); ++dstRow )
					{
						switch ( rbIn.getMode() )
						{
							case BORDER_COPY:
//								printf( "inserting src line %d\n", srcRow );
								rbIn.insert( in->getXIterator( srcRow, inROI.getX() ), klhood, inROI.getWidth() );
								break;
							case BORDER_ZERO:
								rbIn.insertZeros();
								break;
							default:
								break;
						}
//						printf( "calculating row %d\n", dstRow );
//						rb.print(); pressKey();
						sampleBuffer( out->getXIteratorBegin( dstRow ), out->getXIteratorEnd( dstRow ) );
					}
					//postprocessing
//					DBGOUT( DBGALL, "leaving convolve 2d\n" );
				}

				void processBufferInOut( Channel<img_in_type> *in )
				{
					uint kthood = kernel.getTopHood();
					uint kbhood = kernel.getBottomHood();
					uint klhood = kernel.getLeftHood();
					uint krhood = kernel.getRightHood();
					uint srcRow = in->getROI().getY();
					uint dstRow = in->getROI().getY();
					ROI inROI = in->getROI();

					//initialize buffer
					rbOut.init( in->getROI().getWidth(), kthood + 1 );
					rbIn.init( inROI.getWidth() + klhood + krhood, kthood + kbhood + 1 );

					uint i;
					typename Channel<img_in_type>::Iterator1D iit = in->getXIterator( inROI.getY(), inROI.getX() );
					switch ( rbIn.getMode() )
					{
						case BORDER_COPY:
							for ( i = 0; i < kthood; ++i )
							{
								rbIn.insert( iit, klhood, inROI.getWidth() );
							}
							break;
						case BORDER_ZERO:
							for ( i = 0; i < kthood; ++i )
							{
								rbIn.insertZeros();
							}
							break;
						default:
							assert( false );
							break;
					}

					for ( ; srcRow < inROI.getY() + kbhood; ++srcRow )
					{
//						printf( "srcRow/dstRow is %d/%d\n", srcRow, dstRow );
						rbIn.insert( in->getXIterator( srcRow, inROI.getX() ), klhood, inROI.getWidth() );
					}

					//preprocessing - fill the out buffer
					for ( uint outBufRow = 0; outBufRow < kthood + 1; ++outBufRow, ++srcRow )
					{
//						printf( "srcRow/dstRow is %d/%d\n", srcRow, dstRow );
						sampleBuffer( rbOut.beginOfLine(), rbOut.endOfLine() );
						rbIn.insert( in->getXIterator( srcRow, inROI.getX() ), klhood, inROI.getWidth() );
						rbOut.feed();
					}

					for ( ; srcRow < in->getROI().getY() + in->getROI().getHeight(); ++srcRow, ++dstRow )
					{
//						printf( "srcRow/dstRow is %d/%d\n", srcRow, dstRow );
						typename Channel<img_in_type>::Iterator1D iit  = in->getXIteratorBegin( dstRow );
						for ( uint col = 0; col < in->getROI().getWidth(); ++col, ++iit )
						{
							*iit = rbOut( -kthood, col );
						}
						sampleBuffer( rbOut.beginOfLine(), rbOut.endOfLine() );
						rbIn.insert( in->getXIterator( srcRow, inROI.getX() ), klhood, inROI.getWidth() );
						rbOut.feed();
					}
					--srcRow;
					for ( ; dstRow < in->getROI().getY() + in->getROI().getHeight() - 1; ++dstRow )
					{
//						printf( "srcRow/dstRow is %d/%d\n", srcRow, dstRow );
						typename Channel<img_in_type>::Iterator1D iit  = in->getXIteratorBegin( dstRow );
						for ( uint col = 0; col < in->getROI().getWidth(); ++col, ++iit )
						{
							*iit = rbOut( -kthood, col );
						}
						sampleBuffer( rbOut.beginOfLine(), rbOut.endOfLine() );
						switch ( rbIn.getMode() )
						{
							case BORDER_COPY:
//								printf( "inserting src line %d\n", srcRow );
								rbIn.insert( in->getXIterator( srcRow, inROI.getX() ), klhood, inROI.getWidth() );
								break;
							case BORDER_ZERO:
								rbIn.insertZeros();
								break;
							default:
								break;
						}
						rbOut.feed();
					}

					//sample last line
//					printf( "srcRow/dstRow is %d/%d\n", srcRow, dstRow );
					iit  = in->getXIteratorBegin( dstRow );
					for ( uint col = 0; col < in->getROI().getWidth(); ++col, ++iit )
					{
						*iit = rbOut( -kthood, col );
					}
				}


				void processBufferOut( Channel<img_in_type> *in )
				{
					uint kthood = kernel.getTopHood();
					uint kbhood = kernel.getBottomHood();
					uint klhood = kernel.getLeftHood();
					uint srcRow = in->getROI().getY() - kthood;
					uint dstRow = in->getROI().getY();
					rbOut.init( in->getROI().getWidth(), kthood + 1 );

					typedef typename Channel<img_in_type>::Iterator1D InItT;
					InItT iit  = in->getXIteratorBegin( dstRow );
					RingBuffer<InItT> inIts( kthood + kbhood + 1 );
					for ( ; srcRow <= in->getROI().getY() + kbhood; ++srcRow )
					{
						inIts.insert( in->getXIterator( srcRow, in->getROI().getX() - klhood ) );
					}

					//preprocessing - fill the out buffer
					for ( uint outBufRow = 0; outBufRow < kthood + 1; ++outBufRow )
					{
						for ( uint col = 0; col < in->getROI().getWidth(); ++col )
						{
							img_out_type res = img_out_type( kernel.convolve( inIts.begin() ) );
							rbOut( col ) = res;

							for ( typename RingBuffer<InItT>::iterator vit = inIts.begin();
							                vit != inIts.end(); ++vit )
							{
								++( *vit );
							}
						}
						inIts.insert( in->getXIterator( srcRow, in->getROI().getX() - klhood ) );
						rbOut.feed();
					}

					for ( ; srcRow < in->getROI().getY() + in->getROI().getHeight() + kbhood; ++srcRow, ++dstRow )
					{
						typename Channel<img_in_type>::Iterator1D iit  = in->getXIteratorBegin( dstRow );
						for ( uint col = 0; col < in->getROI().getWidth(); ++col, ++iit )
						{
							*iit = rbOut( -kthood, col );

							img_out_type res = img_out_type( kernel.convolve( inIts.begin() ) );
							rbOut( col ) = res;

							for ( typename RingBuffer<InItT>::iterator vit = inIts.begin();
							                vit != inIts.end(); ++vit )
							{
								++( *vit );
							}
						}
						inIts.insert( in->getXIterator( srcRow, in->getROI().getX() - klhood ) );
						rbOut.feed();
					}

					//sample last line
//					printf( "srcRow/dstRow is %d/%d\n", srcRow, dstRow );
					iit  = in->getXIteratorBegin( dstRow );
					for ( uint col = 0; col < in->getROI().getWidth(); ++col, ++iit )
					{
						*iit = rbOut( -kthood, col );
					}
				}

		};

	}

}

#undef DBGCONVOLVE

#endif
