

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGDRAWING_H
#define ADORE_IMGPROCESSINGDRAWING_H

#include <limits>

#include <adrcore/types/aarray.h>
#include <adrcore/types/view.h>
#include <adrcore/types/nview.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/calctype.h>
#include <adrcore/util/templatelimits.h>
#include <adrcore/imgprocessing/bilinearinterpolator.h>
#include <adrcore/imgprocessing/drawingellipse.h>

namespace adore
{

	namespace core
	{

		class LineIterator
		{
			private:
				int x, y;
				int dx, dy, stepx, stepy, fraction, stop;
				bool done;

			public:
				LineIterator( int x0, int y0, int x1, int y1 );

				void init( int x0, int y0, int x1, int y1 );

				int getX()
				{
					return x;
				}

				int getY()
				{
					return y;
				}

				bool hasNext()
				{
					return !done;
				}

				void next();
		};

		template<class typeT>
		void lineBresenham( Channel<typeT> &chan, int x0, int y0, int x1, int y1, typeT value )
		{
			int dy = y1 - y0;
			int dx = x1 - x0;
			int stepx, stepy;
			int offX = chan.getROI().getX();
			int offY = chan.getROI().getY();

			if ( dy < 0 )
			{
				dy = -dy;  stepy = -1;
			}
			else
			{
				stepy = 1;
			}
			if ( dx < 0 )
			{
				dx = -dx;  stepx = -1;
			}
			else
			{
				stepx = 1;
			}
			dy <<= 1;                                                  // dy is now 2*dy
			dx <<= 1;                                                  // dx is now 2*dx

			chan( x0 + offX, y0 + offY )=value;

			if ( dx > dy )
			{
				int fraction = dy - ( dx >> 1 );                       // same as 2*dy - dx
				while ( x0 != x1 )
				{
					if ( fraction >= 0 )
					{
						y0 += stepy;
						fraction -= dx;                   // same as fraction -= 2*dx
					}
					x0 += stepx;
					fraction += dy;                      // same as fraction -= 2*dy
					chan( x0 + offX, y0 + offY )=value;
				}
			}
			else
			{
				int fraction = dx - ( dy >> 1 );
				while ( y0 != y1 )
				{
					if ( fraction >= 0 )
					{
						x0 += stepx;
						fraction -= dy;
					}
					y0 += stepy;
					fraction += dx;
					chan( x0 + offX, y0 + offY )=value;
				}
			}
		}

		template<class typeT>
		void circleBresenham( Channel<typeT> &chan, int x0, int y0, int r, typeT value )
		{
			int offX = chan.getROI().getX() + x0;
			int offY = chan.getROI().getY() + y0;

			int y = r;
			int d = -r;
			int x2m1= -1;
			chan( offX, offY + r )=value;
			chan( offX, offY - r )=value;
			for ( int x = 1; x <= r/sqrt( 2 ); x++ )
			{
				x2m1 += 2;
				d += x2m1;
				if ( d >= 0 )
				{
					y--;
					d -= ( y << 1 );  /* Must do this AFTER y-- */
				}
				chan( offX + x, offY + y )=value;
				chan( offX + y, offY + x )=value;
				chan( offX - x, offY + y )=value;
				chan( offX - y, offY + x )=value;
				chan( offX + x, offY - y )=value;
				chan( offX + y, offY - x )=value;
				chan( offX - x, offY - y )=value;
				chan( offX - y, offY - x )=value;
			}
		}

		template<class typeT>
		void drawLine( Channel<typeT> &chan, int x0, int y0, int x1, int y1, typeT value )
		{
//			DBGOUT( DBGNORMAL, "drawline: (%d, %d)-(%d, %d)\n", x0, y0, x1, y1 );
			int xOff = chan.getROI().getX();
			int yOff = chan.getROI().getY();
			if ( ( x0 + xOff < 0 ) || ( x0 + xOff >= ( int )chan.getWidth() ) ||
			                ( y0 + yOff < 0 ) || ( y0 + yOff >= ( int )chan.getHeight() ) ||
			                ( x1 + xOff < 0 ) || ( x1 + xOff >= ( int )chan.getWidth() ) ||
			                ( y1 + yOff < 0 ) || ( y1 + yOff >= ( int )chan.getHeight() ) )
			{
				DBGOUT( DBGNORMAL, "drawline: (%d, %d)-(%d, %d) not drawable\n", x0, y0, x1, y1 );
				return;
			}
			lineBresenham( chan, x0, y0, x1, y1, value );
		}

		template<class typeT>
		void drawCircle( Channel<typeT> &chan, int x0, int y0, int r, typeT value )
		{
			int xOff = chan.getROI().getX() + x0;
			int yOff = chan.getROI().getY() + y0;
			if ( ( xOff - r < 0 ) || ( xOff + r >= ( int )chan.getWidth() ) ||
			                ( yOff - r < 0 ) || ( yOff + r >= ( int )chan.getHeight() ) )
			{
				DBGOUT( DBGNORMAL, "drawCircle: (%d, %d)-(%d) not drawable\n", x0, y0, r );
				return;
			}
			circleBresenham( chan, x0, y0, r, value );
		}

		/**
			draws a cross at a specific point
		*/
		template<class typeT>
		void markPoint( Channel<typeT> &chan, int x, int y, typeT value )
		{
			const int size = 2;
			drawLine( chan, max( 0, x - size ), y,
			          min( x + size, static_cast<int>( chan.getWidth() - 1 ) ), y, value );
			drawLine( chan, x, max( 0, y - size ), x,
			          min( y + size, static_cast<int>( chan.getWidth() - 1 ) ), value );
		}

		template<class typeT>
		ImagePlane<typeT> *combineImages( const ImagePlane<typeT> * const i1,
		                                  const ImagePlane<typeT> * const i2 )
		{
			typedef ImagePlane<typeT> ImageT;

			uint numChans = i1->getNumChannels();

			assert( i2->getNumChannels() == numChans );

			int w1 = static_cast<int>( i1->getWidth()+0.5 );
			int h1 = static_cast<int>( i1->getHeight()+0.5 );
			int w2 = static_cast<int>( i2->getWidth()+0.5 );
			int h2 = static_cast<int>( i2->getHeight()+0.5 );
			ImageT *img = new ImageT( w1 + w2, max( h1, h2 ), numChans );

			//copy content
			img->setROI( 0, 0, w1, h1 );
			( *img ) = *i1;
			img->setROI( w1, 0, w2, h2 );
			( *img ) = *i2;
			img->resetROI();

			return img;
		}

		template<class typeT>
		void drawViewImage( View<typeT> *v, ImagePlane<typeT> *img )
		{
			assert( img );

			typedef ImagePlane<typeT> ImageT;
			typedef View<typeT> ViewT;

			for ( uint featID = 0; featID < maxFeatureTypeNo; ++featID )
			{
				if ( v->getNumFeaturePoints( featID ) )
					for ( typename ViewT::PointsIterator it = v->getPointsBegin( featID );
					                it != v->getPointsEnd( featID ); ++it )
					{
						FeaturePoint *fp = *it;
						if ( fp )
						{
							FeaturePoint &f = *fp;
							assert( f( 2 ) == 1.0 );
							markPoint( *img, static_cast<int>( f( 0 )+0.5 ),
							           static_cast<int>( f( 1 )+0.5 ), maxLimit<typeT>() );
						}
					}
			}
		}

		/**
			creates an image that contain the image of the view with marker at the
			feature coordinates
		*/
		template<class typeT>
		ImagePlane<typeT> *createViewImage( View<typeT> *v )
		{
			typedef ImagePlane<typeT> ImageT;
			typedef View<typeT> ViewT;

			ImageT *img = NULL;
			if ( v->hasImage() )
			{
				img = new ImageT( *v->getImage() );
			}
			else
			{
				assert( !v->isNormalized() );
				img = new ImageT( static_cast<int>( v->getWidth()+0.5 ),
				                  static_cast<int>( v->getHeight()+0.5 ), 1 );
			}
			assert( img );

			drawViewImage( v, img );

			return img;
		}

		template<class typeT>
		ImagePlane<typeT> *createViewsImage( View<typeT> *v1, View<typeT> *v2, NView *nv )
		{
			typedef ImagePlane<typeT> ImageT;
			typedef View<typeT> ViewT;

			assert( !v1->isNormalized() );
			assert( !v2->isNormalized() );

			ImageT *img = combineImages( v1->getImage(), v2->getImage() );
			int w1 = static_cast<int>( v1->getImage()->getWidth()+0.5 );

			for ( typename NView::CorrIterator it = nv->getPointCorrsBegin();
			                it != nv->getPointCorrsEnd(); ++it )
			{
				assert( *it );

				FeaturePoint *fp1 = ( *it )->getFeature( v1 );
				if ( !fp1 ) continue;
				FeaturePoint *fp2 = ( *it )->getFeature( v2 );
				if ( !fp2 ) continue;
				int x1 = static_cast<int>( ( *fp1 )( 0 ) );
				int y1 = static_cast<int>( ( *fp1 )( 1 ) );
				int x2 = w1 + static_cast<int>( ( *fp2 )( 0 ) );
				int y2 = static_cast<int>( ( *fp2 )( 1 ) );

				drawLine( *img, x1, y1, x2, y2, maxLimit<typeT>() );
				markPoint( *img, x1, y1, maxLimit<typeT>() );
				markPoint( *img, x2, y2, maxLimit<typeT>() );
			}

			return img;
		}

		template<class typeT>
		ImagePlane<typeT> *createViewsImage( View<typeT> *v1, View<typeT> *v2 )
		{
			typedef ImagePlane<typeT> ImageT;
			typedef View<typeT> ViewT;

			assert( !v1->isNormalized() );
			assert( !v2->isNormalized() );

			ImageT *i1 = v1->getImage();
			ImageT *i2 = v2->getImage();
			ImageT *img = combineImages( i1, i2 );
			int w1 = static_cast<int>( i1->getWidth()+0.5 );
			int h1 = static_cast<int>( i1->getHeight()+0.5 );
			int w2 = static_cast<int>( i2->getWidth()+0.5 );
			int h2 = static_cast<int>( i2->getHeight()+0.5 );

			img->setROI( 0, 0, w1, h1 );
			drawViewImage( v1, img );
			img->setROI( w1, 0, w2, h2 );
			drawViewImage( v2, img );
			img->resetROI();

			return img;
		}

		template<class typeT>
        ImagePlane<typeT> *visualizeEpipolarGeometry( const AMatrix & fmat, View<typeT> *v1, View<typeT> *v2 )
		{
			typedef ImagePlane<typeT> ImageT;
			typedef View<typeT> ViewT;

			assert( !v1->isNormalized() );
			assert( !v2->isNormalized() );

			ImageT *i1 = v1->getImage();
			ImageT *i2 = v2->getImage();
			ImageT *img = combineImages( i1, i2 );
			int w1 = static_cast<int>( i1->getWidth()+0.5 );
			int h1 = static_cast<int>( i1->getHeight()+0.5 );
			int w2 = static_cast<int>( i2->getWidth()+0.5 );
			int h2 = static_cast<int>( i2->getHeight()+0.5 );

			AVector pt( 3 );
			int midX = w1 >> 1;
			double x1, y1, x2, y2;
			for ( int row = 0; row < h1; row+=30 )
			{
				pt = midX, row, 1;
				AVector line = matMul( fmat, pt );
				pointAsLine( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w2 - 2, h2 - 2 );
				img->setROI( w1, 0, w2, h2 );
				drawLine( *img, int( x1+0.5 ), int( y1+0.5 ), int( x2+0.5 ), int( y2+0.5 ), maxLimit<typeT>() );

				pt = x1, y1, 1;
				line = matMul_AT( fmat, pt );
				pointAsLine( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w1 - 2, h1 - 2 );
				img->setROI( 0, 0, w1, h1 );
				drawLine( *img, int( x1+0.5 ), int( y1+0.5 ), int( x2+0.5 ), int( y2+0.5 ), maxLimit<typeT>() );
			}
			img->resetROI();

			return img;
		}

		template<class typeT>
        ImagePlane<typeT> * visualizeEpipolarGeometry2( const AMatrix & fmat, View<typeT> * v1, View<typeT> * v2 )
		{
			typedef ImagePlane<typeT> ImageT;
			typedef View<typeT> ViewT;

			assert( !v1->isNormalized() );
			assert( !v2->isNormalized() );

			ImageT *i1 = v1->getImage();
			ImageT *i2 = v2->getImage();
			ImageT *img = combineImages( i1, i2 );
			int w1 = static_cast<int>( i1->getWidth()+0.5 );
			int h1 = static_cast<int>( i1->getHeight()+0.5 );
			int w2 = static_cast<int>( i2->getWidth()+0.5 );
			int h2 = static_cast<int>( i2->getHeight()+0.5 );

			double x1, y1, x2, y2;
			ViewBase::PointsIterator pit = v1->getPointsBegin();
			AVector line( 3 );
			while ( pit != v1->getPointsEnd() )
			{
				FeaturePoint* fp = *pit;
				img->setROI( 0, 0, w1, h1 );
				drawCircle( *img, ( int )( *fp )( 0 ), ( int )( *fp )( 1 ), 4, ( typeT )( maxLimit<typeT>() ) );
				AVector pt( 3 ); pt = ( *fp )( 0 ), ( double )h1 - ( *fp )( 1 ), 1;
				line = matMul( fmat, ( *fp ) );
				img->setROI( w1, 0, w2, h2 );
				bool ok = pointAsLine2( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w2 - 1, h2 - 1 );
				if ( ok )
					drawLine( *img, int( x1+0.5 ), int( y1+0.5 ), int( x2+0.5 ), int( y2+0.5 ), ( typeT )( maxLimit<typeT>() ) );
				++pit;
			}
			pit = v2->getPointsBegin();
			while ( pit != v2->getPointsEnd() )
			{
				FeaturePoint* fp = *pit;
				img->setROI( w1, 0, w2, h2 );
				drawCircle( *img, ( int )( *fp )( 0 ), ( int )( *fp )( 1 ), 4, ( typeT )( maxLimit<typeT>() ) );
				matMul2_AT( fmat, ( *fp ), line );
				img->setROI( 0, 0, w1, h1 );
				bool ok = pointAsLine2( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w1 - 1, h1 - 1 );
				if ( ok )
					drawLine( *img, int( x1+0.5 ), int( y1+0.5 ), int( x2+0.5 ), int( y2+0.5 ), ( typeT )( maxLimit<typeT>() ) );

				++pit;
			}
			img->resetROI();
			return img;
		}


		template<class typeT>
        ImagePlane<typeT> *createViewsImageEpi( View<typeT> *v1, View<typeT> *v2, NView *nv, const AMatrix & fmat, bool uniformly = true )
		{
			assert( v1 );
			assert( v2 );
			typedef ImagePlane<typeT> ImageT;
			typedef View<typeT> ViewT;

			assert( !v1->isNormalized() );
			assert( !v2->isNormalized() );

			ImageT *i1 = v1->getImage();
			ImageT *i2 = v2->getImage();
			assert( i1 );
			assert( i2 );
			ImageT *img = combineImages( i1, i2 );
			assert( img );
			int w1 = static_cast<int>( i1->getWidth()+0.5 );
			int h1 = static_cast<int>( i1->getHeight()+0.5 );
			int w2 = static_cast<int>( i2->getWidth()+0.5 );
			int h2 = static_cast<int>( i2->getHeight()+0.5 );

			AVector pt( 3 );
			if ( uniformly )
			{
				int midX = w1 >> 1;
				double x1, y1, x2, y2;
				for ( int row = 0; row < h1; row+=30 )
				{
					pt = midX, row, 1;
					AVector line = matMul( fmat, pt );
					pointAsLine( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w2 - 2, h2 - 2 );
					img->setROI( w1, 0, w2, h2 );
					drawLine( *img, int( x1+0.5 ), int( y1+0.5 ), int( x2+0.5 ), int( y2+0.5 ), maxLimit<typeT>() );

					pt = x1, y1, 1;
					line = matMul_AT( fmat, pt );
					pointAsLine( line( 0 ), line( 1 ), line( 2 ), x1, y1, x2, y2, w1 - 2, h1 - 2 );
					img->setROI( 0, 0, w1, h1 );
					drawLine( *img, int( x1+0.5 ), int( y1+0.5 ), int( x2+0.5 ), int( y2+0.5 ), maxLimit<typeT>() );
				}
				img->resetROI();
			}


			for ( typename NView::CorrIterator it = nv->getPointCorrsBegin();
			                it != nv->getPointCorrsEnd(); ++it )
			{
				assert( *it );

				FeaturePoint *fp1 = ( *it )->getFeature( v1 );
				FeaturePoint *fp2 = ( *it )->getFeature( v2 );
				assert( fp1 );
				assert( fp2 );
				int x1 = static_cast<int>( ( *fp1 )( 0 ) );
				int y1 = static_cast<int>( ( *fp1 )( 1 ) );
				int x2 = w1 + static_cast<int>( ( *fp2 )( 0 ) );
				int y2 = static_cast<int>( ( *fp2 )( 1 ) );

				drawLine( *img, x1, y1, x2, y2, maxLimit<typeT>() );
				markPoint( *img, x1, y1, maxLimit<typeT>() );
				markPoint( *img, x2, y2, maxLimit<typeT>() );

				if ( !uniformly )
				{
					x2 -= w1;
					double lx1, ly1, lx2, ly2;

					pt = x1, y1, 1;
//					AVector line = matMul( fmat, pt );
                                        AVector line(3);
                                        matMul2( fmat, pt, line );
					pointAsLine( line( 0 ), line( 1 ), line( 2 ), lx1, ly1, lx2, ly2, w2 - 2, h2 - 2 );
					img->setROI( w1, 0, w2, h2 );
					drawLine( *img, int( lx1+0.5 ), int( ly1+0.5 ), int( lx2+0.5 ), int( ly2+0.5 ), maxLimit<typeT>() );

					pt = x2, y2, 1;
					line = matMul_AT( fmat, pt );
					pointAsLine( line( 0 ), line( 1 ), line( 2 ), lx1, ly1, lx2, ly2, w1 - 2, h1 - 2 );
					img->setROI( 0, 0, w1, h1 );
					drawLine( *img, int( lx1+0.5 ), int( ly1+0.5 ), int( lx2+0.5 ), int( ly2+0.5 ), maxLimit<typeT>() );

					img->resetROI();
				}
			}
			return img;
		}

	}
}

#endif
