

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGHARRISFASTHOOD_H
#define ADORE_IMGPROCESSINGHARRISFASTHOOD_H

#include <adrcore/imgprocessing/imgfunctions.h>
#include <adrcore/types/aarray.h>
#include <adrcore/util/gaussians.h>
#include "harriscommon.h"
#include <numeric>

namespace adore
{

	namespace core
	{

		/**
		 * this is called harrisfasthood, because ists like harrisfast.
		 * unlike it, you can pass a hood. Yet for large hoods this is all, but fast.
		 
		*/
		template <class inT>
		class HarrisFastHood
		{
				typedef typename signedT< typename calcT<inT>::type >::type calcT;
			public:
				typedef typename harrisOutT<inT>::ouT ouT;


				HarrisFastHood( uint _hood = 1, float _scale = 1.0 ):
						k( 0.04 ),
						hood( _hood ),
						width( hood + hood + 1 ),
						scale( _scale )
				{
					arr = new float[width * width];
					weights = new float[( width - 1 ) * ( width - 1 )];
					getGaussianWeights2( weights, hood );

					/*for(int y = 0; y < width - 1; ++y)
					{
						int yoff = (width - 1) * y;
						for(int x = 0; x < width - 1; ++x)
						{
							printf("%f ", weights[x + yoff]);
						}
						printf("\n");
					}*/
				}

				~HarrisFastHood()
				{
					delete [] arr;
					delete [] weights;
				}

				void process( Channel<inT> *in, Channel<ouT> *out );

			protected:

				void addG( float a, float b, float c, float d, float weight, double &gxx, double &gyy, double &gxy )
				{
					float xx = b - a + d - c;
					float yy = c - a + d - b;
					gxy += weight * xx * yy;
					gxx += weight * xx * xx;
					gyy += weight * yy * yy;
				}

				void addGArr( double &gxx, double &gyy, double &gxy )
				{
					for ( int y = 0; y < width - 1; ++y )//the <-1 is essential
					{
						int yoff = width * y;
						int wyoff = ( width - 1 ) * y;
						for ( int x = 0; x < width - 1; ++x )//the <-1 is essential
						{
							float a = arr[x + yoff];
							float b = arr[x + 1 + yoff];
							float c = arr[x + yoff + width];
							float d = arr[x + 1 + yoff + width];
							addG( a, b, c, d, weights[x + wyoff], gxx, gyy, gxy );
						}
					}
				}


				void addGSort( double a, double b, double c, double d, float weight, vector<double> &gxx, vector<double> &gyy, vector<double> &gxy )
				{
					double xx = b - a + d - c;
					double yy = c - a + d - b;
					gxy.push_back( weight * xx * yy );
					gxx.push_back( weight * xx * xx );
					gyy.push_back( weight * yy * yy );
				}

				void addGSortArr( double &gxx, double &gyy, double &gxy )
				{
					vector<double> gxxarr; gxxarr.reserve( width * width );
					vector<double> gyyarr; gyyarr.reserve( width * width );
					vector<double> gxyarr; gxyarr.reserve( width * width );
					for ( int y = 0; y < width - 1; ++y )//the <-1 is essential
					{
						int yoff = width * y;
						int wyoff = ( width - 1 ) * y;
						for ( int x = 0; x < width - 1; ++x )//the <-1 is essential
						{
							double a = arr[x + yoff];
							double b = arr[x + 1 + yoff];
							double c = arr[x + yoff + width];
							double d = arr[x + 1 + yoff + width];
							addGSort( a, b, c, d, weights[x + wyoff], gxxarr, gyyarr, gxyarr );
						}
					}
					std::sort( gxxarr.begin(), gxxarr.end() );
					std::sort( gyyarr.begin(), gyyarr.end() );
					std::sort( gxyarr.begin(), gxyarr.end() );
					assert( gxxarr[0] <= gxxarr[1] );
					assert( gyyarr[0] <= gyyarr[1] );
					assert( gxyarr[0] <= gxyarr[1] );
					gxx = std::accumulate( gxxarr.begin(), gxxarr.end(), 0.0 );
					gyy = std::accumulate( gyyarr.begin(), gyyarr.end(), 0.0 );
					gxy = std::accumulate( gxyarr.begin(), gxyarr.end(), 0.0 );
				}


				const float k;
				const int hood;
				const int width;
				const float scale;
				float *arr;
				float *weights;
		};


		template <class inT>
		void HarrisFastHood<inT>::process( Channel<inT> *in, Channel<ouT> *out )
		{
			UsingROI r1( in );
			UsingROI r2( out );
			in->minROIDistance( hood );
			out->fill( 0 );
//			int rX = in->getROI().getX();
//			int rY = in->getROI().getY();
			int rW = in->getROI().getWidth();
			int rH = in->getROI().getHeight();
			out->setCenterdROI( rW, rH );

			typename Channel<inT>::Iterator2D inIt = in->getIteratorBegin();
			typename Channel<ouT>::Iterator2D outIt = out->getIteratorBegin();
			typename Channel<ouT>::Iterator2D outItEnd = out->getIteratorEnd();

//			ouT maxVal = 0, minVal = 0;
			for ( ; outIt != outItEnd; ++outIt, ++inIt )
			{
				//copy the hood into arr
				float *ptr = &arr[0];
				for ( int y = -hood; y <= hood; ++y )
				{
					for ( int x = -hood; x <= hood; ++x )
					{
						*ptr = *inIt.getPointer( x, y );
						++ptr;
					}
				}

				double gxx = 0, gyy = 0, gxy = 0;
				addGArr( gxx, gyy, gxy );
//				addGSortArr( gxx, gyy, gxy );

				const double shift = 0.1; //to stay in a sane range (tested with scaleharrisfinder)
				gxx = shift * scale * gxx;
				gyy = shift * scale * gyy;
				gxy = shift * scale * gxy;

//				printf( "gxx: %f, gyy: %f, gxy: %f\n", gxx, gyy, gxy );

				double trace = gxx + gyy;
//				printf( "%f\n", trace );
				trace *= trace;
//				printf( "%f\n", trace );
				double result = k * trace;
//				printf( "%f\n", result );
//				printf( "%f\n", fgxx * fgyy - fgxy * fgxy - result );
				*outIt = ouT( gxx * gyy - gxy * gxy - result );

//				if ( *outIt > 0 ) printf( "!!!!!!!!!!!!!!!!!!" );
				//printf("%f, %f, %f -> %d\n", gxx, gyy, gxy, *outIt);pressKey();
//				if ( ( *outIt ) > maxVal )maxVal = *outIt;
//				if ( ( *outIt ) < minVal )minVal = *outIt;
			}
//			printf( "min, max : %d, %d\n", minVal, maxVal );pressKey();
		}

		template <>
		inline void HarrisFastHood<float>::process( Channel<float> *in, Channel<float> *out )
		{
			ADORE_UNUSED( in );
			ADORE_UNUSED( out );
			abort();
		}

		template <>
		inline void HarrisFastHood<double>::process( Channel<double> *in, Channel<double> *out )
		{
			ADORE_UNUSED( in );
			ADORE_UNUSED( out );
			abort();
		}


	}

}

#endif
