

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGIMAGEPLANE_H
#define ADORE_IMGPROCESSINGIMAGEPLANE_H

#include <adrcore/imgprocessing/channel.h>
#include <adrcore/persistence/imgloader.h>
#include <adrcore/persistence/imgsaver.h>

namespace adore
{

	namespace core
	{

		template<class imgT> inline int getImageTypeId()
		{
			error ( "type undefined\n" );
		}
		template<> inline int getImageTypeId<uint8_t>()
		{
			return 0;
		}
		template<> inline int getImageTypeId<uint16_t>()
		{
			return 1;
		}
		template<> inline int getImageTypeId<uint32_t>()
		{
			return 2;
		}
		template<> inline int getImageTypeId<int8_t>()
		{
			return 3;
		}
		template<> inline int getImageTypeId<int16_t>()
		{
			return 4;
		}
		template<> inline int getImageTypeId<int32_t>()
		{
			return 5;
		}
		template<> inline int getImageTypeId<float>()
		{
			return 6;
		}
		template<> inline int getImageTypeId<double>()
		{
			return 7;
		}

		template<int id> struct ImageId	{};
		template<> struct ImageId<0>
		{
			typedef uint8_t type;
		};
		template<> struct ImageId<1>
		{
			typedef uint16_t type;
		};
		template<> struct ImageId<2>
		{
			typedef uint32_t type;
		};
		template<> struct ImageId<3>
		{
			typedef int8_t type;
		};
		template<> struct ImageId<4>
		{
			typedef int16_t type;
		};
		template<> struct ImageId<5>
		{
			typedef int32_t type;
		};
		template<> struct ImageId<6>
		{
			typedef float type;
		};
		template<> struct ImageId<7>
		{
			typedef double type;
		};

		class ImageBase
		{
			public:

				ImageBase ( int mode = CHANNEL_SPLIT ) :
						channelMode ( mode ), type_id ( -1 )
				{
				}

				virtual ~ImageBase()
				{
				}

				virtual void setChannelMode ( const int& theValue )
				{
					channelMode = theValue;
				}

				int getChannelMode() const
				{
					return channelMode;
				}

				int getTypeId()
				{
					return type_id;
				}

			protected:

				void setTypeId ( int id )
				{
					type_id = id;
				}

			private:
				int channelMode;
				int type_id;
		};

		template<class T> class Channel;
		template<class T> class ChannelSet;

		template<class T>
		class ImagePlane : public ImageBase, public Channel<T>
		{
//			public:

//				typedef typename Channel<T>::ImgArray ImgArray;

			public:

				ImagePlane ( ImgArray *_img, bool deleteImg = false );

				ImagePlane ( const char *filename, bool deleteImg = true );
				ImagePlane ( const vector<unsigned char>& data, bool deleteImg = true );

				ImagePlane ( int width, int height, int nChans, bool deleteImg = true );

				ImagePlane ( const ImagePlane<T> & ip, bool copyContent = true );

				ImagePlane();

				~ImagePlane();

				void init ( const uint width, const uint height, const uint nChans );

				/**
				 * @return success of operation
				 */
				bool load ( const char *name );
				
				void save ( const char *name ) const;

				void fillAll ( const T val )
				{
					int ch = getChannelMode();
					setChannelMode ( CHANNEL_SPLIT );
					Channel<T>::fill ( val );
					setChannelMode ( ch );
				}

				void operator= ( const ImagePlane<T> &ip );

				void operator= ( const Channel<T> &ip );

				Channel<T>* createChannel ( const uint i ) const
				{
					uint offSet, incStep, numBytes;
					getChannelParam ( int ( i ), offSet, incStep, numBytes );
					Channel<T> *chan = new Channel<T> ( this->getImgArray(), offSet, numBytes, incStep, this->wStep, this->getROI() );
					return chan;
				}

				void createChannels ( ChannelSet<T> &set ) const
				{
					for ( uint i = 0; i < this->getNumChans(); ++i )
					{
						set.add ( createChannel ( i ) );
					}
				}

				void setChannelMode ( const int& channelMode )
				{
					getChannelParam ( channelMode, Channel<T>::offSet, Channel<T>::incStep, Channel<T>::numBytes );
					ImageBase::setChannelMode ( channelMode );
				}

			protected:
				bool deleteImage;

				void getChannelParam ( const int channelmode, uint &offSet, uint &incStep, uint &numBytes ) const;

				void setWStep();

		};

		class UsingChannelMode
		{
			public:
				UsingChannelMode ( ImageBase *i, int cmode ) :
						mode ( i->getChannelMode() ),
						plane ( i )
				{
					plane->setChannelMode ( cmode );
				}

				UsingChannelMode ( ImageBase *i ) :
						mode ( i->getChannelMode() ),
						plane ( i )
				{
				}

				~UsingChannelMode()
				{
					plane->setChannelMode ( mode );
				}
			private:
				int mode;
				ImageBase *plane;
		};


//
//		Specializations
//


		template<class T> ImagePlane<T>::ImagePlane ( ImgArray *_img, bool deleteImg ) :
				ImageBase ( CHANNEL_SPLIT ),
				Channel<T> ( _img, 0, sizeof ( T ), 1 ),
				deleteImage ( deleteImg )
		{
			this->getROI().setWidth ( this->getImgArray()->width );
			this->getROI().setHeight ( this->getImgArray()->height );
			this->setWStep();
			setTypeId ( getImageTypeId<T>() );
		}

		template<class T> ImagePlane<T>::ImagePlane ( int width, int height, int nChans, bool deleteImg ) :
				ImageBase ( CHANNEL_SPLIT ),
				Channel<T> ( NULL, 0, sizeof ( T ), 1 ),
				deleteImage ( deleteImg )
		{
			init ( width, height, nChans );
		}

		template<class T> ImagePlane<T>::ImagePlane() :
				ImageBase ( CHANNEL_SPLIT ),
				Channel<T> ( NULL, 0, sizeof ( T ), 1, 0 ),
				deleteImage ( true )
		{
			setTypeId ( getImageTypeId<T>() );
		}

		template<class T> ImagePlane<T>::ImagePlane ( const ImagePlane &ip, bool copyContent ) :
				ImageBase ( CHANNEL_SPLIT ),
				Channel<T> ( NULL, 0, sizeof ( T ), 1, 0 ),
				deleteImage ( true )
		{
			if ( ip.getImgArray() )
			{
				init ( ip.getWidth(), ip.getHeight(), ip.getNumChannels() );
				if ( copyContent )
				{
					operator= ( ip );
				}
			}
		}

		template<class T> ImagePlane<T>::ImagePlane ( const char *filename, bool deleteImg ) :
				ImageBase ( CHANNEL_SPLIT ),
				Channel<T> ( NULL, 0, sizeof ( T ), 1 ),
				deleteImage ( deleteImg )
		{
			ImgLoader<T> loader;
			int width, height, col;//for the image info
			loader.getInfo ( string ( filename ), width, height, col );

			bool success = true;
			if ( 0 >= width || 0 >= height || 0 >= col )
				success = false;

			ImgArray* img = 0;
			if ( success )
			{
				img = createImgArray<T> ( width, height , col );
				success = loader.loadToBuffer ( string ( filename ), img );
			}

			/*			assert_msg ( success, "Loading image - File %s not found\n", filename );
						ADORE_UNUSED ( success );*/
			this->setImgArray ( img );

			if ( !success ) return; //INVALID STATE!(ZOMBIE)

			this->getROI().setWidth ( img->width );
			this->getROI().setHeight ( img->height );
			this->setWStep();
			setTypeId ( getImageTypeId<T>() );
		}

		template<class T>
		ImagePlane<T>::ImagePlane( const vector<unsigned char>& data, bool deleteImg )
				: ImageBase ( CHANNEL_SPLIT ),
				Channel<T> ( NULL, 0, sizeof ( T ), 1 ),
				deleteImage ( deleteImg )
		{
			PNGPersistence loader;

			int width, height, col;

			bool success = loader.getInfo( &data[0], data.size(), width, height, col );
			if ( 0 >= width || 0 >= height || 0 >= col )
				success = false;

			ImgArray* img = 0;
			if ( success )
			{
				img = createImgArray<T> ( width, height , col );
				success = readPNG( &data[0], data.size(), img->imgData, img->widthStep );
			}

			this->setImgArray ( img );

			if ( !success ) return; //INVALID STATE!(ZOMBIE)

			this->getROI().setWidth ( img->width );
			this->getROI().setHeight ( img->height );
			this->setWStep();
			setTypeId ( getImageTypeId<T>() );
		}

		template<class T> ImagePlane<T>::~ImagePlane()
		{
			if ( deleteImage )
			{
				ImgArray *img = this->getImgArray();
				releaseImgArray ( img );
				this->setImgArray ( NULL );
			}
		}

		template<class T> void ImagePlane<T>::init ( const uint width, const uint height, const uint nChans )
		{
			ImgArray *img = this->getImgArray();
			if ( !img || this->getWidth() != width || this->getHeight() != height || this->getNumChannels() != nChans )
			{
				if ( img && this->deleteImage ) releaseImgArray ( img );
				img = createImgArray<T> ( width, height, nChans );
				this->setImgArray ( img );
				this->setWStep();
				this->deleteImage = true;
			}
			this->setROI ( 0, 0, width, height );
			setChannelMode ( CHANNEL_SPLIT );
			setTypeId ( getImageTypeId<T>() );
		}

		template<class T> void ImagePlane<T>::operator= ( const ImagePlane<T> &ip )
		{
			assert ( this->getROI().getWidth() == ip.getROI().getWidth() );
			assert ( this->getROI().getHeight() == ip.getROI().getHeight() );
			assert ( this->getNumChannels() == ip.getNumChannels() );

			UsingChannelMode usm1 ( ( ImageBase * ) this, CHANNEL_COMBINED );
			UsingChannelMode usm2 ( ( ImageBase * ) &ip, CHANNEL_COMBINED );

			typename Channel<T>::Iterator2D srcIt = ip.getIteratorBegin();
			typename Channel<T>::Iterator2D dstIt = this->getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = ip.getIteratorEnd();
//			typename Channel<T>::Iterator2D dstEnd = this->getIteratorEnd();

//			printf( "%p, %p, %p, %p\n", srcIt.getPointer(), srcEnd.getPointer(), dstIt.getPointer(), dstEnd.getPointer() );
//			int x, y;
//			ip.getCoors( srcIt.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			ip.getCoors( srcEnd.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			this->getCoors( dstIt.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			this->getCoors( dstEnd.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			pressKey();

//			printf( "%p, %p, %p, %p\n", srcIt.getPointer(), srcEnd.getPointer(), dstIt.getPointer(), dstEnd.getPointer() );
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
//				printf( "copying\n" );
				dstIt ( srcIt );

			}
		}

		template<class T> void ImagePlane<T>::operator= ( const Channel<T> &ip )
		{
			assert ( this->getROI().getWidth() == ip.getROI().getWidth() );
			assert ( this->getROI().getHeight() == ip.getROI().getHeight() );
			assert ( this->getChannelMode() >= 0 || this->getNumChans() == 1 );

			typename Channel<T>::Iterator2D srcIt = ip.getIteratorBegin();
			typename Channel<T>::Iterator2D dstIt = this->getIteratorBegin();
			typename Channel<T>::Iterator2D srcEnd = ip.getIteratorEnd();
			typename Channel<T>::Iterator2D dstEnd = this->getIteratorEnd();

//			printf( "%p, %p, %p, %p\n", srcIt.getPointer(), srcEnd.getPointer(), dstIt.getPointer(), dstEnd.getPointer() );
//			int x, y;
//			ip.getCoors( srcIt.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			ip.getCoors( srcEnd.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			this->getCoors( dstIt.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			this->getCoors( dstEnd.getPointer(), x, y );
//			printf( "s1 (%d, %d)\n", x, y );
//			pressKey();

//			printf( "%p, %p, %p, %p\n", srcIt.getPointer(), srcEnd.getPointer(), dstIt.getPointer(), dstEnd.getPointer() );
			for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
			{
//				printf( "copying\n" );
				*dstIt = *srcIt;

			}
		}

		template<class T> void ImagePlane<T>::getChannelParam ( const int channelmode, uint &offSet, uint &incStep, uint &numBytes ) const
		{
			switch ( channelmode )
			{
				case CHANNEL_COMBINED:
					offSet = 0;
					incStep = this->getNumChannels();
					numBytes = sizeof ( T ) * this->getNumChannels();
					break;
				case CHANNEL_SPLIT:
					offSet = 0;
					incStep = 1;
					numBytes = sizeof ( T );
					break;
				default:
					assert ( channelmode >= 0 );
					offSet = channelmode;
					incStep = this->getNumChannels();
					numBytes = sizeof ( T );
					break;
			}
		}
		
		template<class T> bool ImagePlane<T>::load ( const char *filename )
		{
			ImgLoader<T> loader;
			int width, height, col;//for the image info
			loader.getInfo ( string ( filename ), width, height, col );

			if ( 0 >= width || 0 >= height || 0 >= col ) return false;
			init ( width, height, col );

			ImgArray *img = this->getImgArray();
			return loader.loadToBuffer ( string ( filename ), img );
		}

		template<class T>     void ImagePlane<T>::save ( const char *name ) const
		{
			ImgSaver<T> saver;
			ImgArray *img = this->getImgArray();
			saver.saveImgArray ( name, img, this->getROI() );
		}

		template<class T> void ImagePlane<T>::setWStep()
		{
			this->wStep = getImgArrayWidthStep ( this->getImgArray() ) / sizeof ( T );
		}

		template<int id>
		inline ImagePlane<typename ImageId<id>::type> *castImageBase ( ImageBase *i )
		{
			return static_cast<ImagePlane<typename ImageId<id>::type> *> ( i );
		}


	}

}

#endif
