

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORERINGBUFFER2D_H
#define ADORE_CORERINGBUFFER2D_H

#include<vector>
#include<list>
#include <adrcore/imgprocessing/imgconstants.h>

namespace adore
{

	namespace core
	{

		/**
			 
		*/
		template<class TypeT>
		class Ringbuffer2D
		{

			public:
				typedef std::vector<TypeT> LineT;
				typedef typename LineT::iterator lineIterator;
				typedef std::vector<LineT *> BufferT;
				typedef std::list<lineIterator> BeginningsT;
				typedef typename BeginningsT::iterator iterator;

			public:
				Ringbuffer2D( int width = 1, int height = 1 ):
						ring( height ), start( 0 ), mode( BORDER_ZERO )
				{
					initBase( width, height );
				}

				~Ringbuffer2D()
				{
					releaseRing();
				}

				/**
				 * if width and height differ from the current values, the buffer
				 * gets reallocated
				 */
				void init( int width, int height )
				{
					if ( height == getRows() && width == getCols() ) return;
					initBase( width, height );
				}

				/**
				 * copys values into the buffer beginning
				 */
				template<class IteratorT>
				void insert( IteratorT it, int offset, int length )
				{
					assert( getRows() > 0 );
					int rBorder = lineAt( 0 )->size() - length - offset;

//					DBGOUT( DBGNORMAL, "inserting at offset %d %d values (max: %d)\n", offset, length, lineAt( 0 )->size() );
					//take end and put it before begin
					typename LineT::iterator lit = lineAt( 0 )->begin();
					feed();

					for ( int i = 0; i < offset; ++i, ++lit )
					{
						switch ( mode )
						{
							case BORDER_COPY:
								*lit = *it;
								break;
							case BORDER_ZERO:
							default:
								*lit = 0;
								break;
						}
					}
					for ( int i = 0; i < length -1; ++i, ++it, ++lit )
					{
						*lit = *it;
					}
					*lit++ = *it;
					for ( int i = 0; i < rBorder; ++i, ++lit )
					{
						switch ( mode )
						{
							case BORDER_COPY:
								*lit = *it;
								break;
							case BORDER_ZERO:
							default:
								*lit = 0;
								break;
						}
					}
				}

				void insertZeros()
				{
					assert( getRows() > 0 );
					//take end and put it before begin
					typename LineT::iterator lit = lineAt( 0 )->begin();
					typename LineT::iterator litE = lineAt( 0 )->end();
					feed();

					for ( ; lit != litE; ++lit )
					{
						*lit = 0;
					}
				}

				TypeT &operator()( int row, int col )
				{
					return ( *lineAt( row ) )[col];
				}

				TypeT &operator()( int col )
				{
					return ( *lineAt( 0 ) )[col];
				}

				LineT *lineAt( int row )
				{
					return ring[getIdxOfRow( row )];
				}

				/**
				 * Iterate over a line
				 */
				lineIterator beginOfLine( int row = 0 )
				{
					return lineAt( row )->begin();
				}

				lineIterator endOfLine( int row = 0 )
				{
					return lineAt( row )->end();
				}

				void beginOfLines( std::vector<lineIterator> &vec )
				{
					for ( uint i = 0; i < ring.size(); ++i )
					{
						vec.push_back( beginOfLine( i ) );
					}
				}

				/**
				 * Iterate throug the iterators that point to the beginnings
				 * of each line.
				 */
				iterator begin()
				{
					return beginnings.begin();
				}

				int getRows()
				{
					return ring.size();
				}

				int getCols()
				{
					assert( ring[0] );
					return ring[0]->size();
				}

				void print()
				{
					printf( "\n" );
					for ( int i = 0; i < getRows(); ++i )
					{
						for ( int j = 0; j < getCols(); ++j )
						{
//							printf( "%d  ", int( ( *ring[i] )[j] ) );
							printf( "%d  ", int( ( *this )( i, j ) ) );
						}
						printf( "\n" );
					}
					printf( "\n" );
				}

				void feed()
				{
					start = ( start + 1 ) % ring.size();
					//update beginnings
//					void splice( iterator position, list<T, Alloc>& x, iterator i );
					beginnings.splice( beginnings.end(), beginnings, beginnings.begin() );
				}

				void setMode( const BORDER_MODE theValue )
				{
					mode = theValue;
				}

				BORDER_MODE getMode() const
				{
					return mode;
				}

			private:

				void releaseRing()
				{
					for ( typename BufferT::iterator it = ring.begin(); it != ring.end(); ++it )
					{
						delete ( *it );
					}
				}

				int getIdxOfRow( int row )
				{
					return ( start + row ) % ring.size();
				}

				void initBase( int width, int height )
				{
					releaseRing();
					ring.resize( height );
					for ( typename BufferT::iterator it = ring.begin(); it != ring.end(); ++it )
					{
						( *it ) = new LineT( width, 0 );
						beginnings.push_back( ( *it )->begin() );
					}
				}

				BufferT ring;
				BeginningsT beginnings;
				int start;
				BORDER_MODE mode;
		};

	}

}

#endif
