

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORESCALESPACETRAVERSER_H
#define ADORE_CORESCALESPACETRAVERSER_H

#include <adrcore/imgprocessing/imageplane.h>
#include <adrcore/imgprocessing/scalespacer.h>
#include <adrcore/types/ringbuffer.h>

namespace adore
{

	namespace core
	{

		/**
		 * This class traverses a scalespace:
		 * PRESMOOTH
		 * FIRSTSMOOTHS [+] [opt]
		 * SUBSAMPLE
		 * SMOOTHS [+] [opt]
		 *
		 * the images are organized in a ring buffer index 0 points to the newest generated image
		 * the size of the buffer can be set.
			 
		*/
		template<class ConvolverT, class PreConvolverT = ConvolverT>
		class ScaleSpaceTraverser
		{
			public:

				typedef typename ConvolverT::img_in_type img_in_type;
				typedef typename ConvolverT::img_out_type img_out_type;
				typedef ImagePlane<img_in_type> ImageT;
				typedef Channel<img_in_type> ChannelT;
				typedef ScaleSpacer<ConvolverT, PreConvolverT> ScaleSpacerT;

				enum TRAVERSER_STATE
				{
					STATE_PRESMOOTH,
					STATE_FIRSTSMOOTHS,
					STATE_SUBSAMPLE,
					STATE_SUBSAMPLE_AND_SMOOTH,
					STATE_SMOOTH,
					STATE_NOTHING
				};


			public:

				/**
				 * creates a traverser.
				 * @param img the image to smooth and scale repeatedly
				 * @param _state the state to start with
				 * @param fistSmooths the number of smooths before applying the first subsampling
				 * @param smooths the number of smooths between two subsampling stages
				 * @param numImages the number of images buffered internally. minimum is 2.
					 * @param _ownImages if true, only numImages images are maintained and delted at destruction.
					 * if false, a new image gets allocated at each step and these are not deleted at destruction
				 */
				ScaleSpaceTraverser( Channel<img_in_type> *img = NULL, TRAVERSER_STATE _state = STATE_PRESMOOTH,
				                     int firstSmooths = 0, int smooths = 1, int numImages = 2, bool _ownImages = true ):
						rb( numImages, NULL ),
						numFirstSmooths( firstSmooths ),
						numSmooths( smooths ),
						numCurSmooths( numSmooths ),
						state( _state ),
						subsampleState( STATE_SUBSAMPLE_AND_SMOOTH ),
						spacing( 1 ),
						ownImages( _ownImages )
				{
					assert( numImages >= 2 );

					if ( state == STATE_FIRSTSMOOTHS && numFirstSmooths < 1 ) state = STATE_SUBSAMPLE;

					if ( !img ) return;

					initBegin( img );
				}

				ScaleSpaceTraverser( const ScaleSpaceTraverser<ConvolverT> &sst ):
						rb( sst.getNumBufferedImages(), NULL ),
						numFirstSmooths( sst.getNumFirstSmooths() ),
						numSmooths( sst.getNumSmooths() ),
						numCurSmooths( numSmooths ),
						state( STATE_FIRSTSMOOTHS ),
						spacing( sst.getSpacing() )
				{
					if ( state == STATE_FIRSTSMOOTHS && numFirstSmooths < 1 ) state = STATE_SUBSAMPLE;

					Channel<img_in_type> *img = sst( 0 );

					initBegin( img );
				}

				~ScaleSpaceTraverser()
				{
					if ( ownImages ) for ( typename RingBuffer<ChannelT *>::iterator it = rb.begin(); it != rb.end(); ++it )
						{
							if ( *it ) delete *it;
						}
				}

				//todo: adjust image sizes!
				ImageT *next()
				{
					if ( state == STATE_NOTHING )
					{
						DBGOUT( DBGNORMAL, "ScaleSpaceTraverser: STATE_NOTHING\n" );
						if ( numFirstSmooths > 0 )
						{
							state = STATE_FIRSTSMOOTHS;
						}
						else
						{
							state = subsampleState;
						}
						return dynamic_cast<ImageT*>( *rb.begin() );
					}

					ChannelT *&cur = *rb.begin();
					rb.feed();
					ChannelT *&next = *rb.begin();

					assert( cur );

					switch ( state )
					{
						case STATE_PRESMOOTH:
							DBGOUT( DBGNORMAL, "ScaleSpaceTraverser: STATE_PRESMOOTH\n" );
							ensureProperties( &next, cur->getWidth(), cur->getHeight() );
							sp.process( cur, next, ScaleSpacerT::SCALE_PRESMOOTH );
							if ( numFirstSmooths > 0 )
							{
								state = STATE_FIRSTSMOOTHS;
							}
							else
							{
								state = subsampleState;
							}
							break;
						case STATE_FIRSTSMOOTHS:
							DBGOUT( DBGNORMAL, "ScaleSpaceTraverser: STATE_FIRSTSMOOTHS\n" );
							ensureProperties( &next, cur->getWidth(), cur->getHeight() );
							sp.process( cur, next, ScaleSpacerT::SCALE_SMOOTH );
							--numFirstSmooths;
							if ( !numFirstSmooths ) state = subsampleState;
							break;
						case STATE_SUBSAMPLE:
							DBGOUT( DBGNORMAL, "ScaleSpaceTraverser: STATE_SUBSAMPLE\n" );
							ensureProperties( &next, getSubSampleDim( cur->getWidth() ), getSubSampleDim( cur->getHeight() ) );
							sp.process( cur, next, ScaleSpacerT::SCALE_SUBSAMPLE );
							spacing <<= 1;
							if ( numSmooths > 0 )
							{
								state = STATE_SMOOTH;
								numCurSmooths = numSmooths;
							}
							break;
						case STATE_SUBSAMPLE_AND_SMOOTH:
							DBGOUT( DBGNORMAL, "ScaleSpaceTraverser: STATE_SUBSAMPLE_AND_SMOOTH\n" );
							assert( numSmooths > 0 ); //must be consistent
							ssbuffer.init( getSubSampleDim( cur->getWidth() ), getSubSampleDim( cur->getHeight() ), 1 );
							sp.process( cur, &ssbuffer, ScaleSpacerT::SCALE_SUBSAMPLE );
							spacing <<= 1;
							ensureProperties( &next, ssbuffer.getWidth(), ssbuffer.getHeight() );
							sp.process( &ssbuffer, next, ScaleSpacerT::SCALE_SMOOTH );
							if ( numSmooths > 1 )
							{
								state = STATE_SMOOTH;
								numCurSmooths = numSmooths - 1;
							}
							else state = subsampleState;
							break;
						case STATE_SMOOTH:
							DBGOUT( DBGNORMAL, "ScaleSpaceTraverser: STATE_SMOOTH\n" );
							ensureProperties( &next, cur->getWidth(), cur->getHeight() );
							sp.process( cur, next, ScaleSpacerT::SCALE_SMOOTH );
							--numCurSmooths;
							if ( !numCurSmooths ) state = subsampleState;
							break;
						default:
							assert( false );
					}

					return dynamic_cast<ImageT*>( next );
				}

				void setConvolver( ConvolverT *conv )
				{
					sp.setConvolver( conv );
				}

				ConvolverT *getConvolver()
				{
					return sp.getConvolver();
				}

				void setPreConvolver( PreConvolverT *conv )
				{
					sp.setPreConvolver( conv );
				}

				PreConvolverT *getPreConvolver()
				{
					return sp.getPreConvolver();
				}

				int getNumBufferedImages() const
				{
					return rb.size();
				}

				int getNumFirstSmooths() const
				{
					return numFirstSmooths;
				}

				int getNumSmooths() const
				{
					return numSmooths;
				}

				TRAVERSER_STATE getState()
				{
					return state;
				}

				ImageT *operator()( int i ) const
				{
					return dynamic_cast<ImageT*>( rb( i ) );
				}

				ImageT *current()
				{
					return dynamic_cast<ImageT*>( *rb.begin() );
				}

				void setSubSampleMode( TRAVERSER_STATE mode )
				{
					assert( mode == STATE_SUBSAMPLE || mode == STATE_SUBSAMPLE_AND_SMOOTH );
					subsampleState = mode;
				}

				/**
				 * @return the spacing between the pixels.
				 * this first image has spacing 1. After one subsampling it's 2. etc.
				 */
				int getSpacing() const
				{
					return spacing;
				}

				/**
				 * returns the step in scalespace when smoothing with a kernel of level
				 * kernelLevel and a delta of deltaT.
				 * Both values depend on the kernel you passed by the template parameters.
				 */
				float getScaleStep( int kernelLevel, float deltaT )
				{
					return adore::core::getScaleStep( kernelLevel, getSpacing(), deltaT );
				}

			private:

				void initBegin( ChannelT *img )
				{
					if ( ownImages )
					{
						ChannelT *&cur = *rb.begin();
						ensureProperties( &cur, img->getWidth(), img->getHeight() );
						//printf("(%d, %d) | (%d, %d)-(%d, %d)\n", img->getWidth(), img->getHeight(), img->getROI().getX(), img->getROI().getY(), img->getROI().getWidth(), img->getROI().getHeight());
						cur->setROI( img->getROI() );
						//printf("(%d, %d) | (%d, %d)-(%d, %d)\n", cur->getWidth(), cur->getHeight(), cur->getROI().getX(), cur->getROI().getY(), cur->getROI().getWidth(), cur->getROI().getHeight());
						*cur = *img;
					}
					else
					{
						*rb.begin() = img;
					}
				}

				void ensureProperties( ChannelT **i, int w, int h )
				{
					if ( ( !*i ) || ( !ownImages ) )
					{
						*i = new ImagePlane<img_in_type>( w, h, 1 );
						//printf("ensure properties: allocating %p\n", *i);
					}
					else
					{
						ImageT* img = dynamic_cast< ImageT* >( ( *i ) );
						assert( img );
						img->init( w, h, 1 );
						//printf("ensure properties: init %p\n", img);
					}
				}

				ScaleSpacer<ConvolverT, PreConvolverT> sp;
				RingBuffer<ChannelT *> rb;
				int numFirstSmooths;
				int numSmooths;
				int numCurSmooths;
				TRAVERSER_STATE state;
				TRAVERSER_STATE subsampleState;
				ImageT ssbuffer;//a buffer for subsampling

				//statistical variables
				int spacing;

				bool ownImages;
				ImageT *first;//a copy of the channel
		};

	}

}

#endif
