

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGTEMPLATEMATCHING_H
#define ADORE_IMGPROCESSINGTEMPLATEMATCHING_H

#include <adrcore/types/aarray.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/calctype.h>
#include <adrcore/util/avgvar.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/bilinearinterpolator.h>

namespace adore
{

	namespace core
	{

		/**
		  * Functors that can be plugged into the correlation functions below.
		 * These are called for every pixel, this way updating their inner stated.
		 * A final call of "result()" triggers the calculation of the result and returns it.
		 * The funtors have no reset(), so the have currently to be instantiated for every Pixel.
		 */


		/**
		 * calculates the sum of squared differences
		 */
		template <class typeT>
		class SSDFunctor
		{
				typedef typename core::calcT<typeT>::type T;
				typedef typename core::signedT<T>::type sT;
			public:
				typedef T value_type;
			public:
				SSDFunctor() :_result ( 0 )
				{
					;
				}
				void operator() ( const sT templVal, const sT imgVal )
				{
					sT val = templVal - imgVal;
					_result += val * val;
				}
				T result()
				{
					return _result;
				}
			private:
				T _result;
		};

		/**
		 * calculates the normed sum of squared differences, such that the result
		 * lies between 0 and 1.
		 */

		template <class typeT>
		class SSDNormedFunctor
		{
				typedef typename core::calcT<typeT>::type T;
				typedef typename core::signedT<T>::type sT;
			public:
				typedef float value_type;
			public:
				SSDNormedFunctor() :_result ( 0 ), _imgSqr ( 0 ), _templSqr ( 0 )
				{
					;
				}

				void operator() ( const sT templVal, const sT imgVal )
				{
					_imgSqr += imgVal * imgVal;
					_templSqr += templVal * templVal;
					sT val = templVal - imgVal;
					_result += val * val;
				}
				float result()
				{
					return float ( _result ) / sqrtf ( _imgSqr * _templSqr );
				}
			private:
				T _result;
				T _imgSqr;
				T _templSqr;
		};

		template <class typeT>
		class CCORRNormedFunctor
		{
				typedef typename core::calcT<typeT>::type T;
				typedef typename core::signedT<T>::type sT;

			public:
				typedef float value_type;
			public:
				CCORRNormedFunctor() :_corrsum ( 0 ), _normsum1 ( 0 ), _normsum2 ( 0 )
				{
					;
				}

				void operator() ( const sT templVal, const sT imgVal )
				{
					_corrsum  += templVal * imgVal;
					_normsum1 += templVal * templVal;
					_normsum2 += imgVal   * imgVal;
				}
				float result()
				{
					double n1 = double ( _normsum1 ) / double ( _corrsum );
					double n2 = double ( _normsum2 ) / double ( _corrsum );

					//cout<<"corrsum:"<<_corrsum<<" normsum1:"<<_normsum1<<" normsum2:"<<_normsum2<<endl;
					//cout<<"test:"<<(_corrsum) / sqrt(_normsum1 * _normsum2)<<endl;
					return 1.0F / sqrt ( n1 * n2 );

				}
			private:
				sT _corrsum;
				sT _normsum1;
				sT _normsum2;
		};
		/**
		 * calculates a correlation between an image region and a template
		 * The image coordinates may be floats for they are bilinear interplated
		 */
		template<class typeT, class IteratorT, class functorT>
		typename functorT::value_type getCorrelation ( const Channel<typeT> &chan, float srcX, float srcY, IteratorT it, int xHood, int yHood )
		{
			typedef typename core::calcT<typeT>::type T;
			typedef typename core::signedT<T>::type sT;

			BilinearInterpolator<typeT> sampler;
			functorT functor;
			for ( int y = -yHood; y <= yHood; ++y )
			{
				for ( int x = -xHood; x <= xHood; ++x, ++it )
				{
					float newX = srcX + x;
					float newY = srcY + y;
					if ( newX >= 0 && newX <= chan.getWidth() - 1 && newY >= 0 && newY <= chan.getHeight() - 1 )
					{
						sampler.init ( newX, newY );
						sT sit = *it;
						sT sval = sampler.valueOf ( chan );
						functor ( sit, sval );
					}
				}
			}
			return functor.result();
		}

		/**
		 * calculates a correlation between an image region and a template
		 * The image coordinates must be integers. No interpolation is performed
		 */
		template<class typeT, class IteratorT, class functorT>
		typename functorT::value_type getFastCorrelation ( const Channel<typeT> &chan, uint srcX, uint srcY, IteratorT it, int xHood, int yHood )
		{
			typedef typename core::calcT<typeT>::type T;
			typedef typename core::signedT<T>::type sT;

			functorT functor;
			//cout<<"srcX:"<<srcX<<" srcY:"<<srcY<<endl;

			for ( int y = -yHood; y <= yHood; ++y )
			{
				for ( int x = -xHood; x <= xHood; ++x, ++it )
				{
					//cout<<"srcX:"<<srcX<<" x:"<<x<<" srcY:"<<srcY<<" y:"<<y<<endl;
					//assert(int(srcX) > x); assert(int(srcY) > y);


					uint newX = srcX + x;
					uint newY = srcY + y;
					if ( newX >= 0 && newX <= chan.getWidth() - 1 && newY >= 0 && newY <= chan.getHeight() - 1 )
					{
						sT sit = *it;
						sT sval = *chan.getPointer ( newX, newY );
						functor ( sit, sval );
					}
				}
			}
			return functor.result();
		}

		template<class typeT, class IteratorT>
		inline float getFasterCorrelation ( const Channel<typeT> &chan, uint srcX, uint srcY, IteratorT it, int xHood, int yHood )
		{
			float corrsum = 0;
			for ( int y = -yHood; y <= yHood; ++y )
			{
				for ( int x = -xHood; x <= xHood; ++x, ++it )
				{
					uint newX = srcX + x;
					uint newY = srcY + y;

					assert( newX >= 0 && newX <= chan.getWidth() - 1 && newY >= 0 && newY <= chan.getHeight() - 1 );

					float sit = *it;
					float sval = *chan.getPointer ( newX, newY );
					corrsum  += sit * sval;
				}
			}
			return corrsum;
		}

		template<class Iterator>
		void subtractMean ( Iterator begin, int size )
		{
			double avg = getAvg ( begin, size );
			for ( int i = size; i > 0; --i, ++begin )
			{
//				*begin -= typename Iterator::value_type( avg );
				*begin -= avg;
			}
		}

		template<class srcT, class dstT>
		void computeSumAndEnergyLookup ( const srcT &srcImg, dstT &dstImgS, dstT &dstImgE )
		{
            typedef typename dstT::scalar_type T;

			T val = srcImg ( 0, 0 );
			dstImgS ( 0, 0 ) = val;
			dstImgE ( 0, 0 ) = val * val;

			for ( int x = 1; x < ( int )srcImg.getWidth(); ++x )
			{
				val = srcImg ( x, 0 );
				dstImgS ( x, 0 ) = dstImgS ( x - 1, 0 ) + val;
				dstImgE ( x, 0 ) = dstImgE ( x - 1, 0 ) + val * val;
			}

			for ( int y = 1; y < ( int )srcImg.getHeight(); ++y )
			{
				val = srcImg ( 0, y );
				dstImgS ( 0, y ) = dstImgS ( 0, y - 1 ) + val;
				dstImgE ( 0, y ) = dstImgE ( 0, y - 1 ) + val * val;
			}

			for ( int y = 1; y < ( int )srcImg.getHeight(); ++y )
			{
				for ( int x = 1; x < ( int )srcImg.getWidth(); ++x )
				{
					val = srcImg ( x, y );
					dstImgS ( x, y ) = val + dstImgS ( x - 1, y ) - dstImgS ( x - 1, y - 1 ) + dstImgS ( x, y - 1 );
					dstImgE ( x, y ) = val * val + dstImgE ( x - 1, y ) - dstImgE ( x - 1, y - 1 ) + dstImgE ( x, y - 1 );
				}
			}
		}

	}
}


#endif

