

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#ifndef ERRORFUNCTORS2VIEWS_H
#define ERRORFUNCTORS2VIEWS_H

#include <adrcore/types/aarray.h>
#include <adrcore/nview/correspondencecorrectorp.h>
#include <adrcore/nview/correspondencecorrector.h>
#include <adrcore/nview/fmatrixcalculatorpvp.h>

#include <adrcore/util/c++0x.h>

#include <vector>
#include <cmath>
#include <cassert>

using namespace std;

namespace adore
{
	namespace core
	{

		class ProjectiveErrorsWithCorrector
		{
				typedef tuple<PointCorr*, FeaturePoint *, FeaturePoint *> CorrT;
				typedef vector< CorrT > CorrsT;
				CorrespondenceCorrectorP corrector;

			public:
				double operator() ( const AMatrix& H, double thresh, vector<int>& candInliers, CorrsT& corrs,  int numCandidates )
				{
					double error = 0;
					candInliers.clear();
					corrector.prepare ( H );

					for ( int idx = 0; idx < numCandidates; ++idx )
					{
                        const AVector & p1 = *corrs[idx].get<1>();
                        const AVector & p2 = *corrs[idx].get<2>();
                        AVector correct1 = p1;
                        AVector correct2 = p2;
						bool corrected = corrector.correct ( correct1, correct2, H );
						if ( corrected )
						{
							//calculate reprojection error between mates and correct
							double dist = correct1.sqrDist ( p1 ) + correct2.sqrDist ( p2 );
							if ( dist < thresh )
							{
								error += dist;
								candInliers.push_back ( idx );
							}
							else error += thresh;
						}
						else
						{
							error += thresh;
						}
					}
					return error;
				}
		};

		template <class Solution>
		class FundamentalErrorsWithCorrector
		{
				typedef tuple<PointCorr*, FeaturePoint *, FeaturePoint *> CorrT;
				typedef vector< CorrT > CorrsT;
				CorrespondenceCorrector corrector;

			public:

				double operator() ( Solution& sol, const AVector& epiPoint, double thresh, vector<int>& candInliers, CorrsT& corrs,  int numCandidates )
				{
					AMatrix& fmat = sol.getFMat();
					double error = 0;
					candInliers.clear();

					for ( int idx = 0; idx < numCandidates; ++idx )
					{
                        const AVector & p1 = *corrs[idx].get<1>();
                        const AVector & p2 = *corrs[idx].get<2>();
                        AVector correct1 = p1;
                        AVector correct2 = p2;
						correct1 = sol.getCorrectedPoint ( correct1 );
						correct2 = sol.getCorrectedPoint ( correct2 );

						bool o = FMatrixCalculator::isOriented ( fmat, epiPoint, correct1, correct2 );
						if ( !o )
						{
							error += thresh;
							continue;
						}

						bool corrected = corrector ( correct1, correct2, fmat );
						if ( corrected )
						{
							correct1 = sol.getUnCorrectedPoint ( correct1 );
							correct2 = sol.getUnCorrectedPoint ( correct2 );
							double dist = correct1.sqrDist ( p1 ) + correct2.sqrDist ( p2 );
							if ( dist < thresh )
							{
								error += dist;
								candInliers.push_back ( idx );
							}
							else error += thresh;
						}
						else
						{
							error += thresh;
						}
					}
					return error;
				}
		};

		class SampsonH
		{
				typedef tuple<PointCorr*, FeaturePoint *, FeaturePoint *> CorrT;
				typedef vector< CorrT > CorrsT;
				CorrespondenceCorrectorP corrector;

			public:
				double operator() ( const AMatrix& H, double thresh, vector<int>& candInliers, CorrsT& corrs,  int numCandidates )
				{
					double error = 0;
					candInliers.clear();
					AVector Hline ( 9 );
					Hline = H ( 0, 0 ), H ( 0, 1 ), H ( 0, 2 ),
					        H ( 1, 0 ), H ( 1, 1 ), H ( 1, 2 ),
					        H ( 2, 0 ), H ( 2, 1 ), H ( 2, 2 );
					AMatrix A ( 2, 9 ), J ( 2, 4 ), JT ( 4, 2 ), JJT ( 2, 2 );
					AVector e ( 2 ), temp ( 2 );

					for ( int idx = 0; idx < numCandidates; ++idx )
					{
                        const AVector & p1 = *corrs[idx].get<1>();
                        const AVector & p2 = *corrs[idx].get<2>();

						A = 0, 0, 0, -p1 ( 0 ), -p1 ( 1 ), -1, p2 ( 1 ) *p1 ( 0 ), p2 ( 1 ) *p1 ( 1 ), p2 ( 1 ),
						    p1 ( 0 ), p1 ( 1 ), 1, 0, 0, 0, -p2 ( 0 ) *p1 ( 0 ), -p2 ( 0 ) *p1 ( 1 ), -p2 ( 0 );
						matMul2 ( A, Hline, e );
						J =	-H ( 1, 0 ) + p2 ( 1 ) *H ( 2, 0 ),
						    -H ( 1, 1 ) + p2 ( 1 ) *H ( 2, 1 ),
						    0,
						    p1 ( 0 ) *H ( 2, 0 ) + p1 ( 1 ) *H ( 2, 1 ) + H ( 2, 2 ),

						    H ( 0, 0 ) - p2 ( 0 ) *H ( 2, 0 ),
						    H ( 0, 1 ) - p2 ( 0 ) *H ( 2, 1 ),
						    -p1 ( 0 ) *H ( 2, 0 ) - p1 ( 1 ) *H ( 2, 1 ) - H ( 2, 2 ),
						    0;
                        trans ( J, JT );
						matMul2 ( J, JT, JJT );
						assert ( true == invert ( JJT ) );
						matMul2_AT ( JJT, e, temp );
						double dist = dot2 ( temp, e );
						if ( dist < thresh )
						{
							error += dist;
							candInliers.push_back ( idx );
						}
						else error += thresh;
					}
					return error;
				}
		};

		template <class Solution>
		class SampsonF
		{
			private:
				typedef tuple<PointCorr*, FeaturePoint *, FeaturePoint *> CorrT;
				typedef vector< CorrT > CorrsT;

				AVector temp;
				AVector correct1;
				AVector correct2;

				SampsonF& operator= ( const SampsonF& ); // no implementation.
			public:
				SampsonF() : temp ( 3 ), correct1 ( 3 ), correct2 ( 3 ) {}
				SampsonF ( const SampsonF& ) : temp ( 3 ), correct1 ( 3 ), correct2 ( 3 ) {}


				double operator() ( Solution& sol, const AVector& epiPoint, double thresh, vector<int>& candInliers, CorrsT& corrs,  int numCandidates )
				{
					double error = 0;
					candInliers.clear();

					const AMatrix& fmat = sol.getFMat();

					for ( int idx = 0; idx < numCandidates; ++idx )
					{
						const AVector & p1 = *corrs[idx].get<1>();
						const AVector & p2 = *corrs[idx].get<2>();
                        correct1 = p1;
                        correct2 = p2;

						correct1 = sol.getCorrectedPoint ( correct1 );
						correct2 = sol.getCorrectedPoint ( correct2 );
						bool o = FMatrixCalculator::isOriented ( fmat, epiPoint, correct1, correct2 );
						if ( !o )
						{
							error += thresh;
							continue;
						}

						/** p. 287, 11.9 */
						//set nice names
						AVector & Fx1 = temp;
						AVector & FTx2 = temp;

						matMul2 ( fmat, p1, Fx1 );
						double x2Fx1 = dot2 ( p2, Fx1 );
						double dividend = x2Fx1 * x2Fx1;
						double divisor = Fx1 ( 0 ) * Fx1 ( 0 ) + Fx1 ( 1 ) * Fx1 ( 1 );
						matMul2_AT ( fmat, p2, FTx2 );
						divisor +=  FTx2 ( 0 ) * FTx2 ( 0 ) + FTx2 ( 1 ) * FTx2 ( 1 );
						double dist = dividend / divisor;
						dist *= 4.8; // WTF?
						if ( dist < thresh )
						{
							error += dist;
							candInliers.push_back ( idx );
						}
						else error += thresh;
					}
					return error;
				}
		};

	}
}

#endif // ERRORFUNCTORS2VIEWS_H
