

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWFILTERRANSAC2VIEW_H
#define ADORE_NVIEWFILTERRANSAC2VIEW_H

#include <vector>
#include <cassert>

#include "filterransac2viewhelper.h"
#include "filterransac3viewhelper.h"
#include "filterransacprojectivityhelper.h"
#include "errorfunctors2views.h"
#include "errorfunctors3views.h"

using namespace std;

namespace adore
{

	namespace core
	{

		/**
		 * Filters Correspondences via RANSAC,
		 * jointly computes (E)pipolar (G)eometry and (R)adial (D)istortion
		 * @see FilterRANSAC
		 *  
		 */
		template<class Helper>
		class FilterRANSAC : public NVIRANSACClient
		{
			private:
				typedef typename Helper::Solution Solution;
				typedef typename Helper::Calculator Calculator;
				Helper helper;
				double inlierThresh;
				double incFactor;
				int innerSteps;

				Calculator solCalc;

				bool hasBest;
				RANSAC_OPT_METHOD optMethod;

				//for optimize2
				vector<PointCorr *> ocorrs;//for optimize2
				NView::ViewIterator vito;

			private:

				class Inner : public RANSACClient
				{
					private:
						FilterRANSAC *parent;
					public:
						Helper helper;
						float thresh;
						bool hasBest;

					public:
						Inner ( FilterRANSAC *p )
								: parent ( p ),
								helper ( this, getInnerSampleSize<Solution>() ),
								thresh ( 0 ),
								hasBest ( false ) { }

						int getSampleSize() {return helper.numSamples;}

						bool applySample ( const vector<int> &sample )
						{
							assert ( ( int ) sample.size() == helper.numSamples );

							PointCorr *sam[helper.numSamples];

							for ( int i = 0; i < helper.numSamples; ++i )
							{
								sam[i] = helper.getPointCorr ( sample[i] );
								COUT ( DBGALL, "Inner " << toDbgString ( sam[i] ) << endl );
							}

							int numResults = parent->solCalc.calc ( &sam[0], helper.numSamples, parent->vito );

							if ( numResults < 1 ) return false;

							Solution sol = parent->solCalc.getSolution ( 0 );

							bool success = helper.updateSolution ( sol, sample, thresh, false );

							return success;
						}

						bool optimize() {hasBest = true; return false;}

						void ransacDone() {}

						bool success() const { return hasBest; }

				};

			public:
				/**
				* This Constructor takes a threshold for the reprojection error.
				* @param t the treshold.
				*/
				FilterRANSAC ( const double t, const double factor, const int steps ); //value from HZ, p. 291

				/**
				 * simplified version
				 */
				/// t -- value from HZ, p. 291
				FilterRANSAC ( const double t, RANSAC_OPT_METHOD method = INNER_RANSAC );

				/**
					information gained through init() is not cloned
				*/
				FilterRANSAC ( const FilterRANSAC &fr );

				virtual NVIRANSACClient *clone() {return new FilterRANSAC ( *this );}

				void init ( NView::CorrIterator it, int size, NView::ViewIterator vit )
				{
					DBGOUT ( DBGNORMAL, "init FilterRANSAC %p\n", this );
					vito = vit;
					helper.init ( it, size, vit );
					ocorrs.clear();
					ocorrs.reserve ( getNumCandidates() );
				}

				////ransacclient interface functions

				int getSampleSize() {return helper.numSamples;}

				bool applySample ( const vector<int> &sample );

				void ransacDone() {DBGOUT ( DBGNORMAL, "leaving with %d/%d inliers\n", getNumIn(), getNumCandidates() );}

				///an optional lo-step
				bool optimize();

				////ransacclient interface functions - end

				////nviransacclient interface functions
				NViewInfo *createNViewInfo();

				PointCorr *getPointCorr ( int i ) {return helper.getPointCorr ( i );}

				iterator begin() {return helper.begin();}

				iterator end() {return helper.end();}

				////nviransacclient interface functions - end

				int getModelType() const  { return helper.getModelType(); }
				double getThreshold() const { return inlierThresh; }
				void setThreshold( double t ) { inlierThresh = t; }

				void setOptMethod ( const RANSAC_OPT_METHOD theValue )
				{
					optMethod = theValue;
				}

				RANSAC_OPT_METHOD getOptMethod() const
				{
					return optMethod;
				}

				bool success() const
				{
					return hasBest;
				}

				/**
				 * this method allows the access to the solution calculator in case it needs some extra parameters to be set
				 */
				Calculator &getSolutionCalculator()
				{
					return solCalc;
				}

			private:

				bool optimize1();

				bool optimize2();


		};

		typedef FilterRANSAC2ViewHelper<FMatSolution, FundamentalErrorsWithCorrector> Helper2F;
		typedef FilterRANSAC<Helper2F> FilterFMat;

		typedef FilterRANSAC<FilterRANSAC2ViewHelper<FMatSolution, SampsonF> > FilterFMatSampson;

		typedef FilterRANSAC2ViewHelper<RDFMatSolution, FundamentalErrorsWithCorrector> Helper2RDF;
		typedef FilterRANSAC<Helper2RDF>  FilterRDFMat;

		typedef FilterRANSAC3ViewHelper<TTensorSolution, TTWithCorrector> HelperTT;
		typedef FilterRANSAC3ViewHelper<TTensorSolution, TTSampson> HelperTTSampson;
		typedef FilterRANSAC<HelperTT> FilterTTensor;
		typedef FilterRANSAC<HelperTTSampson> FilterTTensorSampson;

		typedef FilterRANSACProjectivityHelper<HSolution, SampsonH> Helper2HSampson;
		typedef FilterRANSAC<Helper2HSampson> FilterProjectivitySampson;

		typedef FilterRANSACProjectivityHelper<HSolution, ProjectiveErrorsWithCorrector> Helper2H;
		typedef FilterRANSAC<Helper2H> FilterProjectivity;

		typedef FilterRANSACProjectivityHelper<HSolutionC, ProjectiveErrorsWithCorrector> Helper2HC;
		typedef FilterRANSAC<Helper2HC> FilterProjectivityFMat;

		typedef FilterRANSACProjectivityHelper<HSolutionC, SampsonH> Helper2HCSampson;
		typedef FilterRANSAC<Helper2HCSampson> FilterProjectivityFMatSampson;

	}
}

#endif
