

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_COREFILTERRANSAC2VIEWHELPER_H
#define ADORE_COREFILTERRANSAC2VIEWHELPER_H

#include <adrcore/nview/correspondencecorrector.h>
#include <adrcore/types/pointcorr.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/debug.h>
#include <adrcore/util/c++0x.h>
#include <adrcore/random/nviransacclient.h>
#include <adrcore/util/divisiondistortion.h>
#include <adrcore/nview/fmatrixcalculatorpvp.h>
#include <adrcore/nview/rdfmatrixcalculator.h>
#include "filterransaccommon.h"

namespace adore
{

	namespace core
	{

		class FMatSolution
		{
			private:
				AMatrix fmat;
			public:
				FMatSolution() :fmat ( 3, 3 ) {;}

                FMatSolution ( const AMatrix & f ) :fmat ( f ) {;}

                inline const AMatrix &getFMat() const {return fmat;}

                inline AMatrix &getFMat() {return fmat;}

				inline AVector getCorrectedPoint ( AVector &vec ) const {return vec;}

				inline AVector getUnCorrectedPoint ( AVector &vec ) const {return vec;}

				NViewInfo *createNViewInfo()
				{
					NViewInfo *info = new NViewInfoFM;
					info->setFMatrix ( fmat );
					return info;
				}

		};

		class RDFMatSolution
		{
			private:
				AMatrix fmat;
				float lambda;
			public:
				RDFMatSolution() :fmat ( 3, 3 ) {;}

                RDFMatSolution ( const AMatrix & fmat, float l ) :fmat ( fmat ), lambda ( l ) {;}

				RDFMatSolution ( tuple<AMatrix, float> s ) :fmat ( s.get<0>() ), lambda ( s.get<1>() ) {;}

                inline const AMatrix &getFMat() const {return fmat;}

                inline AMatrix &getFMat() {return fmat;}

				inline float &getLambda() {return lambda;}

				inline AVector getCorrectedPoint ( AVector &vec ) const
				{
                    AVector tmpPoint = vec;
					undistortRadial<double> ( tmpPoint ( 0 ), tmpPoint ( 1 ), lambda );
					return tmpPoint;
				}

				inline AVector getUnCorrectedPoint ( AVector &vec ) const
				{
                    AVector tmpPoint = vec;
					distortRadial<double> ( tmpPoint ( 0 ), tmpPoint ( 1 ), lambda );
					return vec;
				}

				NViewInfo *createNViewInfo()
				{
					NViewInfo *info = new NViewInfoRDFM;
					info->setFMatrix ( fmat );
					info->setLambda ( lambda );
					return info;
				}
		};

		template<> class SolutionCalculator<FMatSolution>
		{
			private:
				FMatrixCalculatorPVP fcalc;

			public:
//				SolutionCalculator()
//				{
//					fcalc.setBundleAdjust ( false );
//				}

				template<class CorrIterator, class ViewIterator>
				int calcSloppy ( CorrIterator cit, int num, ViewIterator vit )
				{
					return calc ( cit, num, vit );
				}

				template<class CorrIterator, class ViewIterator>
				int calc ( CorrIterator cit, int num, ViewIterator vit )
				{
					return fcalc.calculateFundamentalMatrix ( cit, num, vit );
				}

				FMatSolution getSolutionSloppy ( int i )
				{
					return fcalc.getSolution ( i );
				}

				FMatSolution getSolution ( int i )
				{
					return fcalc.getSolution ( i );
				}
		};



		template<> class SolutionCalculator<RDFMatSolution>
		{
			private:
				RDFMatrixCalculator rdfcalc;
				FMatrixCalculatorPVP fcalc;

			public:

				template<class CorrIterator, class ViewIterator>
				int calcSloppy ( CorrIterator cit, int num, ViewIterator vit )
				{
					return fcalc.calculateFundamentalMatrix ( cit, num, vit );
				}

				template<class CorrIterator, class ViewIterator>
				int calc ( CorrIterator cit, int num, ViewIterator vit )
				{
					return rdfcalc.calculate ( cit, num, vit );
				}

				RDFMatSolution getSolutionSloppy ( int i )
				{
					return RDFMatSolution ( fcalc.getSolution ( i ), 0 );
				}

				RDFMatSolution getSolution ( int i )
				{
					return rdfcalc.getSolution ( i );
				}
		};

		template<> inline int getOuterSampleSize<FMatSolution>() {return 7;}

		template<> inline int getOuterSampleSize<RDFMatSolution>() {return 7;}

		template<> inline int getInnerSampleSize<FMatSolution>() {return 14;}

		template<> inline int getInnerSampleSize<RDFMatSolution>() {return 20;}

		/**
			 
		*/
		template<class SolutionT, template<typename> class ErrorFn>
		class FilterRANSAC2ViewHelper
		{
			public:
				typedef SolutionT Solution;
				typedef SolutionCalculator<Solution> Calculator;

				typedef NVIRANSACClient::InlierCollT InlierCollT;
				typedef NVIRANSACClient::iterator iterator;
				typedef tuple<PointCorr*, FeaturePoint *, FeaturePoint *> CorrT;
				typedef vector< CorrT > CorrsT;
				typedef CorrsT::iterator corr_iterator;

				class CorrTComp
				{
					public:
						bool operator() ( const CorrT &c1, const CorrT &c2 )
						{
							return ( c1.get<0>()->getBadness() < c2.get<0>()->getBadness() );
						}
				};

			private:
				RANSACClient *client;
				AVector epiPoint;
				CorrespondenceCorrector corrector;
				SVDStoraged store3;

			public:
				const int numSamples;
				InlierCollT candInliers;
				InlierCollT inliers;
				CorrsT corrs;
				Solution bestSolution;

				double DBGErr;

			private:
				FilterRANSAC2ViewHelper( const FilterRANSAC2ViewHelper& rhs ) : numSamples ( rhs.numSamples ) { }
			public:
				FilterRANSAC2ViewHelper ( RANSACClient *c, int num ) :client ( c ), epiPoint ( 3 ), store3 ( 3, 3, SVD_VT ), numSamples ( num ) {;}

				template<class CorrIterator, class ViewIterator>
				void init ( CorrIterator it, int size, ViewIterator vit )
				{
					corrs.clear();
					corrs.reserve ( size );
					inliers.clear();
					inliers.reserve ( size );
					candInliers.clear();
					candInliers.reserve ( size );
					ViewBase *v1 = *vit; ++vit;
					ViewBase *v2 = *vit;

					for ( int i = 0; i < size; ++i, ++it )
					{
						PointCorr *corr = *it;
						if ( !corr ) continue;
						FeaturePoint *fp1 = corr->getFeature ( v1 );
						if ( !fp1 ) continue;
						FeaturePoint *fp2 = corr->getFeature ( v2 );
						if ( !fp2 ) continue;
						corrs.push_back ( make_tuple ( corr, fp1, fp2 ) );
					}

					client->setOkay ( false );

					if ( numSamples <= ( int ) corrs.size() ) client->setOkay ( true );
					else
					{
						DBGOUT ( DBGNORMAL, "FR2VHelper: Ranser was not okay (%d > %zd)\n", numSamples, corrs.size() );
					}

					client->setNumCandidates ( corrs.size() );

					client->setNumIn ( -1 );
					client->setError ( FLT_MAX );

					CorrTComp comp;
					sort ( corrs.begin(), corrs.end(), comp );
				}

				PointCorr *getPointCorr ( int i ) {return corrs[i].get<0>();}

				bool updateSolution ( Solution &sol, const vector<int> &base, float thresh, bool opt );

				iterator begin() {return inliers.begin();}

				iterator end() {return inliers.end();}

				MODEL_TYPE getModelType() const  { return MODEL_TYPE_F; }

			private:
				//test all of the indexed points and compute a score (majority vote)
				int isOriented ( Solution &sol, AVector &epiPoint, const vector<int> &in );

				ErrorFn<SolutionT> computeErrors;

		};

	}

}

#endif
