

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming
/// Author: Sergey Cheremukhin


#ifndef ADORE_CORE_FILTERRANSACCP_H
#define ADORE_CORE_FILTERRANSACCP_H

#include <adrcore/mestimate/errorfunctors3d.h>

#include <adrcore/random/ransacclient.h>

#include <adrcore/nview/homographycalculator.h>

#include <adrcore/types/aarray.h>
#include <adrcore/types/nview.h>

#include <cfloat>
#include <cassert>
#include <vector>
using std::vector;

#include <boost/tuple/tuple.hpp>
using boost::tuples::tuple;
using boost::tuples::make_tuple;

namespace adore {
	namespace core {

        class RANSACCPHelper
        {
        protected:
            RANSACClient * client;
        public:
            RANSACCPHelper(RANSACClient * c, int sz) : client(c), sampleSize(sz) {}
            RANSACCPHelper(const RANSACCPHelper& rhs) : client(NULL), sampleSize(rhs.sampleSize) {}
            virtual ~RANSACCPHelper() {}

            const int sampleSize;
        };

        template<class ErrorFn>
        class RANSACCPHelperPointsAndCams : public RANSACCPHelper
        {
        public:
            typedef vector<int>::iterator iterator;

        private:
            HomographyCalculator calc;
            ErrorFn errfn;

            vector<int> inliers;

            AMatrix cam1, cam2;
            AMatrix Hsolution;
            typedef std::pair<AVector*, AVector*> PointPairT;
            vector<PointPairT> points3d;

            typedef tuple<AVector*, AVector*, double> P3dAndBadness;
            class Comp
            {
                public:
                    bool operator() ( const P3dAndBadness & a, const P3dAndBadness & b )
                    {
                        return ( a.get<2>() < b.get<2>() );
                    }
            };

        public:
            RANSACCPHelperPointsAndCams(RANSACClient * client) :
                RANSACCPHelper(client, 2),
                cam1(3, 4), cam2(3, 4), Hsolution(4, 4)
            {}

            void init( const vector<AVector*> & p1, const vector<AVector*> & p2, const AMatrix& c1, const AMatrix& c2, NView* nv1, NView* nv2 )
            {
                cam1 = c1;
                cam2 = c2;

                points3d.clear(); points3d.reserve(p1.size());
                assert(p1.size()== p2.size());

                vector<P3dAndBadness> points; points.reserve(p1.size());
                for(size_t i = 0; i < p1.size(); ++i)
                {
                    AVector * v1 = p1[i], * v2 = p2[i];
                    if(v1 == NULL || v2 == NULL) continue;
                    assert( fabs((*v1)(3)) > FLT_EPSILON && fabs((*v2)(3)) > FLT_EPSILON );

                    (*v1)(0) = (*v1)(0) / (*v1)(3);
                    (*v1)(1) = (*v1)(1) / (*v1)(3);
                    (*v1)(2) = (*v1)(2) / (*v1)(3);
                    (*v2)(0) = (*v2)(0) / (*v2)(3);
                    (*v2)(1) = (*v2)(1) / (*v2)(3);
                    (*v2)(2) = (*v2)(2) / (*v2)(3);
                    (*v1)(3) = (*v2)(3) = 1.0;

                    PointCorr * pc1 = nv1->getPointCorr(v1);
                    PointCorr * pc2 = nv2->getPointCorr(v2);
                    if(pc1 == NULL || pc2 == NULL) continue;

                    points.push_back(make_tuple(v1, v2, pc1->getBadness() + pc2->getBadness()));
                }

                client->setOkay ( false );

                if ( sampleSize <= ( int ) points.size() ) client->setOkay ( true );
                else
                {
                    DBGOUT ( DBGNORMAL, "RANSACCPMetricHelperPointsAndCams: Ranser was not okay (%d > %zd)\n", sampleSize, points.size() );
                }

                client->setNumCandidates ( points.size() );

                client->setNumIn ( -1 );
                client->setError ( FLT_MAX );

                Comp comp;
                sort ( points.begin(), points.end(), comp );

                for(vector<P3dAndBadness>::iterator it = points.begin(); it != points.end(); ++it)
                {
                    points3d.push_back(make_pair(it->get<0>(), it->get<1>()));
                }
            }

            bool calcSolution( const vector<int> & sample, AMatrix& H );

            bool updateSolution( const AMatrix& H, double thresh );

            AVector * getInlier1(size_t i) const
            {
                assert(i < points3d.size());
                return points3d[i].first;
            }

            AVector * getInlier2(size_t i) const
            {
                assert(i < points3d.size());
                return points3d[i].second;
            }

            iterator begin() { return inliers.begin(); }

            iterator end() { return inliers.end(); }

        };

        template<class ErrorFn>
        class RANSACCPHelperOnlyPoints : public RANSACCPHelper
        {
        public:
            RANSACCPHelperOnlyPoints(RANSACClient * client) : RANSACCPHelper(client, 3) {}
        };

	/**
	 * Filters Correspondences via RANSAC,
	 * Is is used to remove outliers when 3D-Points are used for the calculation of
	 * the merging homography. The homography is calculated of Cameras und Points, thus the suffix CP.
	 *  
	 *  
	 */
	template<class Helper>
	class FilterRANSACCP : public RANSACClient
	{
		Helper m_helper;

		vector<AVector*> m_pts1, m_pts2;
		vector<AMatrix*> m_cams1, m_cams2;
		double m_fractionThresh;

		AMatrix m_H; // TODO!
		vector<int> m_candidates;

	public:

		typedef vector<int>::iterator iterator;

                FilterRANSACCP() : m_helper(this), m_fractionThresh( 0.05 ), m_H(4, 4) { setError(DBL_MAX); }

		int getSampleSize()
		{
                    return m_helper.sampleSize;
		}

		bool applySample( const vector<int> &sample );

		void ransacDone() {} // TODO!

		bool optimize() { return false; } // TODO!

                void init(const vector<AVector*> & p1, const vector<AVector*> & p2,
                          const AMatrix & c1, const AMatrix & c2,
                          NView * nv1, NView * nv2)
		{
                    m_helper.init(p1, p2, c1, c2, nv1, nv2);
		}

		void setPoints3D( const vector<AVector*> & p1, const vector<AVector*> & p2 )
		{
			assert( p1.size() == p2.size() );
			m_pts1 = p1;
			m_pts2 = p2;
		}

		void setCameras( const vector<AMatrix*> & c1, const vector<AMatrix*> & c2 )
		{
			assert( c1.size() == c2.size() );
			m_cams1 = c1;
			m_cams2 = c2;
		}

		void setFractionThresh( double theValue )
		{
			m_fractionThresh = theValue;
		}

		double getFractionThresh() const
		{
			return m_fractionThresh;
		}

        iterator begin() { return m_helper.begin(); }

        iterator end() { return m_helper.end(); }

		AVector * getInlier1(uint i) const
		{
            return m_helper.getInlier1(i);
		}

		AVector * getInlier2(uint i) const
		{
            return m_helper.getInlier2(i);
        }

	};

        typedef FilterRANSACCP< RANSACCPHelperPointsAndCams<CPErrorsSimpleDist> > FilterRANSACCP_PC;

	} // core
} // adore

#endif // ADORE_CORE_FILTERRANSACCP_H

