

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_COREFILTERRANSACPROJECTIVITYHELPER_H
#define ADORE_COREFILTERRANSACPROJECTIVITYHELPER_H

#include <adrcore/types/pointcorr.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/debug.h>
#include <adrcore/util/c++0x.h>
#include <adrcore/random/nviransacclient.h>
#include <adrcore/util/divisiondistortion.h>
#include <adrcore/nview/projectivitycalculator.h>
#include <adrcore/nview/projectivitycalculatorfmat.h>
#include "filterransaccommon.h"

namespace adore
{

	namespace core
	{
		class HSolution
		{
			private:
				AMatrix H;
			public:
				HSolution() :H ( 3, 3 ) {;}
                HSolution ( const AMatrix & h ) :H ( h ) {;}
                inline const AMatrix &getH() const {return H;}
                inline AMatrix &getH() {return H;}
				inline AVector getCorrectedPoint ( AVector &vec ) const {return vec;}
				inline AVector getUnCorrectedPoint ( AVector &vec ) const {return vec;}
				NViewInfo *createNViewInfo()
				{
					NViewInfo *info = new NViewInfoP;
					info->setH ( H );
					return info;
				}

		};

		class HSolutionC : public HSolution
		{
			public:
				HSolutionC():HSolution() {;}
				HSolutionC( AMatrix h ) :HSolution( h ) {;}
		};

		template<> class SolutionCalculator<HSolution>
		{
			private:
				ProjectivityCalculator pcalc;
			public:

				template<class CorrIterator, class ViewIterator>
				int calcSloppy ( CorrIterator cit, int num, ViewIterator vit )
				{
					return calc ( cit, num, vit );
				}

				template<class CorrIterator, class ViewIterator>
				int calc ( CorrIterator cit, int num, ViewIterator vit )
				{
					return pcalc.calculateProjectivity ( cit, num, vit );
				}

				HSolution getSolutionSloppy ( int )
				{
					return pcalc.getH ( );
				}

				HSolution getSolution ( int )
				{
					return pcalc.getH ( );
				}
		};

		template<> class SolutionCalculator<HSolutionC>//C for constrained
		{
			private:
				ProjectivityCalculatorFMat pcalc;
			public:

				template<class CorrIterator, class ViewIterator>
				int calcSloppy ( CorrIterator cit, int num, ViewIterator vit )
				{
					return calc ( cit, num, vit );
				}

				template<class CorrIterator, class ViewIterator>
				int calc ( CorrIterator cit, int num, ViewIterator vit )
				{
					return pcalc.calculateProjectivity ( cit, num, vit );
				}

				HSolutionC getSolutionSloppy ( int )
				{
					return pcalc.getH ( );
				}

				HSolutionC getSolution ( int )
				{
					return pcalc.getH ( );
				}

				void setFMatrix( AMatrix &f )
				{
					pcalc.setFMatrix( f );
				}
		};

		template<> inline int getOuterSampleSize<HSolution>() {return 4;}
		template<> inline int getInnerSampleSize<HSolution>() {return 8;}

		template<> inline int getOuterSampleSize<HSolutionC>() {return 3;}
		template<> inline int getInnerSampleSize<HSolutionC>() {return 6;}

		/**
			 
		*/
		template<class SolutionT, class ErrorFn>
		class FilterRANSACProjectivityHelper
		{
			public:
				typedef SolutionT Solution;
				typedef SolutionCalculator<Solution> Calculator;

				typedef NVIRANSACClient::InlierCollT InlierCollT;
				typedef NVIRANSACClient::iterator iterator;
				typedef tuple<PointCorr*, FeaturePoint *, FeaturePoint *> CorrT;
				typedef vector< CorrT > CorrsT;
				typedef CorrsT::iterator corr_iterator;

				class CorrTComp
				{
					public:
						bool operator()( const CorrT &c1, const CorrT &c2 )
						{
							return ( c1.get<0>()->getBadness() < c2.get<0>()->getBadness() );
						}
				};

			private:
				RANSACClient * client;
				SVDStoraged store3;

			public:
				const int numSamples;
				InlierCollT candInliers;
				InlierCollT inliers;
				CorrsT corrs;
				Solution bestSolution;

				double DBGErr;

			public:
				FilterRANSACProjectivityHelper ( RANSACClient *c, int num ) :client ( c ), store3( 3, 3, SVD_VT ), numSamples ( num ) {;}

				template<class CorrIterator, class ViewIterator>
				void init ( CorrIterator it, int size, ViewIterator vit )
				{
					corrs.clear();
					corrs.reserve ( size );
					inliers.clear();
					inliers.reserve ( size );
					candInliers.clear();
					candInliers.reserve ( size );
					ViewBase *v1 = *vit; ++vit;
					ViewBase *v2 = *vit;
					for ( int i = 0; i < size; ++i, ++it )
					{
						PointCorr *corr = *it;
						if ( !corr ) continue;
						FeaturePoint *fp1 = corr->getFeature ( v1 );
						if ( !fp1 ) continue;
						FeaturePoint *fp2 = corr->getFeature ( v2 );
						if ( !fp2 ) continue;
						corrs.push_back ( make_tuple ( corr, fp1, fp2 ) );
					}
					client->setOkay ( false );
					if ( numSamples <= ( int ) corrs.size() ) client->setOkay ( true );
					else
					{
						DBGOUT( DBGNORMAL, "FR2VHelper: Ranser was not okay (%d > %zd)\n", numSamples, corrs.size() );
					}
					client->setNumCandidates ( corrs.size() );
					client->setNumIn ( -1 );
					client->setError ( FLT_MAX );

					CorrTComp comp;
					sort( corrs.begin(), corrs.end(), comp );
				}

				PointCorr *getPointCorr ( int i ) {return corrs[i].get<0>();}

				bool updateSolution ( Solution &sol, const vector<int> &base, float thresh, bool opt );

				iterator begin() {return inliers.begin();}

				iterator end() {return inliers.end();}

				MODEL_TYPE getModelType() const  { return MODEL_TYPE_H; }

			private:
				ErrorFn computeErrors;

		};

	}

}

#endif
