

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_BABAMINFUNCTIONMETRIC_H
#define ADORE_BABAMINFUNCTIONMETRIC_H

#include <adrcore/types/camera.h>
#include <adrcore/types/aarray.h>
#include <adrcore/util/debug.h>
#include <adrcore/util/printmat.h>

namespace adore
{
	namespace core
	{

		/**
		 * This is a LMSparseFunction for LMSparse
		 * It is made for a bundle-adjustment-problem with the following parametrization:
		 * The a-parameters are camera matrices.
		 * A particular a representing the camera matrix A has the form:
		 * a = (A00, A01, A02, A03, A10, A11, A12, A13, A20, A21, A22, A23)
		 * The b-parameters are homogenous 3D-points of the form (X, Y, 1, W).
		 * A particular b for the 3D-point B has the form
		 * b = (X, Y, W)
		 *
		 * Additionally, you can specify whether the _last_ camera is the fixed [I|0]-Camera.
		 * Then, this camera is taken into account by the error computation.
		 * Please make sure to bring your data in the correct order to satify the condition that the identity cam is
		 * the _last_ cam.
		 *
		 *  
		*/
		template <class CamParamT, class PointParamT>
		class BAMinFunction
		{
			public:
				typedef double T_numtype;

			public:
				/**
				 * Constructor.
				 * @param lastCamIsIdentity guess what.
				 * @param homogen should the resulting 2D-points be homogeneous or not?
				 */
				BAMinFunction( CamParamT &cp, PointParamT & pp, bool lastID = false ):
						camera(),
						point( 4 ),
						cParam( cp ),
						pParam( pp ),
						lastId( lastID )
				{
				}

				/**
				 * Destructor.
				 */
				~BAMinFunction()
				{
					;
				}

				void operator()( vector<AVector> &argType1, vector<AVector> &argType2, vector<AVector> &result )
				{
					uint i = 0;
					//printf("processing %d cams\n", argType1.size());
					for ( uint b = 0; b < argType2.size(); ++b )
					{
						for ( uint a = 0; a < argType1.size(); ++a )
						{
							result[i++] = operator()( argType1[a], argType2[b] );
						}
						if ( lastId )
						{
							//printf("lastId is set, processing additional cam\n");
							result[i++] = operator()( argType2[b] );
						}
					}
				}

				/**
				 * Carries out one projection.
				 * @param cam a  camera matrix.
				 * @param pt3D a 3D-Point.
				 * @param result the resulting 2D-point.
				 */
				//arg1 would be a camera and arg2 a 3D-Point
				AVector operator()( AVector &cam, AVector &pt3D, int camIdx = -1 )
				{
					ADORE_UNUSED( camIdx );

					AVector result( 2 );
					cParam.fromParam( cam, camera );
					pParam.fromParam( pt3D, point );

//					printf( "in BAMinFunc:\n" );
//					printMat( camera.getProjMat() );
//					printMat( point );

					AVector2D p2d = camera.project( point );

					double w = p2d( 2 );
					result( 0 ) = p2d( 0 ) / w;
					result( 1 ) = p2d( 1 ) / w;

//					printf( "results:\n" );
//					printMat( result );
//					pressKey();

					return result;
				}

				/**
				 * Carries out one projection with an internal cam (That is: The identity cam).
				 * @param pt3D a 3D-Point.
				 * @param result the resulting 2D-point.
				 * @param cam The index of the internal cam. Leave it zero.
				 */
				//the camera is known but not a parameter to adjust
				//i chooses the camera
				AVector operator()( AVector &pt3D, int cam = 0 )
				{
					assert_msg( cam == 0, "Only one fixed cam is currently supported.\n" );
					AVector result( 2 );

					pParam.fromParam( pt3D, point );

					//multiplication with identity boils down to this:
					double w = point( 2 );
					result( 0 ) = point( 0 ) / w;
					result( 1 ) = point( 1 ) / w;
					return result;
				}

			protected:
				Camera camera;
				AVector3D point;
				CamParamT &cParam;
				PointParamT &pParam;
				bool lastId;
		};

	}

}

#endif
