

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_BABAMINFUNCTIONSIMPLE_H
#define ADORE_BABAMINFUNCTIONSIMPLE_H

#include <adrcore/types/camera.h>
#include <adrcore/types/aarray.h>
#include <adrcore/util/debug.h>
#include <adrcore/util/printmat.h>
#include <adrcore/parameterization/quaternioncamparam.h>

namespace adore
{
	namespace core
	{

		/**
		 * This is a LMSparseFunction for LMSparse
		 * It is made for a bundle-adjustment-problem with the following parametrization:
		 * The a-parameters are camera matrices.
		 * A particular a representing the camera matrix A has the form:
		 * a = (A00, A01, A02, A03, A10, A11, A12, A13, A20, A21, A22, A23)
		 * The b-parameters are homogenous 3D-points of the form (X, Y, 1, W).
		 * A particular b for the 3D-point B has the form
		 * b = (X, Y, W)
		 *
		 * Additionally, you can specify whether the _last_ camera is the fixed [I|0]-Camera.
		 * Then, this camera is taken into account by the error computation.
		 * Please make sure to bring your data in the correct order to satify the condition that the identity cam is
		 * the _last_ cam.
		 *
		 *  
		*/
		class BAMinFunctionSimple
		{
			public:
				typedef double T_numtype;
				enum { CamDim = 12 };

			public:

				inline void operator() ( const vector<AVector> &argType1, const vector<AVector> &argType2, vector<AVector> &result )
				{
					uint i = 0;
					//printf("processing %d cams\n", argType1.size());
					for ( uint b = 0; b < argType2.size(); ++b )
					{
						for ( uint a = 0; a < argType1.size(); ++a )
						{
							operator() ( argType1[a], argType2[b], result[i++] );
						}
					}
				}

				inline void operator() ( const AVector &cam, const AVector &pt3D, AVector &result )
				{
					assert ( cam.rows() == 12 );
					assert ( pt3D.rows() == 3 );

					double w = cam ( 2 ) * pt3D ( 0 ) + cam ( 5 ) * pt3D ( 1 ) + cam ( 8 ) + cam ( 11 ) * pt3D ( 2 );
					if ( fabs ( w ) < FLT_EPSILON )   //TODO?
					{
						w = sqrt( FLT_MAX );
					}
					else
						w = 1.0/ w;

					result ( 0 ) = w * ( cam ( 0 ) * pt3D ( 0 ) + cam ( 3 ) * pt3D ( 1 ) + cam ( 6 ) + cam ( 9 ) * pt3D ( 2 ) );
					result ( 1 ) = w * ( cam ( 1 ) * pt3D ( 0 ) + cam ( 4 ) * pt3D ( 1 ) + cam ( 7 ) + cam ( 10 ) * pt3D ( 2 ) );
				}

				inline void jacobi ( const AVector &cam, const AVector &pt3D, AMatrix &camResult, AMatrix &pointResult )
				{
					assert ( camResult.rows() == 2 );
					assert ( camResult.cols() == 12 );
					assert ( pointResult.rows() == 2 );
					assert ( pointResult.cols() == 3 );
					assert ( cam.rows() == 12 );
					assert ( pt3D.rows() == 3 );

					double w = cam ( 2 ) * pt3D ( 0 ) + cam ( 5 ) * pt3D ( 1 ) + cam ( 8 ) + cam ( 11 ) * pt3D ( 2 );
					if ( fabs ( w ) < FLT_EPSILON )   //TODO?
					{
						w = sqrt( FLT_MAX );
					}
					else
						w = 1.0/ w;//TODO handle the case where w=0 correctly
					double sqrW = - ( w * w );
					double n1 = sqrW * ( cam ( 0 ) * pt3D ( 0 ) + cam ( 3 ) * pt3D ( 1 ) + cam ( 6 ) + cam ( 9 ) * pt3D ( 2 ) );
					double n2 = sqrW * ( cam ( 1 ) * pt3D ( 0 ) + cam ( 4 ) * pt3D ( 1 ) + cam ( 7 ) + cam ( 10 ) * pt3D ( 2 ) );

					//d/dCam
					camResult ( 0, 0 ) = w * pt3D ( 0 );
					camResult ( 0, 1 ) = 0;
					camResult ( 0, 2 ) = n1 * pt3D ( 0 );
					camResult ( 0, 3 ) = w * pt3D ( 1 );
					camResult ( 0, 4 ) = 0;
					camResult ( 0, 5 ) = n1 * pt3D ( 1 );
					camResult ( 0, 6 ) = w;
					camResult ( 0, 7 ) = 0;
					camResult ( 0, 8 ) = n1;
					camResult ( 0, 9 ) = w * pt3D ( 2 );
					camResult ( 0, 10 ) = 0;
					camResult ( 0, 11 ) = n1 * pt3D ( 2 );
					camResult ( 1, 0 ) = 0;
					camResult ( 1, 1 ) = w * pt3D ( 0 );
					camResult ( 1, 2 ) = n2 * pt3D ( 0 );
					camResult ( 1, 3 ) = 0;
					camResult ( 1, 4 ) = w * pt3D ( 1 );
					camResult ( 1, 5 ) = n2 * pt3D ( 1 );
					camResult ( 1, 6 ) = 0;
					camResult ( 1, 7 ) = w;
					camResult ( 1, 8 ) = n2;
					camResult ( 1, 9 ) = 0;
					camResult ( 1, 10 ) = w * pt3D ( 2 );
					camResult ( 1, 11 ) = n2 * pt3D ( 2 );

					//d/dPoint
					pointResult ( 0, 0 ) = w * cam ( 0 ) + n1 * cam ( 2 );
					pointResult ( 0, 1 ) = w * cam ( 3 ) + n1 * cam ( 5 );
					pointResult ( 0, 2 ) = w * cam ( 9 ) + n1 * cam ( 11 );
					pointResult ( 1, 0 ) = w * cam ( 1 ) + n2 * cam ( 2 );
					pointResult ( 1, 1 ) = w * cam ( 4 ) + n2 * cam ( 5 );
					pointResult ( 1, 2 ) = w * cam ( 10 ) + n2 * cam ( 11 );
				}

		};


		/** Rotation, translation and focal length
		  * f(ax,ay,az,tx,ty,tz,f,x,y,z) -> (a,b)
		  */
		class BAMinFunctionSimple_RTF
		{
			public:
				typedef double T_numtype;
				enum { CamDim = 7 };

			public:

				inline void operator() ( const vector<AVector> &argType1, const vector<AVector> &argType2, vector<AVector> &result )
				{
					uint i = 0;
					cout << endl << endl << endl;
					for ( uint a = 0; a < argType1.size(); ++a )
					{
//						cout << "cam. nr.: " << a << endl << argType1[a] << endl;
					}

					for ( uint b = 0; b < argType2.size(); ++b )
					{
						for ( uint a = 0; a < argType1.size(); ++a )
						{
							operator() ( argType1[a], argType2[b], result[i++] );
						}
					}
				}

				inline void operator() ( const AVector &cam, const AVector &pt3D, AVector &result )
				{
					assert ( cam.rows() == 7 );
					assert ( pt3D.rows() == 3 );

//					const double some_small_th(0.000001); // TODO

					double ax = cam(0), ay = cam(1), az = cam(2);
					double tx = cam(3), ty = cam(4), tz = cam(5);
					double f = cam(6);

					double x = pt3D(0), y = pt3D(1), z = pt3D(2);

					double th = sqrt(ax * ax + ay * ay + az * az);

					double r0, r1, r2;

					if(th <= FLT_EPSILON)
					{
						/** R = I + [t]x
						  * (  1  -a3  a2 )
						  * (  a3  1  -a1 )
						  * ( -a2  a1  1  )
						  */
						r0 = f * ( x      - az * y + ay * z - tx);
						r1 = f * ( az * x + y      - ax * z - ty);
						r2 =      -ay * x + ax * y + z      - tz;
					}
					else
					{
						double nx = ax / th, ny = ay / th, nz = az / th;
						double cth = cos(th), sth = sin(th);
						double vnp = nx * x + ny * y + nz * z;
						double nv1c = vnp * (1.0 - cth);

						r0 = f * (x * cth + (ny * z - y * nz) * sth + nx * nv1c - tx);
						r1 = f * (y * cth + (nz * x - z * nx) * sth + ny * nv1c - ty);
						r2 =      z * cth + (nx * y - x * ny) * sth + nz * nv1c - tz;

					}

					if ( fabs ( r2 ) < FLT_EPSILON )   //TODO?
					{
						r2 = sqrt( FLT_MAX );
					}
					else
					{
						r2 = 1.0 / r2;
					}

					result ( 0 ) = r2 * r0;
					result ( 1 ) = r2 * r1;
				}

				inline void jacobi ( const AVector &cam, const AVector &pt3D, AMatrix &camResult, AMatrix &pointResult )
				{
					assert ( camResult.rows() == 2 );
					assert ( camResult.cols() == 7 );
					assert ( pointResult.rows() == 2 );
					assert ( pointResult.cols() == 3 );
					assert ( cam.rows() == 7 );
					assert ( pt3D.rows() == 3 );

//					const double some_small_th(0.000001); // TODO

					double ax = cam(0), ay = cam(1), az = cam(2);
					double tx = cam(3), ty = cam(4), tz = cam(5);
					double f = cam(6);

					double x = pt3D(0), y = pt3D(1), z = pt3D(2);

					double th = sqrt(ax * ax + ay * ay + az * az);

					if(th <= FLT_EPSILON)
					{
						double g1_f = ay * z - az * y + x - tx;
						double g1 = f * g1_f;
						double g2_f = -ax * z + y + az * x - ty;
						double g2 = f * g2_f;
						double h = z + ax * y - ay * x - tz;
						double h2 = h * h;

						double dax = -(y * g1) / h2; // checked 1
						double day = (x * g1) / h2 + (f * z) / h; // checked 1
						double daz = -(f * y) / h;
						double dtx = -f / h;
						double dty = 0;
						double dtz = g1 / h2;
						double df  = g1_f / h;
						double dx  = (ay * g1) / h2 + f / h;
						double dy  = -(ax * g1) / h2 - (az * f) / h;
						double dz  = (ay * f) / h - g1 / h2;

						camResult(0, 0) = dax;
						camResult(0, 1) = day;
						camResult(0, 2) = daz;
						camResult(0, 3) = dtx;
						camResult(0, 4) = dty;
						camResult(0, 5) = dtz;
						camResult(0, 6) = df;

						pointResult(0, 0) = dx;
						pointResult(0, 1) = dy;
						pointResult(0, 2) = dz;

						dax = -(y * g2) / h2 - (f * z) / h; // checked 1
						day = (x * g2) / h2; // checked 1
						daz = (f * x) / h;
						dtx = 0;
						dty = -f / h;
						dtz = g2 / h2;
						df  = g2_f / h;
						dx  = (ay * g2) / h2 + (az * f) / h;
						dy  = f / h - (ax * g2) / h2;
						dz  = -g2 / h2 - (ax * f) / h;

						camResult(1, 0) = dax;
						camResult(1, 1) = day;
						camResult(1, 2) = daz;
						camResult(1, 3) = dtx;
						camResult(1, 4) = dty;
						camResult(1, 5) = dtz;
						camResult(1, 6) = df;

						pointResult(1, 0) = dx;
						pointResult(1, 1) = dy;
						pointResult(1, 2) = dz;

						return;
					}

					{// else
					double cth = cos(th), sth = sin(th), cth1 = 1.0 - cos(th);
					double nx  = ax / th, ny  = ay / th, nz   = az / th;
					double vnp = nx * x + ny * y + nz * z;

					double g1_f = nx*cth1*vnp + sth*(ny*z - nz*y) + cth*x - tx;
					double g1   = f*g1_f; // == diff(g1, f);
					double g2_f = ny*cth1*vnp + sth*(nz*x - nx*z) + cth*y - ty;
					double g2   = f*g2_f; // == diff(g2, f);
					double h    = nz*cth1*vnp + sth*(nx*y - ny*x) + cth*z - tz ;
					double h2   = h * h;

					double dnx_ax    = (1.0 - nx * nx) / th;
					double dnx_ay    = -(nx * ny) / th;
					double dnx_az    = -(nx * nz) / th;

					double dnz_ax    = -(nx * nz) / th;
					double dnz_ay    = -(ny * nz) / th;
					double dnz_az    = (1.0 - nz * nz) / th;

					double dcth1_ax  = nx * sth;
					double dcth1_ay  = ny * sth;
					double dcth1_az  = nz * sth;

					double dvnp_ax   = (x - nx * nz * z - nx * ny * y - nx * nx * x) / th;
					double dvnp_ay   = (y - ny * nz * z - ny * ny * y - nx * ny * x) / th;
					double dvnp_az   = (z - nz * nz * z - ny * nz * y - nx * nz * x) / th;

					double dsth_ax   = nx * cth;
					double dsth_ay   = ny * cth;
					double dsth_az   = nz * cth;

					double nyz       = ny * z;
					double dnyz_ax   = -(nx * ny * z) / th;
					double dnyz_ay   = (z - ny * ny * z) / th;
					double dnyz_az   = -(ny * nz * z) / th;

					double nxy       = nx * y;
					double dnxy_ax   = (y - nx * nx * y) / th;
					double dnxy_ay   = -(nx * ny * y) / th;
					double dnxy_az   = -(nx * nz * y) / th;

					double nzy       = nz * y;
					double dnzy_ax   = -(nx * nz * y) / th;
					double dnzy_ay   = -(ny * nz * y) / th;
					double dnzy_az   = (y - nz * nz * y) / th;

					double nyx       = ny * x;
					double dnyx_ax   = -(nx * ny * x) / th;
					double dnyx_ay   = (x - ny * ny * x) / th;
					double dnyx_az   = -(ny * nz * x) / th;

					double dcth_ax  = -nx * sth;
					double dcth_ay  = -ny * sth;
					double dcth_az  = -nz * sth;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dax
					double dg    = f * (dnx_ax * cth1 * vnp + nx * dcth1_ax * vnp + nx * cth1 * dvnp_ax + dsth_ax * (nyz - nzy) + sth * (dnyz_ax - dnzy_ax) + dcth_ax * x); // checked 1
					double dh_ax =      dnz_ax * cth1 * vnp + nz * dcth1_ax * vnp + nz * cth1 * dvnp_ax + dsth_ax * (nxy - nyx) + sth * (dnxy_ax - dnyx_ax) + dcth_ax * z; // checked 1
					double dax = (dg * h - g1 * dh_ax) / h2;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/day
					dg           = f * (dnx_ay * cth1 * vnp + nx * dcth1_ay * vnp + nx * cth1 * dvnp_ay + dsth_ay * (nyz - nzy) + sth * (dnyz_ay - dnzy_ay) + dcth_ay * x); // checked 1
					double dh_ay =      dnz_ay * cth1 * vnp + nz * dcth1_ay * vnp + nz * cth1 * dvnp_ay + dsth_ay * (nxy - nyx) + sth * (dnxy_ay - dnyx_ay) + dcth_ay * z; // checked 1
					double day = (dg * h - g1 * dh_ay) / h2;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/daz
					dg           = f * (dnx_az * cth1 * vnp + nx * dcth1_az * vnp + nx * cth1 * dvnp_az + dsth_az * (nyz - nzy) + sth * (dnyz_az - dnzy_az) + dcth_az * x); // checked 1
					double dh_az =      dnz_az * cth1 * vnp + nz * dcth1_az * vnp + nz * cth1 * dvnp_az + dsth_az * (nxy - nyx) + sth * (dnxy_az - dnyx_az) + dcth_az * z; // checked 1
					double daz = (dg * h - g1 * dh_az) / h2;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dtx
					double dtx = -f / h; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dty
					double dty = 0; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dtz
					double dtz = g1 / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/df
					double df = (g1_f * h) / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dx
					double dh_dx = nx * nz * cth1 - ny * sth; // checked 1
					double dx = (f * (cth + nx * nx * cth1) * h - g1 * dh_dx) / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dy
					double dh_dy = nx * sth + ny * nz * cth1; // checked 1
					double dy = (f * (nx * ny * cth1 - nz * sth) * h - g1 * dh_dy) / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dz
					double dh_dz = cth + nz * nz * cth1; // checked 1
					double dz = (f * (ny * sth + nx * nz * cth1) * h - g1 * dh_dz) / h2; // checked 1

					camResult(0, 0) = dax;
					camResult(0, 1) = day;
					camResult(0, 2) = daz;
					camResult(0, 3) = dtx;
					camResult(0, 4) = dty;
					camResult(0, 5) = dtz;
					camResult(0, 6) = df;

					pointResult(0, 0) = dx;
					pointResult(0, 1) = dy;
					pointResult(0, 2) = dz;

					double dny_ax    = -(nx * ny) / th;
					double dny_ay    = (1.0 - ny * ny) / th;
					double dny_az    = -(ny * nz) / th;

					double nzx       = nz * x;
					double dnzx_ax   = -(nx * nz * x) / th;
					double dnzx_ay   = -(ny * nz * x) / th;
					double dnzx_az   = (x - nz * nz * x) / th;

					double nxz       = nx * z;
					double dnxz_ax   = (z - nx * nx * z) / th;
					double dnxz_ay   = -(nx * ny * z) / th;
					double dnxz_az   = -(nx * nz * z) / th;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dax
					dg = f * (dny_ax * cth1 * vnp + ny * dcth1_ax * vnp + ny * cth1 * dvnp_ax + dsth_ax * (nzx - nxz) + sth * (dnzx_ax - dnxz_ax) + dcth_ax * y); // checked 1
					dax = (dg * h - g2 * dh_ax) / h2;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/day
					dg = f * (dny_ay * cth1 * vnp + ny * dcth1_ay * vnp + ny * cth1 * dvnp_ay + dsth_ay * (nzx - nxz) + sth * (dnzx_ay - dnxz_ay) + dcth_ay * y); // checked 1
					day = (dg * h - g2 * dh_ay) / h2;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/daz
					dg = f * (dny_az * cth1 * vnp + ny * dcth1_az * vnp + ny * cth1 * dvnp_az + dsth_az * (nzx - nxz) + sth * (dnzx_az - dnxz_az) + dcth_az * y); // checked 1
					daz = (dg * h - g2 * dh_az) / h2;

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dtx
					dtx = 0; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dty
					dty = (-f * h) / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dtz
					dtz = g2 / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/df
					df = (g2_f * h) / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dx
					dx = (f * (nz * sth + nx * ny * cth1) * h - g2 * dh_dx) / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dy
					dy = (f * (cth + ny * ny * cth1) * h - g2 * dh_dy) / h2; // checked 1

					// df(ax, ay, az, tx, ty, tz, f, x, y, z)/dz
					dz = (f * (ny * nz * cth1 - nx * sth) * h - g2 * dh_dz) / h2; // checked 1

					camResult(1, 0) = dax;
					camResult(1, 1) = day;
					camResult(1, 2) = daz;
					camResult(1, 3) = dtx;
					camResult(1, 4) = dty;
					camResult(1, 5) = dtz;
					camResult(1, 6) = df;

					pointResult(1, 0) = dx;
					pointResult(1, 1) = dy;
					pointResult(1, 2) = dz;
					}
				}

		};

		/**
		  * f(a,b,c,d,tx,ty,tz,f1,f2,s,x0,y0,x,y,z) -> (a,b)
		  */
		class BAMinFunctionSimple_Quat_Full
		{
			public:
				typedef double T_numtype;
				enum { CamDim = 12 };
				QuaternionCamParam calc;
				AMatrix cam34;
				AVector temp3, temp4;

			public:
				BAMinFunctionSimple_Quat_Full() : cam34(3, 4), temp3(3), temp4(4) {}

				inline void operator() ( vector<AVector> &argType1, vector<AVector> &argType2, vector<AVector> &result )
				{
					for ( uint b = 0, i = 0; b < argType2.size(); ++b )
					{
						for ( uint a = 0; a < argType1.size(); ++a )
						{
							operator() ( argType1[a], argType2[b], result[i++] );
						}
					}
				}

				inline void operator() ( const AVector &cam, const AVector &pt3D, AVector &result )
				{
					assert ( cam.rows()  == 12 );
					assert ( pt3D.rows() == 3 );

					calc.fromParam(cam, cam34);
					cout << "new cam: " << cam34;

					temp4 = pt3D(0), pt3D(1), pt3D(2), 1.0;

					matMul2(cam34, temp4, temp3);

					double w = temp3(2);
					if ( fabs ( w ) < FLT_EPSILON )   //TODO?
						w = sqrt( FLT_MAX );
					else
						w = 1.0 / w;

					result(0) = temp3(0) / w;
					result(1) = temp3(1) / w;
				}

				inline void jacobi ( const AVector &cam, const AVector &pt3D, AMatrix &camResult, AMatrix &pointResult )
				{
					assert ( camResult.rows() == 2 );
					assert ( camResult.cols() == 12 );
					assert ( pointResult.rows() == 2 );
					assert ( pointResult.cols() == 3 );
					assert ( cam.rows() == 12 );
					assert ( pt3D.rows() == 3 );

					double a = cam(0);
					double b = cam(1);
					double c = cam(2);
					double d = cam(3);
					double l = sqrt(a*a+b*b+c*c+d*d);
					assert(l > FLT_EPSILON);
					a /= l;
					b /= l;
					c /= l;
					d /= l;

					const double tx = cam(4);
					const double ty = cam(5);
					const double tz = cam(6);

					const double f1 = cam(7);
					const double f2 = cam(8);
					const double s  = cam(9);
					const double x0 = cam(10);
					const double y0 = cam(11);

					const double x = pt3D(0);
					const double y = pt3D(1);
					const double z = pt3D(2);

					double u0 = c*d-a*b;
					double u1 = b*d+a*c;
					double u2 = c*d+a*b;
					double u3 = b*c-a*d;
					double u4 = b*d-a*c;
					double u5 = a*d+b*c;
					double u6 = 1.0-2.0*(c*c+b*b);
					double u7 = 1.0-2.0*(d*d+b*b);
					double u8 = 1.0-2.0*(d*d+c*c);

					double g1 = (u6*x0+2.0*u0*s+2.0*u1*f1)*z+(2.0*u2*x0+u7*s+2.0*u3*f1)*y+x*(2.0*u4*x0+2*u5*s+u8*f1)+tz*x0+s*ty+f1*tx;
					double h = u6*z+2.0*u2*y+2.0*u4*x+tz;
					if ( h < FLT_EPSILON ) h = FLT_EPSILON;
					double h2 = h * h;

					double gda = 2.0*(z*(c*f1-b*s) + y*(b*x0-d*f1) + x*(d*s-c*x0));
					double hda = 2.0*b*y-2.0*c*x;
					double fda = (gda * h - g1 * hda) / h2;

					double gdb = 2.0*(z*(-2.0*b*x0-a*s+d*f1) + y*(a*x0-2.0*b*s+c*f1) + x*(d*x0+c*s));
					double hdb = -4.0*b*z+2.0*a*y+2.0*d*x;
					double fdb = (gdb * h - g1 * hdb) / h2;

					double gdc = 2.0*(z*(-2.0*c*x0+d*s+a*f1) + y*(d*x0+b*f1) + x*(-a*x0+b*s-2.0*c*f1));
					double hdc = -4.0*c*z+2.0*d*y-2.0*a*x;
					double fdc = (gdc * h - g1 * hdc) / h2;

					double gdd = 2.0*(z*(c*s+b*f1) + y*(c*x0-2.0*d*s-a*f1) + x*(b*x0+a*s-2.0*d*f1));
					double hdd = 2.0*c*y+2.0*b*x;
					double fdd = (gdd * h - g1 * hdd) / h2;

					double fdtx = f1 / h;
					double fdty = s / h;
					double fdtz = (x0 * h - g1) / h2;

					double gdf1 = 2.0*(z*(b*d+a*c) + y*(b*c-a*d)) + x*(1.0-2.0*(d*d+c*c)) + tx;
					double fdf1 = gdf1 * h / h2;

					double fdf2 = 0;

					double gds = 2.0*(z*(c*d-a*b) + x*(a*d+b*c)) + y*(1.0-2.0*(d*d+b*b)) + ty;
					double fds = gds * h / h2;

					double gdx0 = 2.0*(y*(c*d+a*b) + x*(b*d-a*c)) + z*(1.0-2.0*(c*c+b*b)) + tz;
					double fdx0 = gdx0 * h / h2;

					double fdy0 = 0;

					double gdx = 2.0*(x0*(b*d-a*c) + s*(a*d+b*c)) + f1*(1.0-2.0*(d*d+c*c));
					double hdx = 2.0*b*d-2.0*a*c;
					double fdx = (gdx * h - g1 * hdx) / h2;

					double gdy = 2.0*(x0*(c*d+a*b) + f1*(b*c-a*d)) + s*u7;
					double hdy = 2.0*c*d+2.0*a*b;
					double fdy = (gdy * h - g1 * hdy) / h2;

					double gdz = 2.0*(s*(c*d-a*b) + f1*(b*d+a*c)) + x0*u6;
					double hdz = u6;
					double fdz = (gdz * h - g1 * hdz) / h2;

					camResult(0, 0) = fda;
					camResult(0, 1) = fdb;
					camResult(0, 2) = fdc;
					camResult(0, 3) = fdd;

					camResult(0, 4) = fdtx;
					camResult(0, 5) = fdty;
					camResult(0, 6) = fdtz;

					camResult(0, 7) = fdf1;
					camResult(0, 8) = fdf2;
					camResult(0, 9) = fds;
					camResult(0, 10) = fdx0;
					camResult(0, 11) = fdy0;

					pointResult(0, 0) = fdx;
					pointResult(0, 1) = fdy;
					pointResult(0, 2) = fdz;

					double g2 = (u6*y0+(2.0*c*d-2.0*a*b)*f2)*z+y*((2.0*c*d+2.0*a*b)*y0+u7*f2)+x*((2.0*b*d-2.0*a*c)*y0+(2.0*a*d+2.0*b*c)*f2)+tz*y0+f2*ty;

					gda = 2.0*(-z*b*f2 + x*(d*f2-c*y0) + y*b*y0);
					fda = (gda * h - g2 * hda) / h2;

					gdb = 2.0*(z*(-2.0*b*y0-a*f2) + x*(d*y0+c*f2) + y*(a*y0-2.0*b*f2));
					fdb = (gdb * h - g2 * hdb) / h2;

					gdc = 2.0*(z*(d*f2-2.0*c*y0) + y*d*y0 + x*(b*f2-a*y0));
					fdc = (gdc * h - g2 * hdc) / h2;

					gdd = 2.0*(z*c*f2 + y*(c*y0-2.0*d*f2) + x*(b*y0+a*f2));
					fdd = (gdd * h - g2 * hdd) / h2;

					fdtx = 0;
					fdty = f2 * h / h2;
					fdtz = (y0 * h - g2) / h2;

					fdf1 = 0;

					double gdf2 = 2.0*(z*(c*d-a*b) + x*(a*d+b*c)) + ty + y*(1.0-2.0*(d*d+b*b));
					fdf2 = gdf2 * h / h2;

					fds = 0;

					fdx0 = 0;

					double gdy0 = 2.0*(y*(c*d+a*b) + x*(b*d-a*c)) + tz + z*(1.0-2.0*(c*c+b*b));
					fdy0 = gdy0 * h / h2;

					gdx = 2.0*(y0*(b*d-a*c) + f2*(a*d+b*c));
					fdx = (gdx * h - g2 * hdx) / h2;

					gdy = 2.0*y0*(c*d+a*b) + f2*(1.0-2.0*(d*d+b*b));
					fdy = (gdy * h - g2 * hdy) / h2;

					gdz = 2.0*f2*(c*d-a*b) + y0*(1.0-2.0*(c*c+b*b));
					fdz = (gdz * h - g2 * hdz) / h2;

					camResult(1, 0) = fda;
					camResult(1, 1) = fdb;
					camResult(1, 2) = fdc;
					camResult(1, 3) = fdd;

					camResult(1, 4) = fdtx;
					camResult(1, 5) = fdty;
					camResult(1, 6) = fdtz;

					camResult(1, 7) = fdf1;
					camResult(1, 8) = fdf2;
					camResult(1, 9) = fds;
					camResult(1, 10) = fdx0;
					camResult(1, 11) = fdy0;

					pointResult(1, 0) = fdx;
					pointResult(1, 1) = fdy;
					pointResult(1, 2) = fdz;

				}
		};

		/**
		  * f(a,b,c,d,tx,ty,tz,f,x,y,z) -> (a,b)
		  */
		class BAMinFunctionSimple_Quat_F
		{
			public:
				typedef double T_numtype;
				enum { CamDim = 8 };
				QuaternionCamParam_F calc;
				AMatrix cam34;
				AVector temp3, temp4;

			public:
				BAMinFunctionSimple_Quat_F() : cam34(3, 4), temp3(3), temp4(4) {}

				inline void operator() ( const vector<AVector> &argType1, const vector<AVector> &argType2, vector<AVector> &result )
				{
					for ( uint b = 0, i = 0; b < argType2.size(); ++b )
					{
						for ( uint a = 0; a < argType1.size(); ++a )
						{
							operator() ( argType1[a], argType2[b], result[i++] );
						}
					}
				}

				inline void operator() ( const AVector &cam, const AVector &pt3D, AVector &result )
				{
					assert ( cam.rows()  == 8 );
					assert ( pt3D.rows() == 3 );

					calc.fromParam(cam, cam34);
					cout << "new cam: " << cam34;

					temp4 = pt3D(0), pt3D(1), pt3D(2), 1.0;

					matMul2(cam34, temp4, temp3);

					double z = temp3(2);
					if ( fabs ( z ) < FLT_EPSILON )
						z = sqrt( FLT_MAX );
					else
						z = 1.0 / z;

					result(0) = temp3(0) / z;
					result(1) = temp3(1) / z;
				}

				inline void jacobi ( const AVector &cam, const AVector &pt3D, AMatrix &camResult, AMatrix &pointResult )
				{
					assert ( camResult.rows() == 2 );
					assert ( camResult.cols() == 8 );
					assert ( pointResult.rows() == 2 );
					assert ( pointResult.cols() == 3 );
					assert ( cam.rows() == 8 );
					assert ( pt3D.rows() == 3 );

					double a = cam(0);
					double b = cam(1);
					double c = cam(2);
					double d = cam(3);
					double l = sqrt(a*a+b*b+c*c+d*d);
					assert(l > FLT_EPSILON);
					a /= l;
					b /= l;
					c /= l;
					d /= l;

					const double tx = cam(4);
					const double ty = cam(5);
					const double tz = cam(6);

					const double f = cam(7);

					const double x = pt3D(0);
					const double y = pt3D(1);
					const double z = pt3D(2);

					double g1 = (2.0*b*d+2.0*a*c)*f*z+(2.0*b*c-2.0*a*d)*f*y+(1.0-2.0*(d*d+c*c))*f*x+f*tx;
					double g2 = (2.0*c*d-2.0*a*b)*f*z+(1.0-2.0*(d*d+b*b))*f*y+(2.0*a*d+2.0*b*c)*f*x+f*ty;
					double h = (1.0-2.0*(c*c+b*b))*z+(2.0*c*d+2.0*a*b)*y+(2.0*b*d-2.0*a*c)*x+tz;
					if ( h < FLT_EPSILON ) h = FLT_EPSILON;
					double h2 = h * h;

					double fda = (h * (2.0*c*f*z-2.0*d*f*y) - g1 * (2.0*b*y-2.0*c*x))/ h2;
					double fdb = (h * (2.0*d*f*z+2.0*c*f*y) - g1 * (-4.0*b*z+2.0*a*y+2.0*d*x)) / h2;
					double fdc = (h * (2.0*a*f*z+2.0*b*f*y-4.0*c*f*x) - g1 * (-4.0*c*z+2.0*d*y-2.0*a*x)) / h2;
					double fdd = (h * (2.0*b*f*z-2.0*a*f*y-4.0*d*f*x) - g1 * (2.0*c*y+2.0*b*x)) / h2;

					double fdtx = 0;
					double fdty = 0;
					double fdtz = 0;

					double fdf = 0;

					double fdx = 0;
					double fdy = 0;
					double fdz = 0;

					camResult(0, 0) = fda;
					camResult(0, 1) = fdb;
					camResult(0, 2) = fdc;
					camResult(0, 3) = fdd;

					camResult(0, 4) = fdtx;
					camResult(0, 5) = fdty;
					camResult(0, 6) = fdtz;

					camResult(0, 7) = fdf;

					pointResult(0, 0) = fdx;
					pointResult(0, 1) = fdy;
					pointResult(0, 2) = fdz;


					fda = (h * (2.0*d*f*x-2.0*b*f*z) - g2 * (2.0*b*y-2.0*c*x)) / h2;
					fdb = (h * (-2.0*a*f*z-4.0*b*f*y+2.0*c*f*x) - g2 * (-4.0*b*z+2.0*a*y+2.0*d*x)) / h2;
					fdc = (h * (2.0*d*f*z+2.0*b*f*x) - g2 * (-4.0*c*z+2.0*d*y-2.0*a*x)) / h2;
					fdd = (h * (2.0*c*f*z-4.0*d*f*y+2.0*a*f*x) - g2 * (2.0*c*y+2.0*b*x)) / h2;

					fdtx = 0;
					fdty = 0;
					fdtz = 0;

					fdf = 0;

					fdx = 0;
					fdy = 0;
					fdz = 0;

					camResult(1, 0) = fda;
					camResult(1, 1) = fdb;
					camResult(1, 2) = fdc;
					camResult(1, 3) = fdd;

					camResult(1, 4) = fdtx;
					camResult(1, 5) = fdty;
					camResult(1, 6) = fdtz;

					camResult(1, 7) = fdf;

					pointResult(1, 0) = fdx;
					pointResult(1, 1) = fdy;
					pointResult(1, 2) = fdz;
				}
		};
	}

}

#endif
