

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_MINIMIZATIONLMFUNCTIONJACOBI2_H
#define ADORE_MINIMIZATIONLMFUNCTIONJACOBI2_H

#include<adrcore/types/aarray.h>
#include<adrcore/util/mathutil.h>

namespace adore
{
	namespace core
	{

		/**
		 * This Functor calculates the matrix of partial derivatives known as the Jacobian Matrix.
		 * This version uses an interval to the right of the argument to calculate the ratio.
		 * (Experience tells that this is often a good choice)
		 *  
		 */
		template<class FunctionT>
		class LMFunctionJacobi2
		{
				typedef typename FunctionT::T_numtype numberT;

			public:
				typedef double T_numtype;

			public:
				/**
				 * Contructor.
				 * @param f the function the has to be numerically derived.
				 */
				LMFunctionJacobi2 ( const FunctionT &f, int domainDim, int rangeDim ) :
						func ( f ),
						inDim ( domainDim ),
						outDim ( rangeDim ),
						val1 ( outDim ),
						val2 ( outDim ),
						tmpvalueR ( inDim )
				{
				}

				/**
				 * initialize the Jacobi-Functor
				 * @param f the function that has to be numerically derived.
				 */
				void init ( const FunctionT &f, int domainDim, int rangeDim )
				{
					func = f;
					inDim = domainDim;
					outDim = rangeDim;
				}

				/**
				 * calculate the matrix of partial derivatives numerically.
				 * @param argument the parameter in the domain of the function (must be a vector).
				 * @param result the Jacobian matrix (rangeDim rows and domainDim columns).
				 */
				inline void operator() ( const AArray<numberT, 1> &argument, AArray<numberT, 2> &result /*outdim, inDim*/ ) const;

			protected:
				const FunctionT &func;
				int inDim;
				int outDim;
				mutable AArray<numberT, 1> val1;
				mutable AArray<numberT, 1> val2;
				mutable AArray<numberT, 1> tmpvalueR;
		};

		template<class FunctionT>
		void LMFunctionJacobi2<FunctionT>::operator() ( const AArray<typename LMFunctionJacobi2<FunctionT>::numberT, 1> &argument,
						AArray<typename LMFunctionJacobi2<FunctionT>::numberT, 2> &result ) const
		{
			assert( (int) result.rows() == outDim );
			assert( (int) result.cols() == inDim );

			double delta = 0;
			func ( argument, val1 );
			tmpvalueR = argument;

			//build jacobian as ratios
			for ( int i = 0; i < inDim; ++i )
			{
				delta = MAX ( fabs ( argument ( i ) * 0.0001 ), 0.000001 ); //magic numbers :)
				tmpvalueR ( i ) = argument ( i ) + delta;
				func ( tmpvalueR, val2 );
				for ( int o = 0; o < outDim; ++o )
				{
					result ( o, i ) = ( val2 ( o ) - val1 ( o ) ) /delta; //differences -> first derivation
				}
				tmpvalueR ( i ) = argument ( i );
			}
		}


	}

}

#endif
