

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_AUTOCALIBRATIONAUTOCALIBRATORLINEAR_H
#define ADORE_AUTOCALIBRATIONAUTOCALIBRATORLINEAR_H

#include <adrcore/util/printmat.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/solvesvd.h>
#include <adrcore/nview/autocalibratorcommon.h>
#include <adrcore/nview/quasiaffinator.h>

namespace adore
{
	namespace core
	{

		int AutocalibratorCommon::computeMetricReconstruction ( const AMatrix &Q, NView::CameraIterator camIt, const int numCams, NView::Point3DIterator ptIt, const int numPts )
		{
			//Q is assumed to be known by now
			//all eigenvalues should be positiv
			decomposeQ ( Q, H );

			computeMetricCamReconstruction ( H, camIt, numCams );

			Hinv = H;
			invert ( Hinv );
			computeMetricPointReconstruction ( Hinv, ptIt, numPts );

#ifndef NDEBUG
			//test, whether H is invertible
			DBGOUT ( DBGALL, "Autocalib: Rank of H, the rectifying homography, is: %d\n", getRank ( H ) );
			invert ( Hinv );
			DBGOUT ( DBGALL, "H | invert(invert(H))\n" );
			printMats ( DBGALL, H, Hinv );
			double scale, var;

			//test, whether KK^T = CQC^T for a metric quadric
			AMatrix Qmetric ( 4, 4 );
			setIdentity ( Qmetric );
			Qmetric ( 3, 3 ) =  0.0;
			AMatrix KK ( 3, 3 );
			AMatrix cQ ( 3, 4 );
			AMatrix cQc ( 3, 3 );
			for ( int i = 0; i < numCams; ++i, ++camIt )
			{
				Camera *cam = *camIt;
				cam->decompose();
				AMatrix &K = cam->getCalibrationMatrix();
				matMul2_BT ( K, K, KK );

				AMatrix &c = cam->getProjMat();
				matMul2 ( c, Qmetric, cQ );
				matMul2_BT ( cQ, c, cQc );
				normalizeExponents ( KK.begin(), KK.end() );
				normalizeExponents ( cQc.begin(), cQc.end() );
				scaleEqualityTest ( KK.begin(), cQc.begin(), KK.size(), scale, var );
				DBGOUT ( DBGNORMAL, "AutoCalib-ScaleEQ-Test: scale: %f, var: %f\n", scale, var );
				DBGOUT ( DBGALL, "AutoCalib: (KK | cQc)\n" );
				printMats ( DBGALL, KK, cQc );
			}

#endif
			return 0;
		}

		int AutocalibratorCommon::computeMetricCamReconstruction ( const AMatrix &H, NView::CameraIterator camIt, const int numCams )
		{

			//now H is known, apply to cams and points
			for ( int i = 0; i < numCams; ++i, ++camIt )
			{
				AMatrix &cam = ( *camIt )->getProjMat();
				tmpCam = cam;
				matMul2 ( tmpCam, H, cam );

				if ( !isRightHand ( cam ) )
				{
					( *camIt )->getProjMat() *= -1;
					DBGOUT( DBGNORMAL, "AutocalibratorCommon: Camera was not right handed - fixing\n" );
				}
				assert ( isRightHand ( cam ) );
			}

			return 0;
		}

		int AutocalibratorCommon::computeMetricPointReconstruction ( const AMatrix &Hinv, NView::Point3DIterator ptIt, const int numPts )
		{
			for ( int i = 0; i < numPts; ++i, ++ptIt )
			{
//				DBGOUT( DBGNORMAL, "point going metric:\n (%f, %f, %f, %f)\n(%f, %f, %f)\n",( *pts3D[i] )[0], ( *pts3D[i] )[1], ( *pts3D[i] )[2], ( *pts3D[i] )[3], ( *pts3D[i] )[0] / ( *pts3D[i] )[3], ( *pts3D[i] )[1] / ( *pts3D[i] )[3], ( *pts3D[i] )[2] / ( *pts3D[i] )[3] );
				tmpPoint = **ptIt;
				matMul2 ( Hinv, tmpPoint, **ptIt );
//				DBGOUT( DBGNORMAL, "point beeing metric:\n (%f, %f, %f, %f)\n(%f, %f, %f)\n",( *pts3D[i] )[0], ( *pts3D[i] )[1], ( *pts3D[i] )[2], ( *pts3D[i] )[3], ( *pts3D[i] )[0] / ( *pts3D[i] )[3], ( *pts3D[i] )[1] / ( *pts3D[i] )[3], ( *pts3D[i] )[2] / ( *pts3D[i] )[3] );
//				getchar();
			}

			return 0;
		}

		int AutocalibratorCommon::decomposeQ ( const AMatrix &Q, AMatrix &H )
		{
			AVector &EVal = store4.D;
			AMatrix &EVec = store4.VT;
			svdVT ( Q, store4 );
			transInPlace ( EVec );

			for ( int row = 0; row < 4; ++row )
			{
				for ( int col = 0; col < 3; ++col )
				{
					H ( row, col ) = EVec ( row, col ) * sqrt ( EVal ( col ) );
				}
				H ( row, 3 ) =  0.0;
			}
			H ( 3, 3 ) =  1.0;

			//to prevent left-handed cams; it's a guess
//			AMatrix sub = H( Range( 0, 2 ), Range( 0, 2 ) );
//			if ( det3( sub ) < 0 )
//			{
//				H *= -1;
//			}

			DBGOUT ( DBGALL, "Q decomposes to the following H with rank %d:\n", getRank ( H ) );
			printMat ( DBGALL, H );
			return 0;
		}

		void AutocalibratorCommon::enforceRank3SVD ( AMatrix &Q )
		{
			AVector &EVal = store4.D;
			AMatrix &EVec = store4.VT;
			svdVT ( Q, store4 );

			//setting the lowest singular value to zero
			EVal ( 3 ) =  0;

			AMatrix &tmp44 = Hinv;
			tmp44 ( 0, Range::all() ) = EVec ( 0, Range::all() ) * EVal ( 0 );
			tmp44 ( 1, Range::all() ) = EVec ( 1, Range::all() ) * EVal ( 1 );
			tmp44 ( 2, Range::all() ) = EVec ( 2, Range::all() ) * EVal ( 2 );
			tmp44 ( 3, Range::all() ) = EVec ( 3, Range::all() ) * EVal ( 3 );
			matMul2_AT ( EVec, tmp44, Q );
		}

		int AutocalibratorCommon::countInvisible( const AMatrix &Q, NView &nv )
		{
			//this one need 3D-points to find the correct solution
			assert( nv.getNumPoints3D() > 0 );
			assert( nv.getNumCameras() == 2 );

			decomposeQ ( Q, H );
			int vote = QuasiAffinator::countInvisible ( nv, H );

			return vote;
		}

		void printQEquation ( AMatrix &Q, NView::CameraIterator camIt, int numCams )
		{
			for ( int i = 0; i < numCams; ++i, ++camIt )
			{
				Camera *c = *camIt;
				assert( c );
				AMatrix QPT(4, 3);
				matMul2_BT( Q, c->getProjMat(), QPT );
				AMatrix KKT(3, 3);
				matMul2( c->getProjMat(), QPT, KKT );
				printf( "KKT for Camera %d:\n", i );
				Camera::normalizeK( KKT );
				printMat( KKT );
			}
		}
	}

}

#endif
