

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "badpointremover.h"
#include <adrcore/util/mathutil.h>

namespace adore
{

	namespace core
	{
		void ReproErrorSorter::init( NView &nview )
		{
			if ( nview.getNumPointCorrs() == 0 ) return;

			repros.resize ( nview.getNumPointCorrs() );
			iterator rit = repros.begin();

			for ( NView::CorrIterator cit = nview.getPointCorrsBegin();
			                cit != nview.getPointCorrsEnd(); ++cit, ++rit )
			{
				double reproErr = 0;
				PointCorr *pc = *cit;
				if ( pc->size() < 1 )
				{
					*rit = ReproT ( pc, 0 );
				}
				else
				{
					NView::Point3D *pt3D = nview.getPoint3D ( pc );

					for ( PointCorr::iterator it = pc->begin(); it != pc->end(); ++it )
					{
						FeaturePoint *fp = *it;
						Camera *cam = nview.getCamera ( fp->getView() );
						if ( !cam )
						{
							DBGOUT( DBGNORMAL, "BadPointRemover: missing Camera - stopping\n" );
							return;
						}
						AVector vec = cam->project ( *pt3D );
						inhomogenize<2> ( vec );
						assert ( vec ( 2 ) == 1 );
						assert ( ( *fp ) ( 2 ) == 1 );
						reproErr += fp->sqrDist ( vec );
//						reproErr = MAX( reproErr, fp->sqrDist ( vec ) );
					}
					reproErr /= double ( pc->size() );
					*rit = ReproT ( pc, reproErr );
				}
			}
			assert ( rit == repros.end() );

			ReproComp compi;
			sort ( repros.begin(), repros.end(), compi );
		}

		void BadPointRemover::modify ( NView &nview )
		{
			//find the largest and the second largest error-gap
			//if the largest ist twice-or-so as large as the second largest
			//kick all corrs that have larger error than the one left to the gap
			repros.init( nview );
			ReproIter rit1 = repros.begin();
			ReproIter rit2 = rit1; ++rit2;
			JumpT large( rit1, 0 );
			JumpT larger( rit1, 0 );
			for ( ; rit2 != repros.end(); ++rit1, ++rit2 )
			{
				double gap = ( *rit2 ).get<1>() - ( *rit1 ).get<1>();
				//assert ( gap > 0 );
				if ( gap > larger.get<1>() )
				{
					large = larger;
					larger = JumpT ( rit2, gap );
				}

				//this is not used because we want the second largest gap to the left only
				//that is because we assume crap to the right of a large gap
//				else if ( gap > large.get<1>() )
//				{
//					large = JumpT( rit2, gap );
//				}
			}

			if ( larger.get<1>() < FLT_EPSILON ) return; //everything is fine
			if ( large.get<1>() / larger.get<1>() > goodRatio ) return; //everything is fine

			DBGOUT( DBGNORMAL, "BadPointRemover: removing %td corrs with errors larger than %f\n", std::distance( larger.get<0>(), repros.end() ), larger.get<1>() );

			for ( ReproIter rit = larger.get<0>(); rit != repros.end(); ++rit )
			{
				PointCorr *pc = ( *rit ).get<0>();
				nview.removePointCorrAnd3D ( pc );
			}
		}


		void BadPointRemover2::modify ( NView &nview )
		{
			removed = 0;
			const int numCorrs = nview.getNumPointCorrs();
			ADORE_UNUSED( numCorrs );
			NView::CorrIterator cit = nview.getPointCorrsBegin();
			while ( cit != nview.getPointCorrsEnd() )
			{
				double reproErr = 0;
				PointCorr *pc = *cit; ++cit;
				if ( pc->size() > 0 )
				{
					NView::Point3D *pt3D = nview.getPoint3D ( pc );
					if ( pt3D )
					{
						for ( PointCorr::iterator it = pc->begin(); it != pc->end(); ++it )
						{
							FeaturePoint *fp = *it;
							Camera *cam = nview.getCamera ( fp->getView() );
							if ( !cam )
							{
								DBGOUT( DBGNORMAL, "BadPointRemover2: missing Camera - stopping\n" );
								return;
							}
							AVector vec = cam->project ( *pt3D );
							inhomogenize<2> ( vec );
							assert ( vec ( 2 ) == 1 );
							assert ( ( *fp ) ( 2 ) == 1 );
							reproErr += fp->sqrDist ( vec );
						}
						reproErr /= double ( pc->size() );

						if ( reproErr > reproThresh )
						{
							nview.removePointCorrAnd3D ( pc );
							++removed;
						}
					}
					else
					{
						DBGOUT( DBGNORMAL, "BPR2: no 3D point for PointCorr %p\n", pc );
					}
				}
			}
			DBGOUT( DBGNORMAL, "BPR2 has removed %d/%d corrs\n", removed, numCorrs );
		}

		void BadPointRemover3::modify ( NView &nview )
		{
			removed = 0;
			const int numCorrs = nview.getNumPointCorrs();
			ADORE_UNUSED( numCorrs );
			NView::CorrIterator cit = nview.getPointCorrsBegin();
			while ( cit != nview.getPointCorrsEnd() )
			{
				PointCorr *pc = *cit; ++cit;
				if ( pc->size() > 0 )
				{
					NView::Point3D *pt3D = nview.getPoint3D ( pc );
					if ( pt3D )
					{

						for ( PointCorr::iterator it = pc->begin(); it != pc->end(); ++it )
						{
							FeaturePoint *fp = *it;
							Camera *cam = nview.getCamera ( fp->getView() );
							if ( !cam )
							{
								DBGOUT( DBGNORMAL, "BadPointRemover3: missing Camera - stopping\n" );
								return;
							}
							if ( cam->getSign ( *pt3D ) <= 0 )
							{
								if ( !pretend ) nview.removePointCorrAnd3D ( pc );
								++removed;
								break;
							}
						}
					}
					else
					{
						DBGOUT( DBGNORMAL, "BPR3: no 3D point for PointCorr %p\n", pc );
					}
				}
			}
			DBGOUT( DBGNORMAL, "BPR3 has removed %d/%d corrs\n", removed, numCorrs );
			//assert that at least half of the points lied in front of the cams
//			assert( ( numCorrs >> 1 ) > removed );
		}


	}

}
