

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "correspondencecorrectorp.h"
#include <adrcore/util/printmat.h>
#include <adrcore/util/floatshift.h>

namespace adore
{
	namespace core
	{

		CorrespondenceCorrectorP::CorrespondenceCorrectorP() :
				B ( 3, 3 ),
				C ( 3, 3 )
		{
		}


		bool CorrespondenceCorrectorP::operator() ( AVector2D &p1, AVector2D &p2, const AMatrix &H )
		{
			assert_msg ( p1.size() == 2 || p1 ( 2 ) == 1.0, "(%f, %f, %f)\n", p1 ( 0 ), p1 ( 1 ), p1 ( 2 ) );
			assert_msg ( p2.size() == 2 || p2 ( 2 ) == 1.0, "(%f, %f, %f)\n", p2 ( 0 ), p2 ( 1 ), p2 ( 2 ) );
			return operator() ( p1 ( 0 ), p1 ( 1 ), p2 ( 0 ), p2 ( 1 ), H );
		}

		bool CorrespondenceCorrectorP::operator() ( double &p1x, double &p1y, double &p2x, double &p2y, const AMatrix &H )
		{
			prepare( H );
			return correct( p1x, p1y, p2x, p2y, H );
		}

		bool CorrespondenceCorrectorP::correct ( AVector2D &p1, AVector2D &p2, const AMatrix &H )
		{
			assert_msg ( p1.size() == 2 || p1 ( 2 ) == 1.0, "(%f, %f, %f)\n", p1 ( 0 ), p1 ( 1 ), p1 ( 2 ) );
			assert_msg ( p2.size() == 2 || p2 ( 2 ) == 1.0, "(%f, %f, %f)\n", p2 ( 0 ), p2 ( 1 ), p2 ( 2 ) );
			return correct ( p1 ( 0 ), p1 ( 1 ), p2 ( 0 ), p2 ( 1 ), H );
		}

		bool CorrespondenceCorrectorP::correct( double &p1x, double &p1y, double &p2x, double &p2y, const AMatrix &H )
		{
			//create a Matrix B, that is H, but incorporates a translation
			//of the points p1 and p2 to the origin
			double b9 = p1x * H( 2, 0 ) + p1y * H( 2, 1 ) + H( 2, 2 );
			B( 0, 0 ) = H( 0, 0 ) - p2x * H( 2, 0 );
			B( 0, 1 ) = H( 0, 1 ) - p2x * H( 2, 1 );
			B( 0, 2 ) = p1x * H( 0, 0 ) + p1y * H( 0, 1 ) + H( 0, 2 ) - p2x * b9;

			B( 1, 0 ) = H( 1, 0 ) - p2y * H( 2, 0 );
			B( 1, 1 ) = H( 1, 1 ) - p2y * H( 2, 1 );
			B( 1, 2 ) = p1x * H( 1, 0 ) + p1y * H( 1, 1 ) + H( 1, 2 ) - p2y * b9;

			B( 2, 0 ) = H( 2, 0 );
			B( 2, 1 ) = H( 2, 1 );
			B( 2, 2 ) = b9;

//			DBGOUT( DBGNORMAL, "CorrespondenceCorrectorP: H and B:\n" );
//			printMats( H, B );

			//apply a rotation matrix to B from the right, to set B(2, 1) to zero
			//this is done by applying cosA and sinA from prepare
			C( 0, 0 ) = cosA * B( 0, 0 ) - sinA * B( 0, 1 );
			C( 0, 1 ) = sinA * B( 0, 0 ) + cosA * B( 0, 1 );
			C( 0, 2 ) = B( 0, 2 );

			C( 1, 0 ) = cosA * B( 1, 0 ) - sinA * B( 1, 1 );
			C( 1, 1 ) = sinA * B( 1, 0 ) + cosA * B( 1, 1 );
			C( 1, 2 ) = B( 1, 2 );

			C( 2, 0 ) = cosA * B( 2, 0 ) - sinA * B( 2, 1 );
			assert_msg( fabs( sinA * B( 2, 0 ) + cosA * B( 2, 1 ) ) < 0.0000001, "it is %f\n", sinA * B( 2, 0 ) + cosA * B( 2, 1 ) );
			C( 2, 1 ) = 0;
			C( 2, 2 ) = B( 2, 2 );

//			printMats( B, C );

			//apply a rotation matrix to C from the left, to set B(1, 0) to zero
			double divisor = sqrt( C( 0, 0 ) * C( 0, 0 ) + C( 1, 0 ) * C( 1, 0 ) );
			double cosC = -C( 0, 0 )/divisor;
			double sinC =  C( 1, 0 )/divisor;

			q1 = C( 0, 0 ) * cosC - C( 1, 0 ) * sinC;
			q2 = C( 0, 1 ) * cosC - C( 1, 1 ) * sinC;
			q3 = C( 0, 2 ) * cosC - C( 1, 2 ) * sinC;

			assert_msg( fabs( C( 0, 0 ) * sinC + C( 1, 0 ) * cosC ) < 0.0000001, "it is %f\n", C( 0, 0 ) * sinC + C( 1, 0 ) * cosC );
			q5 = C( 0, 1 ) * sinC + C( 1, 1 ) * cosC;
			q6 = C( 0, 2 ) * sinC + C( 1, 2 ) * cosC;

			q7 = C( 2, 0 );
			q9 = C( 2, 2 );

			prepareGt();

			//sometimes the floating point limits are reached
			if ( reachedInfty() )
			{
				DBGOUT ( DBGNORMAL, "unusual numbers in CorrespondenceCorrectorP:\n" );
				DBGOUT ( DBGNORMAL, "points were: (%f, %f) and (%f, %f)\n", p1x, p1y, p2x, p2y );
				return false;
			}

			//DBGOUT(DBGALL, "calculating roots of g(t)\n");
			if ( !getRootsOfGt ( rootsR, rootsI ) )
			{
				return false;
			}

			//find the minimum of St by comparing its values at the roots of Gt
			double minx = DBL_MAX;
			double miny = DBL_MAX;
			double minErr = DBL_MAX;
			for ( int i = 0; i < 8; ++i )
			{
				//Evaluate only the real roots
				if ( rootsI[i] == 0.0 )
				{
					double x = rootsR[i];
					double y = getY ( x );
					double err = getError( x, y );
//					printf( "R = %.10f   I = %.10f   err = %.10f  x = %.10f  y = %.10f\n", rootsR[i], rootsI[i], err, x, y );
					if ( err < minErr )
					{
						minErr = err;
						minx = x;
						miny = y;
					}
				}
			}
			if ( minErr == DBL_MAX )
			{
				DBGOUT( DBGALL, "CorrespondenceCorrectorP: no real solution. coeffs were: %f, %f, %f, %f, %f, %f, %f, %f, %f\n", p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8] );
				return false;
			}

			//rotate minx and miny to invert the effect of the roation of B
//			DBGOUT( DBGNORMAL, "p1 before: (%.8f, %.8f)\n", p1x, p1y );
			p1x += minx * cosA + miny * sinA;
			p1y += -minx * sinA + miny * cosA;

#ifndef NDEGUG
//			printf( "min: (%.8f, %.8f)[%.8f]\n", minx, miny, minErr );
//			printf( "p1: (%.8f, %.8f)\n", p1x, p1y );
			double debugp2x = p2x;
			double debugp2y = p2y;
#endif

			//apply H to p1 to get p2;
			divisor = p1x * H( 2, 0 ) + p1y * H( 2, 1 ) + H( 2, 2 );
			if ( fabs( divisor ) < 0.00001 )
			{
				DBGOUT( DBGALL, "CorrespondenceCorrectorP: solution sends p2 to infinity. INVESTIGATE!\n" );
				return false;
			}
//			printf( "divisor: %.10f\n", divisor );
			p2x = ( p1x * H( 0, 0 ) + p1y * H( 0, 1 ) + H( 0, 2 ) )/divisor;
			p2y = ( p1x * H( 1, 0 ) + p1y * H( 1, 1 ) + H( 1, 2 ) )/divisor;

#ifndef NDEGUG
			//apply q to get p2 in an alternative fashion and rotate it back
			//the result must be the same as the above
			divisor = minx * q7 + q9;
			double rightx = ( minx * q1 + miny * q2 + q3 )/divisor;
			double righty = ( miny * q5 + q6 )/divisor;
			debugp2x += rightx * cosC + righty * sinC;
			debugp2y += -rightx * sinC + righty * cosC;

			if ( fabs( p2x - debugp2x ) > 0.0000001 )
			{
				DBGOUT( DBGCRITICAL, "CorrespondenceCorrectorP: p2x: %.8f, debugp2x: %.8f\n", p2x, debugp2x );
				printMats( DBGCRITICAL, H, B );
			}
			if ( fabs( p2y - debugp2y ) > 0.0000001 )
			{
				DBGOUT( DBGCRITICAL, "CorrespondenceCorrectorP: p2y: %.8f, debugp2y: %.8f\n", p2y, debugp2y );
				printMats( DBGCRITICAL, H, B );
			}
#endif

			return true;
		}

		void CorrespondenceCorrectorP::prepare( const AMatrix &H )
		{
			//take the negative, because rot^-1 is needed
			double divisor = sqrt( H( 2, 1 ) * H( 2, 1 ) + H( 2, 0 ) * H( 2, 0 ) );
			cosA = -H( 2, 0 )/divisor;
			sinA = H( 2, 1 )/divisor;
		}

		void CorrespondenceCorrectorP::prepareGt()
		{
//			DBGOUT( DBGALL, "initials: %.10f, %.10f, %.10f, %.10f, %.10f, %.10f\n", f1, f2, a, b, c, d );

			q1_2 = q1 * q1;
			q2_2 = q2 * q2;
			q3_2 = q3 * q3;
			q5_2 = q5 * q5;
			q6_2 = q6 * q6;
			q7_2 = q7 * q7;
			q9_2 = q9 * q9;

			double q7_4 = q7_2 * q7_2;

			double q1q3q7 = q1 * q3 * q7;
			double q1q5q7 = q1 * q5 * q7;
			double q1q5q9 = q1 * q5 * q9;
			double q3q7q9 = q3 * q7 * q9;

			double r = q2_2 + q5_2 + q9_2;
			double r_2 = r * r;
			double t = q3*q5 - q2*q6;
			double t_2 = t * t;

			p[8] = ( q9 * q9_2 ) * ( ( -q3_2 -q6_2 ) * q7 * q9 + q1 * q3 * r ) + q1q5q9 * r * t - q7 * ( q9_2 + r ) * t_2;
			p[7] = -4*q3q7q9*q3q7q9*q9 - 4*q6_2*q7_2*q9_2*q9 + 3*q1*q3q7q9*q9*r + q9*r*( q1_2*( q5_2+q9_2 ) + q9_2*r ) - q1q5q7*r*t - 4*q7_2*q9*t_2;
			p[6] = q7 * ( q9 * ( -6 * ( q3_2 + q6_2 )*q7_2*q9 - q1q3q7*( q9_2 - 3*r ) + 4*q9_2*q9*r + 3*q9*r_2 + q1_2*q9*( q5_2 + q9_2 + 3*r ) ) - 5*q1q5q7*q9*t - 2*q7_2*t_2 );
			p[5] = q7_2*( q9*( -4 * ( q3_2+q6_2 )*q7_2 + 4*q9_2*q9_2 + 14*q9_2*r + 3*r_2 ) + q1_2*( -( q5_2 * q9 ) + 3*q9*( q9_2+r ) ) + q1q3q7*( -3*q9_2 + r ) - 3 * q1q5q7 * t );
			p[4] = q7_2*q7 * ( ( -q3_2 - q6_2 )*q7_2 - 3 * q1q3q7*q9 + 16*q9_2*q9_2 + 18*q9_2*r + r_2 + q1_2*( -q5_2 + 3*q9_2 + r ) );
			p[3] = q7_4*( -q1q3q7 + q1_2*q9 + 25*q9_2*q9 + 10*q9*r );
			p[2] = q7_4*q7*( 19*q9_2 + 2*r );
			p[1] = 7*q7_4*q7_2*q9;
			p[0] = q7_4*q7_2*q7;
		}

		/*
		evaluates the function g(t) of algo 12.1 (HZ_Book)
		The variables are named along this algo, so don't blame me
		*/
		bool CorrespondenceCorrectorP::getRootsOfGt ( double rootsR[8], double rootsI[8] )
		{
			int degree = 8;
//			DBGOUT( DBGNORMAL, "roots of: %.10f, %.10f, %.10f, %.10f, %.10f, %.10f, %.10f\n", t[0], t[1], t[2], t[3], t[4], t[5], t[6] );
			bool success = rpoly ( p, degree, rootsR, rootsI );
			//DBGOUT(DBGALL, "success of rootfinding: %s, degree: %d\n", success ? "true" : "false", degree);
			if ( !success )
			{
				DBGOUT ( DBGALL, "Rootfinding failed. Degree is: %d\n", degree );
//				DBGOUT( DBGALL, "roots of: %.10f, %.10f, %.10f, %.10f, %.10f, %.10f, %.10f\n", t[0], t[1], t[2], t[3], t[4], t[5], t[6] );
//				pressKey();
			}
			return success;
		}

		double CorrespondenceCorrectorP::getY ( double x )
		{
//			double dividend = q2*q3 + q5*q6 + q1*q2*x + q4*q5*x;
			double dividend = q2*q3 + q5*q6 + q1*q2*x;
			double divisor = q2_2 + q5_2 + q9_2 + 2*q7*q9*x + q7_2*x*x;
			return -dividend/divisor;
		}

		double CorrespondenceCorrectorP::getError( double x, double y )
		{
			double divisor = q7*x+q9;
			double e1 = ( q1*x+q2*y+q3 )/divisor;
			double e2 = ( q5*y+q6 )/divisor;
			return x*x + y*y + e1*e1 + e2*e2;
		}

	}
}
