

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "globalbundleadjuster.h"

#include <adrcore/nview/baprojective.h>
#include <adrcore/test/assessor.h>
#include <adrcore/nview/ttensorcalculator.h>
#include <adrcore/nview/fmatrixcalculator.h>
#include <adrcore/mestimate/errorfunctors2views.h>

using namespace boost::tuples;

using namespace std;

namespace adore {
	namespace core {

	class SolutionDummy {
		AMatrix fmat;
	public:
		SolutionDummy(const AMatrix& fm) : fmat(fm) {}
		const AMatrix& getFMat() const { return fmat; }
		AMatrix& getFMat() { return fmat; }
		const AVector& getCorrectedPoint(const AVector& v) { return v; }
		const AVector& getUnCorrectedPoint(const AVector& v) { return v; }
	};

	GlobalBundleAdjuster::GlobalBundleAdjuster(uint num_initial_it, double thresh_ratio, bool use_2_corrs) :
		m_num_initial_it(num_initial_it),
		m_thresh_ratio(thresh_ratio),
		m_use_2_corrs(use_2_corrs)
	{
		m_num_initial_it = std::max(m_num_initial_it, 8u);
		m_thresh_ratio = std::max(m_thresh_ratio, 0.01);
		m_thresh_ratio = std::min(m_thresh_ratio, 1.0);
	}

	void GlobalBundleAdjuster::adjust(NViewMap& map)
	{
		BAProjectiveSimple BA;
		BAProjectiveSimpleBlakeZiss BA2; // another cost function.
		Assessor assessor;

		NView * large_nv = stitchAllTripplets(map);
		if(large_nv == NULL) {
			DBGOUT(DBGALL, "GlobalBundleAdjuster failed to stitch all triplets in one NView.\n");
			return;
		}

		// create cameras and 3d points.
		init3D(large_nv);

		// make one guess.
		BA.setMaxIterations(m_num_initial_it);
		BA.modify(*large_nv);
		assessor.assess( *large_nv );
		double sum1 = assessor.getReproErrorSum();
		DBGOUT(DBGNORMAL, "reproerrorsum 1: %f\n", sum1);

		const uint num_all_views = large_nv->getNumViews();
		vector<Pt4_t> first_points(large_nv->getNumPointCorrs());
		vector<C3x4_t> first_cams(num_all_views);

		storeInitialGuessAndReinit(large_nv, first_points, first_cams);

		// make another guess.
		BA.modify(*large_nv);
		assessor.assess( *large_nv );
		double sum2 = assessor.getReproErrorSum();
		DBGOUT(DBGNORMAL, "reproerrorsum 2: %f\n", sum2);

		// if first guess is better as second...
		if(sum1 < sum2) replayInitialGuess(large_nv, first_points, first_cams);

		double threshold = getNormalizedPixel(1.5, *large_nv->getViewsBegin());
		removeBadPoints(large_nv, threshold);

		// try another cost function and smaller threshold.
		BA2.setMaxIterations(1000);
		threshold *= m_thresh_ratio;
		BA2.modify(*large_nv);
		assessor.assess( *large_nv );
		DBGOUT(DBGNORMAL, "reproerrorsum: %f\n", assessor.getReproErrorSum());
		removeBadPoints(large_nv, threshold);
	}

	NView * GlobalBundleAdjuster::stitchAllTripplets(NViewMap& map)
	{
		vector<ViewBase*> all_views;
		vector<NView*> tripplets;

		for(NViewMap::EntryIterator eit = map.getEntriesBegin(); eit != map.getEntriesEnd(); ++eit)
		{
			NView * nv = *eit;
			if(nv->getNumViews() != 3) continue;

			if(all_views.empty()) {
				all_views.insert(all_views.begin(), nv->getViewsBegin(), nv->getViewsEnd());
				tripplets.push_back(nv);
				continue;
			}

			ViewBase * v1 = *nv->getViewsBegin();
			ViewBase * v2 = *(++nv->getViewsBegin());
			ViewBase * v3 = *(++(++nv->getViewsBegin()));
			if(all_views[all_views.size() - 2] != v1 || all_views.back() != v2) // TODO?
			{
				return NULL;
			}
			all_views.push_back(v3);
			tripplets.push_back(nv);
		}

		NView * result = map.createNView( all_views.begin(), all_views.end() );
		if(!result) return NULL;

		for(uint i = 0; i < tripplets.size(); ++i)
		{
			result->addPointCorrs( *tripplets[i] );
		}

		if(!m_use_2_corrs) return result;

		for(NViewMap::EntryIterator eit = map.getEntriesBegin(); eit != map.getEntriesEnd(); ++eit)
		{
			NView * nv2 = *eit;
			if(nv2->getNumViews() != 2) continue;
			ViewBase * v1_2 = *nv2->getViewsBegin();
			ViewBase * v2_2 = *(++nv2->getViewsBegin());

			NView * nv3 = NULL;
			bool is_second = false;
			for(uint i = 0; i < tripplets.size(); ++i) {
				NView * nv = tripplets[i];
				ViewBase * v1 = *nv->getViewsBegin();
				ViewBase * v2 = *(++nv->getViewsBegin());
				ViewBase * v3 = *(++(++nv->getViewsBegin()));

				if(v1 == v1_2 && v2 == v2_2) {
					nv3 = nv;
					break;
				} else if(v2 == v1_2 && v3 == v2_2) {
					nv3 = nv;
					is_second = true;
					break;
				}
			}
			if(!nv3) continue;

			NViewInfoTT * nvinfo = dynamic_cast<NViewInfoTT*>(nv3->getInfo());
			if(!nvinfo) continue;
			ATensor3 tensor(3, 3, 3);
			tensor = nvinfo->getTTensor();
			if(isNullArray(tensor)) continue;

			TTensorCalculator tcalc;
			AVector2D epi1 = tcalc.calculateEpipole(tensor, true);
			AVector2D epi2 = tcalc.calculateEpipole(tensor, false);

			AMatrix fmat(3, 3);
			if(is_second) fmat = tcalc.calculateFundamentalMatrix(tensor, epi1, epi2, TTensorCalculator::FMATRIX_F32);
			else fmat = tcalc.calculateFundamentalMatrix(tensor, epi1, epi2, TTensorCalculator::FMATRIX_F21);

			AVector epipole(3); // epipole in 2nd image
			transInPlace( fmat );
			FMatrixCalculator fcalc;
			fcalc.calculateEpipolarPoint(fmat, epipole);
			transInPlace( fmat );
			inhomogenize<2>( epipole );

			//re-orient fmt in case is has been optimized and lost its sign
			for ( NView::CorrIterator cit = nv2->getPointCorrsBegin(); cit != nv2->getPointCorrsEnd(); ++cit )
			{
				PointCorr *pc = *cit;
				FeaturePoint *fp1 = pc->getFeature( v1_2 );
				FeaturePoint *fp2 = pc->getFeature( v2_2 );
				bool io1 = FMatrixCalculator::isOriented ( fmat, epipole, *fp1, *fp2 );
				bool io2 = FMatrixCalculator::isOriented ( -fmat, epipole, *fp1, *fp2 );
				if ( io1 != io2 )
				{
					if ( io2 ) fmat *= -1;
					break;
				}
			}

			typedef tuple<PointCorr*, FeaturePoint*, FeaturePoint*> CorrT;
			vector<CorrT> corrs;
			corrs.reserve(nv2->getNumPointCorrs());
			for(NView::CorrIterator cit = nv2->getPointCorrsBegin(); cit != nv2->getPointCorrsEnd(); ++cit) {
				PointCorr * pc = *cit;
				FeaturePoint* fp1 = pc->getFeature(v1_2);
				FeaturePoint* fp2 = pc->getFeature(v2_2);
				if(!fp1 || !fp2) continue;
				corrs.push_back(make_tuple(pc, fp1, fp2));
			}

			SolutionDummy sd(fmat);
			FundamentalErrorsWithCorrector<SolutionDummy> filter;
			vector<int> inliers;
			inliers.reserve(nv2->getNumPointCorrs());

			const double thresh2 = 0.5;
			double err = filter(sd, epipole, getNormalizedPixel(thresh2, v1_2), inliers, corrs, corrs.size());
			ADORE_UNUSED( err );

			DBGOUT(DBGNORMAL, "error: %f, num_inliers: %zd of %zd\n", err, inliers.size(), corrs.size());
			vector<FeaturePoint*> corr(2);
			for(uint i = 0; i < inliers.size(); ++i) {
				corr[0] = corrs[inliers[i]].get<1>();
				corr[1] = corrs[inliers[i]].get<2>();
				result->addPointCorr(corr.begin(), corr.end());
			}
		}

		return result;
	}

	void GlobalBundleAdjuster::init3D(NView * nview)
	{
		const uint num_all_views = nview->getNumViews();
		vector<ViewBase*> views(num_all_views);
		nview->getViews(views.begin());
		for(uint i = 0; i < num_all_views; ++i) {
			Camera * cam = nview->getCamera( views[i], true );
			setIdentity( cam->getProjMat() );
			cam->decompose();
			cam->compose();
		}

		for(NView::CorrIterator cit = nview->getPointCorrsBegin(); cit != nview->getPointCorrsEnd(); ++cit)
		{
			PointCorr * pc = *cit;
			AVector3D * p3d = new AVector3D(4);
			FeaturePoint * p2d = *(pc->getBegin());
			(*p3d) = (*p2d)(0), (*p2d)(1), 1.0, 1.0;
			nview->setPoint3D(p3d, pc);
		}
	}

	void GlobalBundleAdjuster::storeInitialGuessAndReinit(NView * nview, std::vector<Pt4_t>& first_points, std::vector<C3x4_t>& first_cams)
	{
		uint i = 0;
		for(NView::CorrIterator cit = nview->getPointCorrsBegin(); cit != nview->getPointCorrsEnd(); ++cit, ++i)
		{
			PointCorr * pc = *cit;
			AVector3D * p3d = nview->getPoint3D(pc);
			first_points[i] = make_tuple( (*p3d)(0), (*p3d)(1), (*p3d)(2), (*p3d)(3) );

			FeaturePoint * p2d = *(pc->getBegin());
			(*p3d) = (*p2d)(0), (*p2d)(1), -1.0, 1.0;
		}

		i = 0;
		for(NView::ViewIterator vit = nview->getViewsBegin(); vit != nview->getViewsEnd(); ++vit, ++i) {
			Camera * cam = nview->getCamera( *vit, false );
			const AMatrix& mat = cam->getProjMat();
			C3x4_t& c = first_cams[i];
			c.get<0>() = make_tuple( mat(0, 0), mat(0, 1), mat(0, 2), mat(0, 3) );
			c.get<1>() = make_tuple( mat(1, 0), mat(1, 1), mat(1, 2), mat(1, 3) );
			c.get<2>() = make_tuple( mat(2, 0), mat(2, 1), mat(2, 2), mat(2, 3) );

			setIdentity( cam->getProjMat() );
			cam->decompose();
			cam->compose();
		}
	}

	void GlobalBundleAdjuster::replayInitialGuess(NView * nview, const std::vector<Pt4_t>& first_points, const std::vector<C3x4_t>& first_cams)
	{
		uint i = 0;
		for(NView::CorrIterator cit = nview->getPointCorrsBegin(); cit != nview->getPointCorrsEnd(); ++cit, ++i)
		{
			PointCorr * pc = *cit;
			AVector3D * p3d = nview->getPoint3D(pc);
			const Pt4_t& p = first_points[i];
			(*p3d) = p.get<0>(), p.get<1>(), p.get<2>(), p.get<3>();
		}

		i = 0;
		for(NView::ViewIterator vit = nview->getViewsBegin(); vit != nview->getViewsEnd(); ++vit, ++i) {
			Camera * cam = nview->getCamera( *vit, false );
			AMatrix& mat = cam->getProjMat();
			const C3x4_t& c = first_cams[i];
			const Pt4_t& p0 = c.get<0>();
			mat(0, Range::all()) = p0.get<0>(), p0.get<1>(), p0.get<2>(), p0.get<3>();
			const Pt4_t& p1 = c.get<1>();
			mat(1, Range::all()) = p1.get<0>(), p1.get<1>(), p1.get<2>(), p1.get<3>();
			const Pt4_t& p2 = c.get<2>();
			mat(2, Range::all()) = p2.get<0>(), p2.get<1>(), p2.get<2>(), p2.get<3>();

			cam->decompose();
			cam->compose();
		}
	}

	void GlobalBundleAdjuster::removeBadPoints(NView * nview, double threshold)
	{
		ViewErrors errors;
		vector< pair<FeaturePoint*, double> > errs;
		errs.reserve(nview->getNumPointCorrs() * nview->getNumViews());

		uint removed = 0;
		for(NView::ViewIterator vit = nview->getViewsBegin(); vit != nview->getViewsEnd(); ++vit)
		{
			errs.clear();
			errors.computeErrors(*vit, *nview);
			errors.getReproErrorsAndFeatures(errs);
			for(uint i = 0; i < errs.size(); ++i) {
				if(errs[i].second > threshold) {
					FeaturePoint * fp = errs[i].first;
					nview->removePointCorrAnd3D(fp);
					++removed;
				}
			}
		}
		DBGOUT(DBGNORMAL, "removed: %d\n", removed);
	}

	} // core
} // adore
