

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "oprcalculator.h"
#include <adrcore/util/printmat.h>
#include <adrcore/util/solvesvd.h>

namespace adore
{

	namespace core
	{

		OPRCalculator::OPRCalculator() :NViewOperator(), metricOnly ( true )
		{
		}

		OPRCalculator::OPRCalculator( const OPRCalculator &opr ):NViewOperator( opr ), metricOnly ( opr.metricOnly )
		{
		}

		OPRCalculator::~OPRCalculator()
		{
		}

		bool OPRCalculator::isQuasiAffine ( NView &nview )
		{
			vector<Camera *> cams( nview.getNumCameras() );
			vector<AVector3D *> pts3D( nview.getNumPoints3D() );

			nview.getCameras ( cams.begin() );
			nview.getPoints3D ( pts3D.begin() );

			return isQuasiAffine ( pts3D, cams );
		}

		bool OPRCalculator::isQuasiAffine ( vector<AVector3D *> &pts3D, vector<Camera *> &cams )
		{
			if ( pts3D.size() < 1 && cams.size() < 1 ) return true;

			vector<double> z ( cams.size() * pts3D.size() );
			vector<int> d ( cams.size() );
			fillZ ( cams, pts3D, z );

			bool success = true;
			d[0] = 1;
			for ( uint p = 0; p < pts3D.size(); ++p )
			{
				double val = d[0] * z[p] * ( *pts3D[p] ) ( 3 );
				if ( val <= 0 )
				{
					success = false;
					break;
				}
			}
			if ( !success ) d[0] = -1;

			int k = 0;
			for ( uint c = 1; c < cams.size(); ++c ) //0 is already set
			{
				d[c] = 1;
				for ( int i = 0; i < 2; ++i )
				{
					int oldK = k;
					success = true;

					for ( uint p = 0; p < pts3D.size(); ++p, k++ )
					{
						double res  = d[c] * z[k] * ( *pts3D[p] ) ( 3 );
						if ( res <= 0 )
						{
							success = false; break;
						}
					}

					if ( success )
					{
						break;
					}
					else
					{
						d[c] = -1;
						k = oldK;
					}
				}
				if ( !success ) return false;
			}
			return true;
		}

		void OPRCalculator::modify ( NView &nview )
		{
			if ( nview.getNumCameras() < 1 ) return;

			if ( !metricOnly )
			{
				vector<Camera *> cams( nview.getNumCameras() );
				vector<AVector3D *> pts3D( nview.getNumPoints3D() );

				nview.getCameras ( cams.begin() );
				nview.getPoints3D ( pts3D.begin() );

				if ( isQuasiAffine ( pts3D, cams ) )
				{
					DBGOUT ( DBGALL, "NView is already affine\n" );
					return;
				}

				AMatrix H ( 4, 4 );
				findTrafo ( cams, pts3D, H );
				printMat ( H );
				assert ( getRank ( H ) == 4 );

				apply ( H, pts3D );

				AMatrix Hinv ( 4, 4 );
				setIdentity ( Hinv );
				Hinv ( 3, 0 ) = -H ( 3, 0 );
				Hinv ( 3, 1 ) = -H ( 3, 1 );
				Hinv ( 3, 2 ) = -H ( 3, 2 );

				printMat ( Hinv );
				apply ( Hinv, cams );

				assert_msg ( isQuasiAffine ( pts3D, cams ), "OPR was failed to achieve\n" );
			}

			makeScaleFactorsPositive ( nview );
			putPointsInFrontMetric ( nview );
		}

		bool OPRCalculator::findTrafo ( vector<Camera *> &cams, vector<AVector3D *> &pts3D, AMatrix &H )
		{
			setIdentity ( H );

			vector<double> z ( cams.size() * pts3D.size() );
			vector<int> d ( cams.size() );
			fillZ ( cams, pts3D, z );

			AMatrix H4 ( 4, 4 );
			H4 = 0;

			bool success = false;
			for ( int r = -1; r <= 1; r += 2 )
				for ( int s = -1; s <= 1; s += 2 )
					for ( int t = -1; t <= 1; t += 2 )
					{
						H4 ( 0, 0 ) =  r;
						H4 ( 0, 1 ) =  s;
						H4 ( 0, 2 ) =  t;
						findInitialD ( pts3D, z, d, H4 );
						if ( goodTransformation ( pts3D, z, d, H4 ) )
						{
							success =  true;
							goto finish;
						}
					}
			H4 = 0;

finish:
			H ( 3, 0 ) = H4 ( 0, 0 );
			H ( 3, 1 ) = H4 ( 0, 1 );
			H ( 3, 2 ) = H4 ( 0, 2 );
			return success;
		}

		void OPRCalculator::fillZ ( vector<Camera *> &cams, vector<AVector3D *> &pts3D, vector<double> &z )
		{
			int k = 0;
			for ( uint c = 0; c < cams.size(); ++c )
			{
				AMatrix &cam = cams[c]->getProjMat();
				for ( uint p = 0; p < pts3D.size(); ++p )
				{
					AVector3D &pt = *pts3D[p];
					z[k++] = cam ( 2, 0 ) * pt ( 0 ) + cam ( 2, 1 ) * pt ( 1 ) + cam ( 2, 2 ) * pt ( 2 ) + cam ( 2, 3 ) * pt ( 3 );
				}
			}
		}

		bool OPRCalculator::goodTransformation ( vector<AVector3D *> &pts3D,
		                vector<double> &z, vector<int> &d, AMatrix &H4 )
		{
			uint cams = z.size() / pts3D.size();
			int k = 0;
			for ( uint c = 1; c < cams; ++c ) //0 is already set
			{
				bool success = true;
				d[c] = 1;
				for ( int i = 0; i < 2; ++i )
				{
					int oldK = k;
					success = true;

					for ( uint p = 0; p < pts3D.size(); ++p, k++ )
					{
						AVector3D &pt = *pts3D[p];
						double val = H4 ( 0, 0 ) * pt ( 0 ) + H4 ( 0, 1 ) * pt ( 1 ) + H4 ( 0, 2 ) * pt ( 2 ) + pt ( 3 );
						double res  = d[c] * z[k] * val;
						if ( res <= 0 )
						{
							success = false; break;
						}
					}

					if ( success )
					{
						break;
					}
					else
					{
						d[c] = -1;
						k = oldK;
					}
				}
				if ( !success ) return false;
			}
			return true;
		}

		void OPRCalculator::findInitialD ( vector<AVector3D *> &pts3D, vector<double> &z, vector<int> &d, AMatrix& H4 )
		{
			bool ret = true;
			d[0] = 1;
			for ( uint p = 0; p < pts3D.size(); ++p )
			{
				AVector3D &pt = *pts3D[p];
				double val = d[0] * z[p] * ( H4 ( 0, 0 ) * pt ( 0 ) + H4 ( 0, 1 ) * pt ( 1 ) + H4 ( 0, 2 ) * pt ( 2 ) + pt ( 3 ) );
				if ( val <= 0 )
				{
					ret = false;
					break;
				}
			}
			if ( !ret ) d[0] = -1;
		}

		void OPRCalculator::apply ( AMatrix &H, vector<Camera *> &cams )
		{
			AMatrix temp34(3, 4);
			//now H is known, apply to cams and points
			for ( uint i = 0; i < cams.size(); ++i )
			{
				matMul2 ( cams[i]->getProjMat(), H, temp34 );
				cams[i]->getProjMat() = temp34;
			}
		}

		void OPRCalculator::apply ( AMatrix &Hinv, vector<AVector3D *> &pts3D )
		{
			AVector temp4(4);
			for ( uint i = 0; i < pts3D.size(); ++i )
			{
				matMul2 ( Hinv, *pts3D[i], temp4 );
				*pts3D[i] = temp4;
			}
		}

		bool OPRCalculator::pointsInFront ( NView &nview )
		{
			int success = true;
			int camIdx = 0;
			for ( NView::CameraIterator cit = nview.getCamerasBegin(); cit != nview.getCamerasEnd(); ++cit, ++camIdx )
			{
				int forbidden = 0;
				Camera *cam = *cit;
				if ( !cam ) return false;

				for ( NView::Point3DIterator pit = nview.getPoints3DBegin(); pit != nview.getPoints3DEnd(); ++pit )
				{
					if ( cam->getSign ( **pit ) <= 0 ) ++forbidden;
				}
				if ( forbidden )
				{
					printf ( "%d of %d Points are NOT in front of Camera %d\n", forbidden, nview.getNumPoints3D(), camIdx );
					success = false;
				}
			}
			return success;
		}

		/*
		TODO
		This method needs to be fixed!
		When testing points if they're in front of a camera, all points are testes.
		Instead, only points visible to this cam must be tested
		*/
		void OPRCalculator::makeScaleFactorsPositive ( NView &nview )
		{
			set<Camera *> camsdone;
			set<Camera *> camstodo;
			set<AVector3D *> pts3Ddone;
			set<AVector3D *> pts3Dtodo;

			DBGOUT ( DBGALL, "Modifying nview with %d views\n", nview.getNumViews() );

			NView::CameraIterator cit = nview.getCamerasBegin();
			//fix the first cam
			if ( !isRightHand ( ( *cit )->getProjMat() ) )
			{
				DBGOUT ( DBGALL, "First cam was left-handed -- corrected\n" );
				( *cit )->getProjMat() = - ( *cit )->getProjMat();
			}
			for ( ; cit != nview.getCamerasEnd(); ++cit )
			{
				camstodo.insert ( *cit );
			}

			for ( NView::Point3DIterator pit = nview.getPoints3DBegin();
			                pit != nview.getPoints3DEnd(); ++pit )
			{
				pts3Dtodo.insert ( *pit );
			}



			bool failed = false;
			bool incomplete = true;
			while ( incomplete )
			{
				uint numCamsDone = camsdone.size();
				uint numPtsDone = pts3Ddone.size();
				fixPoints ( camsdone, pts3Dtodo, pts3Ddone, nview );
				fixCams ( pts3Ddone, camstodo, camsdone, nview );

				if ( camsdone.size() == nview.getNumViews() && pts3Ddone.size() == nview.getNumPoints3D() )
				{
					//success!
					incomplete = false;
					failed = false;
				}
				else if ( camsdone.size() == numCamsDone && numPtsDone == pts3Ddone.size() )
				{
					//no improvement - stuck!
					incomplete = false;
					failed = true;
				}
			}

			if ( failed )
			{
				printf ( "got stuck while making scale factors positive\n" );
			}

#ifndef NDEBUG
			//test the result
			int camcount = 0;
			for ( set<Camera *>::iterator c = camsdone.begin(); c != camsdone.end(); ++c, ++camcount )
			{
				Camera *cam = *c;
				ViewBase *v = nview.getView ( cam );
				int ptcount = 0;
				for ( set<AVector3D *>::iterator p = pts3Dtodo.begin(); p != pts3Dtodo.end(); ++p, ++ptcount )
				{
					AVector3D *pt = *p;
					if ( nview.getFeaturePoint ( pt, v ) )
					{
						AVector2D vector3 = cam->project ( *pt );
						if ( vector3 ( 2 ) < 0 )
							printf ( "failed at camera %d and point %d\n", camcount, ptcount );
					}
				}
			}
#endif
		}

		void OPRCalculator::fixCams ( set<AVector3D *> &pts3Ddone, set<Camera *> &camstodo, set<Camera *> &camsdone, NView &nview )
		{
			printf ( "fixing cams accoding to %zd of %d points\n", pts3Ddone.size(), nview.getNumPoints3D() );

			for ( set<AVector3D *>::iterator p = pts3Ddone.begin(); p != pts3Ddone.end(); ++p )
			{
//				printf( "next point - %d cameras to check\n", camstodo.size() );
				AVector3D *pt = *p;
				int count = 0;
				for ( set<Camera *>::iterator c = camstodo.begin(); c != camstodo.end(); ++count )
				{
//					printf( "cam %d of %d\n", count, camstodo.size() );
					Camera *cam = *c;
					ViewBase *v = nview.getView ( cam );

					if ( nview.getFeaturePoint ( pt, v ) )
					{
//						printf( "known!\n" );
						AVector2D vector3 = cam->project ( *pt );
						if ( vector3 ( 2 ) < 0 )
						{
//							printf( "cam needs fixing\n" );
							cam->getProjMat() = -cam->getProjMat();
						}

						set<Camera *>::iterator cdel = c; ++c;
						camstodo.erase ( cdel );
						camsdone.insert ( cam );
					}
					else ++c;
				}
			}
		}

		void OPRCalculator::fixPoints ( set<Camera *> &camsdone, set<AVector3D *> &pts3Dtodo, set<AVector3D *> &pts3Ddone, NView &nview )
		{
			printf ( "fixing points accoding to %zd of %d cams\n", camsdone.size(), nview.getNumViews() );

			for ( set<Camera *>::iterator c = camsdone.begin(); c != camsdone.end(); ++c )
			{
//				printf( "next cam - %d points to check\n", pts3Dtodo.size() );
				Camera *cam = *c;
				ViewBase *v = nview.getView ( cam );
				int count = 0;
				for ( set<AVector3D *>::iterator p = pts3Dtodo.begin(); p != pts3Dtodo.end(); ++count )
				{
//					printf( "point %d of %d\n", count, pts3Dtodo.size() );
					AVector3D *pt = *p;
					if ( nview.getFeaturePoint ( pt, v ) )
					{
//						printf( "known!\n" );
						AVector2D vector3 = cam->project ( *pt );
						if ( vector3 ( 2 ) < 0 )
						{
//							printf( "point needs fixing\n" );
							( *pt ) ( 0 ) *= -1.0;
							( *pt ) ( 1 ) *= -1.0;
							( *pt ) ( 2 ) *= -1.0;
							( *pt ) ( 3 ) *= -1.0;
						}

						set<AVector3D *>::iterator pdel = p; ++p;
						pts3Dtodo.erase ( pdel );
						pts3Ddone.insert ( pt );
					}
					else ++p;
				}
			}
		}

		void OPRCalculator::putPointsInFrontMetric ( NView &nview )
		{
			if ( nview.getNumCameras() == 0 || nview.getNumPoints3D() == 0 ) return;

			Camera *cam = *nview.getCamerasBegin();
			AVector *pt = *nview.getPoints3DBegin();
			if ( cam && pt && cam->getSign ( *pt ) <= 0 )
			{
				for ( NView::CameraIterator cit = nview.getCamerasBegin(); cit != nview.getCamerasEnd(); ++cit )
				{
					Camera *c = *cit;
					AMatrix &cam = c->getProjMat();
					cam ( 0, 3 ) = -cam ( 0, 3 );
					cam ( 1, 3 ) = -cam ( 1, 3 );
					cam ( 2, 3 ) = -cam ( 2, 3 );
				}
				for ( NView::Point3DIterator pit = nview.getPoints3DBegin(); pit != nview.getPoints3DEnd(); ++pit )
				{
					AVector3D *p = *pit;
					( *p ) ( 3 ) = - ( *p ) ( 3 );
				}

			}
		}

	}
}
