

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "ransacclientmergercp.h"
#include <adrcore/util/matmul2.h>
#include <adrcore/util/printmat.h>
#include <adrcore/nview/merger.h>
#include <adrcore/nview/homographycalculator.h>

namespace adore
{

	namespace core
	{

		RANSACClientMergerCP::RANSACClientMergerCP()
				: RANSACInterface(), oneViewCase( false ),
				fractionThresh( 0.05 ),
				dummyCams( 0 ),
				cams1( &dummyCams ),
				cams2( &dummyCams ),
				pts3D1( NULL ),
				pts3D2( NULL ),
				H( 4, 4 ),
				pt22( 4 )
		{
		}


		RANSACClientMergerCP::~RANSACClientMergerCP()
		{
		}

		void RANSACClientMergerCP::computeError( int resultIdx, vector<int> &inlierFlags, vector<double> &distVals )
		{

//			TODO!!!
			resultIdx = resultIdx; // not used

			int numIn = 0;
			double err = 0.0;

			vector<double> medDist;
			medDist.reserve( pts3D1->size() );

			for ( uint p = 0; p < pts3D1->size(); ++p )
			{
				inlierFlags[p] = false;
//				assert( false );

				AVector3D *p3D2 = ( *pts3D2 )[p];

				double w1 = ( *p3D2 )( 3 );
				if ( fabs( w1 ) < 8 * FLT_EPSILON )
				{
					printf( "RANSACClientMergerCP: fabs(w1) < 8 * FLT_EPSILON\n" );
					inlierFlags[p] = false;
					continue;
				}

				matMul2( H, *p3D2, pt22 );

				double w2 = - pt22( 3 );
				if ( fabs( w2 ) < 8 * FLT_EPSILON )
				{
					printf( "RANSACClientMergerCP: fabs(w2) < 8 * FLT_EPSILON; H - pt2 | pt22:\n" );
					printMat( H );
					printMats( *p3D2, pt22 );
					inlierFlags[p] = false;
					continue;
				}

				double x1 = ( *p3D2 )( 0 ) / w1;
				double y1 = ( *p3D2 )( 1 ) / w1;
				double z1 = ( *p3D2 )( 2 ) / w1;
				double x2 = pt22( 0 ) / w2;
				double y2 = pt22( 1 ) / w2;
				double z2 = pt22( 2 ) / w2;
				double dist = ( x1 - x2 ) * ( x1 - x2 ) + ( y1 - y2 ) * ( y1 - y2 ) + ( z1 - z2 ) * ( z1 - z2 );

				inlierFlags[p] = true;
				distVals[p] = dist;
				medDist.push_back( dist );

				numIn++;
			}

//			printf( "number of inliers (1); %d\n", numIn );

			if ( numIn > 0 )
			{
				//median berechnen
				double median = 0.0;
				sort( medDist.begin(), medDist.end() );
				int half = numIn >> 1;
				if ( half << 1 == numIn )
				{
					//gerade
					median = 0.5 *( medDist[half] + medDist[half - 1] );
				}
				else
				{
					median = medDist[half];
				}

				//all distances inside (1.0 + fractionThresh) * median mark inliers
				double thresh = median * ( 1.0 + fractionThresh );
				for ( uint p = 0; p < pts3D1->size(); ++p )
				{
					if ( !inlierFlags[p] )
					{
						--numIn;
						inlierFlags[p] = false;
					}
					else if ( distVals[p] > thresh )
					{
						--numIn;
						inlierFlags[p] = false;
						err += thresh;
					}
					else err += distVals[p];
				}
			}

//			printf( "number of inliers (2); %d\n", numIn );
			setNumIn( numIn );
			setError( err );
		}

		bool RANSACClientMergerCP::checkPrerequisitsAndInit( int numInliers, const vector< int > & )
		{
			return ( cams1->size() * 11 + numInliers * 3 >= 16 );
		}

		int RANSACClientMergerCP::applySample( const vector<int> &sample )
		{
			assert( cams1->size() * 11 + sample.size() * 3 >= 16 );

			vector< AVector3D* > pts1( sample.size() );
			vector< AVector3D* > pts2( sample.size() );

			//copy sample
			for ( uint i = 0; i < sample.size(); ++i )
			{
				pts1[i] = ( *pts3D1 )[sample[i]];
				pts2[i] = ( *pts3D2 )[sample[i]];
			}

//			printf( "sample size: %d\n", sample.size() );

			if ( oneViewCase )
			{
				HomographyCalculator hcalc;
				H = hcalc.calculateHomographyAlgebraic( *( *cams1 )[0], *( *cams2 )[0], pts1, pts2 );
				return 1;
			}
			else
			{
				H = Merger::calculateHomographyFromCamerasAndPoints( *cams1, *cams2, pts1, pts2 );
				return 1;
			}
			return 0;
		}

		void RANSACClientMergerCP::ransacDone()
		{
			//nothing;
		}

		bool RANSACClientMergerCP::optimize( int /*bestIdx*/ )
		{
			return false;
		}

	}
}
