

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "rdfmatrixcalculator.h"
#include <adrcore/nview/correspondencecorrector.h>
#include <adrcore/util/divisiondistortion.h>

namespace adore
{

	namespace core
	{

		extern AMatrix NullMatrix;
		const float lambdaInaccurateThresh = 0.0001;/*tested value, do not touch*/

		void testSolutions( int n,
		                    const AArray<double, 2> &A,
		                    const AArray<double, 2> &B,
		                    const AArray<double, 1> &lambdas,
		                    const AArray<double, 2> &eigs )
		{
			for ( int i = 0; i < n; ++i )
			{
				AArray<double, 2> C = A - lambdas( i ) * B;
				AArray<double, 1> f = eigs( Range::all(),i );
				AArray<double, 1> result(f.size());
				matMul2( C, f, result );
				double len = dot( result, result );
				assert_msg( len < FLT_EPSILON, "len is: %f\n", len );
				ADORE_UNUSED( len );
			}
		}

		void RDFMatrixCalculator::init( const int numValues )
		{
			numSolutions = 0;
			row = 0;

			if ( (int) D1.rows() != numValues ) D1.resize( numValues,9 );
			if ( (int) D2.rows() != numValues ) D2.resize( numValues,9 );
			if ( (int) D3.rows() != numValues ) D3.resize( numValues,9 );
		}

		int RDFMatrixCalculator::solve( const int numValues, const float minThresh, const float maxThresh )
		{
			if ( row < 9 )
			{
				DBGOUT( DBGCRITICAL, "RDFMatrixCalculator::buildEquation: not enough sane correspondences (%d/9)\n", row );
				return 0;
			}
			else if ( row != numValues )
			{
				D1.resizeAndPreserve( row, 9 );
				D2.resizeAndPreserve( row, 9 );
				D3.resizeAndPreserve( row, 9 );
			}

			createQEPLinearization( D1, D2, D3, A, B );

			numSolutions = solveGEV( A, B, lambdas, eigenvecs, minThresh, maxThresh );
			if ( numSolutions == 0 )
			{
				DBGOUT( DBGNORMAL, "RDFMatrixCalculator - No Solutions in solveGEV\n" );
				return 0;
			}

			//check if one of the lambdas < lambdaInaccurateThresh
			//in that case the solution has been abserved to be inaccurate
			for ( int i = 0; i < numSolutions; ++i )
			{
				if ( fabs( lambdas( i ) ) < lambdaInaccurateThresh )
				{
					int ret = fcalc.calculateFundamentalMatrix8Point( D1 );
					if ( ret != 0 )
					{
						DBGOUT( DBGCRITICAL, "RDFMatrixCalculator: 7-point algo has been invoked. Only one out of three solutions will be provided\n" );
						/*TODO cope with rare 7-point cases*/
					}
					const AMatrix &fundamat = fcalc.getFMatrix1();
					//copy
					eigenvecs( 0, i ) = fundamat( 0, 0 );
					eigenvecs( 1, i ) = fundamat( 0, 1 );
					eigenvecs( 2, i ) = fundamat( 0, 2 );
					eigenvecs( 3, i ) = fundamat( 1, 0 );
					eigenvecs( 4, i ) = fundamat( 1, 1 );
					eigenvecs( 5, i ) = fundamat( 1, 2 );
					eigenvecs( 6, i ) = fundamat( 2, 0 );
					eigenvecs( 7, i ) = fundamat( 2, 1 );
					eigenvecs( 8, i ) = fundamat( 2, 2 );
				}
			}
//			testSolutions( numSolutions, A, B, lambdas, eigenvecs );

			return numSolutions;
		}

		void RDFMatrixCalculator::addCorrespondence( const AVector &fp1, const AVector &fp2 )
		{
			double r1Sqr = fp1( 0 ) * fp1( 0 ) + fp1( 1 ) * fp1( 1 );
			double r2Sqr = fp2( 0 ) * fp2( 0 ) + fp2( 1 ) * fp2( 1 );

			D1( row,0 ) = fp2( 0 ) * fp1( 0 );
			D1( row,1 ) = fp2( 0 ) * fp1( 1 );
			D1( row,2 ) = fp2( 0 );
			D1( row,3 ) = fp2( 1 ) * fp1( 0 );
			D1( row,4 ) = fp2( 1 ) * fp1( 1 );
			D1( row,5 ) = fp2( 1 );
			D1( row,6 ) = fp1( 0 );
			D1( row,7 ) = fp1( 1 );
			D1( row,8 ) = 1;

			D2( row,0 ) = 0;
			D2( row,1 ) = 0;
			D2( row,2 ) = fp2( 0 ) * r1Sqr;
			D2( row,3 ) = 0;
			D2( row,4 ) = 0;
			D2( row,5 ) = fp2( 1 ) * r1Sqr;
			D2( row,6 ) = fp1( 0 ) * r2Sqr;
			D2( row,7 ) = fp1( 1 ) * r2Sqr;
			D2( row,8 ) = r1Sqr + r2Sqr;

			D3( row,0 ) = 0;
			D3( row,1 ) = 0;
			D3( row,2 ) = 0;
			D3( row,3 ) = 0;
			D3( row,4 ) = 0;
			D3( row,5 ) = 0;
			D3( row,6 ) = 0;
			D3( row,7 ) = 0;
			D3( row,8 ) = r1Sqr * r2Sqr;

			++row;
		}

		double RDFMatrixCalculator::getError( AMatrix &fmat, float lambda )
		{
			CorrespondenceCorrector corrector;
			double dx1, dy1, dx2, dy2;
			double x1, y1, x2, y2;
			double dist = 0;

			for ( unsigned row = 0; row < D1.rows(); ++row )
			{
				dx1 = x1 = D1( row,6 );
				dy1 = y1 = D1( row,7 );
				dx2 = x2 = D1( row,2 );
				dy2 = y2 = D1( row,5 );

				undistortRadial<double>( x1, y1, lambda );
				undistortRadial<double>( x2, y2, lambda );

				bool corrected = corrector( x1, y1, x2, y2, fmat );

				if ( corrected )
				{
					distortRadial<double>( x1, y1, lambda );
					distortRadial<double>( x2, y2, lambda );

					//calculate reprojection error between mates and correct
					dist += ( dx1 - x1 )*( dx1 - x1 ) +
					        ( dy1 - y1 )*( dy1 - y1 ) +
					        ( dx2 - x2 )*( dx2 - x2 ) +
					        ( dy2 - y2 )*( dy2 - y2 );
				}
				else
				{
					DBGOUT( DBGCRITICAL, "RDFMatrixCalculator::getError - corr not corrected\n" );
				}
			}
			return dist;
		}

		tuple<AMatrix, float> RDFMatrixCalculator::getSolution( int i )
		{
			AMatrix fundamat( 3, 3 );
			if ( i >= numSolutions ) return make_tuple( NullMatrix, 0 );

			fundamat( 0, 0 ) = eigenvecs( 0, i );
			fundamat( 0, 1 ) = eigenvecs( 1, i );
			fundamat( 0, 2 ) = eigenvecs( 2, i );
			fundamat( 1, 0 ) = eigenvecs( 3, i );
			fundamat( 1, 1 ) = eigenvecs( 4, i );
			fundamat( 1, 2 ) = eigenvecs( 5, i );
			fundamat( 2, 0 ) = eigenvecs( 6, i );
			fundamat( 2, 1 ) = eigenvecs( 7, i );
			fundamat( 2, 2 ) = eigenvecs( 8, i );

			fcalc.enforceRank2( fundamat );

			return make_tuple( fundamat, lambdas( i ) );
		}

	}
}
