

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "camparamchoice.h"

namespace adore
{

	namespace core
	{

		CamParamChoice::CamParamChoice ( const int numCams ) :
				numRotParam ( rotParam.getSize() ),
				numLocParam ( 3 ),
				rotOff ( 0 ),
				locOff ( numRotParam ),
				calibOff ( locOff + numLocParam ),
				numViews ( numCams ),
				flyMode ( FREE ),
				skewMode ( HARD ),
				arMode ( FREE ),
				ppxMode ( FREE ),
				ppyMode ( FREE ),
				flyVal ( numCams ),
				skewVal ( numCams ),
				arVal ( numCams ),
				ppxVal ( numCams ),
				ppyVal ( numCams ),
				tmpVal ( numCams ),
				rotVec ( numRotParam )
		{
			initVector ( flyVal, 0.0 );
			initVector ( skewVal, 0.0 );
			initVector ( arVal, 1.0 );
			initVector ( ppxVal, 0.0 );
			initVector ( ppyVal, 0.0 );

			nParams = new int[maxParameterTypeValue + 1];//this value is from constants.h
		}

		CamParamChoice::CamParamChoice ( const CamParamChoice &cpc ) :
				numRotParam ( rotParam.getSize() ),
				numLocParam ( 3 ),
				rotOff ( 0 ),
				locOff ( numRotParam ),
				calibOff ( locOff + numLocParam ),
				numViews ( cpc.numViews ),
				flyMode ( cpc.flyMode ),
				skewMode ( cpc.skewMode ),
				arMode ( cpc.arMode ),
				ppxMode ( cpc.ppxMode ),
				ppyMode ( cpc.ppyMode ),
				flyVal ( cpc.flyVal ),
				skewVal ( cpc.skewVal ),
				arVal ( cpc.arVal ),
				ppxVal ( cpc.ppxVal ),
				ppyVal ( cpc.ppyVal ),
				tmpVal ( cpc.numViews ),
				rotVec ( numRotParam )
		{
			nParams = new int[maxParameterTypeValue + 1];
		}

		CamParamChoice::~CamParamChoice()
		{
			delete [] nParams;
		}

		int CamParamChoice::getStartRow ( const int camIdx, const int startRow ) const
		{
			assert ( startRow >= 0 );
			int numP = getNumParams();
			int numCP = getNumCONSTParams();
			return startRow + numCP + camIdx * numP;
		}

		void CamParamChoice::getCameraFromAllParams ( const AVector &allParam, Camera *cam,
		                const int camIdx,
		                const int startRow )
		{
			AMatrix &K = cam->getCalibrationMatrix();
			AMatrix &R = cam->getRotationMatrix();
			AVector3D &loc = cam->getLocation();

			getCalibrationMatrixFromAllParams ( allParam, K, camIdx, startRow );
			getRotationMatrixFromAllParams ( allParam, R, startRow );
			getLocationFromAllParams ( allParam, loc, startRow );
		}

		void CamParamChoice::getCalibrationMatrixFromAllParams ( const AVector &allParam, AMatrix &K,
		                const int camIdx, const int startRow ) const
		{
			int startIdx = getStartRow ( camIdx, startRow ) + calibOff;
			getCalibrationMatrixFromParams ( allParam, K, camIdx, startIdx, &allParam, startRow );
		}

		void CamParamChoice::getRotationMatrixFromAllParams ( const AVector &allParam, AMatrix &R,
		                const int camIdx, const int startRow )
		{
			int startIdx = getStartRow ( camIdx, startRow ) + rotOff;
			getRotationMatrixFromParams ( allParam, R, startIdx );
		}

		void CamParamChoice::getLocationFromAllParams ( const AVector &allParam, AVector3D &loc,
		                const int camIdx,
		                const int startRow ) const
		{
			int startIdx = getStartRow ( camIdx, startRow ) + locOff;
			getLocationFromParams ( allParam, loc, startIdx );
		}


		void CamParamChoice::getCalibrationMatrixFromParams ( const AVector &param, AMatrix &K, const int camIdx, int row,
		                const AVector *constParam, int constRow ) const
		{
			setIdentity ( K );

			switch ( flyMode )
			{
				case SOFT_CONST:
				case FREE_CONST:
					K ( 1, 1 ) = ( *constParam ) ( constRow );
					++constRow;
					break;
				case HARD_CONST:
					K ( 1, 1 ) =  flyVal[0];
					break;
				case HARD:
					assert ( camIdx < ( int ) flyVal.size() );
					K ( 1, 1 ) =  flyVal[camIdx];
					break;
				case SOFT:
				case FREE:
                                        K ( 1, 1 ) = param ( row );
					++row;
					break;
				default:
					break;
			}

			switch ( skewMode )
			{
				case SOFT_CONST:
				case FREE_CONST:
					K ( 0, 1 ) = ( *constParam ) ( constRow );
					++constRow;
					break;
				case HARD_CONST:
					K ( 0, 1 ) =  skewVal[0];
					break;
				case HARD:
					assert_msg ( camIdx < ( int ) skewVal.size(), "%d - %zd\n", camIdx, skewVal.size() );
					K ( 0, 1 ) =  skewVal[camIdx];
					break;
				case SOFT:
				case FREE:
					K ( 0, 1 ) = param ( row );
					++row;
					break;
				default:
					break;
			}

			switch ( arMode )
			{
				case SOFT_CONST:
				case FREE_CONST:
					K ( 0, 0 ) = K ( 1, 1 ) * ( *constParam ) ( constRow );
					++constRow;
					break;
				case HARD_CONST:
					K ( 0, 0 ) = K ( 1, 1 ) * arVal[0];
					break;
				case HARD:
					assert ( camIdx < ( int ) arVal.size() );
					K ( 0, 0 ) = K ( 1, 1 ) * arVal[camIdx];
					break;
				case SOFT:
				case FREE:
					K ( 0, 0 ) = K ( 1, 1 ) * param ( row );
					++row;
					break;
				default:
					break;
			}

			switch ( ppxMode )
			{
				case SOFT_CONST:
				case FREE_CONST:
					K ( 0, 2 ) = ( *constParam ) ( constRow );
					++constRow;
					break;
				case HARD_CONST:
					K ( 0, 2 ) =  ppxVal[0];
					break;
				case HARD:
					assert ( camIdx < ( int ) ppxVal.size() );
					K ( 0, 2 ) =  ppxVal[camIdx];
					break;
				case SOFT:
				case FREE:
					K ( 0, 2 ) = param ( row );
					++row;
					break;
				default:
					break;
			}

			switch ( ppyMode )
			{
				case SOFT_CONST:
				case FREE_CONST:
					K ( 1, 2 ) = ( *constParam ) ( constRow );
					++constRow;
					break;
				case HARD_CONST:
					K ( 1, 2 ) =  ppyVal[0];
					break;
				case HARD:
					assert ( camIdx < ( int ) ppyVal.size() );
					K ( 1, 2 ) =  ppyVal[camIdx];
					break;
				case SOFT:
				case FREE:
					K ( 1, 2 ) = param ( row );
					++row;
					break;
				default:
					break;
			}
		}

		void CamParamChoice::getRotationMatrixFromParams ( const AVector &param, AMatrix &R, int row )
		{
			//first, the rotation
			rotVec ( 0 ) = param ( row++ );
			rotVec ( 1 ) = param ( row++ );
			rotVec ( 2 ) = param ( row++ );

			//compute the rotation matrix
			R = rotParam.fromParam ( rotVec );
		}

		void CamParamChoice::getLocationFromParams ( const AVector &param, AVector3D &loc, int row ) const
		{
			//now, the location of the camera is assigned
			loc ( 0 ) = param ( row++ );
			loc ( 1 ) = param ( row++ );
			loc ( 2 ) = param ( row++ );
		}

		void CamParamChoice::getCameraFromParams ( const AVector &param, Camera *cam, const int camIdx, const int row )
		{
			AMatrix &K = cam->getCalibrationMatrix();
			AMatrix &R = cam->getRotationMatrix();
			AVector3D &loc = cam->getLocation();

			getRotationMatrixFromParams ( param, R, row + rotOff );

			getLocationFromParams ( param, loc, row + locOff );

			getCalibrationMatrixFromParams ( param, K, camIdx, row + calibOff );
			//compute camera
			cam->compose();
		}

		void CamParamChoice::getCONSTParam ( AVector &p, int row ) const
		{
			if ( flyMode == FREE_CONST || flyMode == SOFT_CONST )
			{
                                p ( row ) =  flyVal[0];
				++row;
			}
			if ( skewMode == FREE_CONST || skewMode == SOFT_CONST )
			{
                                p ( row ) =  arVal[0];
				++row;
			}
			if ( arMode == FREE_CONST || arMode == SOFT_CONST )
			{
                                p ( row ) =  skewVal[0];
				++row;
			}
			if ( ppxMode == FREE_CONST || ppxMode == SOFT_CONST )
			{
                                p ( row ) =  ppxVal[0];
				++row;
			}
			if ( ppyMode == FREE_CONST || ppyMode == SOFT_CONST )
			{
                                p ( row ) =  ppyVal[0];
				++row;
			}
		}

		void CamParamChoice::getParamFromCalibrationMatrix ( const AMatrix &K, AVector &param, int row ) const
		{
			if ( flyMode == FREE || flyMode == FREE_CONST || flyMode == SOFT || flyMode == SOFT_CONST )
			{
				param ( row ) = K ( 1, 1 );
				++row;
			}
			if ( skewMode == FREE || skewMode == FREE_CONST || skewMode == SOFT || skewMode == SOFT_CONST )
			{
				param ( row ) = K ( 0, 1 );
				++row;
			}
			if ( arMode == FREE || arMode == FREE_CONST || arMode == SOFT || arMode == SOFT_CONST )
			{
				param ( row ) = K ( 0, 0 ) / K ( 1, 1 );
				++row;
			}
			if ( ppxMode == FREE || ppxMode == FREE_CONST || ppxMode == SOFT || ppxMode == SOFT_CONST )
			{
				param ( row ) = K ( 0, 2 );
				++row;
			}
			if ( ppyMode == FREE || ppyMode == FREE_CONST || ppyMode == SOFT || ppyMode == SOFT_CONST )
			{
				param ( row ) = K ( 1, 2 );
				++row;
			}
		}

		void CamParamChoice::getParamFromRotationMatrix ( const AMatrix &R, AVector &param, const int row )
		{
			AVector sub = param ( Range ( row, row + 2 ) );
			rotParam.toParam ( R, sub );

//			assert_msg( anax.isInIntervall( sqrt( getSqrNorm( rotVec ) ) ), "Angle is %f, Please insert nomalization code\n", sqrt( getSqrNorm( rotVec ) ) );
		}

		void CamParamChoice::getParamFromLocation ( const AVector3D &loc, AVector &param, int row ) const
		{
			param ( Range ( row, row + 2 ) ) = loc;
		}

		void CamParamChoice::getParamFromCamera ( Camera *cam, AVector &param, const int row )
		{
			cam->decompose();
			AMatrix &K = cam->getCalibrationMatrix();
			AMatrix &R = cam->getRotationMatrix();
			AVector3D &loc = cam->getLocation();

			getParamFromRotationMatrix ( R, param, row + rotOff );
			getParamFromLocation ( loc, param, row + locOff );
			getParamFromCalibrationMatrix ( K, param, row + calibOff );
		}

		int CamParamChoice::getNumParams() const
		{
			return numRotParam + numLocParam + getNumCalibrationParams();
		}

		int CamParamChoice::getNumCalibrationParams() const
		{
			nParams[HARD] = 0;
			nParams[FREE_CONST] = 0;
			nParams[HARD_CONST] = 0;
			nParams[FREE] = 1;
			nParams[SOFT] = 1;
			nParams[SOFT_CONST] = 0;

			int params = 0;
			params += nParams[flyMode];
			params += nParams[skewMode];
			params += nParams[ppxMode];
			params += nParams[ppyMode];
			params += nParams[arMode];

			return params;
		}

		int CamParamChoice::getNumCONSTParams() const
		{
			nParams[HARD] = 0;
			nParams[FREE_CONST] = 1;
			nParams[HARD_CONST] = 0;
			nParams[FREE] = 0;
			nParams[SOFT] = 0;
			nParams[SOFT_CONST] = 1;

			int params = 0;
			params += nParams[flyMode];
			params += nParams[skewMode];
			params += nParams[ppxMode];
			params += nParams[ppyMode];
			params += nParams[arMode];

			return params;
		}

		void CamParamChoice::print() const
		{
			printf ( "same cameras is %s\n", sameCameras() ?"true":"false" );
			printf ( "fly: " );
			switch ( flyMode )
			{
				case SOFT_CONST:
					printf ( "soft_const" );
					break;
				case SOFT:
					printf ( "soft" );
					break;
				case FREE_CONST:
					printf ( "fc\n" );
					break;
				case HARD_CONST:
					printf ( "hc, %.8f\n", flyVal[0] );
					break;
				case HARD:
					printf ( "h, " );
					for ( int i = 0; i < numViews; ++i ) printf ( "%.8f  ", flyVal[i] );
					printf ( "\n" );
					break;
				case FREE:
					printf ( "f\n" );
					break;
				default:
					break;
			}

			printf ( "skew: " );
			switch ( skewMode )
			{
				case SOFT_CONST:
					printf ( "soft_const" );
					break;
				case SOFT:
					printf ( "soft" );
					break;
				case FREE_CONST:
					printf ( "fc\n" );
					break;
				case HARD_CONST:
					printf ( "hc, %.8f\n", skewVal[0] );
					break;
				case HARD:
					printf ( "h, " );
					for ( int i = 0; i < numViews; ++i ) printf ( "%.8f  ", skewVal[i] );
					printf ( "\n" );
					break;
				case FREE:
					printf ( "f\n" );
					break;
				default:
					break;
			}

			printf ( "ar: " );
			switch ( arMode )
			{
				case SOFT_CONST:
					printf ( "soft_const" );
					break;
				case SOFT:
					printf ( "soft" );
					break;
				case FREE_CONST:
					printf ( "fc\n" );
					break;
				case HARD_CONST:
					printf ( "hc, %.8f\n", arVal[0] );
					break;
				case HARD:
					printf ( "h, " );
					for ( int i = 0; i < numViews; ++i ) printf ( "%.8f  ", arVal[i] );
					printf ( "\n" );
					break;
				case FREE:
					printf ( "f\n" );
					break;
				default:
					break;
			}

			printf ( "ppx: " );
			switch ( ppxMode )
			{
				case SOFT_CONST:
					printf ( "soft_const" );
					break;
				case SOFT:
					printf ( "soft" );
					break;
				case FREE_CONST:
					printf ( "fc\n" );
					break;
				case HARD_CONST:
					printf ( "hc, %.8f\n", ppxVal[0] );
					break;
				case HARD:
					printf ( "h, " );
					for ( int i = 0; i < numViews; ++i ) printf ( "%.8f  ", ppxVal[i] );
					printf ( "\n" );
					break;
				case FREE:
					printf ( "f\n" );
					break;
				default:
					break;

			}

			printf ( "ppy: " );
			switch ( ppyMode )
			{
				case SOFT_CONST:
					printf ( "soft_const" );
					break;
				case SOFT:
					printf ( "soft" );
					break;
				case FREE_CONST:
					printf ( "fc\n" );
					break;
				case HARD_CONST:
					printf ( "hc, %.8f\n", ppyVal[0] );
					break;
				case HARD:
					printf ( "h, " );
					for ( int i = 0; i < numViews; ++i ) printf ( "%.8f  ", ppyVal[i] );
					printf ( "\n" );
					break;
				case FREE:
					printf ( "f\n" );
					break;
				default:
					break;
			}
		}

	}

}
