

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_PARAMETERIZATIONCAMPARAMCHOICE_H
#define ADORE_PARAMETERIZATIONCAMPARAMCHOICE_H

#include <vector>
#include <adrcore/types/camera.h>
#include <adrcore/parameterization/angleaxis.h>
#include <adrcore/parameterization/constants.h>

namespace adore
{
	namespace core
	{

		using namespace std;

		/**
		 * Format of Camera Parameter for one camera:
		 * RotX
		 * RotY
		 * RotZ
		 * LocX
		 * LocY
		 * LocZ
		 * FLY
		 * Skew
		 * AR
		 * PPx
		 * PPy
		 *
		 * Format of Camera Parameter for all cameras:
		 * CONST
		 * RotX_i
		 * RotY_i
		 * RotZ_i
		 * LocX_i
		 * LocY_i
		 * LocZ_i
		 * FLY_i
		 * Skew_i
		 * AR_i
		 * PPx_i
		 * PPy_i
		 *  
		 *
		*/
		class CamParamChoice
		{

			public:
				CamParamChoice( const int numCams = 1 );

				CamParamChoice( const CamParamChoice &cpc );

				~CamParamChoice();

				//all params to camera-parts

				void getCameraFromAllParams( const AVector &allParam, Camera *cam, const int camIdx = 0,
				                             const int startRow = 0 );

				void getCalibrationMatrixFromAllParams( const AVector &allParam, AMatrix &K, const int camIdx = 0,
				                                        const int startRow = 0 ) const ;

				void getRotationMatrixFromAllParams( const AVector &allParam, AMatrix &R, const int camIdx = 0,
				                                     const int startRow = 0 );

				void getLocationFromAllParams( const AVector &allParam, AVector3D &loc, const int camIdx = 0,
				                               const int startRow = 0 ) const ;

				//param to camera-parts

				void getCameraFromParams( const AVector &param, Camera *cam, const int camIdx = 0, const int startRow = 0 );

				void getCalibrationMatrixFromParams( const AVector &param, AMatrix &K, const int camIdx = 0, int startRow = 0, const AVector *constParam = 0, int constRow = 0 ) const ;

				void getRotationMatrixFromParams( const AVector &param, AMatrix &R, const int startRow = 0 );

				void getLocationFromParams( const AVector &param, AVector3D &loc, const int startRow = 0 ) const ;

				//camera-parts to param

				void getParamFromCamera( Camera *cam, AVector &param, const int startRow = 0 );

				void getParamFromCalibrationMatrix( const AMatrix &K, AVector &param, const int startRow = 0 ) const ;

				void getParamFromRotationMatrix( const AMatrix &R, AVector &param, const int startRow = 0 );

				void getParamFromLocation( const AVector3D &loc, AVector &param, const int startRow = 0 ) const ;

				//Params that are const throughout all cameras

				void getCONSTParam( AVector &param, const int startRow = 0 ) const ;

				void setAR( const double val, const int idx = -1 )
				{
					if ( idx < 0 ) for ( uint i = 0; i < arVal.size(); ++i ) arVal[i] = val;
					else arVal[idx] = val;
				}

				void setFLy( const double val, const int idx = -1 )
				{
					if ( idx < 0 ) for ( uint i = 0; i < flyVal.size(); ++i ) flyVal[i] = val;
					else flyVal[idx] = val;
				}

				void setSkew( const double val, const int idx = -1 )
				{
					if ( idx < 0 ) for ( uint i = 0; i < skewVal.size(); ++i ) skewVal[i] = val;
					else skewVal[idx] = val;
				}

				void setPPx( const double val, const int idx = -1 )
				{
					if ( idx < 0 ) for ( uint i = 0; i < ppxVal.size(); ++i ) ppxVal[i] = val;
					else ppxVal[idx] = val;
				}

				void setPPy( const double val, const int idx = -1 )
				{
					if ( idx < 0 ) for ( uint i = 0; i < ppyVal.size(); ++i ) ppyVal[i] = val;
					else ppyVal[idx] = val;
				}

				void setFLyMode( const PARAMETER_TYPE& theValue )
				{
					flyMode = theValue;
				}

				PARAMETER_TYPE getFLyMode() const
				{
					return flyMode;
				}

				void setSkewMode( const PARAMETER_TYPE& theValue )
				{
					skewMode = theValue;
				}

				PARAMETER_TYPE getSkewMode() const
				{
					return skewMode;
				}

				void setARMode( const PARAMETER_TYPE& theValue )
				{
					arMode = theValue;
				}

				PARAMETER_TYPE getARMode() const
				{
					return arMode;
				}

				void setPPxMode( const PARAMETER_TYPE& theValue )
				{
					ppxMode = theValue;
				}

				PARAMETER_TYPE getPPxMode() const
				{
					return ppxMode;
				}

				void setPPyMode( const PARAMETER_TYPE& theValue )
				{
					ppyMode = theValue;
				}

				PARAMETER_TYPE getPPyMode() const
				{
					return ppyMode;
				}

				int getNumParams() const ;

				int getNumCalibrationParams() const ;

				int getNumRotationParams() const
				{
					return numRotParam;
				}

				int getNumLocationParams() const
				{
					return numLocParam;
				}

				int getNumCONSTParams() const ;

				int getNumViews() const
				{
					return numViews;
				}

				/**
				 * are all cameras the same?
				 * This is true, if only constant values are used.
				 * It is not checked, if all HARD values are the same, so use CONST_HARD
				 */
				bool sameCameras() const
				{
					if ( flyMode == FREE ) return false;
					if ( flyMode == HARD ) return false;
					if ( arMode == FREE ) return false;
					if ( arMode == HARD ) return false;
					if ( skewMode == FREE ) return false;
					if ( skewMode == HARD ) return false;
					if ( ppxMode == FREE ) return false;
					if ( ppxMode == HARD ) return false;
					if ( ppyMode == FREE ) return false;
					if ( ppyMode == HARD ) return false;
					return true;
				}

				void print() const ;

				void setNumCams( const int numCams )
				{
					numViews = numCams;
					flyVal.resize( numCams );
					skewVal.resize( numCams );
					arVal.resize( numCams );
					ppxVal.resize( numCams );
					ppyVal.resize( numCams );
					tmpVal.resize( numCams );
				}

			protected:

				AngleAxis rotParam;

				int numRotParam;
				int numLocParam;
				int rotOff;
				int locOff;
				int calibOff;

				int numViews;
				int *nParams;

				PARAMETER_TYPE flyMode;
				PARAMETER_TYPE skewMode;
				PARAMETER_TYPE arMode;
				PARAMETER_TYPE ppxMode;
				PARAMETER_TYPE ppyMode;

				vector<double> flyVal;
				vector<double> skewVal;
				vector<double> arVal;
				vector<double> ppxVal;
				vector<double> ppyVal;
				vector<double> tmpVal;

				AVector rotVec;

				void initVector( vector<double> &vals, const double initVal ) const
				{
					for ( uint i = 0; i < vals.size(); ++i ) vals[i] = initVal;
				}

				int getStartRow( const int camIdx, const int startRow ) const ;
		};

	}

}

#endif
