

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "jpegpersistence.h"
#include <cstring>

using namespace std;

bool JPEGPersistence::getInfo( const char *filename, int& width, int& height, int& channels )
{
	jpeg_decompress_struct cinfo;
	error_mgr jerr; jpeg_std_error( &jerr );
	jerr.error_exit = JPEGPersistence::error_exit;
	
	FILE* file = prepareDecompress( filename, cinfo, jerr );
	if ( !file )
	{
		return false;
	}
	
	width = cinfo.image_width;
	height = cinfo.image_height;
	channels = cinfo.num_components;
	jpeg_destroy_decompress( &cinfo );
	fclose( file );
	return true;
}

bool JPEGPersistence::read( const char *filename, jpgT* buffer, int wstep )
{
	typedef unsigned int uint;
	typedef unsigned long ulong;

	jpeg_decompress_struct cinfo;
	error_mgr jerr; jpeg_std_error( &jerr );
	jerr.error_exit = JPEGPersistence::error_exit;
	
	FILE* file = prepareDecompress( filename, cinfo, jerr );
	
	if ( !file ) return false;
	jpeg_start_decompress ( &cinfo );
	
	uint row_stride = cinfo.output_width * cinfo.output_components;
	JSAMPARRAY buffer_local = ( *cinfo.mem->alloc_sarray )( ( j_common_ptr ) &cinfo, JPOOL_IMAGE, row_stride, 1 );
	uint length = cinfo.image_width * cinfo.num_components;

	uint y = 0;
	while ( cinfo.output_scanline < cinfo.output_height )
	{
		jpeg_read_scanlines( &cinfo, buffer_local, 1 );
		
		unsigned char* ptrSrc = *buffer_local;
		unsigned char* ptrDst = (unsigned char*)buffer + (y * wstep);
		memcpy(ptrDst, ptrSrc, sizeof(jpgT) * length);
		++y;
	}

	jpeg_finish_decompress ( &cinfo );
	jpeg_destroy_decompress ( &cinfo );
	fclose( file );
	return true;
}

FILE* JPEGPersistence::prepareDecompress( const char *filename, jpeg_decompress_struct& cinfo, error_mgr& jerr )
{
	cinfo.err = &jerr;
	jpeg_create_decompress ( &cinfo );
	
	FILE* file;
	if ( ( file = fopen( filename, "rb" ) ) == 0 )
	{
		perror( "JPEGPersistence" );
		fprintf( stderr, "JPEGPersistence: can't open %s\n", filename );
		return false;
	}
	
	if ( setjmp( jerr.setjmp_buffer ) )
	{
		jpeg_destroy_decompress( &cinfo );
		fclose( file ); return 0;
	}
	
	jpeg_stdio_src( &cinfo, file );
	jpeg_read_header( &cinfo, true );
	
	return file;
}

bool JPEGPersistence::write( const char *filename, jpgT* buffer, int nChannels, int width, int height, int wstep, int quality )
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	
	cinfo.err = jpeg_std_error( &jerr );
	jpeg_create_compress( &cinfo );
	
	if ( nChannels == 3 )
	{
		cinfo.in_color_space = JCS_RGB;
	}
	else if ( nChannels == 1 )
	{
		cinfo.in_color_space = JCS_GRAYSCALE;
	}
	else
	{
		fprintf( stderr, "Unsupported Number of Channels! Can't write jpegfile!");
		jpeg_destroy_compress( &cinfo );
		return false;
	}
	
	FILE * outfile;
	if ( ( outfile = fopen( filename, "wb" ) ) == NULL )
	{
		fprintf( stderr, "can't open file %s for writing!", filename);
		return false;
	}
	jpeg_stdio_dest( &cinfo, outfile );
	
	cinfo.image_width = width;
	cinfo.image_height = height;
	cinfo.input_components = nChannels;
	jpeg_set_defaults( &cinfo );
	jpeg_set_quality( &cinfo, quality, false /*strict baseline compatibility*/ );
	
	jpeg_start_compress( &cinfo, TRUE );
	
	JSAMPROW row_pointer[1];        /* pointer to JSAMPLE row[s] */
	//         printf("cinfo.next_scanline is initially set to %d\n", cinfo.next_scanline);
	while ( cinfo.next_scanline < cinfo.image_height )
	{
		row_pointer[0] = buffer + (cinfo.next_scanline * wstep);
		( void ) jpeg_write_scanlines( &cinfo, row_pointer, 1 );
	}
	jpeg_finish_compress( &cinfo );
	fclose( outfile );
	jpeg_destroy_compress( &cinfo );
	
	return true;
}

/* Expanded data source object for memory input */

typedef struct
{
	struct jpeg_source_mgr pub;	/* public fields */
	JOCTET eoi_buffer[2];		/* a place to put a dummy EOI */
} my_source_mgr;

typedef my_source_mgr * my_src_ptr;

static void init_source ( j_decompress_ptr cinfo )
{
	/* No work, since jpeg_memory_src set up the buffer pointer and count.
	 * Indeed, if we want to read multiple JPEG images from one buffer,
	 * this *must* not do anything to the pointer.
	 */
}

/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In this application, this routine should never be called; if it is called,
 * the decompressor has overrun the end of the input buffer, implying we
 * supplied an incomplete or corrupt JPEG datastream.  A simple error exit
 * might be the most appropriate response.
 *
 * But what we choose to do in this code is to supply dummy EOI markers
 * in order to force the decompressor to finish processing and supply
 * some sort of output image, no matter how corrupted.
 */
static boolean fill_input_buffer ( j_decompress_ptr cinfo )
{
	my_src_ptr src = ( my_src_ptr ) cinfo->src;

	/* Create a fake EOI marker */
	src->eoi_buffer[0] = ( JOCTET ) 0xFF;
	src->eoi_buffer[1] = ( JOCTET ) JPEG_EOI;
	src->pub.next_input_byte = src->eoi_buffer;
	src->pub.bytes_in_buffer = 2;

	return true;
}

/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * If we overrun the end of the buffer, we let fill_input_buffer deal with
 * it.  An extremely large skip could cause some time-wasting here, but
 * it really isn't supposed to happen ... and the decompressor will never
 * skip more than 64K anyway.
 */
static void skip_input_data ( j_decompress_ptr cinfo, long num_bytes )
{
	my_src_ptr src = ( my_src_ptr ) cinfo->src;

	if ( num_bytes > 0 )
	{
		while ( num_bytes > ( long ) src->pub.bytes_in_buffer )
		{
			num_bytes -= ( long ) src->pub.bytes_in_buffer;
			( void ) fill_input_buffer( cinfo );
			/* note we assume that fill_input_buffer will never return FALSE,
			* so suspension need not be handled.
			*/
		}
		src->pub.next_input_byte += ( size_t ) num_bytes;
		src->pub.bytes_in_buffer -= ( size_t ) num_bytes;
	}
}

/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */
static void term_source ( j_decompress_ptr cinfo )
{
	/* no work necessary here */
}

/*
 * Prepare for input from a memory buffer.
 */
static void jpeg_memory_src ( j_decompress_ptr cinfo, const JOCTET * buffer, size_t bufsize )
{
	my_src_ptr src;

	/* The source object is made permanent so that a series of JPEG images
		* can be read from a single buffer by calling jpeg_memory_src
		* only before the first one.
		* This makes it unsafe to use this manager and a different source
		* manager serially with the same JPEG object.  Caveat programmer.
		*/
	if ( cinfo->src == NULL )  	/* first time for this JPEG object? */
	{
		cinfo->src = ( struct jpeg_source_mgr * )
					 ( *cinfo->mem->alloc_small ) ( ( j_common_ptr ) cinfo, JPOOL_PERMANENT, sizeof( my_source_mgr ) );
	}

	src = ( my_src_ptr ) cinfo->src;
	src->pub.init_source = init_source;
	src->pub.fill_input_buffer = fill_input_buffer;
	src->pub.skip_input_data = skip_input_data;
	src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
	src->pub.term_source = term_source;

	src->pub.next_input_byte = buffer;
	src->pub.bytes_in_buffer = bufsize;
}

bool JPEGPersistence::getInfo( const char *data, size_t size, int& width, int& height, int& channels )
{
	jpeg_decompress_struct cinfo;
	error_mgr jerr; jpeg_std_error( &jerr );
	jerr.error_exit = JPEGPersistence::error_exit;

	cinfo.err = &jerr;
	jpeg_create_decompress( &cinfo );

	if ( setjmp( jerr.setjmp_buffer ) )
	{
		jpeg_destroy_decompress( &cinfo );
		return false;
	}

	jpeg_memory_src( &cinfo, ( JOCTET* )&data[0], size );
	jpeg_read_header( &cinfo, true );

	width = cinfo.image_width;
	height = cinfo.image_height;
	channels = cinfo.num_components;
	jpeg_destroy_decompress( &cinfo );

	return true;
}

bool JPEGPersistence::read( const char *data, size_t size, jpgT* buffer, int wstep )
{
	jpeg_decompress_struct cinfo;
	error_mgr jerr; jpeg_std_error( &jerr );
	jerr.error_exit = JPEGPersistence::error_exit;

	cinfo.err = &jerr;
	jpeg_create_decompress( &cinfo );

	if ( setjmp( jerr.setjmp_buffer ) )
	{
		jpeg_destroy_decompress( &cinfo );
		return false;
	}

	jpeg_memory_src( &cinfo, ( JOCTET* )data, size );
	jpeg_read_header( &cinfo, true );

	jpeg_start_decompress ( &cinfo );

	typedef unsigned int uint;
	uint row_stride = cinfo.output_width * cinfo.output_components;
	JSAMPARRAY buffer_local = ( *cinfo.mem->alloc_sarray )( ( j_common_ptr ) &cinfo, JPOOL_IMAGE, row_stride + 16, 1 );
	typedef unsigned long ulong;
	ulong addr = ( ulong )*buffer_local;
	ulong old_addr = addr;
	if ( addr % 0x10 ) addr += 0x10 - ( addr % 0x10 );
	*buffer_local = ( JSAMPLE* )addr;

	uint y = 0;
	uint length = cinfo.image_width * 3;
	while ( cinfo.output_scanline < cinfo.output_height )
	{
		jpeg_read_scanlines( &cinfo, buffer_local, 1 );

		unsigned char* ptrSrc = *buffer_local;
		unsigned char* ptrDst = (unsigned char*)buffer + (y * wstep);
		memcpy(ptrDst, ptrSrc, sizeof(jpgT) * length);
		++y;
	}

	*buffer_local = ( JSAMPLE* )old_addr;
	jpeg_finish_decompress ( &cinfo );
	jpeg_destroy_decompress ( &cinfo );

	return true;
}
