

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef SIMPLEKDTREE_H
#define SIMPLEKDTREE_H

#include <adrcore/util/mathutil.h>
#include <adrcore/features/linedistfilter.h>

#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>

#include <vector>
#include <map>
#include <cassert>
#include <algorithm>
#include <limits>

namespace adore {
namespace core {

using namespace std;
using namespace boost;

template<bool> struct compiletime_assert;
template<> struct compiletime_assert<true> {};

/**
  * i is the ith element of T to compare
  */
template <typename T, int i>
struct less_i {
	bool operator() (const T& x, const T& y) const
	{ return (*x)(i) < (*y)(i); }
};


template <typename Key, typename Value, int numAxes> class SimpleKDTree {

public:
	typedef Key key_type;
	typedef Value value_type;
        typedef typename Value::scalar_type work_type;
	typedef vector<work_type> work_vector;

	enum { maxAxesIndex = numAxes - 1 };

	class Node {
		Value* value;
		Key min, max;
		int axis;
		typedef shared_ptr<Node> NodeT;
		NodeT left, right;
	public:
		Node(Value* val, work_type _min, work_type _max, int a) :
				value(val), min(numAxes), max(numAxes), axis(a)
		{
                    if(min.size() != (*value).size())
                    {
                        min.resize((*value).size());
                        max.resize((*value).size());
                    }
                        min = (*value);
                        max = (*value);
			int nextAxis = (axis + 1) % numAxes;
			min(nextAxis) = _min;
			max(nextAxis) = _max;
		}

		Value* getValue() const { return value; }
		const Key& getMin() const { return min; }
		const Key& getMax() const { return max; }
		int getAxis() const { return axis; }

		void setLeft(NodeT node) { left = node; }
		void setRight(NodeT node) { right = node; }
		NodeT getLeft() const { return left; }
		NodeT getRight() const { return right; }
	};

	typedef shared_ptr<Node> NodeT;

	SimpleKDTree() { compiletime_assert<(numAxes >= 1)>(); }

	void build(const vector<Value*>& pts)
	{
		root.reset();
		points = pts;

		work_type bigNegative  = -(numeric_limits<work_type>::max() / 2);
		work_type bigPositive  = numeric_limits<work_type>::max() / 2;
		work_vector min(numAxes), max(numAxes);
		for(int i = 0; i < numAxes; ++i) {
			min[i] = bigPositive;
			max[i] = bigNegative;
		}
		for(size_t i = 0; i < points.size(); ++i) {
			for(int j = 0; j < numAxes; ++j) {
				if((*points[i])(j) < min[j])
					min[j] = (*points[i])(j);
				if((*points[i])(j) > max[j])
					max[j] = (*points[i])(j);
			}
		}

                root = insert(0, points.size(), 0, min, max);
		vector<Value*>().swap(points);
	}

	NodeT getRoot() const { return root; }

private:

	NodeT root;
	vector<Value*> points;

	typedef typename vector<Value*>::iterator iterator;

	template<int axis> uint median(uint first, uint last)
	{
		if(first == last)
			return last;

		sort(points.begin() + first, points.begin() + last, less_i<Value*, axis>());
		iterator begin = points.begin() + first, end = points.begin() + last;
		iterator it = begin + (last - first) / 2;
		if(it == end) {
			assert(false);
			return first;
		}
		return it - points.begin();
	}

	NodeT insert(uint first, uint last, int depth, const work_vector& min, const work_vector& max)
	{
		int axis = depth % numAxes;

		uint m;
		switch(axis) {
		case 0:
			m = median<0>(first, last);
			break;
		case 1:
			m = median<1>(first, last);
			break;
		default: assert(false); return NodeT();
		}

		if(m == last)
			return NodeT();

		Value* val(*(points.begin() + m));
		int nextAxis = axis + 1;
		if(nextAxis >= numAxes)
			nextAxis = 0;

		NodeT node(new Node(val, min[nextAxis], max[nextAxis], axis));
		work_vector temp(max);
		temp[axis] = (*val)(axis);
		node->setLeft(insert(first, m, depth + 1, min, temp));
		node->setRight(insert(m + 1, last, depth + 1, temp, max));
		return node;
	}
};

template <typename Tree> class KDTreeAssessorLine2d {
	const Tree& tree;
	typedef typename Tree::NodeT Node;

public:
	typedef typename Tree::value_type Value;
	typedef typename Tree::key_type Key;

private:
	vector<Node> toDo;
	vector<Value*> result, toMeasure;

	template<typename X> X abs(const X& x) const { return x < 0 ? -x : x; }
	template<typename X> int sign(const X& x) const { return x < 0 ? -1 : (x > 0 ? 1 : 0); }

public:

	KDTreeAssessorLine2d(const Tree& t, size_t num) : tree(t)
	{
		compiletime_assert<(Tree::maxAxesIndex == 1)>();
		toDo.reserve(num);
		result.reserve(num);
		toMeasure.reserve(num);
	}

	const vector<Value*>& assess(const Key& p1, const Key& p2, double thresh)
	{
		toDo.clear(); // assuming: reserved memory is NOT released.
		result.clear();
		toMeasure.clear();

		Key line(3);
		cross(p1, p2, line);
		LineDistance ldist( line( 0 ), line( 1 ), line( 2 ) );
		if ( !ldist.isOkay() )
		{
			DBGOUT( DBGNORMAL, "KDTreeAssessorLine2d: insane line occurred" );
			printMat( DBGNORMAL, line );
			return result;
		}

		toDo.push_back(tree.getRoot());

		while(!toDo.empty()) {

			Node node = toDo.back();
			toDo.pop_back();
			if(NULL == node)
				continue;

			Value* data = node->getValue();
			Node nl = node->getLeft();
			Node nr = node->getRight();
			bool checkL = nl != NULL;
			bool checkR = nr != NULL;
			if(! (checkL || checkR) ) {
				toMeasure.push_back(data);
				continue;
			}

			const Key& pt1 = node->getMin();
			const Key& pt2 = node->getMax();
			double dist1 = ldist.getDistance( pt1 ( 0 ), pt1 ( 1 ) );
			double dist2 = ldist.getDistance( pt2 ( 0 ), pt2 ( 1 ) );
			int sign1 = sign(dist1);
			int sign2 = sign(dist2);

			bool checkThis = true;

			if(!(sign1 == 0 || sign2 == 0)) {
				if(sign1 == sign2) {
					if( (abs(dist1) > thresh) && (abs(dist2) > thresh) )
					{
						checkThis = false;
						if(sign1 > 0) checkL = false;
						else checkR = false;
					}
				}
			}

			if(checkThis)
				toMeasure.push_back(data);

			if(! ( checkL || checkR) )
				continue;

			if(checkL) {
				toDo.push_back(nl);
			}
			if(checkR) {
				toDo.push_back(nr);
			}

		}


		for(size_t i = 0; i < toMeasure.size(); ++i) {
			Value* cand = toMeasure[i];
			double dist = fabs( ldist.getDistance( ( * cand ) ( 0 ), ( * cand ) ( 1 ) ) );
			if(dist <= thresh)
				result.push_back(cand);
		}
		return result;
	}
};

}
}

#endif // SIMPLEKDTREE_H
