

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#ifndef ORIENTATIONHISTOGRAM_H
#define ORIENTATIONHISTOGRAM_H

#include <cmath>
#include <vector>
#include <map>

#include <adrcore/types/aarray.h>
#include <adrcore/util/mathutil.h>

namespace adore
{
	namespace core
	{

		template <typename real_type>
		class OrientationHistogram
		{
			public:
				typedef AArray< real_type, 1 > vec2;

			private:
				enum { max_peaks = 2, buckets_num = 36 };
				real_type buckets[buckets_num];
				
				typedef struct peak2angle_
				{
					double peak;
					real_type angle;
				} peak2angle;
				
				struct peaksComp
				{
					bool operator()(const peak2angle &p2a1, const peak2angle &p2a2)
					{
						return p2a1.peak > p2a2.peak;
					}
				};
				peaksComp pc;

			public:
				OrientationHistogram( )
				{
					clear( );
				}

				void clear( )
				{
					for ( int i = 0; i < buckets_num; ++i ) buckets[i] = real_type( 0 );
				}

				template <typename T> void getHistogram(std::vector<T>& dst) const {
					dst.resize(buckets_num);
					for( int i = 0; i < buckets_num; ++i )
						dst[i] = buckets[i];
				}

				static inline real_type pi() { return M_PI; }
				static inline real_type pi_2() { return M_PI * 2; }

				void add( const vec2& val, real_type factor = real_type( 1 ) )
				{
					real_type magnitude = sqrt( val( 0 )*val( 0 ) + val( 1 )*val( 1 ) ) * factor;
					real_type angle = atan2( val( 1 ), val( 0 ) ); /// y, x
					if ( 0 > angle ) angle += pi_2();
					real_type real_idx = angle / pi_2() * real_type( buckets_num );
					int idx = int( floor( real_idx ) );
					real_type neighbourWeight = real_idx - real_type( idx );
					buckets[idx] += magnitude * ( real_type( 1 ) - neighbourWeight );
					buckets[( idx + 1 ) % buckets_num] += magnitude * neighbourWeight;
				}

				void getMainOrientations( std::vector<real_type> &retval) const
				{
					real_type max( 0 );
					double vals[3], inc, peak;
					for ( int i = 0; i < buckets_num; ++i ) {
						if ( max < buckets[i] ) 
						  max = buckets[i];
					}
					real_type _80p_of_max = max * real_type( 0.80 ); // 80%
					
					retval.clear();
					retval.reserve(buckets_num);
					
					int peaksPos = 0;
					peak2angle peaks[buckets_num];//a collection of magnitudes and angles.
					
					for ( int i = 0; i < buckets_num; ++i )
					{
						if ( _80p_of_max < buckets[i] )
						{
							vals[0] = buckets[( i + buckets_num - 1 ) % buckets_num];
							vals[1] = buckets[i];
							if ( vals[0] > vals[1] ) continue;
							vals[2] = buckets[( i + 1 ) % buckets_num];
							if ( vals[2] > vals[1] ) continue;
							if ( fitParabola( vals[0], vals[1], vals[2], inc, peak ) )
							{
								const real_type angle = ( real_type( i ) + inc ) * pi_2() / buckets_num;
								peaks[peaksPos].angle = angle;
								peaks[peaksPos].peak = peak;
								++peaksPos;
							}
						}
					}
					std::sort(peaks, peaks + peaksPos, pc);
					//check sorting
					assert((peaksPos < 2) || peaks[0].peak >= peaks[1].peak);
					
					const int numToCopy = std::min(peaksPos, (int)max_peaks);
					for ( int i = 0; i < numToCopy; ++i )
						retval.push_back( peaks[i].angle );
				}
		};

	} // core
} // adore

#endif // ORIENTATIONHISTOGRAM_H
