

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "surffinder.h"
#include "surfdescriptor.h"

#include <adrcore/types/featurepoint.h>
#include <adrcore/surf/integralimage.h>

#include <adrcore/imgprocessing/imgfunctions.h>

#include <memory>
#include <cassert>

namespace adore
{
	namespace core
	{

		template <typename T> T abs ( const T& value ) { return value >= 0 ? value : -value; }

		SurfFinder::SurfFinder( uint id, uint octaves, uint intervals, double threshold, bool doubleSize,
		                        int maxP, float minD, float minF ) :
				id( id ),
				m_octaves( octaves ),
				m_intervals( intervals ),
				m_threshold( threshold ),
				m_doubleSize( doubleSize ),
				maxPts( maxP ),
				minDist( minD ),
				minFraction( minF ),
				descriptorCreator( NULL ),
				otherDecrCreator( NULL ),
				ownDescrCreator( false ),
				candidates( false ),
				store( 3, 3, 3 )
		{
			if ( 0 >= m_threshold ) m_threshold = 0.001;
			if ( m_octaves < 1 || m_octaves > MaxOctaves ) m_octaves = 1;
			if ( m_intervals < 3 || m_intervals > MaxIntervals ) m_intervals = 3;

			initFilterSizes();
		}

		SurfFinder::SurfFinder( const SurfFinder & rhs ) :
				id( rhs.id ),
				m_octaves( rhs.m_octaves ),
				m_intervals( rhs.m_intervals ),
				m_threshold( rhs.m_threshold ),
				m_doubleSize( rhs.m_doubleSize ),
				maxPts( rhs.maxPts ),
				minDist( rhs.minDist ),
				minFraction( rhs.minFraction ),
				descriptorCreator( NULL ),
				otherDecrCreator( NULL ),
				ownDescrCreator( false ),
				candidates( false ),
				store( 3, 3, 3 )
		{
			if ( NULL != rhs.descriptorCreator )
			{
				descriptorCreator =
				        static_cast<SurfDescriptor*>( rhs.descriptorCreator->clone() );
				ownDescrCreator = true;
			}
			else if ( NULL != rhs.otherDecrCreator )
			{
				otherDecrCreator = rhs.otherDecrCreator->clone();
				ownDescrCreator = true;
			}

			initFilterSizes();
		}

		void SurfFinder::initFilterSizes()
		{
			int first = 9;
			int step = 6;
			if ( m_doubleSize ) first = 15;
			for ( int i = 0; i < 4; ++i )
			{
				for ( int j = 0; j < 4; ++j )
				{
					filter_sizes[i][j] = first + step * j;
				}
				first += step;
				step *= 2;
			}
		}

		FeatureFinder * SurfFinder::clone()
		{
			return new SurfFinder( *this );
		}

		void SurfFinder::setDescriptorCreator( DescriptorCreator * theValue, bool own )
		{
			if ( descriptorCreator && ownDescrCreator ) delete descriptorCreator;
			else if ( otherDecrCreator && ownDescrCreator ) delete otherDecrCreator;
			descriptorCreator = NULL;
			otherDecrCreator = NULL;
			ownDescrCreator = false;
			
			if(!theValue) return;

			descriptorCreator = dynamic_cast<SurfDescriptor*>( theValue );
			if ( !descriptorCreator ) otherDecrCreator = theValue;
			ownDescrCreator = own;
		}

		SurfFinder::~SurfFinder()
		{
			if ( ownDescrCreator )
			{
				delete descriptorCreator;
				delete otherDecrCreator;
			}
		}

		uint SurfFinder::findFeaturesDerived( ViewT * v )
		{
			using namespace std;

			//check input sanity
			if ( NULL == v ) return 0;
			view = v;
			ImageT* img = view->getImage();
			if ( NULL == img ) return 0;
			offX = img->getROI().getX(); offY = img->getROI().getY();
			height = img->getROI().getHeight(), width = img->getROI().getWidth();

			//convert to grayscale if necessary
			if ( 3 == img->getNumChannels() )
			{
				gray.init( width, height, 1 );
				convgray.process( img, &gray );
				img = &gray;
			}
			ImageT *img4DescriptorCreator = img;
			candidates.init( 0, 0, width, height );

			//double images size if required
			if ( m_doubleSize )
			{
				doubled.init( width * 2, height * 2, 1 );
				upsampler.process(img, &doubled);
				img = &doubled;
			}

			integralImage.init(img);
			
			if ( descriptorCreator )
			{
				if ( m_doubleSize ) descriptorCreator->prepare( img4DescriptorCreator );
				else descriptorCreator->prepare( img4DescriptorCreator, &integralImage );
			}

			num_feats = 0;

			int max_interval = 2;
			for ( uint i = 2; i < m_intervals; ++i )
			{
				if ( filter_sizes[0][i] < width && filter_sizes[0][i] < height )
				{
					max_interval = i + 1;
				}
			}
			if ( max_interval < 3 ) return 0;

			int max_octave = 1;
			for ( uint i = 1; i < m_octaves; ++i )
				if ( filter_sizes[i][max_interval - 1] < width && filter_sizes[i][max_interval - 1] < height )
					max_octave = i + 1;

			DetImg det0( width, height, 1 ), det1( width, height, 1 ), det2( width, height, 1 );
			DetImages dets = { { &det0, &det1, &det2 } };

			bool has_dets[3] = { false, false, false };
			int cur_det = 0;
			for ( int o = 0; o < max_octave; ++o )
			{
				for ( int i = 0; i < max_interval; ++i )
				{
					if ( i < cur_det && has_dets[i] ) continue;
					else if ( has_dets[cur_det] ) continue;
					fastHessian( integralImage, filter_sizes[o][i], *dets[cur_det] );
					has_dets[cur_det] = true;
					++cur_det;
					if ( 3 == cur_det )
					{
						findExtrema<greater_equal<DetStoreT> > ( m_threshold, dets, o, i, integralImage );
						findExtrema<less_equal<DetStoreT> > ( -m_threshold, dets, o, i, integralImage );
						if ( 2 == i )
						{
							DetImg* temp = dets[0];
							dets[0] = dets[1]; dets[1] = dets[2];
							dets[2] = temp;
						}
						else if ( 3 == i )
						{
							DetImg* temp = dets[1];
							dets[1] = dets[2]; dets[2] = temp;
						}
						--cur_det;
						has_dets[cur_det] = false;
					}
				}
				if ( 3 == max_interval )
				{
					--cur_det;
					has_dets[cur_det] = false;
				}
			}

// 			DBGOUT ( DBGALL, "SurfFinder: %d features before filtering\n", candidates.getSize());
			filterFeatures ( candidates, minDist, minFraction, maxPts );
// 			DBGOUT ( DBGALL, "SurfFinder: %d features after filtering\n", candidates.getSize());

			FeatureGrid::Iterator it = candidates.getBegin();
			FeatureGrid::Iterator endIt = candidates.getEnd();
			while ( it != endIt )
			{
				FeaturePoint * fp = *it;
				(*fp)(0) += float(offX);
				(*fp)(1) += float(offY);
				fp->setView ( view );
				
// 				printf("SurfFinder: adding feature point\n");
// 				std::cout << *fp << std::endl;

				FeaturePoint *fpSame = view->pick ( ( *fp ) ( 0 ), ( *fp ) ( 1 ), id );
				if ( fpSame )
				{
					DBGOUT ( DBGALL, "SurfFinder: point occupied by scale %f (tried %f)\n", fpSame->getScale(), fp->getScale() );
				}

				if ( view->add ( fp ) != GRID_SUCCESS )
				{
					DBGOUT ( DBGALL, "SurfFinder: view->add() failed\n" );
					++it;
					delete fp;
					continue;
				}
				if ( descriptorCreator ) descriptorCreator->makeDescriptor( *fp );
				++num_feats;
				++it;
			}

			if ( otherDecrCreator )
				otherDecrCreator->performAction ( view );

			return num_feats;
		}

		void SurfFinder::fastHessian( const SurfIntegralImageT& image, uint filter_sz, DetImg& det ) const
		{
			int width_n = width - filter_sz;//removed +1 because integralimage evaluates a closed interval
			int height_n = height - filter_sz;

			uint lobe_min = filter_sz / 3;
			uint lobe_max = lobe_min * 2 - 1;
			uint border_max = ( filter_sz - lobe_max ) / 2;
			uint border_in = 1;
			uint border_out = ( filter_sz - ( lobe_max + 2 ) ) / 2;
			uint out_in_lmin = border_out + border_in + lobe_min;

			double factor = 1.0 / ( filter_sz * filter_sz ); factor *= factor;

			//the border of width (filter_sz / 2) is ignored
			Image1DIterator<DetStoreT> linedst = det.getXIterator( filter_sz / 2, filter_sz / 2 );
			uint step = det.getWStep();

			for ( int y = 0; y < height_n; ++y, linedst.nextLine( step ) )
			{
				Image1DIterator<DetStoreT> it = linedst;
				for ( int x = 0; x < width_n; ++x, ++it )
				{
					double XX =
					        image.getSum( x, y + border_max, filter_sz, lobe_max )
					        - 3 * image.getSum( x + lobe_min, y + border_max, lobe_min, lobe_max );
					double YY =
					        image.getSum( x + border_max, y, lobe_max, filter_sz )
					        - 3 * image.getSum( x + border_max, y + lobe_min, lobe_max, lobe_min );
					double XY =
					        image.getSum( x + border_out, y + border_out, lobe_min, lobe_min )
					        - image.getSum( x + out_in_lmin, y + border_out, lobe_min, lobe_min )
					        - image.getSum( x + border_out, y + out_in_lmin, lobe_min, lobe_min )
					        + image.getSum( x + out_in_lmin, y + out_in_lmin, lobe_min, lobe_min );

					int lap_sign = ( XX + YY >= 0 ? 1 : -1 );
					double determinant = ( XX * YY - XY * XY * 0.81 ) * factor;
					*it = determinant > m_threshold ? determinant * lap_sign : 0;
				}
			}
		}

		template<typename Arg2NotExtremum>
		void SurfFinder::findExtrema( double thresh, DetImages& dets, int cur_oct, int cur_int, const SurfIntegralImageT& image )
		{
			/// second pixel for biggest filter
			int shift = filter_sizes[cur_oct][cur_int] / 2 + 1;

			Image1DIterator<DetStoreT> pt0, pt1, pt2;
			Arg2NotExtremum arg2NotExtremum;
			workT extremum, x, y, s;
			FeaturePoint* fp;
			bool ok;

			for ( int row = shift; row < height - shift; ++row )
			{
				for ( int col = shift; col < width - shift; ++col )
				{

					extremum = ( *dets[1] )( col, row ); // x, y
					if ( arg2NotExtremum ( thresh, extremum ) ) continue;
					ok = true;

					// TOP + BOTTOM
					pt0 = dets[1]->getXIterator( row - 1, col - 1 );
					pt2 = dets[1]->getXIterator( row + 1, col - 1 );
					for ( int i = 0; i < 3 && ok; ++i, ++pt0, ++pt2 )
					{
						if ( arg2NotExtremum ( *pt0, extremum ) ) ok = false;
						if ( arg2NotExtremum ( *pt2, extremum ) ) ok = false;
					}
					if ( !ok ) continue;

					// LEFT
					if ( arg2NotExtremum ( ( *dets[1] )( col - 1, row ), extremum ) )
						continue;
					// RIGHT
					if ( arg2NotExtremum ( ( *dets[1] )( col + 1, row ), extremum ) )
						continue;

					// ANOTHER LEVELS
					for ( int row2 = row - 1; row2 <= row + 1; ++row2 )
					{
						pt0 = dets[0]->getXIterator( row2, col - 1 );
						pt2 = dets[2]->getXIterator( row2, col - 1 );
						for ( int col2 = col - 1; col2 <= col + 1 && ok; ++col2, ++pt0, ++pt2 )
						{
							if ( arg2NotExtremum ( *pt0, extremum ) || arg2NotExtremum ( *pt2, extremum ) )
								ok = false;
						}
					}
					if ( !ok ) { ++col; continue; } // small optimization

					// copy 3x3x3
					for ( int i = 0; i < 3; ++i )
					{
						pt0 = dets[0]->getXIterator( row - 1 + i, col - 1 );
						pt1 = dets[1]->getXIterator( row - 1 + i, col - 1 );
						pt2 = dets[2]->getXIterator( row - 1 + i, col - 1 );
						for ( int j = 0; j < 3; ++j )
						{
							store( 0, i, j ) = *pt0; ++pt0;
							store( 1, i, j ) = *pt1; ++pt1;
							store( 2, i, j ) = *pt2; ++pt2;
						}
					}

					x = col, y = row;
					if ( goodContrast()  && notAtEdge( cur_oct, cur_int, col, row, image )
					                && subCoorQuadratic( x, y ) )
					{
						s = 1.2 * ( ( filter_sizes[cur_oct][cur_int - 1] ) / 9.0 );
						if ( m_doubleSize ) { x = (x-0.5)/2.0; y = (y-0.5)/2.0; s /= 2.0; }

// 						fp = new FeaturePoint( x + 0.5, y + 0.5, id );
						fp = new FeaturePoint( x, y, id );
						fp->setScale( static_cast<float>( s ) );
						fp->setStrength( static_cast<float>( extremum ) );
						candidates.add ( fp );
					}

					++col; // small optimization
				}
			}
		}

		/**
		 * The iterator must be:
		 * (-1, -1) (0, -1) (1, -1)
		 * (-1,  0) (0,  0) (1,  0)
		 * (-1,  1) (0,  1) (1,  1)
		 */
		bool SurfFinder::subCoorQuadratic( workT & x, workT & y ) const
		{
			float a = store( 1, 0, 0 );
			float b = store( 1, 0, 1 );
			float c = store( 1, 0, 2 );
			float d = store( 1, 1, 0 );
			float sub = store( 1, 1, 1 ); //DC-part
			float e = store( 1, 1, 2 );
			float f = store( 1, 2, 0 );
			float g = store( 1, 2, 1 );
			float h = store( 1, 2, 2 );
			float r[8] = {a - sub, b - sub, c - sub, d - sub, e - sub, f - sub, g - sub, h - sub};
			bool hasworked = adore::core::subCoorQuadratic( a, b, r );
			x += a; y += b;
			return hasworked;
		}

		bool SurfFinder::goodContrast() const
		{
//			AVector d3d( 3 ), temp( 3 );
//			deriv3D ( d3d );
//			/// remove low contrast features
//			double contr = matMul2( d3d, d3d );
//			double test = store( 1, 1, 1 ) + 0.5 * contr;
//			if ( 7.68 > abs( test ) ) return false;
			return true;
		}

		bool SurfFinder::notAtEdge( int cur_oct, int cur_int, int x, int y, const SurfIntegralImageT& image ) const
		{
			/// eliminate edge responses
			uint sz = filter_sizes[cur_oct][cur_int - 1];
			uint lobe_min = sz / 3;
			uint lobe_max = lobe_min * 2 - 1;
			uint border_max = ( sz - lobe_max ) / 2;

			double factor = 1.0 / ( sz * sz );
			int xx = x - sz / 2;
			int yy = y - sz / 2;
			double XX =
			        image.getSum( xx, yy + border_max, sz, lobe_max )
			        - 3 * image.getSum( xx + lobe_min, yy + border_max, lobe_min, lobe_max );
			double YY =
			        image.getSum( xx + border_max, yy, lobe_max, sz )
			        - 3 * image.getSum( xx + border_max, yy + lobe_min, lobe_max, lobe_min );
			double tr2 = ( XX + YY ) * factor; tr2 *= tr2;
			double ratio = tr2 / store( 1, 1, 1 );
			if ( 12.1 < abs( ratio ) ) return false;
			return true;
		}

		void SurfFinder::deriv3D ( AVector& b ) const
		{
			b( 0 ) = ( store( 1, 1, 2 ) - store( 1, 1, 0 ) ) * 0.5;
			b( 1 ) = ( store( 1, 2, 1 ) - store( 1, 0, 1 ) ) * 0.5;
			b( 2 ) = ( store( 2, 1, 1 ) - store( 0, 1, 1 ) ) * 0.5;
		}

	} // core
} // adore

