

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#include "delaunay.h"

#include <vector>
#include <cassert>
#include <cmath>
#include <algorithm>
#include <limits>
#include <adrcore/util/debug.h>

#include "geometry2.h"

namespace adore
{
	namespace core
	{
		static const double epsilon = 0.0000000;

		/**	Comparator class that orders 2D points by one of their coordinates
		*/
		class AxisCmp2D
		{
				unsigned axis;
				Point2D* source;

			public:
                AxisCmp2D( Point2D* s, unsigned a ): axis( a ), source( s ) {}

				bool operator()( unsigned lv, unsigned rv )
				{
					return source[lv][axis] < source[rv][axis];
                }
		};

		/**	Subdivide the entire point set into 2 subset of roughly the same size.
		*	The entire set is subdivided by an axis-aligned splitting line.
		*	of the axis.
		*
		*	@param src Source array of points
		*	@param pts Indices of the point set members
		*	@param axis Axis of separation
		*	@param a Output subset of points to the negative side of the splitting line
		*	@param a Output subset of points to the positive side of the splitting line
		*
		*	@return Coordinate of the splitting line
		*/
		inline double pointsetPartition( Point2D* src, std::vector<unsigned>& pts, unsigned axis, std::vector<unsigned>& a,std::vector<unsigned>& b )
		{
			std::vector<unsigned> temp( pts );

			AxisCmp2D cmp( &src[0],axis );
			std::sort( temp.begin(),temp.end(),cmp );
			a.assign( temp.begin(),temp.begin()+temp.size()/2 );
			b.assign( temp.begin()+temp.size()/2,temp.end() );
			double alpha = 0.5 * ( src[temp[temp.size()/2-1]][axis]+src[temp[temp.size()/2]][axis] );

			return alpha;

		}

		/**	Make the first simplex of the simplex wall.
		*	The function gets the main set of points and two subset, calculated
		*	before by the partitioning function.
		*
		*	The algorithm runs as follows:
		*	1. Look in the entire point set for the point nearest to the splitting line.
		*	2. Look for the second point that is nearest to the first point
		*	but lies on the other side of the splitting line
		*	3. Look for the third point so that the triangle formed by three
		*	points has the minimal circumradius.
		*	4. Return all three points
		*
		*	@param src Source array of points
		*	@param pts Indices of the point set members
		*	@param axis Axis of separation
		*	@param alpha Splitting line coordinate
		*	@param a Subset of points to the negative side of the splitting line
		*	@param a Subset of points to the positive side of the splitting line
		*
		*	@return Constructed simplex
		*/
		inline Simplex2D makeFirstSimplex( Point2D* src, std::vector<unsigned>& pts, std::vector<unsigned>& a, std::vector<unsigned>& b, unsigned axis, double alpha )
		{
			double dist = std::numeric_limits<double>::infinity();
			double temp;
			unsigned index[3] = {0,0,0};

			// look for the point closest to the splitting plane
			for ( unsigned i = 0; i < pts.size(); ++i )
			{
				temp = std::abs( src[pts[i]][axis]-alpha );
				if ( temp < dist )
				{
					index[0] = pts[i];
					dist = temp;
				}
			}

			dist = std::numeric_limits<double>::infinity();
			bool exist = false;
			if ( src[index[0]][axis] < alpha )
			{
				for ( unsigned i = 0; i < pts.size(); ++i )
				{
					if ( src[pts[i]][axis]>=alpha )
					{
						temp = ( src[pts[index[0]]]-src[pts[i]] )();
						if ( temp<dist )
						{
							index[1] = pts[i];
							dist = temp;
							exist = true;
						}
					}
				}
			}
			else
			{
				for ( unsigned i = 0; i < pts.size(); ++i )
				{
					if ( src[pts[i]][axis]<alpha )
					{
						temp = ( src[pts[index[0]]]-src[pts[i]] )();
						if ( temp<dist )
						{
							index[1] = pts[i];
							dist = temp;
							exist = true;
						}
					}
				}
			}

			if ( !exist )
			{
				Simplex2D result( 0,0,0 );
				return result;
			}

			dist = std::numeric_limits<double>::infinity();
			exist = false;
			for ( unsigned i = 0; i < pts.size(); ++i )
			{
				if ( pts[i]!=index[0] && pts[i]!= index[1] )
				{
					temp = circumradius( src,index[0],index[1],pts[i] );
					assert( temp!=dist && "More than 3 points on same circumcircle!" );
					if ( temp<dist )
					{
						index[2] = pts[i];
						dist = temp;
						exist = true;
					}
				}
			}

			if ( !exist )
			{
				Simplex2D result( 0,0,0 );
				return result;
			}

			Simplex2D result( index[0],index[1],index[2] );
//			Point2D c = circumcenter( src,index[0],index[1],index[2] );
			return result;
		}

		/**	Test if the face vertices are in the set
		*	@param pts Point set
		*	@param f Face to test
		*/
		inline bool contains( std::vector<unsigned>& pts, Face2D f )
		{
			bool first = 0;
			bool second = 0;
			for ( unsigned i = 0; i < pts.size(); ++i )
			{
				first |= f[0] == pts[i];
				second |= f[1] == pts[i];
			}
			return first && second;
		}

		/**	Delaunay distance from the face to the point
		*	Delaunay distance is defined as a signed circumradius
		*	of the triangle formed by the edge f and the point p.
		*	The distance is positive if the point p is in the same
		*	halfspace relative to the edge f as the circumcenter and
		*	negative otherwise.
		*	@param src Source array of points
		*	@param f Face
		*	@param p Point
		*	@return Delaunay distance
		*/
		inline double distance( Point2D* src, Face2D f, unsigned p )
		{
			Circle2D c = circumcircle( src,f[0],f[1],p );

			if ( halfspace( src,f,src[p],epsilon )==halfspace( src,f,c.center,epsilon ) )
				return c.radius;
			else
				return -c.radius;
		}

		/**	Normalize the simplex so that its area is positive
		*	@param src Source array of points
		*	@param t Simplex to normalize
		*/
		inline void normalize( Point2D* src, Simplex2D* t )
		{
			assert( t );

			// swap any 2 vertices if area is negative
			// to make the winding of the simplex counterclockwise
			if ( area( src,( *t )[0],( *t )[1],( *t )[2] ) < 0.0 )
			{
				unsigned temp = ( *t )[0];
				( *t )[0] = ( *t )[1];
				( *t )[1] = temp;
			}

			assert( area( src,( *t )[0],( *t )[1],( *t )[2] ) > 0.0 );
		}

		/**	Given the face f find the point p that minimizes Delaunay
		*	distance with the edge f. This function can be later
		*	accelerated with the Uniform Grid technique.
		*	@param src Source array of points
		*	@param pts Point set
		*	@param f Face
		*/
		inline Simplex2D makeSimplex( Point2D* src, std::vector<unsigned>& pts, Face2D f )
		{
			assert( src[f[0]]!=src[f[1]] );

			double dist = std::numeric_limits<double>::infinity();
			double temp;
			unsigned index = 0;
			bool exist = false;
			Simplex2D result( 0,0,0 );
			for ( unsigned i = 0; i < pts.size(); ++i )
			{
				if ( pts[i] != f[0] && pts[i] != f[1] && halfspace( src,f,src[pts[i]],epsilon ) > 0 )
				{
					temp = distance( src,f,pts[i] );
					assert( temp!=dist && "More than 3 points on same circumcircle!" );
					if ( temp < dist )
					{
						dist = temp;
						index = pts[i];
						exist = true;
					}
				}
			}
			if ( exist )
			{
				result[0] = f[0];
				result[1] = f[1];
				result[2] = index;
			}

			return result;
		}

		/**	Recursive dewall function for calculation of Delaunay triangulation.
		*	@param src Source array of points
		*	@param pts Point set indices
		*	@param axis Axis of subdivision
		*	@param afl Active Face List
		*	@param result Receiver of calculated simplexes
		*/
		void dewall( Point2D* src, std::vector<unsigned>& pts, unsigned axis, FaceList2D& afl, SimplexList2D& result )
		{
			Face2D temp;
			FaceList2D AFL[3];

			std::vector<unsigned> a;
			std::vector<unsigned> b;

			double alpha = pointsetPartition( src,pts,axis,a,b );

			if ( afl.empty() )
			{
				Simplex2D t = makeFirstSimplex( src,pts,a,b,axis,alpha );
				if ( !t.nullSimplex() )
				{
					normalize( src,&t );

					temp[0] = t[0];
					temp[1] = t[1];

					if ( contains( a,temp ) )
						AFL[1].insert( temp );
					else if ( contains( b,temp ) )
						AFL[2].insert( temp );
					else
						AFL[0].insert( temp );

					temp[0] = t[1];
					temp[1] = t[2];

					if ( contains( a,temp ) )
						AFL[1].insert( temp );
					else if ( contains( b,temp ) )
						AFL[2].insert( temp );
					else
						AFL[0].insert( temp );

					temp[0] = t[2];
					temp[1] = t[0];
					if ( contains( a,temp ) )
						AFL[1].insert( temp );
					else if ( contains( b,temp ) )
						AFL[2].insert( temp );
					else
						AFL[0].insert( temp );

					result.insert( t );
				}
			}
			else
			{
				afl.start();
				while ( afl.next( temp ) )
				{
					if ( contains( a,temp ) )
						AFL[1].insert( temp );
					else if ( contains( b,temp ) )
						AFL[2].insert( temp );
					else
						AFL[0].insert( temp );
				}
			}

			while ( !AFL[0].empty() )
			{
				Face2D f = AFL[0].extract();
				Simplex2D t = makeSimplex( src,pts,f );
				if ( !t.nullSimplex() )
				{
					normalize( src,&t );
					result.insert( t );

					temp[0] = t[0];
					temp[1] = t[1];
					if ( temp!=f )
					{
						if ( contains( a,temp ) )
							AFL[1].update( temp );
						else if ( contains( b,temp ) )
							AFL[2].update( temp );
						else
							AFL[0].update( temp );
					}

					temp[0] = t[1];
					temp[1] = t[2];
					if ( temp!=f )
					{
						if ( contains( a,temp ) )
							AFL[1].update( temp );
						else if ( contains( b,temp ) )
							AFL[2].update( temp );
						else
							AFL[0].update( temp );
					}

					temp[0] = t[2];
					temp[1] = t[0];
					if ( temp!=f )
					{
						if ( contains( a,temp ) )
							AFL[1].update( temp );
						else if ( contains( b,temp ) )
							AFL[2].update( temp );
						else
							AFL[0].update( temp );
					}
				}
			}

			axis = ( axis+1 )%2;

			if ( !AFL[1].empty() )
				dewall( src,a,axis,AFL[1],result );
			if ( !AFL[2].empty() )
				dewall( src,b,axis,AFL[2],result );
		}

		void delaunay2( double* points, unsigned count, addsimplex receiver, void* extra )
		{
#ifndef NDEBUG
			if ( count < 3 ) printf( "count in delaunay is %d\n", count );
#endif
			assert( points );
			assert( receiver );
			assert( count >= 3 );

			Point2D* src = reinterpret_cast<Point2D*>( points );
			SimplexList2D result( receiver,extra );

			if ( count == 3 )
			{
				DBGOUT( DBGNORMAL, "count in delaunay is 3\n" );
				Simplex2D t( 1,2,3 );
				normalize( src,&t );
				result.insert( t );
			}
			else
			{
				FaceList2D afl;
				std::vector<unsigned> pts( count );
				for ( unsigned i = 0; i < count; ++i )
					pts[i] = i;

				dewall( src,pts,0,afl,result );
			}
		}
	}
}
