

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWCAMERA_H
#define ADORE_NVIEWCAMERA_H

#include <adrcore/util/mathutil.h>
#include <adrcore/util/matmul2.h>
#include <vector>

namespace adore
{
	namespace core
	{

		/**
		* A Camera, represented by a 3x4-Matrix
		*  
		*/
		class Camera
		{
			public:
				/**
				 * Constructor.
				 */
				Camera();

				/**
				 * Copy-Constructor.
				 */
				Camera ( const Camera &cam );

				explicit Camera ( const AMatrix &cam );

				/**
				 * Destructor.
				 */
				virtual ~Camera();

				/**
				 * Projects a point using this Camera.
				 * @param p3D a homogeneous 3D-point.
				 * @return a homogeneous 2D-point that will hold the result.
				 */
                inline const AVector project ( const AVector &p3D ) const
				{
					// AVector val = matMul ( getProjMat(), p3D );
					AVector val(3);
					matMul2 ( getProjMat(), p3D, val );
					distort ( val );
					return val;
				}
                inline void project ( const AVector &p3D, AVector &val ) const
				{
					matMul2 ( getProjMat(), p3D, val );
					distort ( val );
				}


				void compose();

				void decompose() const;

				AMatrix &getProjMat() { return projMat; }

				const AMatrix &getProjMat() const { return projMat; }

				/**
				 * this returns the rotation matrix of the camera
				 * that is the rotation the camera performs on the environment
				 * The cameras rotation in the world frame is the inverse/transposed
				 */
				AMatrix &getRotationMatrix() {return R;}

				const AMatrix &getRotationMatrix() const {return R;}

				AMatrix &getCalibrationMatrix() {return K;}

				const AMatrix &getCalibrationMatrix() const {return K;}

				AMatrix &getCalibrationMatrixInverse() {return Kinv;}

				const AMatrix &getCalibrationMatrixInverse() const {return Kinv;}

				AVector &getLocation() {return loc;}

				const AVector &getLocation() const {return loc;}

				double getSkew() const {return K ( 0, 1 );}
				double getPPX() const {return K ( 0, 2 );}
				double getPPY() const {return K ( 1, 2 );}
				double getAspectX() const {return K ( 0, 0 );}
				double getAspectY() const {return K ( 1, 1 );}
				//height / width
				double getAspectRatio() const {return getAspectY() / getAspectX();}

				inline const Camera &operator= ( const Camera &c )
				{
					projMat = c.getProjMat();
					return *this;
				}

				/**
				 * calculates the oriented camera center c of a camera P = [p1, p2, p3, p4].
				 * c = [|p2, p3, p4|,|p3, p4, p1|,|p4, p1, p2|,|p1, p2, p3|]
				 *
				 * inherently normalizes exponents
				 *
				 *(not const, because the projMat is normalized)
				 */
				const AVector calculateCameraCenter();
				void calculateCameraCenter ( AVector &result );

				double getDepth ( const AVector &pt ) const;

				int getSign ( const AVector &pt ) const;

				static void normalizeK ( AMatrix &K ) /*const*/;
				
				virtual void writeAt(std::ostream &os) const
				{
					os << projMat;
				}
				
			private:

				//double camData[12];
				AMatrix projMat;
				mutable AMatrix K;
				mutable AMatrix Kinv;
				mutable AMatrix R;

				mutable AVector3D loc;
				mutable AMatrix temp33;
				mutable AVector temp3;

				inline void getValuesForDepthCalculation ( const AVector &pt, double &detM, double &T, double &w, double &m ) const;

				//distorts the projected 2D point according to the camera model
				virtual void distort ( AVector &coor ) const {;}
		};

		bool isIdentityCam ( Camera *c );

        bool isIdentityCam ( const AMatrix & cam );
		
		inline std::ostream& operator<<( std::ostream& stream, Camera const& obj )
		{
			obj.writeAt( stream );
			return stream;
		}

	}

}

#endif
