

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "view.h"
#include <adrcore/util/printmat.h>
#include <adrcore/util/numerator.h>
#include <adrcore/util/scaleshifttrafo.h>

namespace adore
{
	namespace core
	{

		Numerator<uint> viewNumerator;

		uint nextViewId()
		{
			return viewNumerator();
		}

		void leastViewId( uint id )
		{
			viewNumerator.atLeast( id );
		}

		ViewBase::ViewBase():
				x( -1 ),
				y( -1 ),
				width( -1 ),
				height( -1 ),
				sameDist( 0.75 ),
				normalized( false ),
				trafo( 3, 3 ),
				invTrafo( 3, 3 ),
				id( nextViewId() ),
				pointMode( POINT_DELETE )
		{
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
			{
				features[i] = NULL;
				triangles[i] = NULL;
			}
			setIdentity( trafo );
			setIdentity( invTrafo );
		}

		ViewBase::ViewBase( ViewBase &view ):
				x( view.getX() ),
				y( view.getY() ),
				width( view.getWidth() ),
				height( view.getHeight() ),
				sameDist( view.getSameDist() ),
				normalized( false ),
				trafo( 3, 3 ),
				invTrafo( 3, 3 ),
				id( view.getId() ),
				pointMode( POINT_DELETE )
		{
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
			{
				features[i] = NULL;
				triangles[i] = NULL;
			}
//			printf( "Copy-Ctor in View\n" );
			initialize();
		}

		ViewBase::~ViewBase()
		{
//			DBGOUT( DBGNORMAL, "feature size: %d\n", features.getSize() );
			reset();
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
			{
				delete features[i];
				delete triangles[i];
			}

		}

		void ViewBase::copy( ViewBase *view )
		{
			DBGOUT( DBGNORMAL, "ViewBase copies ViewBase\n" );
			assert( view );
			x = view->getX();
			y = view->getY();
			width = view->getWidth();
			height = view->getHeight();
			sameDist = view->getSameDist();
			normalized = view->getNormalized();
			trafo = view->getTrafo();
			invTrafo = view->getInvTrafo();
			id = view->getId();
			pointMode = view->getPointMode();
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
			{
				delete features[i];
				if ( view->features[i] )
				{
					DBGOUT( DBGNORMAL, "Creatign feature container %d\n", i );
					features[i] = new FeatureContainer( *view->features[i] );
					DBGOUT( DBGNORMAL, "sizes %d - %d\n", view->features[i]->getSize(), features[i]->getSize() );
				}
				delete triangles[i];
				if ( view->triangles[i] ) triangles[i] = new vector<unsigned>( *view->triangles[i] );
			}
		}

		void ViewBase::reset()
		{
			//DBGOUT(DBGNORMAL, "resetting view containing %d features\n", getNumAllFeaturePoints());
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
			{
				if ( features[i] ) features[i]->clean();
				if ( triangles[i] ) triangles[i]->clear();
			}
			segments.clear();
		}

		FeaturePoint *ViewBase::pick( coorT x, coorT y, int featsID, double maxDist )
		{
			if ( maxDist < 0 ) maxDist = sameDist;
			if ( !features[featsID] ) return NULL;
			return features[featsID]->pick( x, y, maxDist );
		}

		FeaturePoint *ViewBase::pick( coorT x, coorT y, int featsID )
		{
			if ( !features[featsID] ) return NULL;
			return features[featsID]->pick( x, y, sameDist );
		}

		void ViewBase::pickNear ( double x, double y, std::vector<FeaturePoint*>& points, int featsID, double maxDist ) const
		{
			if ( maxDist < 0 ) maxDist = sameDist;
			points.clear();
			if ( !features[featsID] ) return;
			features[featsID]->pickNear(x, y, maxDist, points);
		}

		GRID_ERR ViewBase::add( FeaturePoint *const vec, bool checkExistence )
		{
			uint id = vec->getFeatureTypeId();
			assert_msg( id < maxFeatureTypeNo, "%d >= %d\n", id, maxFeatureTypeNo );

			if ( !features[id] )
			{
				features[id] = new FeatureContainer( ( pointMode == POINT_DELETE ) );
				features[id]->init( x, y, width, height );
			}
			FeatureContainer *fc = features[id];

//			printf( "(%f, %f)-(%f, %f)\n", x, y, width, height );
//			DBGOUT( DBGNORMAL, "feature size: %d\n", features.getSize() );
			assert( vec->getView() == NULL || vec->getView() == this );
			vec->setView( this );
			return fc->add( vec, checkExistence );
		}

		GRID_ERR ViewBase::movePoint( FeaturePoint *nP, double newX, double newY, double maxDist )
		{
//			printf( "VIew pointMoved: sameDist: %f, maxDist: %f\n", sameDist, maxDist );
			uint id = nP->getFeatureTypeId();

			assert( id < maxFeatureTypeNo );
			assert( features[id] );
			assert( nP->getView() == this );

			FeatureContainer *fc = features[id];
			assert( nP->getView() == NULL || nP->getView() == this );

			bool oldDeleteMode = fc->getDeletePoints();
			fc->setDeletePoints ( false );

			//if thi assertion failes, make sure to correct it also below at the assingment of newX/Y
			assert( ( *nP ) ( 2 ) == 1.0 );

			double oldx = ( *nP ) ( 0 );
			double oldy = ( *nP ) ( 1 );
			FeaturePoint *rem = fc->remove( oldx, oldy, maxDist );

#ifndef NDEBUG
			if ( rem != nP )
			{
				printf( "View: wrong point at distance %f removed (sameDist: %f, maxDist: %f)\n", sqrtf( nP->sqrDist( *rem ) ), sameDist, maxDist );
				exit( 1 );
			}
#endif
			assert( rem == nP );//point must exist
			ADORE_UNUSED( rem );

			( *nP ) ( 0 ) = newX;
			( *nP ) ( 1 ) = newY;

			GRID_ERR retVal = fc->add( nP, true /*checkExistence*/ );
			printGridError( retVal );

			assert( retVal == GRID_SUCCESS );

			fc->setDeletePoints ( oldDeleteMode );
			return retVal;
		}

		void ViewBase::remove( const FeaturePoint * const nP )
		{
			assert( nP->getView() == this );
			assert( ( *nP )( 2 ) == 1.0 );
			remove( ( *nP )( 0 ), ( *nP )( 1 ), nP->getFeatureTypeId() );
		}

		void ViewBase::remove( const double x, const double y, const uint id )
		{
			assert( id < maxFeatureTypeNo );

			if ( !features[id] ) return;
			FeatureContainer *fc = features[id];
			fc->remove( x, y, sameDist );
		}


		bool ViewBase::exists( FeaturePoint *const vec )
		{
			uint id = vec->getFeatureTypeId();
			assert( id < maxFeatureTypeNo );
			return features[id]->exists( vec );
		}

		void ViewBase::initialize()
		{
//			printf( "ViewBase::initializing (%f, %f, %f, %f, %f)\n", x, y, width, height, sameDist );
			assert( width > 0 );
			assert( height > 0 );
//			printf( "ViewBase normtrafo\n" );
			getNormalizationTrafo( width, height, trafo );
//			printf( "ViewBase invtrafo\n" );
			invTrafo = trafo;
//			printf( "ViewBase invNormtrafo\n" );
			invertNormalizationMatrix( invTrafo );
//			printf( "ViewBase grid\n" );
			for ( uint i = 0; i < maxFeatureTypeNo; ++i ) if ( features[i] ) features[i]->init( x, y, width, height );
//			printf( "ViewBase is leaving initialize\n" );
		}

		void ViewBase::normalize()
		{
//			printf( "normalize?\n" );
			if ( normalized )
			{
				DBGOUT( DBGALL, "View already normalized\n" );
				return;
			}
			DBGOUT( DBGALL, "View %p gets normalized! Trafo is\n", this );
			printMat( DBGALL, trafo );
			x = trafo( 0, 0 ) * x + trafo( 0, 2 );
			y = trafo( 1, 1 ) * y + trafo( 1, 2 );
			width *= trafo( 0, 0 );
			height *= trafo( 1, 1 );
			sameDist *= trafo( 0, 0 );//two points are the same point, if there distance is less than this
//			printMat( trafo );
//			FeaturePoint *fp = *features[0]->getBegin();
//			printf( "View - normalize: %f, %f, %f\n", ( *fp )( 0 ), ( *fp )( 1 ), ( *fp )( 2 ) );
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
				if ( features[i] )
					features[i]->transform( trafo( 0, 2 ), trafo( 1, 2 ), trafo( 0, 0 ), trafo( 1, 1 ) );
//			printf( "View - normalize: %f, %f, %f\n", ( *fp )( 0 ), ( *fp )( 1 ), ( *fp )( 2 ) );
			for ( uint i = 0; i < segments.size(); ++i )
			{
				segments.at( i )->transform( trafo( 0, 2 ), trafo( 1, 2 ), trafo( 0, 0 ), trafo( 1, 1 ) );
			}
			normalized = true;
		}

		void ViewBase::denormalize()
		{
			if ( !normalized ) return;
			x = invTrafo( 0, 0 ) * x + invTrafo( 0, 2 );
			y = invTrafo( 1, 1 ) * y + invTrafo( 1, 2 );
			width *= invTrafo( 0, 0 );
			height *= invTrafo( 1, 1 );
			sameDist *= invTrafo( 0, 0 );//two points are the same point, if there distance is less than this
//			printMat( invTrafo );
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
				if ( features[i] )
					features[i]->transform( invTrafo( 0, 2 ), invTrafo( 1, 2 ), invTrafo( 0, 0 ), invTrafo( 1, 1 ) );
			for ( uint i = 0; i < segments.size(); ++i )
			{
				segments.at( i )->transform( invTrafo( 0, 2 ), invTrafo( 1, 2 ), invTrafo( 0, 0 ), invTrafo( 1, 1 ) );
			}
			normalized = false;
		}

		void ViewBase::setPointMode ( const POINT_MODE& theValue )
		{
			pointMode = theValue;
			for ( uint i = 0; i < maxFeatureTypeNo; ++i )
			{
				if ( features[i] )
				{
					features[i]->setDeletePoints( pointMode == POINT_DELETE );
				}
			}
		}

	}
}
