

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "mathutil.h"
#include "debug.h"
#include "printmat.h"
#include <adrcore/util/storageorder.h>
#include <adrcore/util/solvesvd.h>
#include <adrcore/random/random.h>
namespace adore
{
	namespace core
	{

		double HUBER_VALUE ( 0.01 );//normalized
		const double HUBER_RECOMMENDED ( 12 );//not normalized, use getNormalizedPixel on this

		int getRealRootsOfCubic ( double a, double b, double c, double *r1, double *r2, double *r3 )
		{
			double Q = ( a * a - 3 * b ) / 9.0;
			double R = ( 2.0 * a * a * a - 9.0 * a * b + 27.0 * c ) / 54.0;
			double athird = a / 3.0;
			if ( ( R * R ) < ( Q * Q * Q ) )
			{
				//three roots
				double theta = acos ( R/sqrt ( Q*Q*Q ) ) / 3.0;
				double sqrtQ = -2.0 * sqrt ( Q );

				*r1 = sqrtQ * cos ( theta ) - athird;
				*r2 = sqrtQ * cos ( theta + 2.0 * M_PI / 3.0 ) - athird;
				*r3 = sqrtQ * cos ( theta - 2.0 * M_PI / 3.0 ) - athird;
				return 3;
			}
			else
			{
				//one root
				double A = -signum ( R ) * cbrt ( fabs ( R ) +sqrt ( R * R - Q * Q * Q ) );
				if ( A == 0.0 ) *r1 = - athird;
				else
				{
					*r1 = A + Q / A - athird;
				}
				return 1;
			}
			return 0;
		}

		double intPow ( double val, int exp )
		{
			double pow = 1.0;

			if ( exp != 0 )
			{
				if ( exp < 0 )
				{
					if ( val == 0.0 )
					{
						return ( pow );
					}
					exp = -exp;
					val = 1.0/val;
				}
				for ( ; ; )
				{
					if ( exp & 01 )
						pow *= val;
					if ( exp >>= 1 )
						val *= val;
					else
						break;
				}
			}
			return ( pow );
		}

		int intPow ( int val, uint exp )
		{
			int pow = 1;

			if ( exp != 0 )
			{
				for ( ; ; )
				{
					if ( exp & 01 )
						pow *= val;
					if ( exp >>= 1 )
						val *= val;
					else
						break;
				}
			}
			return ( pow );
		}

		void pointAsLine ( double a, double b, double c, double &x1, double &y1, double &x2, double &y2, double maxX2, double maxY2 )
		{
			if ( fabs ( a ) < fabs ( b ) )
			{
				x1 = 0;
				y1 = -c/b;
				x2 = maxX2;
				y2 = -a/b*x2-c/b;
			}
			else
			{
				y1 = 0;
				x1 = -c/a;
				y2 = maxY2;
				x2 = -b/a*y2-c/a;
			}
		}


		bool intersect2lines( const AVector & l1, const AVector & l2, AVector & result, double min, double max, int idx )
		{
			if ( min >= max || ( idx != 0 && idx != 1 ) )
				return false;
			cross( l1, l2, result );
			if ( abs( result( 2 ) ) < numeric_limits<double>::epsilon() )
				return false;
			result /= result( 2 );
			if ( result( idx ) < min || result( idx ) > max )
				return false;
			return true;
		}

		bool pointAsLine2( double a, double b, double c, double & x1, double & y1, double & x2, double & y2,
		                   double width, double height, double x, double y )
		{
			AVector line( 3 );
			line = a, b, c;
			return pointAsLine2(line, x1, y1, x2, y2, width, height, x, y);
		}

		bool pointAsLine2( const AVector& line, AVector& p1, AVector& p2,
						   double width, double height, double x, double y )
		{
			return pointAsLine2(line, p1(0), p1(1), p2(0), p2(1), width, height, x, y);
		}

		bool pointAsLine2( const AVector& line, double &x1, double &y1, double &x2, double &y2,
						   double width, double height, double x, double y )
		{
			AVector top( 3 ), right( 3 ), bottom( 3 ), left( 3 );
			cross( x, y, 1, x + width, y, 1, top );
			cross( x + width, y, 1, x + width, y + height, 1, right );
			cross( x + width, y + height, 1, x, y + height, 1, bottom );
			cross( x, y + height, 1, x, y, 1, left );

			bool first_found = false;
			AVector test( 3 );
			bool ok = intersect2lines( top, line, test, x, x + width, 0 );
			if ( ok )
			{
				first_found = true;
				x1 = test( 0 ), y1 = test( 1 );
			}
			ok = intersect2lines( right, line, test, y, y + height, 1 );
			if ( ok && !first_found )
			{
				first_found = true;
				x1 = test( 0 ), y1 = test( 1 );
			}
			else if ( ok )
			{
				x2 = test( 0 ), y2 = test( 1 );
				return true;
			}
			ok = intersect2lines( bottom, line, test, x, x + width, 0 );
			if ( ok && !first_found )
			{
				first_found = true;
				x1 = test( 0 ), y1 = test( 1 );
			}
			else if ( ok )
			{
				x2 = test( 0 ), y2 = test( 1 );
				return true;
			}
			ok = intersect2lines( left, line, test, y, y + height, 1 );
			if ( ok && !first_found )
				return false;
			else if ( ok )
			{
				x2 = test( 0 ), y2 = test( 1 );
				return true;
			}
			return false;
		}


		double box_muller ( double m, double s )	/* normal random variate generator */
		{
			/* mean m, standard deviation s */
			double x1, x2, w, y1;
			static double lastM;
			static double lastS;
			static double y2;
			static int use_last = 0;

			if ( use_last && lastM == m && lastS == s )
			{
				y1 = y2;
				use_last = 0;
			}
			else
			{
				do
				{
					double rval = double ( intRand() ) / ( RAND_MAX + 1.0 );
					x1 = 2.0 * rval - 1.0;
					rval = double ( intRand() ) / ( RAND_MAX + 1.0 );
					x2 = 2.0 * rval - 1.0;
					w = x1 * x1 + x2 * x2;
				}
				while ( w >= 1.0 );

				w = sqrt ( ( -2.0 * log ( w ) ) / w );
				y1 = x1 * w;
				y2 = x2 * w;
				use_last = 1;
			}

			return ( m + y1 * s );
		}


		double valueToInterval ( double value, double lower, double higher )
		{
			assert ( higher > lower );
			double d = higher - lower;

			if ( value > higher )
			{
				int n = int ( ceil ( ( value - higher ) /d ) );
				return value - n * d;
			}
			else if ( value < lower )
			{
				int n = int ( ceil ( ( lower - value ) /d ) );
				return value + n * d;
			}
			return value;
		}


		bool fitParabola ( double left, double middle, double right, double &delta, double &value )
		{
			double a = ( ( left + right ) - 2.0 * middle ) / 2.0;
			delta = value = 0.0;

			// Not a parabola
			if ( a == 0.0 )
			{
				//DBGOUT(DBGALL, "fit this! (%f, %f, %f)", left, middle, right);
				delta = 0;
				value = middle;
				return false;
			}

			double c = ( ( ( left - middle ) / a ) - 1.0 ) / 2.0;
			double b = middle - c * c * a;

			if ( c < -0.5 || c > 0.5 ) return false;

			delta = c;
			value = b;

			return ( true );
		}

		void makeDiagOfROfRQPosisitve ( AArray<double, 2> &R, AArray<double, 2> &Q )
		{
//			bool right = isRightHand( Q );

//			int numMods = 0; //to preserve Handedness of Q, only 0 or two modifications are allowed
			if ( R ( 2, 2 ) < 0 )
			{
				R ( 0, 2 ) = -R ( 0, 2 );
				R ( 1, 2 ) = -R ( 1, 2 );
				R ( 2, 2 ) = -R ( 2, 2 );
				Q ( 2, 0 ) = -Q ( 2, 0 );
				Q ( 2, 1 ) = -Q ( 2, 1 );
				Q ( 2, 2 ) = -Q ( 2, 2 );
//				++numMods;
			}
			if ( R ( 1, 1 ) < 0 )
			{
				R ( 0, 1 ) = -R ( 0, 1 );
				R ( 1, 1 ) = -R ( 1, 1 );
				Q ( 1, 0 ) = -Q ( 1, 0 );
				Q ( 1, 1 ) = -Q ( 1, 1 );
				Q ( 1, 2 ) = -Q ( 1, 2 );
//				++numMods;
			}
			if ( R ( 0, 0 ) < 0 )
//				if ( ( right && numMods == 1 ) || ( !right && ( numMods == 0 || numMods == 2 ) ) )
				{
					R ( 0, 0 ) = -R ( 0, 0 );
					Q ( 0, 0 ) = -Q ( 0, 0 );
					Q ( 0, 1 ) = -Q ( 0, 1 );
					Q ( 0, 2 ) = -Q ( 0, 2 );
				}

		}

		void invertUpperTriangular3x3Matrix ( AArray<double, 2> &mat )
		{
			double a = mat ( 0, 0 );
			double b = mat ( 0, 1 );
			double c = mat ( 0, 2 );
			double d = mat ( 1, 1 );
			double e = mat ( 1, 2 );
			double f = mat ( 2, 2 );
			mat ( 0, 0 ) = 1.0/a;
			mat ( 0, 1 ) = -b/ ( a*d );
			mat ( 0, 2 ) = ( e*b-c*d ) / ( f*d*a );
			mat ( 1, 1 ) = 1.0/d;
			mat ( 1, 2 ) = -e/ ( f*d );
			mat ( 2, 2 ) = 1.0/f;
		}
		
		bool invert22 (AArray<double, 2>& mat)
		{
			double a = mat(0, 0);
			double b = mat(0, 1);
			double c = mat(1, 0);
			double d = mat(1, 1);
			double divisor = a*d - b*c;
			if(fabs(divisor) < FLT_EPSILON) return false;
			mat(0, 0) = d / divisor;
			mat(0, 1) = -b / divisor;
			mat(1, 0) = -c / divisor;
			mat(1, 1) = a / divisor;
			return true;
		}
		
		bool invert22 (AArray<float, 2>& mat)
		{
			float a = mat(0, 0);
			float b = mat(0, 1);
			float c = mat(1, 0);
			float d = mat(1, 1);
			float divisor = a*d - b*c;
			if(fabs(divisor) < FLT_EPSILON) return false;
			mat(0, 0) = d / divisor;
			mat(0, 1) = -b / divisor;
			mat(1, 0) = -c / divisor;
			mat(1, 1) = a / divisor;
			return true;
		}


		const AArray<double, 1> getVectorOfMinors3x4 ( const AArray<double, 2> &in )
		{
			AArray<double, 1> out ( 4 );
			getVectorOfMinors3x4 ( in, out );
			return out;
		}

		void getVectorOfMinors3x4 ( const AArray<double, 2> &in, AArray<double, 1> &out )
		{
			assert ( out.rows() == 4 );
			AArray<double, 2> mat33 ( 3, 3 );//TODO

			mat33 ( Range::all(), 0 ) = in ( Range::all(), 1 );// OK7
			mat33 ( Range::all(), 1 ) = in ( Range::all(), 2 );
			mat33 ( Range::all(), 2 ) = in ( Range::all(), 3 );

			out ( 0 ) = -det3 ( mat33 );

			mat33 ( Range::all(), 0 ) = in ( Range::all(), 0 );

			out ( 1 ) = det3 ( mat33 );

			mat33 ( Range::all(), 1 ) = in ( Range::all(), 1 );

			out ( 2 ) = -det3 ( mat33 );

			mat33 ( Range::all(), 2 ) = in ( Range::all(), 2 );

			out ( 3 ) = det3 ( mat33 );
		}


		void pseudoInvert ( const AMatrix &m, AMatrix &result )
		{
			AMatrix temp ( m.rows(), m.cols() );
			SVDStoraged store ( m, SVD_U | SVD_VT );
			AMatrix &U  = store.U;
			AMatrix &V_T = store.VT;
			AVector &sing = store.D;

			svdUVT ( m, store );
                        for(unsigned i = 0; i < sing.rows(); ++i)
			{
				if(fabs(sing(i)) > DBL_EPSILON)
					sing(i) = 1.0 / sing(i);
			}

			matMul2Diag ( sing, V_T, temp );
			matMul2_AT_BT ( temp, U, result );
		}


		AMatrix pseudoInvertDeficient ( const AMatrix &m )
		{
			assert ( m.rows() < m.cols() );
//			printf( "pseudoInvertDeficient - begin, inverting matrix\n" );
//			cout << m;
			SVDStoraged store ( m, SVD_U | SVD_VT );
			AMatrix &U  = store.U;
			AMatrix &V_T = store.VT;
			AVector &sing = store.D;

			svdUVT ( m, store );

//			cout << "V_T" << endl;
//			printMat ( V_T );
//			cout << "U" << endl;
//			printMat ( U );

                        for ( unsigned j = 0; j < U.cols(); ++j )
			{
				double d = sing ( j );
//				printf ( "sing: %f\n", d );
				if ( d > DBL_EPSILON ) d = 1.0 / d;
				else d = 0;
//				printf ( "sing: %f\n", d );
                                for ( unsigned i = 0; i < U.rows(); ++i )
				{
					U ( i, j ) = d * U ( i, j );
				}
			}

//			cout << "U'" << endl;
//			printMat ( U );

			//subVT is the part of V_T the is not multiplied by zero
			AMatrix subV_T = V_T ( Range ( 0, U.rows() - 1 ), Range::all() );// OK7

//			cout << "subV_T" << endl;
//			printMat( subV_T );

			AMatrix result(subV_T.cols(), U.rows());
			matMul2_AT_BT ( subV_T, U, result );
			return result;//matMul_AT_BT ( subV_T, U );
		}

		void normalize ( AVector &mat, double len )
		{
			if ( fabs(len) < FLT_EPSILON ) return;

			double length = sqrt ( getSqrNorm ( mat ) );
			length /= len;

			if ( fabs(length) < FLT_EPSILON ) return;

                        for ( unsigned row = 0; row < mat.rows(); ++row )
			{
				mat ( row ) = mat ( row ) / length;
			}
		}
		void normalize ( AMatrix &mat, double len )
		{
			if ( fabs(len) < FLT_EPSILON ) return;

			double length = sqrt ( getSqrNorm ( mat.begin(), mat.end() ) );
			length /= len;

			if ( fabs(length) < FLT_EPSILON ) return;

                        for ( unsigned row = 0; row < mat.rows(); ++row )
			{
                                for ( unsigned col = 0; col < mat.cols(); ++col )
				{
					mat ( row, col ) = mat ( row, col ) / length;
				}
			}
		}
		void normalize ( ATensor3 &mat, double len )
		{
			if ( fabs(len) < FLT_EPSILON ) return;

			double length = sqrt ( getSqrNorm ( mat.begin(), mat.end() ) );
			length /= len;

			if ( fabs(length) < FLT_EPSILON ) return;
			for ( ATensor3::iterator it = mat.begin(); it != mat.end(); ++it )
				*it /= length;
		}

//		modified Gram-Schmidt
		void MGS ( AMatrix &AQ, AMatrix &R ) // TODO
		{
			R = 0;
			int n = AQ.cols();
			AVector colK( AQ.rows() ), colJ( AQ.rows() );
			for ( int k = 0; k < n; ++k )
			{
				colK = AQ ( Range::all(), k );
				double normColK = getSqrNorm ( colK );
				R ( k, k ) = normColK;
				colK /= normColK;
				AQ ( Range::all(), k ) = colK;

				for ( int j = k + 1; j < n; ++j )
				{
					colJ = AQ ( Range::all(), j );
					R ( k, j ) = dot ( colK, colJ );
					colJ = colJ - colK * R ( k, j );
					AQ ( Range::all(), j ) = colJ;
				}
			}
		}


		int computeConvexHull ( const vector<AVector2D> &input, vector<AVector> &output )
		{
			output.resize ( input.size() +1 );

			int n=input.size();
//			the output array H[] will be used as the stack
			int    bot=0, top= ( -1 );  // indices for bottom and top of the stack
			int    i;                // array scan index

			// Get the indices of points with min x-coord and min|max y-coord
			int minmin = 0, minmax;
			float xmin = input[0] ( 0 );
			for ( i=1; i<n; i++ )
				if ( input[i] ( 0 ) != xmin ) break;
			minmax = i-1;
			if ( minmax == n-1 )       // degenerate case: all x-coords == xmin
			{
				output[++top] = input[minmin];
				if ( input[minmax] ( 1 ) != input[minmin] ( 1 ) ) // a nontrivial segment
					output[++top] = input[minmax];
				output[++top] = input[minmin];           // add polygon endpoint
				return top+1;
			}

			// Get the indices of points with max x-coord and min|max y-coord
			int maxmin, maxmax = n-1;
			float xmax = input[n-1] ( 0 );
			for ( i=n-2; i>=0; i-- )
				if ( input[i] ( 0 ) != xmax ) break;
			maxmin = i+1;

			// Compute the lower hull on the stack H
			output[++top] = input[minmin];      // push minmin point onto stack
			i = minmax;
			while ( ++i <= maxmin )
			{
				// the lower line joins P[minmin] with P[maxmin]
				if ( isLeft ( input[minmin], input[maxmin], input[i] ) >= 0 && i < maxmin )
					continue;          // ignore P[i] above or on the lower line

				while ( top > 0 )      // there are at least 2 points on the stack
				{
					// test if P[i] is left of the line at the stack top
					if ( isLeft ( output[top-1], output[top], input[i] ) > 0 )
						break;         // P[i] is a new hull vertex
					else
						top--;         // pop top point off stack
				}
				output[++top] = input[i];       // push P[i] onto stack
			}

			// Next, compute the upper hull on the stack H above the bottom hull
			if ( maxmax != maxmin )    // if distinct xmax points
				output[++top] = input[maxmax];  // push maxmax point onto stack
			bot = top;                 // the bottom point of the upper hull stack
			i = maxmin;
			while ( --i >= minmax )
			{
				// the upper line joins P[maxmax] with P[minmax]
				if ( isLeft ( input[maxmax], input[minmax], input[i] ) >= 0 && i > minmax )
					continue;          // ignore P[i] below or on the upper line

				while ( top > bot )  // at least 2 points on the upper stack
				{
					// test if P[i] is left of the line at the stack top
					if ( isLeft ( output[top-1], output[top], input[i] ) > 0 )
						break;         // P[i] is a new hull vertex
					else
						top--;         // pop top point off stack
				}
				output[++top] = input[i];       // push P[i] onto stack
			}
			if ( minmax != minmin )
				output[++top] = input[minmin];  // push joining endpoint onto stack
			return top+1;
		}

		/**
		 * interprets x and y as radius and tangens and calculates
		 * appropriate sin and cos values
		 * (strahlensatz)
		 */
		void tan2sincos ( float y, float x, float &sin, float &cos )
		{
			if ( x == 0 )
			{
				if ( y == 0 ) {sin = 0; cos = 0; return;}
				sin = signum ( y );
				cos = 0;
				return;
			}

			if ( y == 0 )
			{
				sin = 0;
				cos = signum ( x );
				return;
			}

			float xx = x * x;
			float yy = y * y;
			float sinsin = ( yy ) / ( xx + yy );
			if ( y > 0 ) sin = sqrt ( sinsin );
			else sin = -sqrt ( sinsin );

			cos = x * sin / y;
		}

		float sigmoid ( float x, float steepness, float inflectionpoint )
		{
			return 1.0F/ ( 1.0F+expf ( -steepness* ( x - inflectionpoint ) ) );
		}
		
		bool atInfty( const double x, const double y, const double w )
		{
			if ( fabs( w ) < FLT_EPSILON )
			{
//				DBGOUT( RECTBUG, "last coor criterion -> infty\n" );
//				printMat( RECTBUG, a );
				return true;
			}
			if ( fabs( x ) > FLT_EPSILON && fabs( w/x ) < FLT_EPSILON )
			{
//				DBGOUT( RECTBUG, "x criterion -> infty\n" );
//				printMat( RECTBUG, a );
				return true;
			}
			if ( fabs( y ) > FLT_EPSILON && fabs( w/y ) < FLT_EPSILON )
			{
//				DBGOUT( RECTBUG, "y criterion -> infty\n" );
//				printMat( RECTBUG, a );
				return true;
			}
			return false;
		}
		
		void getDistanceToAndAlongLine(const double sx, const double sy, 
					  const double dirX, const double dirY, 
					  const double x, const double y, 
					  double &distTo, double &distAlong)
		{
			const double diffX = x - sx;
			const double diffY = y - sy;

			double dirLen = sqrt(dirX * dirX + dirY * dirY);
			double dirXn = dirX / dirLen;
			double dirYn = dirY / dirLen;
			distAlong = diffX * dirXn + diffY * dirYn;
			
			double diffLenSqr = diffX * diffX + diffY * diffY;
			distTo = sqrt(diffLenSqr - distAlong * distAlong);
			
// 			const double denom = dirX * dirX + dirY * dirY;
// 			distTo    = (diffY*dirX - diffX*dirY)/denom;
// 			distAlong = (diffX*dirX + diffY*dirY)/denom;
		}
	}
}//namespace
