

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORETHREAD_H
#define ADORE_CORETHREAD_H

#include <pthread.h>
#include <semaphore.h>
#include <stdlib.h>
#include <stdio.h>
#include <cstring>
#include <vector>
#include <adrcore/util/debug.h>
#include <adrcore/types/heaparray.h>
#include <cassert>

namespace adore
{

	namespace core
	{

		/**
			tries to figure out a good number of parallel threads
		*/
		uint figureNumThreads();
		void setNumThreads( uint n );
		uint getNumThreads();

		//debug functions
		uint getThreadsCreated();
		void incThreadsCreated();

		class Thread;

		/**
			a thread that stops may notify a listener
		*/
		class StopListener
		{
			public:
				virtual ~StopListener() {;}
				virtual void notify( Thread * ) = 0;
		};

		class JobThread;

		class JobFinishedListener
		{
			public:
				virtual ~JobFinishedListener() {;}
				virtual void jobNotify( JobThread * ) = 0;
		};

		/**
		 * A very basic Thread
		 * Derive and overload execute.
		 * Inside execute, shallStop() should be polled to know when to terminate the thread
			 
		*/
		class Thread
		{
			private:
				bool running;
				pthread_t id;
				pthread_attr_t attr;
				bool stopSoon;
				StopListener *listener;

			public:
				Thread();
				virtual ~Thread();

				int start( bool detached = 0, bool sscope = 0 );

				virtual void stop()
				{
					stopSoon = true;
				}

				void setStopListener ( StopListener* theValue )
				{
					listener = theValue;
				}

				StopListener* getStopListener() const
				{
					return listener;
				}

				bool isRunning()
				{
					return running;
				}

				void join()
				{
					pthread_join( id, NULL );
				}

			protected:

				void run();
				static void *entryPoint( void* );
				bool shallStop() {return stopSoon;}

				virtual void execute() = 0;
		};


		class Mutex
		{
			public:
				Mutex()
				{
					if ( pthread_mutex_init ( &mut, NULL ) ) {printf( "mutex error\n" );exit( 1 );}
				}

				~Mutex()
				{
					int err = pthread_mutex_destroy ( &mut );
					if ( err ) {printf( "mutex error - destroy - %s\n", std::strerror( err ) );exit( 1 );}
				}

				void lock()
				{
					if ( pthread_mutex_lock ( &mut ) ) {printf( "mutex error\n" );exit( 1 );}
				}

				void trylock()
				{
					if ( pthread_mutex_trylock ( &mut ) ) {printf( "mutex error\n" );exit( 1 );}
				}

				void unlock()
				{
					if ( pthread_mutex_unlock ( &mut ) ) {printf( "mutex error\n" );exit( 1 );}
				}

				pthread_mutex_t &getPThreadMutex() {return mut;}

			private:
				pthread_mutex_t mut;
		};

		class Lock
		{
			public:
				Lock( Mutex &m ):mut( m )
				{
//					printf( "locking mutex\n" );
					mut.lock();
				}
				~Lock()
				{
//					printf( "unlocking mutex\n" );
					mut.unlock();
				}

			private:
				Mutex &mut;
		};

		class Condition
		{
			protected:
				pthread_cond_t cond;

			public:
				Condition  () { pthread_cond_init( &cond, NULL ); }

				~Condition () { pthread_cond_destroy( &cond ); }

				void wait ( Mutex & m )
				{
					pthread_cond_wait( &cond, &m.getPThreadMutex() );
				}

				void signal ()
				{
					pthread_cond_signal( &cond );
				}

				void broadcast ()
				{
					pthread_cond_broadcast( &cond );
				}
		};

		class Semaphore
		{
			private:
				sem_t sem;

			public:
				Semaphore( int count )
				{
					sem_init( &sem, 0, count );
				}

				~Semaphore()
				{
					sem_destroy( &sem );
				}

				void increment()
				{
					sem_post( &sem );
				}

				void decrement()
				{
					sem_wait( &sem );
				}

				bool tryDecrement()
				{
					int val = sem_trywait( &sem );
					return( val != -1 );
				}

				int getValue()
				{
					int val = 0;
					sem_getvalue( &sem, &val );
					return val;
				}
		};

		class Job
		{
			public:
				virtual ~Job() {;}

				virtual void execute() = 0;
		};

		/**
		 * A thread that takes jobs
		 * this one starts itself immediately
		 */
		class JobThread : public Thread
		{
			private:
				Job *job;
				Semaphore runSem; //this semaphore stops the thread while it waits on new jobs
				bool idle;
				JobFinishedListener *joblistener;
				Mutex joinMutex; //to join jobs

			public:
				JobThread():Thread(), job( NULL ), runSem( 0 ), idle( true )
				{
					Thread::start();
				}

				~JobThread()
				{
//					printf( "dtor %p:\n", this );
					stop();
				}

				virtual void stop();

				/**
				 * the purpose of this mehtod is, to make the use of this
				 * class in loops easier, if getJob is used to re-use
				 * an associated job.
				 * This way you can associate a job without running it.
				 * NEVER use it, once doJob has been called, since this function
				 * may replace the job while it is performed.
				 */
				void setJob( Job *j )
				{
					job = j;
				}

				/**
				 * return the last used job object, in case you want to re-use it
				 */
				Job *getJob() {return job;}

				/**
				 * perform a job
				 * don't try to start more than one job. That would create a mess.
				 */
				void doJob( Job *j );

				bool isIdle();

				void setJobFinishedListener ( JobFinishedListener* theValue )
				{
					joblistener = theValue;
				}

				JobFinishedListener* getJobFinishedListener() const
				{
					return joblistener;
				}

				void joinJob();

			protected:

				void setIdle( bool val );

				void execute();
		};

		/**
			a threadpool
			this one is not threadsave itself, so beware
		*/
		class ThreadPool : public JobFinishedListener
		{
			public:

			private:
				HeapArray<JobThread> threads;
				Semaphore semaphore;

			public:
				ThreadPool( uint nThreads );

				~ThreadPool() {join();}

				uint getNumThreads() {return threads.size();}

				JobThread *getThread( int idx ) {return &threads[idx];}

				/**
				 * blocks, if none is idle
				 * */
				JobThread *getIdle();

				/**
					overloaded from JobFinishedListener
				 */
				void jobNotify( JobThread *t );


				void joinJobs();

			private:

				/**
				 * this stops all threads (ending their posix counterparts). The threadpool will be unusable afterwards, since
				 * there is currently no way to recreate the threads
				 */
				void join();

		};


	}
}

#undef THDBG

#endif
