

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "linesegmentfinder.h"

#include <adrcore/imgprocessing/rgbtogray.h>
#include <adrcore/surf/integralimage.h>

#include <utility>
#include <list>

namespace adore
{
	namespace core
	{
		using namespace std;
		using namespace boost;

		LineSegmentFinder::LineSegmentFinder( double minLength, double maxDist, int filterSz, int typeID ) :
				featureID( typeID ),
				workImage( NULL ),
				workIntImage( NULL ),
				grayDeletor( NULL ),
				width( 0 ), height( 0 ),
				minSegmentLength( minLength ),
				pointLineMaxDist( maxDist ),
				filterSize( filterSz ),
				filterMax( 0.0 ),
				filterMin( 0.0 )
		{
			if ( 2 > filterSz ) filterSz = 2;
			if ( 99 < filterSz ) filterSz = 99;
		}

		LineSegmentFinder::LineSegmentFinder( const LineSegmentFinder& rhs ) :
				featureID( rhs.featureID ),
				workImage( NULL ),
				workIntImage( NULL ),
				grayDeletor( NULL ),
				width( 0 ), height( 0 ),
				minSegmentLength( rhs.minSegmentLength ),
				pointLineMaxDist( rhs.pointLineMaxDist ),
				filterSize( rhs.filterSize ),
				filterMax( 0.0 ),
				filterMin( 0.0 )
		{
		}

		FeatureFinder * LineSegmentFinder::clone()
		{
			return new LineSegmentFinder( *this );
		}

		unsigned int LineSegmentFinder::findFeaturesDerived( ViewT * v )
		{
			uint num = 0;
			if ( !initImages( v ) )
				return num;
			haarImagesT imgs = getHaarImages();
			suppress( imgs );
			vector<segmentT > lines = find( imgs );
			for ( uint i = 0; i < lines.size(); i++ )
			{
				FeatureLine* fl = new FeatureLine( lines[i].first( 0 ), lines[i].first( 1 ),
				                                   lines[i].second( 0 ), lines[i].second( 1 ), featureID );
				ViewBase::segmentT segment( fl );
				segment->setView( v );
				v->addSegment( segment );
			}

#ifndef NDEBUG
			cout << "LineSegmentFinder found " << lines.size() << " line segments" << endl;
#endif

			return lines.size();
		}

		bool LineSegmentFinder::initImages( ViewT * v )
		{
			ImagePlane<imgT> * chan = v->getImage();
			if ( !chan )
				return false;
			width = chan->getWidth();
			height = chan->getHeight();
			ImagePlane<imgT> * gray = NULL;
			if ( chan->getNumChannels() == 3 )
			{
				gray = new ImagePlane<imgT> ( width, height, 1 );
				RGBtoGray<imgT> converter;
				Channel<imgT>::Set in ( true );
				chan->createChannels ( in );
				converter.process ( in, gray );
				chan = gray;
			}
			grayDeletor.reset( gray );
			workImage = chan;
			workIntImage.reset( new IImageT( workImage ) );
			return true;
		}

		LineSegmentFinder::haarImagesT LineSegmentFinder::getHaarImages()
		{
			shared_ptr<ImageHaarT> imgM( new ImageHaarT( width, height, 1 ) );
			shared_ptr<ImageHaarT> imgO( new ImageHaarT( width, height, 1 ) );
			imgM->fill( ImageHaarT::value_type( 0 ) );
			imgO->fill( ImageHaarT::value_type( 0 ) );
			AArray<double, 1> resp( 2 );
			double magsum = 0.0;
			for ( int y = filterSize / 2; y < height - filterSize / 2 - filterSize % 2; ++y )
			{
				for ( int x = filterSize / 2; x < width - filterSize / 2 - filterSize % 2; ++x )
				{
					resp = responseXandY<double>( x, y );
					double mag = sqrt( resp( 0 ) * resp( 0 ) + resp( 1 ) * resp( 1 ) );
					double orient = atan2( resp( 1 ), resp( 0 ) );
					*( imgM->getPointer( x, y ) ) = mag;
					*( imgO->getPointer( x, y ) ) = orient;
					magsum += mag;
				}
			}
			magsum /= width * height;
			filterMin = 0.5 * magsum;
			filterMax = 2.5 * filterMin;
			return make_pair( imgM, imgO );
		}

		template <typename T> AArray< T, 1 > LineSegmentFinder::responseXandY( int x, int y )
		{
			int step = filterSize / 2;
			int topX = x - step;
			int topY = y - step;

			AArray< T, 1 > resp( 2 );
			if ( 0 == filterSize % 2 )
			{
				T white = workIntImage->getSum( topX, topY, filterSize, filterSize );
				resp( 0 ) = white - ( T )2.0 * workIntImage->getSum( topX, topY, step, filterSize );
				resp( 1 ) = white - ( T )2.0 * workIntImage->getSum( topX, topY, filterSize, step );
			}
			else
			{
				resp( 0 ) = workIntImage->getSum( x + 1, topY, step, filterSize ) - workIntImage->getSum( topX, topY, step, filterSize );
				resp( 1 ) = workIntImage->getSum( topX, y + 1, filterSize, step ) - workIntImage->getSum( topX, topY, filterSize, step );
			}
			return resp;
		}

		void LineSegmentFinder::suppress( LineSegmentFinder::haarImagesT& imgs )
		{
			std::auto_ptr<ImageHaarT> imgSupp( new ImageHaarT( width, height, 1 ) );
			imgSupp->fill( ImageHaarT::value_type( 0 ) );
			for ( int y = 0; y < height; ++y )
			{
				for ( int x = 0; x < width; ++x )
				{
					double magnitude = *( imgs.first->getPointer( x, y ) );
					if ( magnitude < filterMin )
						continue;

					double orientation = *( imgs.second->getPointer( x, y ) );
					EdgeCell cell( orientation );
					int xx1 = x, yy1 = y, xx2 = x, yy2 = y;
					cell.next( xx1, yy1 );
					cell = -cell;
					cell.next( xx2, yy2 );

					if ( xx1 >= 0 && xx1 < width && yy1 >= 0 && yy1 < height )
					{
						if ( *( imgs.first->getPointer( xx1, yy1 ) ) >= magnitude )
							continue;
					}

					if ( xx2 >= 0 && xx2 < width && yy2 >= 0 && yy2 < height )
					{
						if ( *( imgs.first->getPointer( xx2, yy2 ) ) >= magnitude )
							continue;
					}
					*( imgSupp->getPointer( x, y ) ) = *( imgs.first->getPointer( x, y ) );
				}
			}
			imgs.first.reset( imgSupp.release() );
		}

		vector< LineSegmentFinder::segmentT > LineSegmentFinder::find( haarImagesT & imgs )
		{
			vector<segmentT> lines;
			for ( int y = 0; y < height; ++y )
			{
				for ( int x = 0; x < width; ++x )
				{
					double magnitude = *( imgs.first->getPointer( x, y ) );
					if ( filterMax <= magnitude )
					{
						list<std::pair<int, int> > coors;
						coors.push_back( make_pair( x, y ) );
						*imgs.first->getPointer( x, y ) = 0.0;

						double orient = *( imgs.second->getPointer( x, y ) );
						double orient_ortho = orient - M_PI * 0.5;
						EdgeCell ec( orient_ortho );
						int xx = x, yy = y;
						bool ok = true;
						do
						{
							ok = nextCell( xx, yy, ec, orient, filterMin, imgs );
							if ( ok )
							{
								coors.push_front( make_pair( xx, yy ) );
								*imgs.first->getPointer( xx, yy ) = 0.0;
							}
						}
						while ( ok );

						ec = -ec;
						xx = x, yy = y;
						ok = true;
						do
						{
							ok = nextCell( xx, yy, ec, orient, filterMin, imgs );
							if ( ok )
							{
								coors.push_back( make_pair( xx, yy ) );
								*imgs.first->getPointer( xx, yy ) = 0.0;
							}
						}
						while ( ok );

						if ( minSegmentLength > coors.size() )
							continue;
						vector<pair<int, int> > chain( coors.begin(), coors.end() );
						vector<pair<size_t, size_t> > filteredLines = filterChain( chain, 0, chain.size() );
						for ( size_t i = 0; i < filteredLines.size(); ++i )
						{
							AVector p1( 3 ), p2( 3 );
							bool ok = approximateLine( chain, filteredLines[i].first, filteredLines[i].second, p1, p2 );
							if ( !ok ) continue;
							lines.push_back( ( make_pair( p1, p2 ) ) );
						}
					}
				}
			}
			return lines;
		}

		bool LineSegmentFinder::nextCell( int& x, int& y, EdgeCell& ec, double orientation, double min, haarImagesT & imgs )
		{
			int oldX = x, oldY = y;
			double lastDiff = M_PI * 0.25;
			int goodX = -1, goodY = -1;

			ec.next( x, y );
			double test = filteredOrientation( x, y, orientation, lastDiff, min, imgs );
			if ( 0.0 != test )
			{
				lastDiff = test;
				goodX = x, goodY = y;
			}

			x = oldX, y = oldY;
			ec.left( x, y );
			test = filteredOrientation( x, y, orientation, lastDiff, min, imgs );
			if ( 0.0 != test )
			{
				lastDiff = test;
				goodX = x, goodY = y;
			}

			x = oldX, y = oldY;
			ec.right( x, y );
			test = filteredOrientation( x, y, orientation, lastDiff, min, imgs );
			if ( 0.0 != test )
			{
				lastDiff = test;
				goodX = x, goodY = y;
			}

			if ( -1 != goodX )
			{
				x = goodX;
				y = goodY;
				return true;
			}
			return false;
		}

		double LineSegmentFinder::filteredOrientation( int x, int y, double orientation, double minDiff, double min, const haarImagesT & imgs ) const
		{
			if ( x < 0 || x >= width || y < 0 || y >= height )
				return 0.0;
			double orHere = *( imgs.second->getPointer( x, y ) );
			double diff = abs( orHere - orientation );
			if ( diff < M_PI * 0.25 && diff < minDiff && *( imgs.first->getPointer( x, y ) ) > min )
				return diff;
			return 0.0;
		}

		vector< pair<size_t, size_t> > LineSegmentFinder::filterChain( const vector<pair<int, int> >& chain, size_t first, size_t last ) const
		{
			vector< pair<size_t, size_t> > retval;
			double tmp = ( double )last - ( double )first;
			if ( minSegmentLength > tmp )
				return retval;
			if ( 3.0 > tmp )
				return retval;

			double x1 = chain[first].first, y1 = chain[first].second;
			double x2 = chain[last - 1].first, y2 = chain[last - 1].second;
			double ort1 = y2 - y1, ort2 = x1 - x2;
			double len = sqrt( ort1 * ort1 + ort2 * ort2 );
			double maxDst = 0.0;
			size_t maxDstIndex = first;
			for ( size_t i = first + 1; i < last - 1; ++i )
			{
				double x0 = x1 - chain[i].first, y0 = y1 - chain[i].second;
				tmp = abs( x0 * ort1 + y0 * ort2 ) / len;
				if ( maxDst < abs( tmp ) )
				{
					maxDst = abs( tmp );
					maxDstIndex = i;
				}
			}
			if ( pointLineMaxDist >= maxDst )
			{
				retval.push_back( make_pair( first, last ) );
				return retval;
			}
			vector< pair<size_t, size_t> > retvalL = filterChain( chain, first, maxDstIndex );
			if ( !retvalL.empty() )
				retval.insert( retval.end(), retvalL.begin(), retvalL.end() );
			vector< pair<size_t, size_t> > retvalR = filterChain( chain, maxDstIndex, last );
			if ( !retvalR.empty() )
				retval.insert( retval.end(), retvalR.begin(), retvalR.end() );
			return retval;
		}

		bool LineSegmentFinder::approximateLine( const vector<pair<int, int> >& chain, size_t first, size_t last, AVector& p1, AVector& p2 ) const
		{
			double xmean = 0, ymean = 0;
			for ( size_t i = first; i < last; ++i )
			{
				xmean += chain[i].first;
				ymean += chain[i].second;
			}
			int size = ( int )last - ( int )first;
			xmean /= size;
			ymean /= size;

			AMatrix pts( size, 2 );
			for ( size_t i = first; i < last; ++i )
			{
				pts( ( int )( i - first ), 0 ) = ( double )chain[i].first - xmean;
				pts( ( int )( i - first ), 1 ) = ( double )chain[i].second - ymean;
			}

			AMatrix pts2( 2, 2 );
			matMul2_AT_A( pts, pts2 );

			double a = pts2( 0, 0 ), b = pts2( 0, 1 );
			double c = pts2( 1, 0 ), d = pts2( 1, 1 );
			double t1 = ( a + d ) * 0.5;
			double t2 = sqrt( 4.0 * b * c + ( a - d ) * ( a - d ) ) * 0.5;
			double lmax = std::max( t1 + t2, t1 - t2 );
			pts2( 0, 0 ) -= lmax;
			pts2( 1, 1 ) -= lmax;
			double A = -pts2( 0, 1 ), B = pts2( 0, 0 );
			const double eps = std::numeric_limits< double >::epsilon();
			if ( abs( A ) < eps && abs( B ) < eps )
			{
				A = -pts2( 1, 1 ), B = pts2( 1, 0 );
			}
			if ( abs( A ) < eps && abs( B ) < eps )
				return false;

			double l = sqrt( A * A + B * B );
			A /= l; B /= l;
			double C = -( A * xmean + B * ymean );
			AVector v1( 2 ); v1 = xmean, ymean;
			AVector v2( 2 ); v2 = -A * C, -B * C;
			normalize( v2, 1.0 );

			AVector tmp( 2 );
			tmp = pts( 0, 0 ), pts( 0, 1 );
			double ab = dot( tmp, v2 );
			double bb = dot( v2, v2 );
			p1 = ab / bb * v2( 0 ) + xmean, ab / bb * v2( 1 ) + ymean, 1;
			tmp = pts( size - 1, 0 ), pts( size - 1, 1 );
			ab = dot( tmp, v2 );
			p2 = ab / bb * v2( 0 ) + xmean, ab / bb * v2( 1 ) + ymean, 1;
			return true;
		}
	}
}
