

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "nviewdependencygraph.h"

namespace adore
{

	namespace core
	{

		void NViewDependencyGraph::takeRoots()
		{
			//put the last entries, which are the roots, into the roots vector for convenient access
			uint numViews = 0;
			roots.clear();
			VertexIter it, it_end;
			tie( it_end, it ) = vertices( graph );
			if ( it == it_end ) return;
			--it_end; --it;
			NView *nv = ( graph )[*it];
			if ( nv )
			{
				roots.push_back( *it );
				numViews = nv->getNumViews();
			}
			--it;
			for ( ; it != it_end; --it )
			{
				NView *nv = ( graph )[*it];
				if ( nv && ( nv->getNumViews() == numViews ) )
				{
					roots.push_back( *it );
				}
				else break;
			}
		}

		void NViewDependencyGraph::create( NViewMap &map, ViewKnitter &knit )
		{
			typedef ViewKnitter::Tuple Tuple;
			typedef ViewKnitter::VTuple VTuple;

			//add all nviews, smallest first
			for ( int s = 0; s < knit.numSteps(); ++s )
			{
				for ( NViewMap::NViewIterator it = map.getEntriesBegin(); it != map.getEntriesEnd(); ++it )
				{
					NView *nv = *it;
					if ( nv && ( ( int )nv->getNumViews() == knit.getNumViewsOfUnion( s ) ) )
					{
						Vertex vert = add_vertex( nv, graph );
						nvToVertex[nv] = vert;
						DBGOUT( DBGNORMAL, "NViewDependencyGraph: adding %p / %d\n", nv, nv->getNumViews() );
					}
				}
			}

			if ( nvToVertex.empty() ) return;

			takeRoots();

			for ( int s = knit.numSteps() - 1; s >= 1; --s )
			{
				for ( ViewKnitter::iterator it = knit.begin( s ); it != knit.end( s ); ++it )
				{
					Tuple &tup = *it;

					VTuple &vtupA = *tup.get<0>();
					VTuple &vtupB = *tup.get<1>();
					VTuple &vtupAB = *tup.get<2>();

					NView *nvA = map.getNView( vtupA.begin(), vtupA.end() );
					NView *nvB = map.getNView( vtupB.begin(), vtupB.end() );
					NView *nvAB = map.getNView( vtupAB.begin(), vtupAB.end() );

					//the ViewKnitter plan does not need to be executed completely
					if ( !nvA || !nvB || !nvAB ) continue;

//					printf( "[%d][%d]->[%d]\n", nvA->getNumViews(), nvB->getNumViews(), nvAB->getNumViews() );

					Vertex &nvAV = nvToVertex[nvA];
					Vertex &nvBV = nvToVertex[nvB];
					Vertex &nvABV = nvToVertex[nvAB];

					add_edge( nvABV, nvAV, graph );
					add_edge( nvABV, nvBV, graph );
				}
			}
		}

		void NViewDependencyGraph::create( NViewMap &map )
		{
			typedef vector<NView *> NViewTuple;
			typedef std::map<NView *, int> VotesColl;
			typedef VotesColl::iterator votes_iter;

			if ( map.getNumEntries() < 1 ) return;

			std::map<NView *, int> nviewVotes;
			std::map<ViewBase *, NViewTuple> viewToNViews;

			//init data structures
			for ( NViewMap::ViewIterator it = map.getAllViewsBegin(); it != map.getAllViewsEnd(); ++it )
			{
				NViewTuple tup; tup.reserve( 50 );
				viewToNViews[*it] = tup;
			}

			//create vertices
			uint largestNumViews = 0;
			uint smallestNumViews = INT_MAX;
			for ( NViewMap::NViewIterator it = map.getEntriesBegin(); it != map.getEntriesEnd(); ++it )
			{
				NView *nv = *it;
				Vertex vert = add_vertex( nv, graph );
				nvToVertex[nv] = vert;
				nviewVotes[nv] = 0;
				largestNumViews = MAX( largestNumViews, nv->getNumViews() );
				smallestNumViews =  MIN( smallestNumViews, nv->getNumViews() );
			}

			//reconstruct
			//the assumption is that the NViews were merged from NViews of the same size
			//e.g. (2, 2)->3; (3, 3)->4; (4, 4)->6
			uint s = smallestNumViews;
			while ( s < largestNumViews )
			{
				//reset the view to nview mapping
				for ( NViewMap::ViewIterator it = map.getAllViewsBegin(); it != map.getAllViewsEnd(); ++it )
				{
					viewToNViews[*it].clear();
				}

				//remember which view is member of which nview at size s
				for ( NViewMap::NViewIterator it = map.getEntriesBegin(); it != map.getEntriesEnd(); ++it )
				{
					NView *nv = *it; assert( nv );
					if ( nv->getNumViews() == s )
					{
						for ( NView::ViewIterator vit = nv->getViewsBegin(); vit != nv->getViewsEnd(); ++vit )
						{
							viewToNViews[*vit].push_back( nv );
						}
					}
				}

				int count = 0;
				do
				{
					++s;
					count = 0;
					for ( NViewMap::NViewIterator it = map.getEntriesBegin(); it != map.getEntriesEnd(); ++it )
					{
						NView *nv = *it; assert( nv );
						if ( nv->getNumViews() == s )
						{
							++count;//meaning success at size s

							//reset the votes
							for ( std::map<NView *, int>::iterator sit = nviewVotes.begin(); sit != nviewVotes.end(); ++sit )
							{
								( *sit ).second = 0;
							}

							//vote
							for ( NView::ViewIterator vit = nv->getViewsBegin(); vit != nv->getViewsEnd(); ++vit )
							{
								NViewTuple &tup = viewToNViews[*vit];
								for ( NViewTuple::iterator tit = tup.begin(); tit != tup.end(); ++tit )
								{
									++nviewVotes[*tit];
								}
							}

							//the two nviews with the most votes are the parents of nv
							NView *nviews[2] = {NULL/*most votes*/, NULL/*second most votes*/};
							int votes[2] = {0, 0};
							for ( votes_iter voit = nviewVotes.begin(); voit != nviewVotes.end(); ++voit )
							{
								NView *nv = ( *voit ).first;
								int vote = ( *voit ).second;
								if ( vote > votes[0] )
								{
									votes[1] = votes[0];
									votes[0] = vote;
									nviews[1] = nviews[0];
									nviews[0] = nv;
								}
								else if ( vote > votes[1] )
								{
									votes[1] = vote;
									nviews[1] = nv;
								}
							}

							assert( nviews[0] );
							assert( nviews[1] );

							Vertex &nvAV = nvToVertex[nviews[0]];
							Vertex &nvBV = nvToVertex[nviews[1]];
							Vertex &nvABV = nvToVertex[nv];

							add_edge( nvABV, nvAV, graph );
							add_edge( nvABV, nvBV, graph );
						}
					}
				}
				while ( count == 0 );
			}

			takeRoots();
		}

		string NViewDependencyGraph::toString()
		{
			ostringstream oss;

			VertexIter vi, vi_end;
			for ( tie( vi, vi_end ) = vertices( graph ); vi != vi_end; ++vi )
			{
				NView *nv = graph[*vi];
				oss << "( " << nv << "/" << nv->getNumViews() << " ) - ";

				AdjacencyIter avi, avi_end;
				for ( tie( avi, avi_end ) = adjacent_vertices( *vi, graph ); avi != avi_end; ++avi )
				{
					nv = graph[*avi];
					oss << "( " << nv << "/" << nv->getNumViews() << " )";
				}
				oss << endl;
			}

			return oss.str();
		}

	}

}
