

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_GLCAMERA_H
#define ADORE_GLCAMERA_H

#include "glmath.h"
#include <cstdio> 

namespace adore
{
	namespace visual
	{
		/**	A virtual camera with controls which imitate a real camera
		 *
		 *	 
		 */
		class GLCamera
		{
				Vector3D eyeloc;	///< Eye location
				Vector3D spine;		///< turning axis
				Vector3D u;			///< Vector pointing right
				Vector3D v;			///< Vector pointing up
				Vector3D n;			///< Vector pointing in

				float fov;			///< Field of view angle
				float aspect;		///< Viewport aspect ratio
				float near_dist;	///< Near clipping plane
				float far_dist;		///< Far clipping plane

			public:
				GLCamera():fov(M_PI), aspect(1.0F), near_dist(10), far_dist(1000){}
				
				/**	Set up the camera
				 *	@param eye Camera eye location
				 *	@param aim Camera aim location
				 *	@param up A vector pointing up
				 */
				void set ( Vector3D eye, Vector3D aim, Vector3D up );

				/**	Set up the camera
				 *	@param eye Camera eye location
				 *	@param rot Camera orientation
				 */
				void set ( Vector3D eye, Quaternion rot );

				/**
				 * for persistence
				 */
				void set ( Vector3D _eyeloc, Vector3D _spine, Vector3D _u, Vector3D _v, Vector3D _n );

				/**	Roll the camera
				 *	@param angle Roll angle in radians
				 */
				void roll ( float angle );

				/**	Increase the pitch
				 *	@param angle Pitch angle in radians
				 */
				void pitch ( float angle );

				/**	Yaw the camera
				 *	@param angle Yaw angle in radians
				 */
				void yaw ( float angle );

				/**	Turn the camera around the spine axis
				 *	@param angle Turning angle in radians
				 */
				void turn ( float angle );

				/**	Tilt the camera
				 *	@param angle angle of tilting
				 *	@param gap the minimum angle between the view direction and the spine
				 */
				void tilt ( float angle, float gap );

				/**	Rotate the camera around the axis in camera space
				 *	@param axis Axis of rotation
				 *	@param angle Rotation angle in radians
				 */
				void rotate ( Vector3D axis, float angle );

				/**	Slide the camera
				 *	@param delta Sliding offset in camera space
				 */
				void slide ( Vector3D delta );

				/**	Set perspective projection parameters
				 *	@param fov_ Field of view
				 *	@param sapect_ Aspect ratio
				 *	@param near_dist_ Near clipping plane
				 *	@param far_dist_ Far clipping plane
				 */
				void perspective ( float fov_, float aspect_, float near_dist_, float far_dist_ );

				/**	Set OpenGL GL_MODELVIEW matrix
				 */
				void setOpenGLModelView() const;

				/**	Set OpenGL GL_PROJECTION matrix
				 */
				void setOpenGLProjection() const;

				/**	Get the matrix of camera transformation
				 */
				void getMatrix ( Matrix4x4& modelViewProj ) const;

				/**	Project the vector in camera space
				 */
				void projectVector ( Vector3D src, Vector3D& dst ) const;

				/**	Unproject the vector from camera space
				 */
				void unprojectVector ( Vector3D src, Vector3D& dst ) const;

				/** Build a ray for intersection testing.
				 *	The ray starts from the eye and goes through the point with screen
				 *	coordinates x and y and which lies on the near clipping plane.
				 *	@param x X screen coordinate
				 *	@param y Y screen coordinate
				 *	@param org Origin of the ray
				 *	@param dir Direction vector of the ray (always normalized)
				 */
				void buildRay ( int x, int y, Vector3D& org, Vector3D& dir ) const;

				//acessors below
				
				Vector3D &getEyeLoc() {return eyeloc;}
				Vector3D &getSpine() {return spine;}
				Vector3D &getU() {return u;}
				Vector3D &getV() {return v;}
				Vector3D &getN() {return n;}

				void setFov ( float theValue )
				{
					fov = theValue;
				}


				float getFov() const
				{
					return fov;
				}

				void setAspect ( float theValue )
				{
					aspect = theValue;
				}


				float getAspect() const
				{
					return aspect;
				}

				void setNear_dist ( float theValue )
				{
					near_dist = theValue;
				}


				float getNear_dist() const
				{
					return near_dist;
				}

				void setFarDist ( float theValue )
				{
					far_dist = theValue;
				}


				float getFarDist() const
				{
					return far_dist;
				}
				
				void print()
				{
					printf("=== GLCamera:\n");
					printf("eyeloc: %f, %f, %f\n", eyeloc[0], eyeloc[1], eyeloc[2]);
					printf("spine: %f, %f, %f\n", spine[0], spine[1], spine[2]);
					printf("u: %f, %f, %f\n", u[0], u[1], u[2]);
					printf("v: %f, %f, %f\n", v[0], v[1], v[2]);
					printf("n: %f, %f, %f\n", n[0], n[1], n[2]);
				}

		};
	}
}

#endif	// ADORE_GLCAMERA_H

