

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#include "gldelaunayrenderer.h"

#include <list>
#include <functional>
#include <algorithm>

#include <GL/gl.h>
#include <adrcore/surface/delaunay.h>

#include "glplaneloader.h"
#include "glavatar.h"

#include <adrcore/util/clock.h>

#include <adrcore/types/nviewmap.h>
#include <adrcore/random/random.h>
namespace adore
{
	namespace visual
	{
		// define randomization parameters
		// set range to zero to disable randomization
		// range == 0.001 is usually enough for everything
		static const double inv_rand_max = 1.0 / static_cast<double>(RAND_MAX);
		static const double range = 0.001;
		static const double r = range * inv_rand_max;

		void simplex(unsigned* v, void* extra)
		{
			assert(extra);
			std::list<unsigned>& faces = *reinterpret_cast<std::list<unsigned>*>(extra);
			faces.push_back(v[0]);
			faces.push_back(v[1]);
			faces.push_back(v[2]);
		}

		class comparator : public std::binary_function<Face&,Face&,bool>
		{
			Vertex* vertex_data;

			float perimeter_sqr(const Face& f)
			{
				float a = (vertex_data[f.v1].pos-vertex_data[f.v2].pos).lengthSqr();
				float b = (vertex_data[f.v2].pos-vertex_data[f.v3].pos).lengthSqr();
				float c = (vertex_data[f.v3].pos-vertex_data[f.v1].pos).lengthSqr();
				return a + b + c;
			};

			float max_edge_sqr(const Face& f)
			{
				float a = (vertex_data[f.v1].pos-vertex_data[f.v2].pos).lengthSqr();
				float b = (vertex_data[f.v2].pos-vertex_data[f.v3].pos).lengthSqr();
				float c = (vertex_data[f.v3].pos-vertex_data[f.v1].pos).lengthSqr();
				if(a > b && a > c)
					return a;
				else if(b > c)
					return b;
				else return c;
			};

			float dbl_area_sqr(const Face& f)
			{
				Vector3D a = vertex_data[f.v2].pos-vertex_data[f.v1].pos;
				Vector3D b = vertex_data[f.v3].pos-vertex_data[f.v1].pos;
				Vector3D area2;
				area2.cross(a,b);
				return area2.lengthSqr();
			};

		public:
			comparator(Vertex* vd)
				:vertex_data(vd) {};

			bool operator() (const Face& lv, const Face& rv)
			{
				return perimeter_sqr(lv)+dbl_area_sqr(lv) < perimeter_sqr(rv)+dbl_area_sqr(rv);
// 				return dbl_area_sqr(lv) < dbl_area_sqr(rv);
			};
		};

		void GLDelaunayRenderer::clear()
		{
			DBGOUT(DBGNORMAL, "GLDelaunayRenderer::clear\n");
			for(unsigned i = 0; i < data.size(); ++i)
			{
				delete[] data[i].pixels;
				delete data[i].entity;
			}
			data.clear();
			nv_points.clear();
			vertex_map.clear();
			entity = 0;
			maxdistance = 0.0f;
			minobjectsize = std::numeric_limits<float>::infinity();
		}

		void GLDelaunayRenderer::triangulate(unsigned index, std::vector<core::Point3D>& vertices)
		{
			assert(index < data.size() && data[index].entity);

			// Create and lock mesh vertices
			data[index].entity->mesh.createVertices(vertices.size());
			Vertex* mesh_vertices = data[index].entity->mesh.lockVertices(0,vertices.size());
			if(!mesh_vertices) return;
			assert(mesh_vertices && "Cannot lock mesh vertices");

			// Copy mesh vertices
			for(unsigned i = 0; i < vertices.size(); ++i)
			{
				// Copy position
				mesh_vertices[i].pos.x = static_cast<float>(vertices[i][0]);
				mesh_vertices[i].pos.y = static_cast<float>(vertices[i][1]);
				mesh_vertices[i].pos.z = static_cast<float>(vertices[i][2]);

				// Store index of the correspondent NVPoint (in nv_points array)
				mesh_vertices[i].extra = vertex_map[nv_points[i]];
			}

			double inv_width = 1.0 / data[index].width;
			double inv_height = 1.0 / data[index].height;

			// Query and convert feature points
			GLFeatureLoader loader(data[index].view);
			std::vector<double> points(loader.count()*2);
			loader.load(&points[0]);

			// Process vertices
			unsigned src = 0;
			double x, y;
			std::vector<double> drpts;
			std::vector<unsigned> drpts_map;
			drpts.reserve(points.size()*2);
			drpts_map.reserve(points.size());
			DBGOUT(DBGNORMAL, "GLDelaunayRenderer: numAllPoints is %d\n", data[index].view->getNumAllFeaturePoints());
			core::ViewBase::PointsIterator b = data[index].view->getPointsBegin(0);
			core::ViewBase::PointsIterator e = data[index].view->getPointsEnd(0);
			int iters = 0;
			for(core::ViewBase::PointsIterator iter = b; iter != e; ++iter)
			{
				core::AVector3D* p = nv_ref->getPoint3D(*iter);
				if(p)
				{
					x = points[src*2];
					y = data[index].height - points[src*2 + 1];

					// Get corresponding mesh vertex
					int dst = vertex_map[p];

					// Randomize and assign points for Delaunay triangulation
					drpts.push_back(x + x * r * static_cast<double>(intRand()));
					drpts.push_back(y + y * r * static_cast<double>(intRand()));

					drpts_map.push_back(dst);

					// Assign texture coordinates
					mesh_vertices[dst].tex_u = static_cast<float>(x * inv_width);
					mesh_vertices[dst].tex_v = static_cast<float>(y * inv_height);
					
					++iters;
				}
				++src; 
			}
			DBGOUT(DBGNORMAL, "%d iters\n", iters);

			// Triangulate points
			std::list<unsigned> faces;
			if(drpts.size()/2 < 3) return;
			core::delaunay2(&drpts[0],drpts.size()/2,simplex,reinterpret_cast<void*>(&faces));

			// Create and lock faces, create and lock normals
			unsigned num_triangles = faces.size()/3;
			DBGOUT(DBGNORMAL,"Triangulation %d: %zd features, %d faces\n",index,drpts.size()/2,num_triangles);

			data[index].entity->mesh.createFaces(num_triangles);
			Face* mesh_faces = data[index].entity->mesh.lockFaces(0,num_triangles);
			assert(mesh_faces && "Cannot lock mesh faces");
			data[index].entity->mesh.createNormals(num_triangles);
			Vector3D* mesh_normals = data[index].entity->mesh.lockNormals(0,num_triangles);
			assert(mesh_normals && "Cannot lock mesh normals");

			// Process faces and face normals
			std::list<unsigned>::iterator faceiter = faces.begin();
			for(unsigned i = 0; i < num_triangles; ++i)
			{
				// Assign mesh triangle vertices
				mesh_faces[i].v1 = drpts_map[*faceiter++];
				mesh_faces[i].v2 = drpts_map[*faceiter++];
				mesh_faces[i].v3 = drpts_map[*faceiter++];
				mesh_faces[i].normal = i;

				/// TODO!
				if(mesh_faces[i].v1 >= (int)vertices.size() ||
							  mesh_faces[i].v2 >= (int)vertices.size() ||
							  mesh_faces[i].v3 >= (int)vertices.size())
				{
					DBGOUT(DBGNORMAL,"!!!!!!!!!! %d, %d, %d\n", mesh_faces[i].v1, mesh_faces[i].v2, mesh_faces[i].v3);
					mesh_faces[i].v1 = mesh_faces[i].v2 = mesh_faces[i].v3 = 0;
					data[index].entity->mesh.unlockNormals();
					data[index].entity->mesh.unlockFaces();
					data[index].entity->mesh.unlockVertices();
					return;
				}
				// compute edges vectors
				Vector3D e1 = mesh_vertices[mesh_faces[i].v1].pos -
					mesh_vertices[mesh_faces[i].v2].pos;
				Vector3D e2 = mesh_vertices[mesh_faces[i].v2].pos -
					mesh_vertices[mesh_faces[i].v3].pos;
//				Vector3D e3 = mesh_vertices[mesh_faces[i].v3].pos -
//					mesh_vertices[mesh_faces[i].v1].pos;

				// Compute normal vector
				mesh_normals[i].cross(e1,e2);
				mesh_normals[i].normalize();
			}

			// sort triangles by the longest edge of every triangle
			comparator cmp(mesh_vertices);
			std::sort(mesh_faces,mesh_faces+num_triangles,cmp);

			// Unlock all
			data[index].entity->mesh.unlockNormals();
			data[index].entity->mesh.unlockFaces();
			data[index].entity->mesh.unlockVertices();
		}

		void GLDelaunayRenderer::loadCamera(unsigned index)
		{
			assert(nv_ref && camavatar && index < data.size());
			core::Camera* cam = nv_ref->getCamera(data[index].view);
			if(cam)
			{
				GLCameraLoader loader(cam);
				Vector3D center(centroid[0],centroid[1],centroid[2]);
				data[index].eye = loader.location() - center;
				data[index].rot = loader.orientation();
			}
		}

		bool GLDelaunayRenderer::loadCameraAvatar(unsigned index)
		{
			if(!entity)
			{
				return false; //TODO!!!!!!!!!!!!
			}
			float scale = 2.0f*pensize;
			assert(camavatar && index < data.size());
			Vector3D center(centroid[0],centroid[1],centroid[2]);

			// uncomment this line if you want camera size to be proportional
			// to the object size
			//scale* = camavatar->mesh.getBoundingSphere() / minobjectsize;

			camavatar->world.identity();
			camavatar->world.setScale(Vector3D(scale,scale,scale));
			camavatar->world.setRotation(entity->world.getRotation() * data[index].rot);
			camavatar->world.setTranslation(entity->world.getRotation() * data[index].eye);
			return true;
		}

		void GLDelaunayRenderer::drawCamera(unsigned index)
		{
			assert(nv_ref && camavatar && index < data.size());
			loadCameraAvatar(index);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glPushMatrix();
			if((int)index == highlighted)
				glColor3ub(0x80,0x80,0x80);
			else
				glColor3ub(0xff,0xff,0xff);
			glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
			camavatar->render();
			glPopMatrix();
			glPushMatrix();
			if((int)index == highlighted)
				glColor3ub(0xff,0x00,0x00);
			else
				glColor3ub(0x00,0x00,0x00);
			glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
			camavatar->render();
			glPopMatrix();
			glEnable(GL_TEXTURE_2D);
			glEnable(GL_LIGHTING);
		}

		void GLDelaunayRenderer::select(Vector3D& org, Vector3D& dir)
		{
			float min_t = std::numeric_limits<float>::infinity();
			int result = -1;
			for(unsigned i = 0; i < data.size(); ++i)
			{
				if(loadCameraAvatar(i))
				{
					float temp = camavatar->hit(org,dir);
					if(temp < min_t)
					{
						min_t = temp;
						result = i;
					}
				}
			}
			highlighted = result;

			GLRenderer::select(org,dir);
		}

		GLDelaunayRenderer::GLDelaunayRenderer()
			: nv_ref(0),
			centroid(0.0,0.0,0.0),
			camavatar(0),
			highlighted(-1),
			maxdistance(0.0f),
			minobjectsize(std::numeric_limits<float>::infinity())
		{
		}

		GLDelaunayRenderer::~GLDelaunayRenderer()
		{
			attach((core::NView*)0);
			clear();
		}

		void GLDelaunayRenderer::attach(core::NView* nv)
		{
			nv_ref = nv;
#ifndef NDEBUG
			if(nv_ref)
			{
				DBGOUT(DBGNORMAL,"NView:  %p (%d Views, %d PointCorrs)\n",nv_ref,nv_ref->getNumViews(),nv_ref->getNumPointCorrs());
			}
			else
			{
				DBGOUT(DBGNORMAL,"NULL-NView attached\n");
			}
#endif
		}

		void GLDelaunayRenderer::init()
		{
// 			GLRenderer::getCamera().print();
// 			if(entity) entity->world.print();
			GLRenderer::init();

			if(!nv_ref) return;

			camavatar = new GLEntity;//TODO leak?
			GLAvatar::camera(camavatar->mesh);
			camavatar->drawframe = true;
			camavatar->mesh.buildBoundingVolumes();

			// Query vertices
			int num_points3D = nv_ref->getNumPoints3D();
			nv_points.clear();
			nv_points.reserve(num_points3D);
			DBGOUT(DBGVERBOSE,"Total %d vertices in NView\n",num_points3D);
			vertex_map.clear();
			NView::Point3DIterator p3d_it = nv_ref->getPoints3DBegin();
			NView::Point3DIterator p3d_end = nv_ref->getPoints3DEnd();
			for( ; p3d_it != p3d_end; ++p3d_it) {
				vertex_map[*p3d_it] = nv_points.size();
				nv_points.push_back(*p3d_it);
			}

			// Calculate centroid and prepare vertices
			centroid = core::Point3D(0.0,0.0,0.0);//TODO do not assume object contains (0,0,0)
			std::vector<core::Point3D> vertices;
			vertices.resize(num_points3D);

			core::AVector3D* pnt_temp = 0;
			double inv_w;
			unsigned index = 0;
			std::vector<core::AVector3D*>::iterator pnt_iter = nv_points.begin();
			std::vector<core::AVector3D*>::iterator pnt_end = nv_points.end();
			for( ; pnt_iter != pnt_end; ++pnt_iter) {
				// copy vertex
				pnt_temp = *pnt_iter;
				inv_w = 1.0 / (*pnt_temp)(3);
				vertices[index][0] = (*pnt_temp)(0) * inv_w;
				vertices[index][1] = (*pnt_temp)(1) * inv_w;
				vertices[index][2] = (*pnt_temp)(2) * inv_w;
				// sum up coordinates
				centroid = centroid + vertices[index];
				++index;
			}

			// Find centroid and shift vertices
			if(num_points3D)
				centroid = centroid / static_cast<double>(num_points3D);

			for(int i = 0; i < num_points3D; ++i)
				vertices[i] = vertices[i] - centroid;

			// reset reference scales
			maxdistance = 0.0f;
			minobjectsize = std::numeric_limits<float>::infinity();

			// Create each GLMesh
			for(unsigned i = 0; i < data.size(); ++i) {
				// Triangulate mesh
				{
					triangulate(i, vertices);
				}
				// Assign mesh texture
				assert(data[i].entity); // TODO TODO TODO
				DBGOUT(DBGNORMAL, "modifying entity %p, index %d\n", data[i].entity, i);
				DBGOUT(DBGNORMAL, "GLDelaunayRenderer::reinit  -  setting texture (%d/%d, %d/%d)\n", data[i].width, data[i].entity->texture.getWidth(),data[i].height, data[i].entity->texture.getHeight());
				data[i].entity->texture.set(data[i].width,data[i].height,data[i].pixels);
				data[i].entity->texture.setMode(TEX_REPLACE);
				// Load camera
				loadCamera(i);

				// Clear image cache
				delete[] data[i].pixels;
				data[i].pixels = 0;

				data[i].entity->mesh.buildBoundingVolumes();
				float minobjsz = data[i].entity->mesh.getBoundingSphere();
				if(minobjsz < minobjectsize)
					minobjectsize = minobjsz;
				float maxd = data[i].eye.lengthSqr();
				if(maxdistance < maxd)
					maxdistance = maxd;
			}
			maxdistance = std::sqrt(maxdistance);
			if(maxdistance < 10) maxdistance = 10;

			perspective(fieldofview);
		}
		
		unsigned GLDelaunayRenderer::getNumTriangulations()
		{
			return data.size();
		}

		void GLDelaunayRenderer::setActiveTriangulation(unsigned index)
		{
			if(index < data.size())
			{
				DBGOUT(DBGVERBOSE,"Active triangulation: %d\n",index);
				entity = data[index].entity;
			}
		}

		void GLDelaunayRenderer::setActiveCamera(unsigned index)
		{
			if(index < data.size())
			{
				//core::Camera* cam = nv_ref->getCamera(data[index].view);
				camera.set(data[index].eye,data[index].rot);
			}
		}

		void GLDelaunayRenderer::select(const core::ViewBase* const v, const bool reset)
		{
			for(unsigned i = 0; i < data.size(); ++i)
			{
				if(data[i].view==v)
				{
					selectView(i, reset);
					DBGOUT(DBGVERBOSE,"View selected: %p\n",v);
					return;
				}
			}
			DBGOUT(DBGNORMAL,"NView doesn't have the View: %p\n",v);
		}

		void GLDelaunayRenderer::paint()
		{
			// draw object
			GLRenderer::paint();

			// draw cameras
			if(displayflags & DISPLAY_CAMERAS)
			{
				if(!nv_ref) return;
				if(nv_ref->getCamerasBegin() == nv_ref->getCamerasEnd()) return;
				for(unsigned i = 0; i < data.size(); ++i)
					drawCamera(i);
			}
		}

		void GLDelaunayRenderer::pick(int x, int y)
		{
			int last_point = -1;
			int last_camera = -1;

			if(displayflags & DISPLAY_CAMERAS)
				last_camera = highlighted;

			if(entity)
				last_point = entity->highlighted;


			GLRenderer::pick(x,y);


			if(displayflags & DISPLAY_CAMERAS && highlighted >= 0 && highlighted!=last_camera)
			{
				core::Camera* cam = nv_ref->getCamera(data[highlighted].view);
				DBGOUT(DBGNORMAL,"Camera selected %p\n",cam);
				notifyListeners(cam);
			}

			if(entity && entity->highlighted >= 0 && entity->highlighted!=last_point)
			{
				int index = entity->mesh.extra(entity->highlighted);

				if(index!=entity->highlighted)
					DBGOUT(DBGSERIOUS,"Abnormal vertex (index %d, extra: %d)\n", entity->highlighted,index);

				if(index>=0 && index < (int)nv_points.size())
				{
					adore::core::AVector3D* nvp = nv_points[index];
					if(nvp)
					{
						DBGOUT(DBGNORMAL,"NVPoint selected %p\n",nvp);
						notifyListeners(nvp);
					}
				}
				else
				{
					DBGOUT(DBGSERIOUS,"Vertex (index %d, extra: %d) doesn't correspond to NVPoint!!!\n", entity->highlighted,index);
				}
			}
		}

		void GLDelaunayRenderer::perspective(float fov)
		{
			if(entity)
			{
				float span = entity->mesh.getBoundingSphere();
				if(span == 0.0f) {
					/// special case: it is no object here!
					float minx = 1000, miny = 1000, minz = 1000;
					float maxx = -1000, maxy = -1000, maxz = -1000;
					for(uint i = 0; i < data.size(); ++i) {
						if(data[i].eye.x < minx) minx = data[i].eye.x;
						if(data[i].eye.x > maxx) maxx = data[i].eye.x;
						if(data[i].eye.y < miny) miny = data[i].eye.y;
						if(data[i].eye.y > maxy) maxy = data[i].eye.y;
						if(data[i].eye.z < minz) minz = data[i].eye.z;
						if(data[i].eye.z > maxz) maxz = data[i].eye.z;
					}
					float x = maxx - minx, y = maxy - miny, z = maxz - minz;
					span = sqrt(x * x + y * y + z * z);
					const float spanfactor = 32.0f;
					float nplane = 0.5f;
					float fplane = 100.0f;
					nplane = span / spanfactor;
					fplane = span * spanfactor;
					fieldofview = fov;
					camera.perspective(fieldofview,aspect,nplane,fplane);
					camera.setOpenGLProjection();
					return;
				}
			}
			GLRenderer::perspective(fov);
		}

		void GLDelaunayRenderer::translate(const float delta[3], bool absolute)
		{
			GLRenderer::translate(delta,absolute);
		}

		void GLDelaunayRenderer::getTranslation(float (& /*v3*/)[3], uint /*idx*/) {
/*			if(!entity) return;
			
			GLEntity* cam = camavatar;///< camera avatar
			camavatar = new GLEntity;
			loadCameraAvatar(idx);
			Vector3D v3d = camavatar->world.getTranslation();
			delete camavatar; camavatar = cam;

			v3[0] = v3d.x;
			v3[1] = v3d.y;
			v3[2] = v3d.z;
			cout << "x: " << v3[0] <<", y: " << v3[1]<<", z: "<<v3[2] << endl;*/
		}

		void GLDelaunayRenderer::rotate(const float axis[3], float angle, bool absolute)
		{
			GLRenderer::rotate(axis,angle,absolute);
/*			std::cout << "Rotate\n";

			float v31[3], v32[3], v33[3];
			getTranslation(v31, 0);
			getTranslation(v32, 1);
			getTranslation(v33, 2);
			float d1 = sqrt(v31[0]*v31[0] + v31[1]*v31[1] + v31[2]*v31[2]);
			float d2 = sqrt(v32[0]*v32[0] + v32[1]*v32[1] + v32[2]*v32[2]);
			float d3 = sqrt(v33[0]*v33[0] + v33[1]*v33[1] + v33[2]*v33[2]);
			cout << "d1: " << d1 << ", d2: " << d2 << ", d3: " << d3 << endl;*/
		}

		void GLDelaunayRenderer::scale(const float factor[3], bool absolute)
		{
			GLRenderer::scale(factor,absolute);
		}

		void GLDelaunayRenderer::identity()
		{
			GLRenderer::identity();
		}

		void GLDelaunayRenderer::selectView(unsigned index, const bool reset)
		{
			setActiveTriangulation(index);//sets entity
			setActiveCamera(index);
			if(reset) entity->world.identity();
		}

		void GLDelaunayRenderer::safeView()
		{
			if(entity)
			{
				entity->world.identity();
				entity->mesh.buildBoundingVolumes();
				float radius = entity->mesh.getBoundingSphere();
				Vector3D eye(1.0f,1.0f,1.0f);
				eye.normalize();
				eye *= 2.0f * radius;
				Vector3D aim(0.0f,0.0f,0.0f);
				Vector3D up(0.0f,0.0f,1.0f);
				camera.set(eye,aim,up);
				DBGOUT(DBGVERBOSE,"Default camera is set (bounding sphere = %f)\n",radius);
			}
			else
			{
				DBGOUT(DBGVERBOSE,"Entity is not ready\n");
			}
		}

		void GLDelaunayRenderer::save(const char* filename)
		{
			if(nv_points.empty())
				return;

			unsigned index = 0;
			double inv_w;
			Point3D temp;
			Point3D lower(FLT_INF,FLT_INF,FLT_INF);
			Point3D upper(-FLT_INF,-FLT_INF,-FLT_INF);
			std::vector<Point3D> vertices(nv_points.size());
			std::vector<core::AVector3D*>::iterator pnt_iter;
			for(pnt_iter = nv_points.begin(); pnt_iter!=nv_points.end(); ++pnt_iter)
			{
				inv_w = 1.0 / (**pnt_iter)(3);
				vertices[index][0] = (**pnt_iter)(0) * inv_w;
				vertices[index][1] = (**pnt_iter)(1) * inv_w;
				vertices[index][2] = (**pnt_iter)(2) * inv_w;

				if(vertices[index][0] < lower[0])
					lower[0] = vertices[index][0];
				if(vertices[index][1] < lower[1])
					lower[1] = vertices[index][1];
				if(vertices[index][2] < lower[2])
					lower[2] = vertices[index][2];
				if(vertices[index][0] > upper[0])
					upper[0] = vertices[index][0];
				if(vertices[index][1] > upper[1])
					upper[1] = vertices[index][1];
				if(vertices[index][2] > upper[2])
					upper[2] = vertices[index][2];

				++index;
			}

			Point3D diff = upper - lower;
			diff[0] = std::abs(diff[0]);
			diff[1] = std::abs(diff[1]);
			diff[2] = std::abs(diff[2]);
			double scale;
			if(diff[0] > diff[1] && diff[0] > diff[2])
				scale = 1.0 / diff[0];
			else if(diff[1] > diff[2])
				scale = 1.0 / diff[1];
			else
				scale = 1.0 / diff[2];

			std::ofstream out(filename);
			for(unsigned i = 0; i<vertices.size();++i)
			{
				out<<scale*(vertices[i][0]+lower[0])<<' ';
				out<<scale*(vertices[i][1]+lower[1])<<' ';
				out<<scale*(vertices[i][2]+lower[2])<<'\n';
			}
		}
	}
}
