

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_GLMATH_H
#define ADORE_GLMATH_H

#include <cmath>
#include <memory>
#include <cassert>
#include <cstring>
#include <limits>

#include <adrcore/util/debug.h>

namespace adore
{
	namespace visual
	{
		/**	Various mathematical constants
		 */
		const float	PI				=	3.14159265358979323846f;
		const float	TWO_PI			=	2.0f * PI;
		const float	HALF_PI			=	0.5f * PI;
		const float	QUARTER_PI		=	0.25f * PI;
		const float	E				=	2.71828182845904523536f;
		const float	SQRT_2			=	1.41421356237309504880f;
		const float	SQRT_3			=	1.73205080756887729352f;
		const float	SQRT_1_OVER_2	=	0.70710678118654752440f;
		const float	SQRT_1_OVER_3	=	0.57735026918962576450f;
		const float	DEG2RAD			=	PI / 180.0f;
		const float	RAD2DEG			=	180.0f / PI;
		const float	FLT2CLR			=	255.0f;
		const float	CLR2FLT			=	1.0f / 255.0f;
		const float	FLT_INF			=	1e30f;
		const float	FLT_EPS			=	1.192092896e-07f;

		/**	Average treshold values for comparing vectors and matrices
		*/
		const float VECTOR_EPSILON = 0.001f;
		const float MATRIX_INVERSE_EPSILON = 1e-14f;
		const float MATRIX_EPSILON	= 1e-6f;

		/**	3D vector
		 *
		 *	A standard mathematical 3 dimensional vector. May also represent
		 *	a 2D vector or 2D point in homogeneous coordinates.
		 */
		class Vector3D
		{
		public:
			float x;		///< x-coordinate
			float y;		///< y-coordinate
			float z;		///< z-coordinate

			/**	Default constructor
			 */
			Vector3D();

			/**	Constructor with initialization
			 */
			Vector3D(const float x_, const float y_, const float z_);

			/**	Set vector components
			 */
			void set(const float x_, const float y_, const float z_);

			/**	Set all components to zero
			 */
			void zero();

			/**	Access vector components as an array
			 */
			float operator[](const int index) const;

			/**	Access vector components as an array
			 */
			float& operator[](const int index);

			/** Get inverse vector
			 */
			Vector3D operator-() const;

			/**	Vector-vector dot product
			 */
			float operator*(const Vector3D& a) const;

			/**	Vector-scalar product
			 */
			Vector3D operator*(const float a) const;

			/**	Vector-scalar division
			 */
			Vector3D operator/(const float a) const;

			/**	Componentwise vector addition
			 */
			Vector3D operator+(const Vector3D& a) const;

			/**	Componentwise vector subtraction
			 */
			Vector3D operator-(const Vector3D& a) const;

			/**	Componentwise vector addition
			 */
			Vector3D& operator+=(const Vector3D& a);

			/**	Componentwise vector subtraction
			 */
			Vector3D& operator-=(const Vector3D& a);

			/**	Componentwise vector division
			 */
			Vector3D& operator/=(const Vector3D& a);

			/** Vector-scalar division
			 */
			Vector3D& operator/=(const float a);

			/**	Vector-scalar product
			 */
			Vector3D& operator*=(const float a);

			/**	Equality operator, exact compare - no epsilon
			 */
			bool operator==(const Vector3D& a) const;
			
			const Vector3D&operator=(const Vector3D& a);

			/**	Inequality operator, exact compare - no epsilon
			 */
			bool operator!=(const Vector3D& a) const;

			/**	Scalar-vector product
			 */
			friend Vector3D operator*(const float a, const Vector3D b);

			/**	Compare 2 vectors exact, no epsilon
			*/
			bool compare(const Vector3D& a) const;	

			/**	Compare 2 vectors with epsilon, a good reference value is @c VECTOR_EPSILON
			 */
			bool compare(const Vector3D& a, const float epsilon) const;

			/**	Vector-vector cross product
			 */
			Vector3D cross(const Vector3D& a) const;

			/**	Assign self to the cross product of @c a and @c b
			*/
			Vector3D& cross(const Vector3D& a, const Vector3D& b);

			/**	Vector magnitude
			 */
			float length() const;

			/**	Vector magnitude squared
			 */
			float lengthSqr() const;

			/**	Normalize makes the vector a unit length and returns original length
			 */
			float normalize();

			/**	Cap vector length
			 */
			Vector3D& truncate(float length);

			/**	Componentwise clamp between min and max
			 */
			void clamp(const Vector3D &min, const Vector3D &max);

			/**	Snap to closest integer value
			 */
			void snap();

			/**	Snap towards integer-floor
			 */
			void snapInt();

			/**	Compute vector yaw (angle with XZ-plane) in radians
			 */
			float toYaw() const;

			/** Compute vector pitch (angle with YZ-plane) in radians
			 */
			float toPitch() const;

			/**	Rotate vector counterclockwise about the x-axis
			 *	@param angle Angle in radians
			 */
			void rotateAboutX(float angle);

			/**	Rotate vector counterclockwise about the y-axis
			 *	@param angle Angle in radians
			 */
			void rotateAboutY(float angle);

			/**	Rotate vector counterclockwise about the z-axis
			 *	@param angle Angle in radians
			 */
			void rotateAboutZ(float angle);

			/**	Rotate vector counterclockwise about an arbitrary axis
			 *	@param angle Angle in radians
			 *	@param axis Axis to rotate about
			 */
			void rotate(float angle, const Vector3D& axis);

			/**	Linearly interpolate between 2 vectors
			*/
			void lerp(const Vector3D& v1, const Vector3D& v2, const float t);

			/**	Access this vector as array of floats
			 */
			const float* toArray() const;

			/**	Access this vector as array of floats
			 */
			float* toArray();
		};

		inline Vector3D::Vector3D()
		{
			zero();
		}

		inline Vector3D::Vector3D(const float x_, const float y_, const float z_)
		{
			x = x_;
			y = y_;
			z = z_;
		}

		inline void Vector3D::set(const float x_, const float y_, const float z_)
		{
			x = x_;
			y = y_;
			z = z_;
		}

		inline void Vector3D::zero()
		{
			x = y = z = 0.0f;
		}

		inline float Vector3D::operator[](const int index) const
		{
			return (&x)[index];
		}
	
		inline float& Vector3D::operator[](const int index)
		{
			return (&x)[index];
		}

		inline Vector3D Vector3D::operator-() const
		{
			return Vector3D(-x,-y,-z);
		}

		inline float Vector3D::operator*(const Vector3D& a) const
		{
			return x*a.x+y*a.y+z*a.z;
		}

		inline Vector3D Vector3D::operator*(const float a) const
		{
			return Vector3D(x*a,y*a,z*a);
		}

		inline Vector3D Vector3D::operator/(const float a) const
		{
			float inv_a = 1.0f / a;
			return Vector3D(x*inv_a,y*inv_a,z*inv_a);
		}

		inline Vector3D Vector3D::operator+(const Vector3D& a) const
		{
			return Vector3D(x+a.x,y+a.y,z+a.z);
		}

		inline Vector3D Vector3D::operator-(const Vector3D& a) const
		{
			return Vector3D(x-a.x,y-a.y,z-a.z);
		}

		inline Vector3D& Vector3D::operator+=(const Vector3D& a)
		{
			x += a.x;
			y += a.y;
			z += a.z;

			return *this;
		}

		inline Vector3D& Vector3D::operator-=(const Vector3D& a)
		{
			x -= a.x;
			y -= a.y;
			z -= a.z;

			return *this;
		}

		inline Vector3D& Vector3D::operator/=(const Vector3D& a)
		{
			x /= a.x;
			y /= a.y;
			z /= a.z;

			return *this;
		}

		inline Vector3D& Vector3D::operator/=(const float a)
		{
			float inv_a = 1.0f / a;

			x *= inv_a;
			y *= inv_a;
			z *= inv_a;

			return *this;
		}

		inline Vector3D& Vector3D::operator*=(const float a)
		{
			x *= a;
			y *= a;
			z *= a;

			return *this;
		}

		inline bool Vector3D::operator==(const Vector3D& a) const
		{
			return compare(a);
		}
		
		inline const Vector3D &Vector3D::operator=(const Vector3D& a)
		{
			x = a.x;
			y = a.y;
			z = a.z;
			return *this;
		}

	
		inline bool Vector3D::operator!=(const Vector3D& a) const
		{
			return !compare(a);
		}

		inline Vector3D operator*(const float a, const Vector3D b)
		{
			return Vector3D(b.x*a,b.y*a,b.z*a);
		}

		inline bool Vector3D::compare(const Vector3D& a) const
		{
			return (x==a.x)&&(y==a.y)&&(z==a.z);
		}

		inline bool Vector3D::compare(const Vector3D& a, const float epsilon) const
		{
			if(std::fabs(x-a.x)>epsilon)
				return false;
			if(std::fabs(y-a.y)>epsilon)
				return false;
			if(std::fabs(z-a.z)>epsilon)
				return false;
			return true;
		}

		inline Vector3D Vector3D::cross(const Vector3D& a) const
		{
			return Vector3D(y*a.z - z*a.y, z*a.x - x*a.z, x*a.y - y*a.x);
		}

		inline Vector3D& Vector3D::cross(const Vector3D& a, const Vector3D& b)
		{
			x = a.y*b.z - a.z*b.y;
			y = a.z*b.x - a.x*b.z;
			z = a.x*b.y - a.y*b.x;

			return *this;
		}

		inline float Vector3D::length() const
		{
			return std::sqrt(x*x+y*y+z*z);
		}

		inline float Vector3D::lengthSqr() const
		{
			return x*x+y*y+z*z;
		}

		inline float Vector3D::normalize()
		{
			float sqr_length = x*x + y*y + z*z;
			float inv_length = 1.0 / std::sqrt(sqr_length);
			x *= inv_length;
			y *= inv_length;
			z *= inv_length;
			return inv_length * sqr_length;
		}

		inline Vector3D& Vector3D::truncate(float length)
		{
			float length2;
			float ilength;

			if (!length)
				zero();
			else
			{
				length2 = lengthSqr();
				if (length2 > length*length)
				{
					ilength = length * 1.0 / std::sqrt(length2);
					x *= ilength;
					y *= ilength;
					z *= ilength;
				}
			}

			return *this;
		}

		inline void Vector3D::clamp(const Vector3D &min, const Vector3D &max)
		{
			if(x<min.x)
				x = min.x;
			else if(x>max.x)
				x = max.x;
			if(y<min.y)
				y = min.y;
			else if(y>max.y)
				y = max.y;
			if(z<min.z)
				z = min.z;
			else if(z>max.z)
				z = max.z;
		}

		inline void Vector3D::snap()
		{
			x = std::floor(x + 0.5f);
			y = std::floor(y + 0.5f);
			z = std::floor(z + 0.5f);
		}

		inline void Vector3D::snapInt()
		{
			x = float(int(x));
			y = float(int(y));
			z = float(int(z));
		}

		inline float Vector3D::toYaw() const
		{
			float yaw;

			if((y==0.0f)&&(x==0.0f))
				yaw = 0.0f;
			else
			{
				yaw = RAD2DEG * std::atan2(y,x);
				if(yaw<0.0f)
					yaw += 360.0f;
			}

			return yaw;
		}

		inline float Vector3D::toPitch() const
		{
			float forward;
			float pitch;

			if((x==0.0f)&&(y==0.0f))
			{
				if(z>0.0f)
					pitch = 90.0f;
				else
					pitch = 270.0f;
			}
			else
			{
				forward = std::sqrt(x*x + y*y);
				pitch = RAD2DEG*std::atan2(z,forward);
				if(pitch<0.0f)
					pitch += 360.0f;
			}

			return pitch;
		}

		inline void Vector3D::rotateAboutX(float angle)
		{
			float s = std::sin(angle);
			float c = std::cos(angle);

			float ny = c*y - s*z;
			float nz = c*z + s*y;

			y = ny;
			z = nz;
		}

		inline void Vector3D::rotateAboutY(float angle)
		{
			float s = std::sin(angle);
			float c = std::cos(angle);

			float nx = c*x + s*z;
			float nz = c*z - s*x;

			x = nx;
			z = nz;
		}

		inline void Vector3D::rotateAboutZ(float angle)
		{
			float s = std::sin(angle);
			float c = std::cos(angle);

			float nx = c*x - s*y;
			float ny = c*y + s*x;

			x = nx;
			y = ny;
		}

		inline void Vector3D::rotate(float angle, const Vector3D& axis)
		{
			float s = std::sin(angle);
			float c = std::cos(angle);

			float k = 1.0f - c;

			float nx = x*(c + k*axis.x*axis.x) + y*(k*axis.x*axis.y - s*axis.z)	+ z*(k*axis.x*axis.z + s*axis.y);
			float ny = x*(k*axis.x*axis.y + s*axis.z) + y*(c + k*axis.y*axis.y)	+ z*(k*axis.y*axis.z - s*axis.x);
			float nz = x*(k*axis.x*axis.z - s*axis.y) + y*(k*axis.y*axis.z + s*axis.x) + z*(c + k*axis.z*axis.z);

			x = nx;
			y = ny;
			z = nz;
		}

		inline void Vector3D::lerp(const Vector3D& v1, const Vector3D& v2, const float t)
		{
			if(t<=0.0f)
				(*this) = v1;
			else if(t>=1.0f)
				(*this) = v2;
			else
				(*this) = v1 + t*(v2 - v1);
		}

		inline const float* Vector3D::toArray() const
		{
			return &x;
		}

		inline float* Vector3D::toArray()
		{
			return &x;
		}

		/**	4D vector
		 *
		 *	A standard mathematical 4 dimensional vector. May also represent
		 *	a 3D vector or 3D point in homogeneous coordinates.
		 */
		class Vector4D
		{
		public: 
			float x;		///< x-coordinate
			float y;		///< y-coordinate
			float z;		///< z-coordinate
			float w;		///< w-coordinate

			/**	Default constructor
			 */
			Vector4D();

			/**	Constructor with initialization
			 */
			Vector4D(const float x_, const float y_, const float z_, const float w_);

			/**	Set vector components
			 */
			void set(const float x_, const float y_, const float z_, const float w_);

			/**	Set all components to zero
			 */
			void zero();

			/**	Access vector components as an array
			 */
			float operator[](const int index) const;

			/**	Access vector components as an array
			 */
			float& operator[](const int index);

			/** Get inverse vector
			 */
			Vector4D operator-() const;

			/**	Vector-vector dot product
			 */
			float operator*(const Vector4D& a) const;

			/**	Vector-scalar product
			 */
			Vector4D operator*(const float a) const;

			/**	Vector-scalar division
			 */
			Vector4D operator/(const float a) const;

			/**	Componentwise vector addition
			 */
			Vector4D operator+(const Vector4D& a) const;

			/**	Componentwise vector subtraction
			 */
			Vector4D operator-(const Vector4D& a) const;

			/**	Componentwise vector addition
			 */
			Vector4D& operator+=(const Vector4D& a);

			/**	Componentwise vector subtraction
			 */
			Vector4D& operator-=(const Vector4D& a);

			/**	Componentwise vector division
			 */
			Vector4D& operator/=(const Vector4D& a);

			/** Vector-scalar division
			 */
			Vector4D& operator/=(const float a);

			/**	Vector-scalar product
			 */
			Vector4D& operator*=(const float a);

			/**	Equality operator, exact compare - no epsilon
			 */
			bool operator==(const Vector4D& a) const;

			/**	Inequality operator, exact compare - no epsilon
			 */
			bool operator!=(const Vector4D& a) const;

			/**	Scalar-vector product
			 */
			friend Vector4D operator*(const float a, const Vector4D b);

			/**	Compare 2 vectors exact, no epsilon
			 */
			bool compare(const Vector4D& a) const;

			/**	Compare 2 vectors with epsilon, a good reference value is @c VECTOR_EPSILON
			 */
			bool compare(const Vector4D& a, const float epsilon) const;

			/**	Vector magnitude
			 */
			float length() const;

			/**	Vector magnitude using fast square root function
			 */
			float lengthFast() const;

			/**	Vector magnitude squared
			 */
			float lengthSqr() const;

			/**	Normalize makes the vector a unit length and returns original length
			 */
			float normalize();

			/**	Access the 3D part of this vector as a separate vector
			 */
			const Vector3D& toVector3D() const;

			/**	Access the 3D part of this vector as a separate vector
			 */
			Vector3D& toVector3D();

			/**	Linearly interpolate between 2 vectors
			 */
			void lerp(const Vector4D& v1, const Vector4D& v2, const float t);

			/**	Access this vector as array of floats
			 */
			const float* toArray() const;

			/**	Access this vector as array of floats
			 */
			float* toArray();
		};

		inline Vector4D::Vector4D()
		{
		}

		inline Vector4D::Vector4D(const float x_, const float y_, const float z_, const float w_)
		{
			x = x_;
			y = y_;
			z = z_;
			w = w_;
		}

		inline void Vector4D::set(const float x_, const float y_, const float z_, const float w_)
		{
			x = x_;
			y = y_;
			z = z_;
			w = w_;
		}

		inline void Vector4D::zero()
		{
			x = y = z = w = 0.0f;
		}

		inline float Vector4D::operator[](const int index) const
		{
			return (&x)[index];
		}

		inline float& Vector4D::operator[](const int index)
		{
			return (&x)[index];
		}

		inline Vector4D Vector4D::operator-() const
		{
			return Vector4D(-x,-y,-z,-w);
		}

		inline float Vector4D::operator*(const Vector4D& a) const
		{
			return x*a.x+y*a.y+z*a.z+w*a.w;
		}

		inline Vector4D Vector4D::operator*(const float a) const
		{
			return Vector4D(x*a,y*a,z*a,w*a);
		}

		inline Vector4D Vector4D::operator/(const float a) const
		{
			float inv_a = 1.0f / a;
			return Vector4D(x*inv_a,y*inv_a,z*inv_a,w*inv_a);
		}

		inline Vector4D Vector4D::operator+(const Vector4D& a) const
		{
			return Vector4D(x+a.x,y+a.y,z+a.z,w+a.w);
		}

		inline Vector4D Vector4D::operator-(const Vector4D& a) const
		{
			return Vector4D(x-a.x,y-a.y,z-a.z,w-a.w);
		}

		inline Vector4D& Vector4D::operator+=(const Vector4D& a)
		{
			x += a.x;
			y += a.y;
			z += a.z;
			w += a.w;

			return *this;
		}

		inline Vector4D& Vector4D::operator-=(const Vector4D& a)
		{
			x -= a.x;
			y -= a.y;
			z -= a.z;
			w -= a.w;

			return *this;
		}

		inline Vector4D& Vector4D::operator/=(const Vector4D& a)
		{
			x /= a.x;
			y /= a.y;
			z /= a.z;
			w /= a.w;

			return *this;
		}

		inline Vector4D& Vector4D::operator/=(const float a)
		{
			float inv_a = 1.0f / a;
			x *= inv_a;
			y *= inv_a;
			z *= inv_a;
			w *= inv_a;

			return *this;
		}

		inline Vector4D& Vector4D::operator*=(const float a)
		{
			x *= a;
			y *= a;
			z *= a;
			w *= a;

			return *this;
		}

		inline bool Vector4D::operator==(const Vector4D& a) const
		{
			return compare(a);
		}

		inline bool Vector4D::operator!=(const Vector4D& a) const
		{
			return !compare(a);
		}

		inline Vector4D operator*(const float a, const Vector4D b)
		{
			return Vector4D(a*b.x,a*b.y,a*b.z,a*b.w);
		}

		inline bool Vector4D::compare(const Vector4D& a) const
		{
			return (x==a.x)&&(y==a.y)&&(z==a.z)&&(w==a.w);
		}

		inline bool Vector4D::compare(const Vector4D& a, const float epsilon) const
		{
			if(std::fabs(x-a.x)>epsilon)
				return false;
			if(std::fabs(y-a.y)>epsilon)
				return false;
			if(std::fabs(z-a.z)>epsilon)
				return false;
			if(std::fabs(w-a.w)>epsilon)
				return false;
			return true;
		}

		inline float Vector4D::length() const
		{
			return std::sqrt(x*x+y*y+z*z+w*w);
		}

		inline float Vector4D::lengthSqr() const
		{
			return x*x+y*y+z*z+w*w;
		}

		inline float Vector4D::normalize()
		{
			float sqr_length = x*x + y*y + z*z + w*w;
			float inv_length = 1.0f / std::sqrt(sqr_length);
			x *= inv_length;
			y *= inv_length;
			z *= inv_length;
			w *= inv_length;
			return inv_length * sqr_length;
		}

		inline const Vector3D& Vector4D::toVector3D() const
		{
			return *reinterpret_cast<const Vector3D*>(this);
		}

		inline Vector3D& Vector4D::toVector3D()
		{
			return *reinterpret_cast<Vector3D*>(this);
		}

		inline void Vector4D::lerp(const Vector4D& v1, const Vector4D& v2, const float t)
		{
			if(t <= 0.0f)
				(*this) = v1;
			else if(t>=1.0f)
				(*this) = v2;
			else
				(*this) = v1 + t*(v2 - v1);
		}

		inline const float* Vector4D::toArray() const
		{
			return &x;
		}

		inline float* Vector4D::toArray()
		{
			return &x;
		}

		/**	4x4 row-major square matrix
		 */
		class Matrix4x4
		{
			Vector4D mat[4];						///< rows of the matrix

		public:
			static const Matrix4x4 mat_zero;		///< 4x4 zero matrix
			static const Matrix4x4 mat_identity;	///< 4x4 identity matrix

			/**	Default constructor
			 */
			Matrix4x4();

			/**	Constructor with rowwise initialization
			 */
			explicit Matrix4x4(
				const Vector4D& x, const Vector4D& y,
				const Vector4D& z, const Vector4D& w);

			/**	Constructor with explicit componentwise initialization
			 */
			explicit Matrix4x4(
				const float m11, const float m12, const float m13, const float m14,
				const float m21, const float m22, const float m23, const float m24,
				const float m31, const float m32, const float m33, const float m34,
				const float m41, const float m42, const float m43, const float m44);

			/**	Constructor with componentwise initialization
			 */
			explicit Matrix4x4(const float src[4][4]);

			/**	Access matrix rows as an array of vectors
			 */
			const Vector4D&	operator[](int index) const;

			/**	Access matrix rows as an array of vectors
			 */
			Vector4D& operator[](int index);

			/**	Matrix negation
			 */
			Matrix4x4 operator-() const;

			/**	Matrix-scalar multiplication
			 */
			Matrix4x4 operator*(const float a) const;

			/**	Matrix-vector(column) right multiplication
			 */
			Vector4D operator*(const Vector4D& vec) const;

			/**	3D point transformation with perspective division
			 */
			Vector3D operator*(const Vector3D& pt) const;

			/**	Matrix multiplication
			 */
			Matrix4x4 operator*(const Matrix4x4& a) const;

			/**	Matrix addition
			 */
			Matrix4x4 operator+(const Matrix4x4& a) const;

			/**	Matrix subtraction
			 */
			Matrix4x4 operator-(const Matrix4x4& a) const;

			/**	Matrix-scalar multiplication
			 */
			Matrix4x4& operator*=(const float a);

			/**	Matrix multiplication
			 */
			Matrix4x4& operator*=(const Matrix4x4& a);

			/**	Matrix addition
			 */
			Matrix4x4& operator+=(const Matrix4x4& a);

			/**	Matrix subtraction
			 */
			Matrix4x4& operator-=(const Matrix4x4& a);

			/**	Scalar-matrix multiplication
			 */
			friend Matrix4x4 operator*(const float a, const Matrix4x4& mat);

			/**	3D point transformation with perspective division
			 */
			friend Vector3D operator*(const Vector3D& pt, const Matrix4x4& mat);

			/**	Matrix-vector(column) right multiplication
			 */
			friend Vector4D operator*(const Vector4D& vec, const Matrix4x4& mat);

			/**	3D point transformation and assignment with perspective division
			 */
			friend Vector3D& operator*=(Vector3D& pt, const Matrix4x4& mat);

			/**	Matrix-vector(column) right multiplication with assignment
			 */
			friend Vector4D& operator*=(Vector4D& vec, const Matrix4x4& mat);

			/**	Compare 2 matrices exact, no epsilon
			 */
			bool compare(const Matrix4x4& a) const;

			/**	Compare 2 matrices with epsilon, a good reference value is @c MATRIX_EPSILON
			 */
			bool compare(const Matrix4x4& a, const float epsilon) const;

			/**	Equality operator, exact compare - no epsilon
			 */
			bool operator==(const Matrix4x4& a) const;

			/**	Inequality operator, exact compare - no epsilon
			 */
			bool operator!=(const Matrix4x4& a) const;

			/**	Assign the matrix to zero matrix
			 */
			void zero();

			/**	Assign the matrix to identity matrix
			 */
			void identity();

			/**	Check, if the matrix is identity matrix
			 */
			bool isIdentity(const float epsilon = MATRIX_EPSILON) const;

			/**	Check, if the matrix is symmetric
			 */
			bool isSymmetric(const float epsilon = MATRIX_EPSILON) const;

			/**	Check, if the matrix is diagonal
			 */
			bool isDiagonal(const float epsilon = MATRIX_EPSILON) const;

			/**	Check, if the matrix is rotated
			 */
			bool isRotated() const;

			/**	Find the trace of the matrix
			 */
			float trace() const;

			/**	Find the determinant of the matrix
			 */
			float determinant() const;

			/**	Find the transpose of the matrix
			 */
			Matrix4x4 transpose() const;

			/**	Assign the matrix to its transpose
			 */
			Matrix4x4& transposeSelf();

			/**	Find the inverse of the matrix
			 */
			Matrix4x4 inverse() const;

			/**	Assign the matrix to its inverse
			 *	@return @c false if determinant is zero
			 */
			bool inverseSelf();

			/**	Multiply transpose of the matrix with argument
			 */
			Matrix4x4 transposeMultiply(const Matrix4x4& b) const;

			/**	Project vector using the matrix
			 */
			void projectVector(Vector4D src, Vector4D& dst) const;

			/**	Unproject vector using the matrix, works only for symmetric matrices
			 */
			void unprojectVector(Vector4D src, Vector4D& dst) const;

			/**	Access this matrix as array of floats
			 */
			const float* toArray() const;

			/**	Access this matrix as array of floats
			 */
			float* toArray();
		};

		inline Matrix4x4::Matrix4x4()
		{
		}

		inline  Matrix4x4::Matrix4x4(const Vector4D& x, const Vector4D& y, const Vector4D& z, const Vector4D& w)
		{
			mat[0] = x;
			mat[1] = y;
			mat[2] = z;
			mat[3] = w;
		}

		inline  Matrix4x4::Matrix4x4(
				const float m11, const float m12, const float m13, const float m14,
				const float m21, const float m22, const float m23, const float m24,
				const float m31, const float m32, const float m33, const float m34,
				const float m41, const float m42, const float m43, const float m44)
		{
			mat[0][0] = m11; mat[0][1] = m12; mat[0][2] = m13; mat[0][3] = m14;
			mat[1][0] = m21; mat[1][1] = m22; mat[1][2] = m23; mat[1][3] = m24;
			mat[2][0] = m31; mat[2][1] = m32; mat[2][2] = m33; mat[2][3] = m34;
			mat[3][0] = m41; mat[3][1] = m42; mat[3][2] = m43; mat[3][3] = m44;
		}

		inline const Vector4D&	Matrix4x4::operator[](int index) const
		{
			return mat[index];
		}

		inline Vector4D& Matrix4x4::operator[](int index)
		{
			return mat[index];
		}

		inline Matrix4x4 Matrix4x4::operator-() const
		{
			return Matrix4x4(
				-mat[0].x, -mat[0].y, -mat[0].z, -mat[0].w,
				-mat[1].x, -mat[1].y, -mat[1].z, -mat[1].w,
				-mat[2].x, -mat[2].y, -mat[2].z, -mat[2].w,
				-mat[3].x, -mat[3].y, -mat[3].z, -mat[3].w );
		}

		inline Matrix4x4 Matrix4x4::operator*(const float a) const
		{
			return Matrix4x4(
				mat[0].x * a, mat[0].y * a, mat[0].z * a, mat[0].w * a,
				mat[1].x * a, mat[1].y * a, mat[1].z * a, mat[1].w * a,
				mat[2].x * a, mat[2].y * a, mat[2].z * a, mat[2].w * a,
				mat[3].x * a, mat[3].y * a, mat[3].z * a, mat[3].w * a );
		}

		inline Vector4D Matrix4x4::operator*(const Vector4D& vec) const
		{
			return Vector4D(
				mat[0].x * vec.x + mat[0].y * vec.y + mat[0].z * vec.z + mat[0].w * vec.w,
				mat[1].x * vec.x + mat[1].y * vec.y + mat[1].z * vec.z + mat[1].w * vec.w,
				mat[2].x * vec.x + mat[2].y * vec.y + mat[2].z * vec.z + mat[2].w * vec.w,
				mat[3].x * vec.x + mat[3].y * vec.y + mat[3].z * vec.z + mat[3].w * vec.w );
		}

		inline Vector3D Matrix4x4::operator*(const Vector3D& pt) const
		{
			float s = mat[3].x * pt.x + mat[3].y * pt.y + mat[3].z * pt.z + mat[3].w;
			if (s == 0.0f)
				return Vector3D( 0.0f, 0.0f, 0.0f );

			if (s == 1.0f)
				return Vector3D(
					mat[0].x * pt.x + mat[0].y * pt.y + mat[0].z * pt.z + mat[0].w,
					mat[1].x * pt.x + mat[1].y * pt.y + mat[1].z * pt.z + mat[1].w,
					mat[2].x * pt.x + mat[2].y * pt.y + mat[2].z * pt.z + mat[2].w );
			else
			{
				float inv_s = 1.0f / s;
				return Vector3D(
					(mat[0].x * pt.x + mat[0].y * pt.y + mat[0].z * pt.z + mat[0].w) * inv_s,
					(mat[1].x * pt.x + mat[1].y * pt.y + mat[1].z * pt.z + mat[1].w) * inv_s,
					(mat[2].x * pt.x + mat[2].y * pt.y + mat[2].z * pt.z + mat[2].w) * inv_s );
			}
		}

		inline Matrix4x4 Matrix4x4::operator*(const Matrix4x4& a) const
		{
			Matrix4x4 result;
			const float* mat_a = reinterpret_cast<const float *>(this);
			const float* mat_b = reinterpret_cast<const float *>(&a);
			float* mat_r = reinterpret_cast<float *>(&result);

			for(int i = 0; i < 4; i++ )
			{
				for(int j = 0; j < 4; j++ )
				{
					*mat_r = mat_a[0] * mat_b[0 * 4 + j]
							+ mat_a[1] * mat_b[1 * 4 + j]
							+ mat_a[2] * mat_b[2 * 4 + j]
							+ mat_a[3] * mat_b[3 * 4 + j];
					mat_r++;
				}
				mat_a += 4;
			}
			return result;
		}

		inline Matrix4x4 Matrix4x4::operator+(const Matrix4x4& a) const
		{
			return Matrix4x4( 
				mat[0].x + a[0].x, mat[0].y + a[0].y, mat[0].z + a[0].z, mat[0].w + a[0].w,
				mat[1].x + a[1].x, mat[1].y + a[1].y, mat[1].z + a[1].z, mat[1].w + a[1].w,
				mat[2].x + a[2].x, mat[2].y + a[2].y, mat[2].z + a[2].z, mat[2].w + a[2].w,
				mat[3].x + a[3].x, mat[3].y + a[3].y, mat[3].z + a[3].z, mat[3].w + a[3].w );
		}

		inline Matrix4x4 Matrix4x4::operator-(const Matrix4x4& a) const
		{
			return Matrix4x4( 
				mat[0].x - a[0].x, mat[0].y - a[0].y, mat[0].z - a[0].z, mat[0].w - a[0].w,
				mat[1].x - a[1].x, mat[1].y - a[1].y, mat[1].z - a[1].z, mat[1].w - a[1].w,
				mat[2].x - a[2].x, mat[2].y - a[2].y, mat[2].z - a[2].z, mat[2].w - a[2].w,
				mat[3].x - a[3].x, mat[3].y - a[3].y, mat[3].z - a[3].z, mat[3].w - a[3].w );
		}

		inline Matrix4x4& Matrix4x4::operator*=(const float a)
		{
			mat[0].x *= a; mat[0].y *= a; mat[0].z *= a; mat[0].w *= a;
			mat[1].x *= a; mat[1].y *= a; mat[1].z *= a; mat[1].w *= a;
			mat[2].x *= a; mat[2].y *= a; mat[2].z *= a; mat[2].w *= a;
			mat[3].x *= a; mat[3].y *= a; mat[3].z *= a; mat[3].w *= a;
			return *this;
		}

		inline Matrix4x4& Matrix4x4::operator*=(const Matrix4x4& a)
		{
			float* mat_a = reinterpret_cast<float*>(this);
			const float* mat_b = reinterpret_cast<const float*>(&a);
			float dst[4];
			for(int i = 0; i < 4; i++ )
			{
				for(int j = 0; j < 4; j++ )
				{
					dst[j]  = mat_a[0] * mat_b[0 * 4 + j]
							+ mat_a[1] * mat_b[1 * 4 + j]
							+ mat_a[2] * mat_b[2 * 4 + j]
							+ mat_a[3] * mat_b[3 * 4 + j];
				}

				mat_a[0] = dst[0];
				mat_a[1] = dst[1];
				mat_a[2] = dst[2];
				mat_a[3] = dst[3];

				mat_a += 4;
			}
			return *this;
		}

		inline Matrix4x4& Matrix4x4::operator+=(const Matrix4x4& a)
		{
			mat[0].x += a[0].x; mat[0].y += a[0].y; mat[0].z += a[0].z; mat[0].w += a[0].w;
			mat[1].x += a[1].x; mat[1].y += a[1].y; mat[1].z += a[1].z; mat[1].w += a[1].w;
			mat[2].x += a[2].x; mat[2].y += a[2].y; mat[2].z += a[2].z; mat[2].w += a[2].w;
			mat[3].x += a[3].x; mat[3].y += a[3].y; mat[3].z += a[3].z; mat[3].w += a[3].w;
			return *this;
		}

		inline Matrix4x4& Matrix4x4::operator-=(const Matrix4x4& a)
		{
			mat[0].x -= a[0].x; mat[0].y -= a[0].y; mat[0].z -= a[0].z; mat[0].w -= a[0].w;
			mat[1].x -= a[1].x; mat[1].y -= a[1].y; mat[1].z -= a[1].z; mat[1].w -= a[1].w;
			mat[2].x -= a[2].x; mat[2].y -= a[2].y; mat[2].z -= a[2].z; mat[2].w -= a[2].w;
			mat[3].x -= a[3].x; mat[3].y -= a[3].y; mat[3].z -= a[3].z; mat[3].w -= a[3].w;
			return *this;
		}

		inline Matrix4x4 operator*(const float a, const Matrix4x4& mat)
		{
			return mat * a;
		}

		inline Vector4D operator*(const Vector4D& vec, const Matrix4x4& mat)
		{
			return mat * vec;
		}

		inline Vector3D operator*(const Vector3D& pt, const Matrix4x4& mat)
		{
			return mat * pt;
		}

		inline Vector4D& operator*=(Vector4D& vec, const Matrix4x4& mat)
		{
			vec = mat * vec;
			return vec;
		}

		inline Vector3D& operator*=(Vector3D& pt, const Matrix4x4& mat)
		{
			pt = mat * pt;
			return pt;
		}

		inline bool Matrix4x4::compare(const Matrix4x4& a) const
		{
			const float* mat_a = reinterpret_cast<const float *>(mat);
			const float* mat_b = reinterpret_cast<const float *>(a.mat);
			for(int i = 0; i < 4*4; ++i)
				if (mat_a[i] != mat_b[i])
					return false;
			return true;
		}

		inline bool Matrix4x4::compare(const Matrix4x4& a, const float epsilon) const
		{
			const float* mat_a = reinterpret_cast<const float *>(mat);
			const float* mat_b = reinterpret_cast<const float *>(a.mat);
			for(int i = 0; i < 4*4; ++i)
				if (std::fabs(mat_a[i] - mat_b[i]) > epsilon)
					return false;
			return true;
		}

		inline bool Matrix4x4::operator==(const Matrix4x4& a) const
		{
			return compare(a);
		}

		inline bool Matrix4x4::operator!=(const Matrix4x4& a) const
		{
			return !compare(a);
		}

		inline void Matrix4x4::zero()
		{
			std::memset(mat,0,sizeof(Matrix4x4));
		}

		inline void Matrix4x4::identity()
		{
			*this = mat_identity;
		}

		inline bool Matrix4x4::isIdentity(const float epsilon) const
		{
			return compare(mat_identity,epsilon);
		}

		inline bool Matrix4x4::isSymmetric(const float epsilon) const
		{
			for(int i = 1; i < 4; ++i)
				for(int j = 0; j < i; ++j)
					if (std::fabs(mat[i][j] - mat[j][i]) > epsilon)
						return false;
			return true;
		}

		inline bool Matrix4x4::isDiagonal(const float epsilon) const
		{
			for(int i = 0; i < 4; ++i)
				for(int j = 0; j < 4; ++j)
					if(i != j && std::fabs(mat[i][j]) > epsilon)
						return false;
			return true;
		}

		inline bool Matrix4x4::isRotated() const
		{
			if (!mat[0][1] && !mat[0][2] &&
				!mat[1][0] && !mat[1][2] &&
				!mat[2][0] && !mat[2][1] )
				return false;
			else
				return true;
		}

		inline float Matrix4x4::trace() const
		{
			return (mat[0][0] + mat[1][1] + mat[2][2] + mat[3][3]);
		}

		inline Matrix4x4 Matrix4x4::inverse() const
		{
			Matrix4x4 inv_mat;
	
			inv_mat = *this;
			bool result = inv_mat.inverseSelf();
			ADORE_UNUSED( result );
			assert(result);
			return inv_mat;
		}

		inline Matrix4x4 Matrix4x4::transposeMultiply(const Matrix4x4& b) const
		{
			return Matrix4x4(
				mat[0].x * b[0].x + mat[1].x * b[1].x + mat[2].x * b[2].x + mat[3].x * b[3].x,
				mat[0].x * b[0].y + mat[1].x * b[1].y + mat[2].x * b[2].y + mat[3].x * b[3].y,
				mat[0].x * b[0].z + mat[1].x * b[1].z + mat[2].x * b[2].z + mat[3].x * b[3].z,
				mat[0].x * b[0].w + mat[1].x * b[1].w + mat[2].x * b[2].w + mat[3].x * b[3].w,

				mat[0].y * b[0].x + mat[1].y * b[1].x + mat[2].y * b[2].x + mat[3].y * b[3].x,
				mat[0].y * b[0].y + mat[1].y * b[1].y + mat[2].y * b[2].y + mat[3].y * b[3].y,
				mat[0].y * b[0].z + mat[1].y * b[1].z + mat[2].y * b[2].z + mat[3].y * b[3].z,
				mat[0].y * b[0].w + mat[1].y * b[1].w + mat[2].y * b[2].w + mat[3].y * b[3].w,

				mat[0].z * b[0].x + mat[1].z * b[1].x + mat[2].z * b[2].x + mat[3].z * b[3].x,
				mat[0].z * b[0].y + mat[1].z * b[1].y + mat[2].z * b[2].y + mat[3].z * b[3].y,
				mat[0].z * b[0].z + mat[1].z * b[1].z + mat[2].z * b[2].z + mat[3].z * b[3].z,
				mat[0].z * b[0].w + mat[1].z * b[1].w + mat[2].z * b[2].w + mat[3].z * b[3].w,

				mat[0].w * b[0].x + mat[1].w * b[1].x + mat[2].w * b[2].x + mat[3].w * b[3].x,
				mat[0].w * b[0].y + mat[1].w * b[1].y + mat[2].w * b[2].y + mat[3].w * b[3].y,
				mat[0].w * b[0].z + mat[1].w * b[1].z + mat[2].w * b[2].z + mat[3].w * b[3].z,
				mat[0].w * b[0].w + mat[1].w * b[1].w + mat[2].w * b[2].w + mat[3].w * b[3].w
				);
		}

		inline void Matrix4x4::projectVector(Vector4D src, Vector4D& dst) const
		{
			dst.x = src * mat[0];
			dst.y = src * mat[1];
			dst.z = src * mat[2];
			dst.w = src * mat[3];
		}

		inline void Matrix4x4::unprojectVector(Vector4D src, Vector4D& dst) const
		{
			dst = mat[0] * src.x + mat[1] * src.y + mat[2] * src.z + mat[3] * src.w;
		}

		inline const float* Matrix4x4::toArray() const
		{
			return mat[0].toArray();
		}

		inline float* Matrix4x4::toArray()
		{
			return mat[0].toArray();
		}

		/**	An algebraic quaternion
		 *
		 *	A quaternion is a hypercomplex number, usually represented 
		 *	as w + i*x + j*y + k*z in its algebraic form or as a tupple
		 *	{vector(x,y,z), scalar w}.
		 *
		 *	The following rules apply:
		 *
		 *	i*i = j*j = k*k = -1 @n
		 *	i*j = -j*i = k @n
		 *	j*k = -k*j = i @n
		 *	k*i = -i*k = j @n
		 *
		 *	Quaternions can be used to represent and interpolate rotations.
		 *	A quaternion q written as q = [v,s], where s = cos(phi/2),
		 *	v = (x,y,z) * sin(phi/2) / |(x,y,z)| represents a rotation
		 *	about an axis (0,0,0)-(x,y,z) by an angle phi.
		 *
		 *	An important rule: if two rotations are represented with
		 *	quaternions, then their product represents a combination of these
		 *	rotations. Note that quaternions are not commutative: r*t != t*r.
		 */
		class Quaternion
		{
		public:
			float x;			///< factor by "i"
			float y;			///< factor by "j"
			float z;			///< factor by "k"
			float w;			///< scalar

			/**	Constructor
			 */
			Quaternion();

			/**	Constructor with initialization
			 */
			Quaternion(float x_, float y_, float z_, float w_);

			/**	Set quaternion components
			 */
			void set(float x_, float y_, float z_, float w_);

			/**	Make a quaternion from axis-angle representation of rotation
			 */
			void axisAngle(Vector3D axis, float angle);

			/**	Make a quaternion from matrix representation of rotation
			 */
			void matrix(const Matrix4x4& rotmat);

			/**	Access quaternion components as an array
			 */
			float operator[](int index) const;

			/**	Access quaternion components as an array
			 */
			float& operator[](int index);

			/** Get inverse quaternion for addition
			 */
			Quaternion operator-() const;

			/**	Quaternion addition
			 */
			Quaternion operator+(const Quaternion& a) const;

			/**	Quaternion multiplication
			 */
			Quaternion operator*(const Quaternion& a) const;

			/**	Rotate a vector
			 */
			Vector3D operator*(const Vector3D& a) const;

			/**	Quaternion-scalar multiplication
			 */
			Quaternion operator*(float a) const;

			/**	Quaternion subtraction
			 */
			Quaternion operator-(const Quaternion& a) const;

			/**	Assignment operator
			 */
			Quaternion& operator=(const Quaternion& a);

			/**	Quaternion addition with assignment
			 */
			Quaternion& operator+=(const Quaternion& a);

			/**	Quaternion subtraction with assignment
			 */
			Quaternion& operator-=(const Quaternion& a);

			/**	Quaternion multiplication with assignment
			 */
			Quaternion& operator*=(const Quaternion& a);
	
			/**	Quaternion-scalar multiplication with assignment
			 */
			Quaternion& operator*=(float a);

			/**	Quaternion-scalar multiplication
			 */
			friend Quaternion operator*(const float a, const Quaternion& b);

			/**	Rotate a vector
			 */
			friend Vector3D operator*(const Vector3D& a, const Quaternion& b);

			/**	Equality operator, exact compare - no epsilon
			 */
			bool operator==(const Quaternion& a) const;

			/**	Inequality operator, exact compare - no epsilon
			 */
			bool operator!=(const Quaternion& a) const;

			/**	Compare 2 quaternions exact, no epsilon
			 */
			bool compare(const Quaternion& a) const;

			/**	Compare 2 quaternions with epsilon
			 */
			bool compare(const Quaternion& a, const float epsilon) const;

			/** Get inverse quaternion for multiplication
			 */
			Quaternion inverse() const;

			/** Get inverse of the unit quaternion for multiplication
			 */
			Quaternion unitInverse() const;

			/**	Quaternion magnitude
			 */
			float length() const;

			/**	Normalize makes quaternion a unit length
			 */
			Quaternion& normalize();

			/**	Get the matrix form for rotation
			 */
			Matrix4x4 toMatrix() const;

			/**	Access this quaternion as array of floats
			 */
			const float* toArray() const;

			/**	Access this quaternion as array of floats
			 */
			float* toArray();
		};

		inline Quaternion::Quaternion()
		{
		}

		inline Quaternion::Quaternion(float x_, float y_, float z_, float w_)
		{
			x = x_;
			y = y_;
			z = z_;
			w = w_;
		}

		inline void Quaternion::set(float x_, float y_, float z_, float w_)
		{
			x = x_;
			y = y_;
			z = z_;
			w = w_;
		}

		inline void Quaternion::axisAngle(Vector3D axis, float angle)
		{
			axis.normalize();
			float halfangle = angle / 2;
			float s = std::sin(halfangle);
			x = axis.x * s;
			y = axis.y * s;
			z = axis.z * s;
			w = std::cos(halfangle);
		}

		inline float Quaternion::operator[]( int index ) const
		{
			assert((index >= 0) && (index < 4));
			return (&x)[index];
		}

		inline float& Quaternion::operator[]( int index )
		{
			assert((index >= 0) && (index < 4));
			return (&x)[index];
		}

		inline Quaternion Quaternion::operator-() const
		{
			return Quaternion(-x,-y,-z,-w);
		}

		inline Quaternion Quaternion::operator+(const Quaternion& a) const
		{
			return Quaternion(x + a.x,y + a.y,z + a.z,w + a.w);
		}

		inline Quaternion Quaternion::operator-(const Quaternion& a) const
		{
			return Quaternion(x - a.x,y - a.y,z - a.z,w - a.w);
		}

		inline Quaternion Quaternion::operator*(const Quaternion& a) const
		{
			return Quaternion(	w*a.x + x*a.w + y*a.z - z*a.y,
						w*a.y + y*a.w + z*a.x - x*a.z,
						w*a.z + z*a.w + x*a.y - y*a.x,
						w*a.w - x*a.x - y*a.y - z*a.z );
		}

		inline Vector3D Quaternion::operator*(const Vector3D& a) const
		{
			float	wx, wy, wz;
			float	xx, yy, yz;
			float	xy, xz, zz;
			float	x2, y2, z2;

			x2 = x + x;	y2 = y + y;	z2 = z + z;

			xx = x * x2;	xy = x * y2;	xz = x * z2;
					yy = y * y2;	yz = y * z2;
							zz = z * z2;
			wx = w * x2;	wy = w * y2;	wz = w * z2;

			return Vector3D(
				(1.0f-(yy+zz))*a.x	+	(xy-wz)*a.y		+	(xz+wy)*a.z,
				(xy + wz)*a.x		+	(1.0f-(xx+zz))*a.y	+	(yz-wx)*a.z,
				(xz - wy)*a.x		+	(yz + wx)*a.y		+	(1.0f-(xx+yy))*a.z );
		}

		inline Quaternion Quaternion::operator*(float a) const
		{
			return Quaternion(x * a,y * a,z * a,w * a);
		}

		inline Quaternion operator*(const float a, const Quaternion& b)
		{
			return b * a;
		}

		inline Vector3D operator*(const Vector3D &a, const Quaternion& b)
		{
			return b * a;
		}

		inline Quaternion& Quaternion::operator=(const Quaternion &a)
		{
			x = a.x;
			y = a.y;
			z = a.z;
			w = a.w;
			return *this;
		}

		inline Quaternion& Quaternion::operator+=(const Quaternion& a)
		{
			x += a.x;
			y += a.y;
			z += a.z;
			w += a.w;
			return *this;
		}

		inline Quaternion& Quaternion::operator-=(const Quaternion& a)
		{
			x -= a.x;
			y -= a.y;
			z -= a.z;
			w -= a.w;
			return *this;
		}

		inline Quaternion& Quaternion::operator*=(const Quaternion& a)
		{
			*this = *this * a;
			return *this;
		}

		inline Quaternion& Quaternion::operator*=(float a)
		{
			x *= a;
			y *= a;
			z *= a;
			w *= a;
			return *this;
		}
	
		inline bool Quaternion::operator==(const Quaternion& a) const
		{
			return compare(a);
		}

		inline bool Quaternion::operator!=(const Quaternion& a) const
		{
			return !compare(a);
		}

		inline bool Quaternion::compare(const Quaternion& a) const
		{
			return ((x == a.x) && (y == a.y) && (z == a.z) && (w == a.w));
		}

		inline bool Quaternion::compare(const Quaternion& a, const float epsilon) const
		{
			if (std::fabs(x - a.x) > epsilon)
				return false;
			if (std::fabs(y - a.y) > epsilon)
				return false;
			if (std::fabs(z - a.z) > epsilon)
				return false;
			if (std::fabs(w - a.w) > epsilon)
				return false;
			return true;
		}

		inline Quaternion Quaternion::inverse() const
		{
			float len = length();
			assert(len);

			len = 1.0f / len;
			return Quaternion(-x * len,-y * len,-z * len,w * len);
		}

		inline Quaternion Quaternion::unitInverse() const
		{
			return Quaternion(-x,-y,-z,w);
		}

		inline float Quaternion::length() const
		{
			float len;
			len = x * x + y * y + z * z + w * w;
			return std::sqrt(len);
		}

		inline Quaternion& Quaternion::normalize()
		{
			float len;
			float inv_len;

			len = length();
			if (len)
			{
				inv_len = 1.0 / len;
				x *= inv_len;
				y *= inv_len;
				z *= inv_len;
				w *= inv_len;
			}
			return *this;
		}

		inline const float* Quaternion::toArray() const
		{
			return &x;
		}

		inline float* Quaternion::toArray()
		{
			return &x;
		}
	}
}

#endif	// ADORE_GLMATH_H
