

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#include "glmesh.h"

#include <cmath>
#include <limits>
#include <iostream>
#include <GL/gl.h>
#include <cstdio>

using namespace std;

namespace adore
{
	namespace visual
	{
		GLMesh::GLMesh()
		{
			displaylist = 0;

			vertices = 0;

			normals = 0;
			faces = 0;

			numvert = 0;
			numnorm = 0;
			numface = 0;

			allocvert = 0;
			allocnorm = 0;
			allocface = 0;

			lockvert = false;	
			locknorm = false;
			lockface = false;

			dirty = false;

			lod = 1.1f;
			bsphere = -1.0f;

		}

		GLMesh::~GLMesh()
		{
			if(displaylist)
				glDeleteLists(displaylist,2);
			delete[] vertices;
			delete[] normals;
			delete[] faces;
		}

		void GLMesh::setVertices(Vertex* vertexdata, int startindex, int count)
		{
			if(startindex >= 0 && (startindex+count) < numvert)
			{
				std::memcpy(vertexdata,vertices+startindex,sizeof(Vertex)*count);
				dirty = true;
			}
		}

		void GLMesh::setNormals(Vector3D* normaldata, int startindex, int count)
		{
			if(startindex >= 0 && (startindex+count) < numnorm)
			{
				std::memcpy(normaldata,normals+startindex,sizeof(Vector3D)*count);
				dirty = true;
			}
		}

		void GLMesh::setFaces(Face* facedata, int startindex, int count)
		{
			if(startindex >= 0 && (startindex+count) < numvert)
			{
				std::memcpy(facedata,faces+startindex,sizeof(Face)*count);
				dirty = true;
			}
		}

		void GLMesh::createVertices(int count)
		{
			if(!count) return;
			assert(!lockvert && count > 0);
			if(count > allocvert)
			{
				Vertex* temp = new Vertex[count];
				std::memcpy(temp,vertices,sizeof(Vertex)*numvert);
				delete[] vertices;
				vertices = temp;
				allocvert = count;
				dirty = true;
			}
			numvert = count;
		}

		void GLMesh::createNormals(int count)
		{
			assert(!locknorm && count > 0);
			if(count > allocnorm)
			{
				Vector3D* temp = new Vector3D[count];
				std::memcpy(temp,normals,sizeof(Vector3D)*numnorm);
				delete[] normals;
				normals = temp;
				allocnorm = count;
				dirty = true;
			}
			numnorm = count;
		}

		void GLMesh::createFaces(int count)
		{
		assert(!lockface);
		assert(count > 0);
			if(count > allocface)
			{
				Face* temp = new Face[count];
				std::memcpy(temp,faces,sizeof(Face)*numface);
				delete[] faces;
				faces = temp;
				allocface = count;
				dirty = true;
			}
			numface = count;
		}

		int GLMesh::getNumVertices()
		{
			return numvert;
		}

		int GLMesh::getNumFaces()
		{
			return numface;
		}

		int GLMesh::getNumNormals()
		{
			return numnorm;
		}

		Vertex* GLMesh::lockVertices(int startindex, int count)
		{
			assert(!lockvert && "Multiple lock is not allowed");
			if(count > 0 && startindex >= 0 && (startindex+count) <= numvert)
			{
				lockvert = true;
				return vertices + startindex;
			}
			return NULL;
		}

		Vector3D* GLMesh::lockNormals(int startindex, int count)
		{
			assert(!locknorm && "Multiple lock is not allowed");
			if(startindex >= 0 && (startindex+count) <= numnorm)
			{
				locknorm = true;
				return normals + startindex;
			}
			else
				return 0;
		}

		Face* GLMesh::lockFaces(int startindex, int count)
		{
			assert(!lockface && "Multiple lock is not allowed");
			if(startindex >= 0 && (startindex+count) <= numface)
			{
				lockface = true;
				return faces + startindex;
			}
			else
				return 0;
		}

		void GLMesh::unlockVertices()
		{
			lockvert = false;
			dirty = true;
			//current implementation does nothing
		}

		void GLMesh::unlockNormals()
		{
			locknorm = false;
			dirty = true;
			//current implementation does nothing
		}

		void GLMesh::unlockFaces()
		{
			lockface = false;
			dirty = true;
			//current implementation does nothing
		}

		void GLMesh::setLevelOfDetail(float value)
		{
			lod = value;
			dirty = true;
		}

		void GLMesh::buildDisplayList()
		{
			// calculate number of visible triangles
			int numvisible = static_cast<int>(numface * lod);
			if(numvisible<0)
				numvisible = 0;
			if(numvisible>numface)
				numvisible = numface;	

			// calculate bounding volumes
			buildBoundingVolumes();

			if(!displaylist)
				displaylist = glGenLists(2);

			// draw faces
			glNewList(displaylist,GL_COMPILE);

			glBegin(GL_TRIANGLES);
	
			for(int i = 0; i< numvisible; ++i)
			{
				glNormal3fv(normals[faces[i].v1].toArray());

				glTexCoord2fv(&vertices[faces[i].v1].tex_u);
				glVertex3fv(vertices[faces[i].v1].pos.toArray());

				glTexCoord2fv(&vertices[faces[i].v2].tex_u);
				glVertex3fv(vertices[faces[i].v2].pos.toArray());

				glTexCoord2fv(&vertices[faces[i].v3].tex_u);
				glVertex3fv(vertices[faces[i].v3].pos.toArray());
			}

			glEnd();

			glEndList();

			//draw normals
			glNewList(displaylist+1,GL_COMPILE);

			glBegin(GL_LINES);

			Vector3D start, stop;

			for(int i = 0; i< numvisible; ++i)
			{
				start.zero();
				start += vertices[faces[i].v1].pos;
				start += vertices[faces[i].v2].pos;
				start += vertices[faces[i].v3].pos;
				start /= 3;

				stop = start + normals[faces[i].normal];

				glVertex3fv(start.toArray());
				glVertex3fv(stop.toArray());
			}

			glEnd();

			glEndList();
			dirty = false;
		}

		void GLMesh::buildBoundingVolumes()
		{
			int numvisible = static_cast<int>(numface * lod);

			// clamp
			if(numvisible<0)
				numvisible = 0;
			if(numvisible>numface)
				numvisible = numface;	

			float length;
			bsphere = 0.0f;
			int num = getNumVertices();
			for(int i = 0; i < num; ++i)
			{
				length = vertices[i].pos.lengthSqr();
				bsphere = length > bsphere ? length : bsphere;
			}
			bsphere = std::sqrt(bsphere);
			return;
			for(int i = 0; i < numvisible; ++i)
			{
				length = vertices[faces[i].v1].pos.lengthSqr();
				bsphere = length > bsphere ? length : bsphere;
				length = vertices[faces[i].v2].pos.lengthSqr();
				bsphere = length > bsphere ? length : bsphere;
				length = vertices[faces[i].v3].pos.lengthSqr();
				bsphere = length > bsphere ? length : bsphere;
			}

			bsphere = std::sqrt(bsphere);
		}

		void GLMesh::draw()
		{
			if(dirty)
				buildDisplayList();

			glCallList(displaylist);
		}

		void GLMesh::drawNormals()
		{
			if(dirty)
				buildDisplayList();

			glCallList(displaylist+1);
		}

		void GLMesh::drawVertexRange(int index, int count)
		{
			assert(index >= 0 && count > 0 && index+count <= numvert && "invalid range" );

			glBegin(GL_POINTS);
			for(int i = 0; i < count; ++i)
				glVertex3fv(vertices[index+i].pos.toArray());
			glEnd();
		}

		void GLMesh::drawFaceRange(int index, int count)
		{
			assert(index >= 0 && count > 0 && index+count <= numface && "invalid range" );
			glBegin(GL_TRIANGLES);
			for(int i = 0; i< count; ++i)
			{
				int addr = index + i;
				glNormal3fv(normals[faces[addr].normal].toArray());
				glTexCoord2f(vertices[faces[addr].v1].tex_u,vertices[faces[addr].v1].tex_v);
				glVertex3fv(vertices[faces[addr].v1].pos.toArray());
				glVertex3fv(vertices[faces[addr].v2].pos.toArray());
				glVertex3fv(vertices[faces[addr].v3].pos.toArray());
			}
			glEnd();
		}

		void GLMesh::drawBillboardRange(int index, int count, GLMesh& mesh, float scl)
		{
			if(index < 0 || count <= 0 || index+count > numvert) {
				return;
			}

			float world[16];

			glMatrixMode(GL_MODELVIEW);

			for(int i = 0; i < count; ++i)
			{
				int addr = index + i;
				glPushMatrix();
				glTranslatef(vertices[addr].pos.x,vertices[addr].pos.y,vertices[addr].pos.z);
				glGetFloatv(GL_MODELVIEW_MATRIX,world);

				// commented out entries need not to be changed
				world[0] = scl;			world[4] = 0.0f;		world[8] = 0.0f;		//world[12] = world[12];
				world[1] = 0.0f;		world[5] = scl;			world[9] = 0.0f;		//world[13] = world[13];
				world[2] = 0.0f;		world[6] = 0.0f;		world[10] = scl;		//world[14] = world[14];
				//world[3] = world[3];	world[7] = world[7];	world[11] = world[11];	world[15] = world[15];

				glLoadMatrixf(world);
				mesh.draw();
				glPopMatrix();
			}
		}

		int GLMesh::select(Vector3D org, Vector3D dir, float* tret)
		{
			float t_min = std::numeric_limits<float>::infinity();
			float t[2] = { 0.0f,0.0f };

			// test against bounding sphere
			Vector3D center;
			center.zero();
			if(lineSphereIntersect(center,bsphere,org,dir,t))
			{
				if(t[0]<0.0f && t[1] < 0.0f)
					return -1;
				else
					t_min = t[0] > t[1] ? t[0] : t[1];
			}
			else
				return -1;

			int index_min = -1;
			for(int i = 0; i < numface; ++i)
			{
				if(lineTriangleIntersect(vertices[faces[i].v1].pos,vertices[faces[i].v2].pos,
					vertices[faces[i].v3].pos,org,dir,t) && t[0] > 0.0f && t[0] < t_min)
				{
					t_min = t[0];
					index_min = i;
				}
			}

			if(index_min>=0 && tret)
				*tret = t_min;
			return index_min;
		}

		int GLMesh::select(Vector3D org, Vector3D dir, float radius)
		{
			float t_min = std::numeric_limits<float>::infinity();
			int index_min = -1;
			float t[2] = { 0.0f,0.0f };

			// test against bounding sphere
			Vector3D center;
			center.zero();
			if(lineSphereIntersect(center,bsphere,org,dir,t))
			{
// 				printf("GLMesh::select() bsphere intersected\n");
				if(t[0]<0.0f && t[1] < 0.0f)
					return -1;
				else
					t_min = t[0] > t[1] ? t[0] : t[1];
			}
			else
			{
// 				printf("GLMesh::select() bsphere not intersected\n");
				return -1;
			}

			for(int i = 0; i < numvert; ++i)
			{
				switch(lineSphereIntersect(vertices[i].pos,radius,org,dir,t))
				{
				case 0:
				default:
					break;
				case 2:
					if(t[1]>0.0f && t[1] < t[0])
						t[0] = t[1];
					// no break statement here!
				case 1:
					if(t[0] > 0.0f && t[0] < t_min)
					{
						t_min = t[0];
						index_min = i;
					}
					break;
				}
			}
			return index_min;
		}

		int GLMesh::extra(int index)
		{
			assert(index >= 0 && index < numvert);
			return vertices[index].extra;
		}

		float GLMesh::getBoundingSphere()
		{
			return bsphere;
		}

		int GLMesh::lineSphereIntersect(
				Vector3D& center,
				float& radius,
				Vector3D& org,
				Vector3D& dir,
				float t[2])
		{
			float b, c, discrm;
			Vector3D pmc = org - center;
			b = dir * pmc;
			c = pmc * pmc - radius * radius;
			discrm = b * b - c;
			if(discrm > 0.0f)
			{
				if(t)
				{
					double d = std::sqrt(discrm);
					t[0] = - b + d;
					t[1] = - b - d;
				}
				return 2;
			}
			else if(std::fabs(discrm) < std::numeric_limits<float>::epsilon())
				// the line is tangent to the sphere
			{
				if(t)
					t[0] = -b;
				return 1;
			}
			else
				return 0;
		}
		
		bool GLMesh::lineTriangleIntersect(
				Vector3D& a,
				Vector3D& b,
				Vector3D& c,
				Vector3D& org,
				Vector3D& dir,
				float* t)
		{
			Vector3D tvec;
			Vector3D pvec;
			Vector3D qvec;

			float u;
			float v;
			float det;

			// find vectors for two edges sharing first vertex
			Vector3D edge1 = b - a;
			Vector3D edge2 = c - a;

			// begin calculating determinant - also used to calculate U parameter
			pvec.cross(dir,edge2);

			// if determinant is near zero, ray lies in plane of triangle
			det = edge1 * pvec;
			if (det < MATRIX_EPSILON)
				return false;

			// calculate distance from vert0 to ray origin
			tvec = org - a;

			// calculate U parameter and test bounds
			u = tvec * pvec;
			if (u < 0.0f || u > det)
				return false;

			// prepare to test V parameter
			qvec.cross(tvec,edge1);

			// calculate V parameter and test bounds
			v = dir * qvec;
			if (v < 0.0f || u + v > det)
				return false;

			// calculate t, scale parameters, ray intersects triangle
			*t = (edge2 * qvec) / det;
			return true;
		}
	}
}
