

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef GLPLANELOADER_H
#define GLPLANELOADER_H

#include <adrcore/types/aarray.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/types/view.h>
#include <adrcore/types/camera.h>


#include "glcamera.h"

namespace adore
{
	namespace visual
	{
		class GLFeatureLoader
		{
			core::ViewBase* view;

		public:
			GLFeatureLoader(core::ViewBase* vb);
			void load(double* points);
			unsigned count();
		};

		class GLCameraLoader
		{
			core::Camera* camera;

		public:
			GLCameraLoader(core::Camera* camera);
			Vector3D location();
			Quaternion orientation();
			void load(GLCamera& glcamera);
			void load(Vector3D& eye, Vector3D& aim, Vector3D& up);
		};

		/**	Loads textures from several color planes stored as IplImage structures
		 *
		 *	1 color plane = grayscale image @n
		 *	2 color planes = grayscale image + alpha channel @n
		 *	3 color planes = RGB image @n
		 *	4 color planes = RGB image + alpha channel @n
		 *
		 *	@warning The loader assumes that all color planes have the same dimensions
		 *
		 *	 
		 */
		template<class T> class GLPlaneLoader
		{
		protected:
			typename core::Channel<T>::Set &planes;			///< Source color planes

		public:
			/**	Constructor
			 *	@param in Input color planes
			 */
			GLPlaneLoader(typename core::Channel<T>::Set& in)
				:planes(in)
			{
			};

			/**	Load image into the buffer
			 *	@param pixels Memory buffer
			 */
			void load(T* pixels)
			{
				unsigned width = planes[0]->getWidth();                   // original width
				unsigned height = planes[0]->getHeight();                 // original height
				unsigned isrc;
				unsigned idst;
				T temp;

				// Flatten color planes
				switch(planes.size())
				{
				case 1:     // grayscale image
					for(unsigned i = 0; i < height; ++i)
					{
						for(unsigned j = 0; j < width; ++j)
						{
							temp = *planes[0]->getPointer(j, i);
							idst = 4*((height-i-1)*width+j);
							pixels[idst] = temp;
							pixels[idst + 1] = temp;
							pixels[idst + 2] = temp;
							pixels[idst + 3] = 0xff;
						}
					}
					break;
				case 2:     // grayscale image + alpha chaneel
					for(unsigned i = 0; i < height; ++i)
					{
						for(unsigned j = 0; j < width; ++j)
						{
							isrc = planes[0]->getOffset(j, i);
							idst = 4*((height-i-1)*width+j);
							temp = *planes[1]->getPointer(isrc);
							pixels[idst] = temp;
							pixels[idst + 1] = temp;
							pixels[idst + 2] = temp;
							pixels[idst + 3] = *planes[0]->getPointer(isrc);
						}
					}
					break;
				case 3:     // RGB image
					for(unsigned i = 0; i < height; ++i)
					{
						for(unsigned j = 0; j < width; ++j)
						{
							isrc = planes[0]->getOffset(j, i);
							idst = 4*((height-i-1)*width+j);
							pixels[idst] = *planes[0]->getPointer(isrc);
							pixels[idst + 1] = *planes[1]->getPointer(isrc);
							pixels[idst + 2] = *planes[2]->getPointer(isrc);
							pixels[idst + 3] = 0xff;
						}
					}
					break;
				case 4:     // RGB image + alpha channel
					for(unsigned i = 0; i < height; ++i)
					{
						for(unsigned j = 0; j < width; ++j)
						{
							isrc = planes[0]->getOffset(j, i);
							idst = 4*((height-i-1)*width+j);
							pixels[idst] = *planes[0]->getPointer(isrc);
							pixels[idst + 1] = *planes[1]->getPointer(isrc);
							pixels[idst + 2] = *planes[2]->getPointer(isrc);
							pixels[idst + 3] = *planes[3]->getPointer(isrc);
						}
					}
					break;
				default:
					break;
				}
			};
		};
	}
}

#endif	// GLPLANELOADER_H
