

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_GLTEXTURE_H
#define ADORE_GLTEXTURE_H

#include "glrgba.h"

namespace adore
{
	namespace visual
	{
		/**	Describes OpenGL texturing mode
		 */
		enum TEXMODE
		{
			TEX_REPLACE,				///< replace material color with texture color
			TEX_MODULATE,				///< modulate material color with texture color
			TEX_DECAL,					///< blends between material and texture
			TEX_BLEND					///< blends between material and color argument
		};

		/**	OpenGL filtering mode
		 */
		enum FILTERMODE
		{
			FILTER_NEAREST,				///< no filtering
			FILTER_LINEAR,				///< bilinear filtering
			FILTER_NEAREST_MM_NEAREST,	///< nearest mipmapping without filtering
			FILTER_LINEAR_MM_NEAREST,	///< nearest mipmapping with bilinear filtering
			FILTER_NEAREST_MM_LINEAR, 	///< linear mipmapping without filtering
			FILTER_LINEAR_MM_LINEAR		///< trilinear filtering
		};

		/**	OpenGL texture holder
		 *
		 *	 
		 */
		class GLTexture
		{
			unsigned texid;				///< OpenGL texture id
			int envmode;				///< environment mode
			RGBA envcolor;				///< environment color argument
			int filter;					///< texture filtering
			int width;					///< texture width, must be power of 2
			int height;					///< texture width, must be power of 2
			RGBA* image;				///< texture image data
			mutable bool dirty;			///< texture requires update

		public:
			/**	Constructor
			 */
			GLTexture();

			/**	Destructor
			 */
			~GLTexture();

			/**	Create texture
			 *	@param w Texture width, must be power of 2
			 *	@param h Texture height, must be power of 2
			 */
			void create(int w, int h);

			/**	Set texture data
			 *	@param w Texture width, must be power of 2
			 *	@param h Texture height, must be power of 2
			 *	@param pixels Pointer to the pixel data to copy from
			 */
			void set(int w, int h, RGBA* pixels);

			/**	Lock the texture
			 *	@warning Any changes will not show up before unlocking
			 */
			RGBA* lock();

			/**	Unlock the texture
			 */
			void unlock();

			/**	Get texture width
			 */
			int getWidth() const;

			/**	Get texture height
			 */
			int getHeight() const;

			/**	Get texture environment mode
			 */
			int getMode() const;

			/**	Set texture environment mode
			 *	@param mode Mode as appears in TEXMODE enumeration
			 */
			void setMode(int mode);

			/**	Get color argument
			*/
			RGBA getColorArgument() const;

			/**	Set color argument
			 */
			void setColorArgument(RGBA arg);

			/**	Get horizontal repeat
			 */
			bool getHorizontalRepeat();

			/**	Set horizontal repeat
			 */
			void setHorizontalRepeat(bool rep);

			/**	Get vertical repeat
			 */
			bool getVerticalRepeat();

			/**	Set vertical repeat
			 */
			void setVerticalRepeat(bool rep);

			/** Get texture filtering mode
			 *	@return One of the FILTERMODE members which states current mode
			 */
			int getFilter();

			/**	Set texture filtering mode
			*	@param filt One of the FILTERMODE members which states mode to set
			*/
			void setFilter(int filt);

			/**	Update internal details
			 *	Load image in video memory, rebuild mipmaps
			 */
			void update() const;

			/**	Set this texture as current OpenGL texture
			 */
			void setOpenGLTexture() const;

			/**	Check, if the argument is a power of 2
			 */
			static bool isPower2(int x);

			/**	Get the closest power of 2 that is greater than x
			 *	@param x Argument, must be positive
			 */
			static int ceilPower2(int x);

			/**	Get the closest power of 2 that is less than x
			 *	@param x Argument, must be positive
			 */
			static int floorPower2(int x);

		private:
			/**	Copy constructor is closed
			 */
			GLTexture(const GLTexture& rv);

			/**	Assignment operator is closed
			 */
			GLTexture& operator=(const GLTexture& rv);
		};
	}
}

#endif	// ADORE_GLTEXTURE_H
