

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_GLTRANSFORM_H
#define ADORE_GLTRANSFORM_H

#include "glmath.h"
#include <cstdio>

namespace adore
{
	namespace visual
	{
		/**	Represents local coordinate transformation
		 *
		 *	The formula is W = R * (S * V) + T where @n
		 *	@b V - initial vertex @n
		 *	@b W - transformed vertex @n
		 *	@b S - scaling factor (Sx,Sy,Sz) @n
		 *	@b R - rotation matrix derived from the form of the quaternion @n
		 *	@b T - translation @n
		 *
		 *	 
		 */
		class GLTransform
		{
			Vector3D tr;			///< translation vector
			Vector3D scl;			///< scale factors
			Quaternion rot;			///< rotation (unit quaternion)
			Matrix4x4 total;		///< total transformation
			Matrix4x4 invert;		///< inverse of the total transformation
			bool dirty;				///< total transformation requires update

		public:
			/**	Constructor
			 */
			GLTransform();

			/**	Get Translation
			 */
			const Vector3D& getTranslation() const;

			/**	Get Scale
			 */
			const Vector3D& getScale() const;

			/**	Get Rotation
			 */
			const Quaternion& getRotation() const;

			/**	Set Translation
			 */
			void setTranslation(Vector3D t);

			/**	Set Scale
			 */
			void setScale(Vector3D s);

			/**	Set Rotation
			 */
			void setRotation(Quaternion r);

			/**	Load identity transform - S = (1,1,1), T=(0,0,0), R=(1,0,0,0)
			 */
			void identity();

			/**	Build total transform matrix
			 */
			void build();

			/**	Project Vector using this transform
			 */
			void projectVector(Vector4D& src, Vector4D& dst);

			/**	Unproject Vector using this transform
			 */
			void unprojectVector(Vector4D& src, Vector4D& dst);

			/**	Set OpenGL GL_MODELVIEW matrix
			 */
			void setOpengGLMatrix();
			
			const GLTransform &operator=(const GLTransform &t);
			
			void print()
			{
				printf("=== GLTransform\n");
				printf("trans: %f, %f, %f\n", tr[0], tr[1], tr[2]);
				printf("rot: %f, %f, %f, %f\n", rot[0], rot[1], rot[2], rot[3]);
				printf("scale: %f, %f, %f\n", scl[0], scl[1], scl[2]);
			}
		};
	}
}

#endif	// ADORE_GLTRANSFORM_H
