

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "rendernview.h"
#include <adrcore/parameterization/angleaxis.h>

namespace adore {
namespace visual {

RenderNView::RenderNView()
 : RenderItem(),
 displayCams(true),
 displayDots(true),
 centroidDots(3),
 centroidCams(3)
{
        lowColor[0] = 1.0;
        lowColor[1] = 1.0;
        lowColor[2] = 1.0;
        lowColor[3] = 1.0;
        
        highColor[0] = 0.5;
        highColor[1] = 0.5;
        highColor[2] = 0.5;
        highColor[3] = 0.5;
        
        dotScale = 1.0;
        
        lastHighCam = -1;
        lastHighDot = -1;
        
        createDotDL();
        createCamDL();
}


RenderNView::~RenderNView()
{
}

void RenderNView::createDotDL() {
    dotID = glGenLists(1);
    glNewList(dotID,GL_COMPILE);
    drawDot();
    glEndList();
}

void RenderNView::createCamDL() {
    camID = glGenLists(1);
    glNewList(camID,GL_COMPILE);
    drawCam();
    glEndList();
}

void RenderNView::drawDot() {
    float d = 1.0f / sqrt(2.0);
    glTranslatef(0.0f ,0.0f, 0.0f);
    glBegin(GL_TRIANGLE_FAN);
    glVertex3f( 0.0f,  0.0f, 0.0f);
    glVertex3f( 1.0f,  0.0f, 0.0f);
    glVertex3f(    d,     d, 0.0f);
    glVertex3f( 0.0f,  1.0f, 0.0f);
    glVertex3f(   -d,     d, 0.0f);
    glVertex3f(-1.0f,  0.0f, 0.0f);
    glVertex3f(   -d,    -d, 0.0f);
    glVertex3f( 0.0f, -1.0f, 0.0f);
    glVertex3f(    d,    -d, 0.0f);
    glVertex3f( 1.0f,  0.0f, 0.0f);
    glEnd();            
}

void RenderNView::drawCam() {
    glTranslatef(0.0f ,0.0f, 0.0f);
    glBegin(GL_TRIANGLE_FAN);
    glColor3f(1.0f,0.5f,0.0f);                        //orange
    glVertex3f(  0.0f,  0.0f, 0.0f);
    glColor3f(0.0f,0.0f,1.0f);                        //blue
    glVertex3f(1.0f, 1.0f, 2.5f);
    glVertex3f(-1.0f, 1.0f, 2.5f);
    glVertex3f(-1.0f, -1.0f, 2.5f);
    glVertex3f(1.0f, -1.0f, 2.5f);
    glVertex3f(1.0f, 1.0f, 2.5f);
    glEnd();
}

void RenderNView::render()
{
    renderSelect(0);
}

void RenderNView::renderSelect(GLuint i)
{
    if(displayDots)
    {
//         DBGOUT(DBGNORMAL, "RenderNView: rendering %d dots\n", points.size());
        for(vector<glPoint>::iterator it = points.begin(); it != points.end(); ++it, ++i)
        {
            if((*it).disabled) continue;
            glPushName(i);
            
            glPushMatrix();
            glColor4f((*it).r, (*it).g, (*it).b, (*it).a);
            glTranslatef((*it).x, (*it).y, (*it).z);
            fakeBillboardBegin();
            glScalef(dotScale, dotScale, 1.0);
            glCallList(dotID);
            fakeBillboardEnd();
            glPopMatrix();
            
            glPopName();
        }
    }
        
    if(displayCams)
    {
//         DBGOUT(DBGNORMAL, "RenderNView: rendering %d cams\n", cams.size());
        for(vector<glCamera>::iterator it = cams.begin(); it != cams.end(); ++it, ++i)
        {
//             DBGOUT(DBGNORMAL, "RenderNView: rendering cam at position (%f, %f, %f)\n", (*it).x, (*it).y, (*it).z);
            
            if((*it).disabled) continue;
            glPushName(i);
            glPushMatrix();
            glTranslatef((*it).x, (*it).y, (*it).z);
            glRotated((*it).angle, (*it).rx, (*it).ry, (*it).rz);
            glScalef(dotScale, dotScale,dotScale);
            glCallList(camID);
            glPopMatrix();
            glPopName();
        }
    }
}

void RenderNView::attach(NView*nv) 
{
    DBGOUT(DBGNORMAL, "RenderNView: NView Attached - (pt: %d, cam: %d)\n", nv->getNumPoints3D(), nv->getNumCameras());
    
    drawPoints(nv->getPoints3DBegin(), nv->getNumPoints3D());
    drawCams(nv->getCamerasBegin(), nv->getNumCameras());
    
    DBGOUT(DBGNORMAL, "RenderNView: (pt: %zd, cam: %zd) drawn\n", this->points.size(), this->cams.size());
}


void RenderNView::drawPoints(NView::Point3DIterator it, uint size)
{
    if(size < 1) return;
    points.resize(size);
    
    double sumX = 0.0;
    double sumY = 0.0;
    double sumZ = 0.0;
    int nPoints = 0;
    
    for(uint i = 0; i < size; ++i, ++it)
    {
        AVector pt = **it;
        
        if(pt(3) != 0)
        {
            points[i].x = pt(0) / pt(3);
            points[i].y = pt(1) / pt(3);
            points[i].z = pt(2) / pt(3);
            sumX += points[i].x;
            sumY += points[i].y;
            sumZ += points[i].z;
            points[i].setColor(lowColor);
            points[i].disabled = false;
            points[i].point = *it;
            ++nPoints;
//             printf("3D-Point: (%f, %f, %f)\n", points[i].x, points[i].y, points[i].z);
        }
        else
        {
            points[i].disabled = true;
        }
    }
    
    sumX /= double(nPoints);
    sumY /= double(nPoints);
    sumZ /= double(nPoints);
    
    //setting lookat
    centroidDots(0) = sumX;
    centroidDots(1) = sumY;
    centroidDots(2) = sumZ;

}

void RenderNView::drawCams(NView::CameraIterator it, uint size)
{
    if(size < 1) return;
    cams.resize(size);
    
    double sumX = 0.0;
    double sumY = 0.0;
    double sumZ = 0.0;
    int nPoints = 0;
    
    //put cameras in the appropriate list
    DBGOUT(DBGNORMAL, "cams size is %zd\n", cams.size());
    adore::core::AngleAxis anax;
    for(uint i = 0; i < size; ++i, ++it)
    {
        Camera *c = *it;
        c->decompose();
        
        if(!isRightHand(c->getRotationMatrix()))
        {
            DBGOUT(DBGNORMAL, "not right handed camera\n");
            cams[i].disabled = true;
            continue;
        }
        
        cams[i].disabled = false;
        
        AVector3D loc = c->getLocation();
        
        cams[i].x = loc(0);
        cams[i].y = loc(1);
        cams[i].z = loc(2);
        AVector2D axis = anax.matToAngleAxis(c->getRotationMatrix(), cams[i].angle);
        cams[i].angle *= -180.0 / M_PI;
        cams[i].rx = axis(0);
        cams[i].ry = axis(1);
        cams[i].rz = axis(2);
        
        cams[i].setColor(lowColor);
        cams[i].cam = c;
        
    }
    
    sumX /= double(nPoints);
    sumY /= double(nPoints);
    sumZ /= double(nPoints);
    
    //setting lookat
    centroidCams(0) = sumX;
    centroidCams(1) = sumY;
    centroidCams(2) = sumZ;
}

void RenderNView::fakeBillboardBegin()
{
    float modelview[16];
    
    // save the current modelview matrix
    glPushMatrix();

    // get the current modelview matrix
    glGetFloatv(GL_MODELVIEW_MATRIX , modelview);

    // undo all rotations
    // beware all scaling is lost as well 
    modelview[ 0] = 1.0;
    modelview[ 5] = 1.0;
    modelview[10] = 1.0;
    modelview[1] = modelview[2] = modelview[4] = 0.0;
    modelview[6] = modelview[8] = modelview[9] = 0.0;

    // set the modelview with no rotations
    glLoadMatrixf(modelview);
}

void RenderNView::fakeBillboardEnd()
{
    // restore the previously 
    // stored modelview matrix
    glPopMatrix();
}

}}
