

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "objectselectdialog.h"
#include "controller.h"
#include "mainwindow.h"
#include "inspector.h"
#include "objectselectsavedialog.h"
#include "featuresview.h"
#include "glwidget.h"
#include "nviewlist.h"

#include <adrcore/persistence/persist.h>
#include <adrcore/persistence/persist_boost.h>

#include <adrcore/util/clock.h>
#include <adrvisual/glvisualizerdots.h>
#include <adrvisual/glnviewrenderer.h>

#include <adrcore/nview/normalizer.h>
#include <adrcore/nview/modifier.h>

#include "calcthread.h"
#include "progressdialog.h"

using namespace adore::core;

Controller::Controller( MainWindow* w, const string& configFile )
		:	mainwin( w ),
		fView( mainwin->getFeaturesView() ),
		glwidget( mainwin->getGLWidget() ),
		nViewSetList( mainwin->getSetsView() ),
		lastSelected( 0 ),
		thread( 0 ),
		delete3DPoints( false )
{
	mainwin->setController( this );

	glwidget->setListener( this );
	glwidget->set_GL_Config( configFile.c_str() );
	connect(glwidget, SIGNAL(standardViewRequested()), this, SLOT(toStandardCamera()));
	connect( glwidget, SIGNAL( deletePoint3D( bool ) ),
			 this, SLOT( deletePoint3D( bool ) ) );

	connect( fView, SIGNAL( sigViewSelected( const ViewT* const, bool ) ),
	         this, SLOT( slotViewSelected( const ViewT* const, bool ) ) );
	connect( fView, SIGNAL( sigCameraSelected( const ViewT* const ) ),
	         this, SLOT( slotCameraSelected( const ViewT* const ) ) );

	connect( nViewSetList, SIGNAL( sigItemSelected( NView* ) ),
	         this, SLOT( nViewSelected( NView* ) ) );
	connect( nViewSetList, SIGNAL( sigItemDoubleClicked( NView* ) ),
	         this, SLOT( nViewDoubleClicked( NView* ) ) );
}

void Controller::addViews( const QStringList& fileList, const QDir& dir )
{
	resetAll();

	QStringList::const_iterator ii = fileList.constBegin();
	QStringList::const_iterator end = fileList.constEnd();
	for ( ; ii != end; ++ii )
	{
		QString file = dir.canonicalPath() + QDir::separator() + ( *ii );
		QByteArray ascii( file.toLatin1() );
		char* data = ascii.data();
		ImagePlane<uchar>* img;
		try { img = new ImagePlane<uchar>( data ); }
		catch ( ... ) { img = 0; printf( "Controller: problem with image io\n" ); }
		if ( !img ) printf( "Image %s not found\n", data );
		else
		{
			ViewT* v = new ViewT( img, true );
			v->setImagePath( string( data ) );
			v->initialize();
			fView->addView( v );
			nviewmap.addView( v );
		}
	}
}

void Controller::addViews( NViewMap& nviewmap )
{
	for ( NViewMap::ViewIterator it = nviewmap.getAllViewsBegin(); it != nviewmap.getAllViewsEnd(); ++it )
	{
		ViewT *v = ( ViewT * ) *it;
		fView->addView( v );
		nviewmap.addView( v );
	}
}

void Controller::resetMap()
{
	nViewSetList->clear();
	nviewmap.setViewResetMode( FORGET_VIEW | RESET_VIEW );
	int mode = nviewmap.getMasterNView().getViewResetMode();
	nviewmap.getMasterNView().setViewResetMode( mode & ~FORGET_VIEW );
	nviewmap.reset();
	nviewmap.getMasterNView().setViewResetMode( mode );
	lastSelected = 0;
}

void Controller::resetViews()
{
	int oldVMode = nviewmap.getViewResetMode();
	int oldNMode = nviewmap.getNViewResetMode();
	nviewmap.setViewResetMode( DELETE_VIEW );
	nviewmap.setNViewResetMode( KEEP_NVIEW );
	nviewmap.reset();
	nviewmap.setViewResetMode( oldVMode );
	nviewmap.setNViewResetMode( oldNMode );
	mainwin->getFeaturesView()->reset();
}

void Controller::to3D( bool multithreaded )
{
	vector<ViewT *> views;
	fView->getSelectedViews( views );
	batch::Config cfg( mainwin->getConfig() );
	thread = new CalcThread( nviewmap, views, cfg, multithreaded );
	connect( thread, SIGNAL( finished() ), this, SLOT( slotThreadFinished() ) );

	resetAllButViews();
	bool nondemo = MainWindow::NonDemo == mainwin->getMode();
	ProgressDialog dlg( thread, mainwin, nondemo );
	dlg.exec();
}

void Controller::slotThreadFinished()
{
	QMutexLocker locker( &mutex );

	CalcThread* calc_thread = qobject_cast<CalcThread*>( sender() );
	if ( !calc_thread ) return;
	if ( calc_thread != thread ) return;

	if ( calc_thread->stopped() )
	{
		delete thread;
		thread = 0;
		return;
	}

	const NViewMap& nviewmap_ref( thread->getNViewMap() );
	int cnt = nviewmap_ref.getNumEntries();
	if ( !cnt )
	{
		NViewMap::ViewIterator ii = nviewmap_ref.getAllViewsBegin();
		while ( ii != nviewmap_ref.getAllViewsEnd() )
		{
			nviewmap.addView( *ii );
			++ii;
		}
	}
	NViewMap::EntryIterator nvit = nviewmap_ref.getEntriesBegin();
	while ( nvit != nviewmap_ref.getEntriesEnd() )
	{
		NView* nv = *nvit;
		nviewmap.setNView( nv->getViewsBegin(), nv->getViewsEnd(), nv );
		++nvit;
	}

	delete thread;
	thread = 0;

	fView->featuresChanged();
	fillNViewSetTreeView( nviewmap );
}

void Controller::fillNViewSetTreeView( NViewMap &nviewmap )
{
	printf( "--- filling tree view ---\n" );
	typedef NViewDependencyGraph::RootIter root_iter;
	typedef NViewDependencyGraph::AdjacencyIter adj_iter;

	class _
	{
			NViewList* l;
		public: _( NViewList* _l ) : l( _l ) { l->setUpdatesEnabled( false ); }
			~_() { l->setUpdatesEnabled( true ); }
	};
	_ __( nViewSetList ); // enable if going out of scope

	nViewSetList->clear();

	NViewDependencyGraph nvdg( nviewmap );
	cout << nvdg.toString() << endl;
	printf( "num nviews: %d, numRoots in graph: %d\n", nviewmap.getNumEntries(), nvdg.numRoots() );
	if ( nvdg.numRoots() == 0 ) return;

	lastSelected = 0;
	for ( root_iter rit = nvdg.getRootsBegin(); rit != nvdg.getRootsEnd(); ++rit )
	{
		NView *nv = nvdg.getNView( *rit );
		QString setString = nv->isNormalized() ? "(+N)" : "(-N)";
		lastSelected = nViewSetList->newRow( setString, nv );

		NView* last = 0;
		for ( adj_iter it = nvdg.getAdjacentBegin( *rit ); it != nvdg.getAdjacentEnd( *rit ); ++it )
		{
			last = fillNViewSetTreeView( nvdg, *it, lastSelected, last );
		}
	}

	nViewSetList->update();
	nViewSetList->expandAll();
	lastSelected = nViewSetList->selectFirst();
}

NView* Controller::fillNViewSetTreeView( NViewDependencyGraph &nvdg, NViewDependencyGraph::Vertex nv, NView* parent, NView* below )
{
	typedef NViewDependencyGraph::AdjacencyIter adj_iter;
	NView *nview = nvdg.getNView( nv );

	NView* elem = 0;
	if ( below ) elem = nViewSetList->newRow( parent, below, "set", nview );
	else elem = nViewSetList->newRow( parent, "set", nview );

	NView* last = 0;
	for ( adj_iter it = nvdg.getAdjacentBegin( nv ); it != nvdg.getAdjacentEnd( nv ); ++it )
	{
		last = fillNViewSetTreeView( nvdg, *it, elem, last );
	}
	return last;
}

void Controller::point3DChosen( AVector3D *pt )
{
	if ( pt == 0 ) return;
	if ( lastSelected == 0 ) return;

	if ( delete3DPoints )
	{
		PointCorr *pc = lastSelected->getPointCorr ( pt );
		if ( !pc ) return;
		lastSelected->removePointCorrAnd3D ( pc );
		if ( glwidget ) glwidget->nViewSelected( lastSelected, fView->getFocusedView(), false/*do not reset view transformation*/ );
	}
	else
	{
		PointCorr* corr = lastSelected->getPointCorr( pt );
		fView->highlightPointCorr( corr );
	}
}

void Controller::cameraChosen( Camera *cam )
{
	if ( lastSelected == 0 ) return;
	if ( cam == 0 ) return;
	fView->highlightView( lastSelected->getView( cam ) );
}

void Controller::nViewSelected( NView* nview )
{
	if ( !nview )
	{
		printf( "NULL-object clicked\n" );
		lastSelected = 0;
		return;
	}

	lastSelected = nview;
	fView->nViewSelected( lastSelected );
	if ( glwidget ) glwidget->nViewSelected( lastSelected, fView->getFocusedView() );
}

void Controller::nViewDoubleClicked( NView* nview )
{
	if ( !nview )
	{
		printf( "NULL-object double-clicked\n" );
		return;
	}
	Inspector* ins = new Inspector( nview, mainwin );
	ins->show();
}

void Controller::save( const QString& s, bool simple, bool embed, bool descr )
{

	/// TODO: remove it! only test
// 	NViewMap::EntryIterator nit = nviewmap.getEntriesBegin();
// 	while(nit != nviewmap.getEntriesEnd()) {
// 		NView* nv = *nit;
// 		NView::Point3DIterator pit = nv->getPoints3DBegin();
// 		while(pit != nv->getPoints3DEnd()) {
// 			nv->removePointCorrAnd3D(nv->getPointCorr(*pit));
// 			pit = nv->getPoints3DBegin();
// 		}
// 		++nit;
// 	}

	using namespace persist;
	Persist<PersistBoost> database( s.toStdString() );
	vector<string> names( database.objectNames() );
	string obj_name( default_obj_name );

	if ( !simple )
	{
		ObjectSelectSaveDialog objdia;
		objdia.setNames( names );
		objdia.unselect();
		if ( objdia.exec() == QDialog::Rejected ) return;
		if ( !objdia.getLineEdit().empty() )
			obj_name = objdia.getLineEdit();
		else if ( !objdia.getSelected().empty() )
			obj_name = objdia.getSelected();
	}

	database.saveNViewMap( nviewmap, obj_name, simple, Options( embed, descr ) );
}

void Controller::load( const QString& s )
{
	using namespace persist;
	Persist<PersistBoost> database( s.toStdString() );
	vector<string> names( database.objectNames() );
	string obj_name( default_obj_name );
	if ( !names.empty() )
	{
		ObjectSelectDialog objdia;
		int width = qMax( objdia.width(), mainwin->width() );
		objdia.resize( width / 2, objdia.height() );
		objdia.setNames( names );
		if ( objdia.exec() == QDialog::Rejected ) return;
		obj_name = objdia.getSelected();
	}

	resetViews(); resetMap();
	bool ok = database.loadNViewMap( nviewmap, obj_name );
	if ( !ok ) { resetMap(); return; }

	fView->setUpdatesEnabled( false );
	glwidget->setUpdatesEnabled( false );
	addViews( nviewmap );
	fView->setUpdatesEnabled( true );
	glwidget->setUpdatesEnabled( true );
	fillNViewSetTreeView( nviewmap );
}

void Controller::setDeviation( int ticks )
{
	if ( !glwidget ) return;
	glwidget->setDeviation(ticks);
}

void Controller::setPointSize( int ticks )
{
	if ( !glwidget ) return;
	glwidget->setPointSize(ticks);
}

void Controller::slotCameraSelected( const ViewT * const v )
{
	if ( !v || !glwidget ) return;
	if ( !lastSelected || 0 == lastSelected->getNumCameras() )
		return;
	Camera* cam = lastSelected->getCamera( const_cast<ViewT*>( v ), false );
	if ( 0 == cam ) return;

	GLWidget::vis_type vis_t = glwidget->visualizerType();
	if ( GLWidget::vis_delaunay != vis_t ) return;
	GLNViewRenderer<imgTypeT>* vis = 0;
	vis = dynamic_cast<GLNViewRenderer<imgTypeT>*>( glwidget->visualizer() );
	if ( !vis ) return;

	vis->select( v, true/*reset the view trafo*/ );
	glwidget->updateGL();
	fView->focusView( v );
}

void Controller::slotViewSelected( const ViewT * const v, bool selected )
{
	if ( !v ) return;
	fView->selectView( v, !selected );
}

void Controller::toStandardCamera()
{
	if ( !lastSelected || 0 == lastSelected->getNumCameras() )
		return;
	NView::CameraIterator ii = lastSelected->getCamerasBegin();
	while ( ii != lastSelected->getCamerasEnd() )
	{
		Camera* cam = *ii;
		cam->decompose();
		AMatrix rotmat( cam->getRotationMatrix() );
		if ( isRightHand( rotmat ) )
		{
			slotCameraSelected( static_cast<const ViewT*>( lastSelected->getView( cam ) ) );
			return;
		}
		++ii;
	}
}

void Controller::resetAll()
{
	mainwin->getFeaturesView()->reset();
	resetGL();
	resetMap();
	resetViews();
}

void Controller::resetHighlights()
{
	fView->highlightView( 0 );
	fView->focusView( 0 );
}

void Controller::resetGL()
{
	if ( glwidget ) glwidget->nViewSelected( 0 );
	lastSelected = 0;
}


void Controller::resetAllButViews()
{
	//resetMap resets the views, but that does not touch the normalization
	//to be on the save side, we force a denormalization of all views
	nviewmap.getMasterNView().normalize();
	nviewmap.getMasterNView().denormalize();

	resetHighlights();
	resetGL();
	resetMap();

	fView->resetPreservingViews();
}

void Controller::deletePoint3D( bool d )
{
	delete3DPoints = d;
}
