

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include <QtGui/QColor>

#include "corrgl2d.h"
#include "pixmapplane.h"
#include "qglhelper.h"

#include <adrcore/types/pointcorr.h>

#include <map>

using namespace std;
using namespace adore::core;

CorrGL2D::CorrGL2D( PointCorr * c, const std::vector<ImagePtr>& images ) :
		corr( c )
{
	createContent( images );
}

CorrGL2D::~ CorrGL2D()
{
}

void CorrGL2D::createContent( const std::vector<ImagePtr>& images )
{
	if ( corr->getSize() < 1 )
		return;

	//sort the points according to the order of the views
	vector<FeaturePoint*> orderedFP;
	uint cnt = images.size();
	for ( uint i = 0; i < cnt; i++ )
	{
		ViewBase* v = images[i]->getView();
		PointCorr::Iterator end = corr->getEnd();
		for ( PointCorr::Iterator pit = corr->getBegin(); pit != end; ++pit )
		{
			if ( ( *pit )->getView() == v )
				orderedFP.push_back( *pit );
		}
	}

	class _lineStripGL : public Item2d
	{
		public:
			_lineStripGL( const QColor& col, const vector<QPointF>& pts ) :
					r( ( float )col.red() / 255 ), g( ( float )col.green() / 255 ),
					b( ( float )col.blue() / 255 ), points( pts )
			{
				if ( 2 > points.size() ) return;
				rect = QRectF( points.at( 0 ), points.at( 1 ) ).normalized();
				for ( uint i = 2 ; i < points.size(); ++i )
					rect |= QRectF( points.at( i - 1 ), points.at( i ) ).normalized();
			}

			virtual void paint() const
			{
				if ( 2 > points.size() ) return;
				glColor3f( r, g, b );
				for ( uint i = 0 ; i < points.size() - 1; ++i )
					drawLine( points.at( i ), points.at( i + 1 ) );
			}
		private:
			float r, g, b;
			QRectF rect;
			vector<QPointF> points;

			void drawLine( const QPointF & p0, const QPointF & p1 ) const
			{
				if(!QGLHelper::isMultisamplingAvailable()) {
					boost::shared_ptr<GLEnabler> enabler(new GLEnabler(GL_BLEND));
					glBlendFunc(GL_SRC_ALPHA,  GL_ONE_MINUS_SRC_ALPHA);
					glColor4d(0.0, 1.0, 0.0, 1.0);
					QGLHelper::aaGlLineWidth(3.0);
					QGLHelper::aaGlBeginLines();
					QGLHelper::aaGlVertex2(p0.x(), p0.y());
					QGLHelper::aaGlVertex2(p1.x(), p1.y());
					QGLHelper::aaGlEndLines();
					return;
				}
				double thOver2 = 0.5;
				glBegin( GL_QUADS );
				glVertex3d( p0.x() - thOver2, p0.y() - thOver2, 0.7 );
				glVertex3d( p1.x() + thOver2, p1.y() - thOver2, 0.7 );
				glVertex3d( p1.x() + thOver2, p1.y() + thOver2, 0.7 );
				glVertex3d( p0.x() - thOver2, p0.y() + thOver2, 0.7 );
				glEnd();
			}
	};

	class _6AngleGL : public Item2d
	{
		public:
			_6AngleGL( const QColor& col ) :
					radius( 3.0 ),
					r( ( float )col.red() / 255 ), g( ( float )col.green() / 255 ),
					b( ( float )col.blue() / 255 )
			{}
			virtual QRectF boundingRect() const
			{
				return QRectF( 0, 0, radius, radius );
			}
			virtual void paint() const
			{
				static float x1 = radius * 0.500000000000000111;
				static float y1 = radius * 0.8660254037844385966;
				glColor3f( r, g, b );
				glPushMatrix();
				glTranslatef( x(), y(), 0.8 );
				glBegin( GL_POLYGON );
				glVertex3f( radius, 0.0, 0.0 );
				glVertex3f( x1, y1, 0.0 );
				glVertex3f( -x1, y1, 0.0 );
				glVertex3f( -radius, 0.0, 0.0 );
				glVertex3f( -x1, -y1, 0.0 );
				glVertex3f( x1, -y1, 0.0 );
				glEnd();
				glPopMatrix();
			}
		private:
			const float radius;
			float r, g, b;
	};

	struct ViewData
	{
		float scale;
		float xOff;
		float yOff;
	};

	ViewData data[cnt];
	std::map<ViewBase*, void*> view2data;
	for ( uint i = 0; i < cnt; ++i )
	{
		ViewBase* view = images[i]->getView();
		assert( view );
		if ( view->isNormalized() )
		{
			AMatrix trafo = view->getInvTrafo();
			data[i].scale = ( float )trafo( 0, 0 );
			data[i].xOff  = ( float )trafo( 0, 2 ) + images[i]->x();
			data[i].yOff  = ( float )trafo( 1, 2 ) + images[i]->y();
		}
		else
		{
			data[i].scale = ( float )1.0;
			data[i].xOff = images[i]->x();
			data[i].yOff = images[i]->y();
		}
		view2data[view] = &data[i];
	}

	uint num_pts = orderedFP.size();
	vector<QPointF> points;
	for ( uint i = 0; i < num_pts; ++i )
	{
		Child item( ( Item2d* )( new _6AngleGL( i % 2 ? Qt::white : Qt::blue ) ) );
		FeaturePoint* pt = orderedFP[i];
		float x = ( *pt )( 0 ) / ( *pt )( 2 );
		float y = ( *pt )( 1 ) / ( *pt )( 2 );
		ViewData* d = ( ViewData* )view2data[pt->getView()];
		x = d->scale * x + d->xOff;
		y = d->scale * y + d->yOff;
		item->setPosition( QPointF( x, y ) );
		children.push_back( item );
		points.push_back( QPointF( x, y ) );
	}
	Child item( ( Item2d* )( new _lineStripGL( Qt::green, points ) ) );
	children.push_back( item );
}

void CorrGL2D::paint() const
{
	vector<Child>::const_iterator it = children.begin();
	for ( ; it != children.end(); ++it )
		( *it )->paint();
}

