

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "corrlayeritem.h"
#include "pixmapplane.h"
#include "qglhelper.h"

#include <cassert>
#include <adrcore/types/nview.h>
#include <adrcore/types/pointcorr.h>

using namespace adore::core;
using namespace std;

CorrLayerItem::CorrLayerItem( NView * nv, const std::vector<ImagePtr>& images ) :
		nview( nv ),
		pointCorrsList( 0 ),
		lineCorrsList( 0 ),
		pCorrsVisible( true ),
		lCorrsVisible( true )
{
	createContent( images );
}

CorrLayerItem::~ CorrLayerItem()
{
	if ( pointCorrsList )
	{
		glDeleteLists( pointCorrsList, 1 );
	}
	if ( lineCorrsList )
	{
		glDeleteLists( lineCorrsList, 1 );
	}
}

void CorrLayerItem::paint() const
{
	boost::shared_ptr<GLEnabler> enabler;
	if(!QGLHelper::isMultisamplingAvailable()) {
		glDisable(GL_DEPTH_TEST);
		enabler.reset(new GLEnabler(GL_BLEND));
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	glColor4d(1.0, 0.0, 0.0, 1.0);
	if ( pointCorrsList && pCorrsVisible )
	{
		glCallList( pointCorrsList );
	}
	if ( lineCorrsList && lCorrsVisible )
	{
		glCallList( lineCorrsList );
	}
}

void CorrLayerItem::createContent( const std::vector<ImagePtr>& images )
{
	uint cnt = images.size();
	if ( 2 > cnt )
		return;

	struct ViewData
	{
		float scale;
		float xOff;
		float yOff;
	};
	ViewData data[cnt];
	map<ViewBase*, void*> offsets;
	for ( uint i = 0; i < cnt; ++i )
	{
		ViewBase* view = images[i]->getView();
		assert( view );
		if ( view->isNormalized() )
		{
			AMatrix trafo = view->getInvTrafo();
			data[i].scale = ( float )trafo( 0, 0 );
			data[i].xOff  = ( float )trafo( 0, 2 ) + images[i]->x();
			data[i].yOff  = ( float )trafo( 1, 2 ) + images[i]->y();
		}
		else
		{
			data[i].scale = 1.0f;
			data[i].xOff = images[i]->x();
			data[i].yOff = images[i]->y();
		}
		offsets[view] = &data[i];
	}

	pointCorrsList = glGenLists( 1 );
	glNewList( pointCorrsList, GL_COMPILE );

	NView::CorrIterator it = nview->getPointCorrsBegin();
	NView::CorrIterator end = nview->getPointCorrsEnd();
	while ( it != end )
	{
		FeaturePoint* pt = 0;
		double linePt1[2] = {-1, -1};
		double linePt2[2] = {-1, -1};

		//sort the points according to the order of the views
		vector<FeaturePoint*> orderedFP;
		for ( uint i = 0; i < cnt; i++ )
		{
			ViewBase* v = images[i]->getView();
			PointCorr::Iterator end = ( *it )->getEnd();
			for ( PointCorr::Iterator pit = ( *it )->getBegin(); pit != end; ++pit )
			{
				if ( ( *pit )->getView() == v )
					orderedFP.push_back( *pit );
			}
		}

		uint num = orderedFP.size();
		for ( uint i = 0; i < num; i++ )
		{
			pt = orderedFP[i];
			if ( pt )
			{
				ViewData* d( static_cast<ViewData*>( offsets[pt->getView()] ) );
				float x = ( *pt )( 0 ) / ( *pt )( 2 );
				float y = ( *pt )( 1 ) / ( *pt )( 2 );
				linePt2[0] = d->scale * x + d->xOff;
				linePt2[1] = d->scale * y + d->yOff;

				if ( linePt1[0] != -1 && linePt1[1] != -1 )
				{
					drawLineGl( QPointF( linePt1[0], linePt1[1] ), QPointF( linePt2[0], linePt2[1] ), 1.0 );
				}
				linePt1[0] = linePt2[0];
				linePt1[1] = linePt2[1];
			}
		}

		++it;
	}
	glEndList();

	lineCorrsList = glGenLists( 1 );
	glNewList( lineCorrsList, GL_COMPILE );

	glTranslated( 0.0, 0.0, 0.4 );
	NView::LineCorrIterator lit = nview->getLineCorrsBegin();
	NView::LineCorrIterator lend = nview->getLineCorrsEnd();
	while ( lit != lend )
	{
		FeatureLine* l = 0;
		double lineL1[2] = {-1, -1};
		double lineL2[2] = {-1, -1};

		//sort the lines according to the order of the views
		vector<FeatureLine*> orderedFL;
		for ( uint i = 0; i < cnt; i++ )
		{
			ViewBase* v = images[i]->getView();
			LineCorr::Iterator flend = ( *lit )->getEnd();
			for ( LineCorr::Iterator flit = ( *lit )->getBegin(); flit != flend; ++flit )
			{
				if ( ( *flit )->getView() == v )
					orderedFL.push_back( *flit );
			}
		}

		uint num = orderedFL.size();
		for ( uint i = 0; i < num; i++ )
		{
			l = orderedFL[i];
			if ( l )
			{
				ViewData* d( static_cast<ViewData*>( offsets[l->getView()] ) );
				const AVector& pt1 = l->getPt1();
				const AVector& pt2 = l->getPt2();
				double x = 0.5 * ( pt1( 0 ) + pt2( 0 ) );
				double y = 0.5 * ( pt1( 1 ) + pt2( 1 ) );
				lineL2[0] = d->scale * x + d->xOff;
				lineL2[1] = d->scale * y + d->yOff;

				if ( lineL1[0] != -1 && lineL1[1] != -1 )
				{
					glColor3f( 1.0, 1.0, 1.0 );
					drawLineGl( QPointF( lineL1[0], lineL1[1] ) - QPointF( 1, 1 ),
					            QPointF( lineL1[0], lineL1[1] ) + QPointF( 1, 1 ), 2.0 );
					drawLineGl( QPointF( lineL2[0], lineL2[1] ) - QPointF( 1, 1 ),
					            QPointF( lineL2[0], lineL2[1] ) + QPointF( 1, 1 ), 2.0 );
					glColor3f( 0.0, 1.0, 1.0 );
					drawLineGl( QPointF( lineL1[0], lineL1[1] ), QPointF( lineL2[0], lineL2[1] ), 1.0 );
				}
				lineL1[0] = lineL2[0];
				lineL1[1] = lineL2[1];
			}
		}

		++lit;
	}
	glTranslated( 0.0, 0.0, -0.4 );

	glEndList();
}


void CorrLayerItem::drawLineGl( const QPointF & p0, const QPointF & p1, double thickness )
{
	if(!QGLHelper::isMultisamplingAvailable()) {
		QGLHelper::aaGlLineWidth(thickness);
		QGLHelper::aaGlBeginLines();
		QGLHelper::aaGlVertex2(p0.x(), p0.y());
		QGLHelper::aaGlVertex2(p1.x(), p1.y());
		QGLHelper::aaGlEndLines();
		return;
	}
	double thOver2 = thickness * 0.5;
	glBegin( GL_QUADS );
	glVertex3d( p0.x() - thOver2, p0.y() - thOver2, 0.0 );
	glVertex3d( p1.x() + thOver2, p1.y() - thOver2, 0.0 );
	glVertex3d( p1.x() + thOver2, p1.y() + thOver2, 0.0 );
	glVertex3d( p0.x() - thOver2, p0.y() + thOver2, 0.0 );
	glEnd();
}
