

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include <QGLWidget>

#include <cassert>
#include <map>

#include "featureslayer.h"
#include "qglhelper.h"

using namespace adore::core;

FeaturesLayer::FeaturesLayer( ViewT * v, bool listm )
		: view( v ),
		listMode( listm ),
		feature( 0 ),
		featList( 0 ),
		lineList( 0 ),
		featuresVisible( true ),
		linesVisible( true )
{
	assert( view );

	double width = v->getWidth();
	double height = v->getHeight();
	bool norm = v->getNormalized();
	if ( norm )
	{
		double scale = v->getInvTrafo()( 0, 0 );
		width = ceil( scale * v->getWidth() );
		height = ceil( scale * v->getHeight() );
	}
	size = QSizeF( width, height );
	updateFeatures();
}

FeaturesLayer::~ FeaturesLayer()
{
	if ( listMode )
	{
		if ( featList )
			glDeleteLists( featList, 1 );
		if ( lineList )
			glDeleteLists( lineList, 1 );
	}
	if ( feature )
		glDeleteLists( feature, 1 );
}

void FeaturesLayer::updateFeatures()
{
	if ( !feature )
		feature = glMakeFeatureList();
	if ( listMode )
	{
		if ( featList )
			glDeleteLists( featList, 1 );
		featList = glGenLists( 1 );
		if ( lineList )
			glDeleteLists( lineList, 1 );
		lineList = glGenLists( 1 );
	}

	float scale, xOff, yOff;
	ViewBase* v = view->getBase();
	if ( v->isNormalized() )
	{
		adore::core::AMatrix trafo = view->getInvTrafo();
		scale = trafo( 0, 0 );
		xOff  = trafo( 0, 2 );
		yOff  = trafo( 1, 2 );
	}
	else
	{
		scale = 1.0;
		xOff = 0;
		yOff = 0;
	}

	ViewBase::PointsIterator pit = view->getPointsBegin();
	ViewBase::PointsIterator end = view->getPointsEnd();
	float x_old = 0.0f, y_old = 0.0f;

	if ( listMode )
		glNewList( featList, GL_COMPILE );
	else
	{
		offsets.clear();
		offsets.reserve( view->getNumAllFeaturePoints() );
	}

	for ( ; pit != end; ++pit )
	{
		FeaturePoint* fp = *pit;
		float _x = ( *fp )( 0 ) / ( *fp )( 2 );
		float _y = ( *fp )( 1 ) / ( *fp )( 2 );
		float x = scale * _x + xOff;
		float y = scale * _y + yOff;
		if ( listMode )
		{
			glTranslatef( x - x_old, y - y_old, 0.0 );
			glCallList( feature );
		}
		else
			offsets.push_back( xy( x - x_old, y - y_old ) );
		x_old = x, y_old = y;
	}
	if ( listMode )
	{
		glTranslatef( -x_old, -y_old, 0.0f );
		glEndList(); //featList
	}

	if ( listMode )
	{
		glNewList( lineList, GL_COMPILE );
		size_t numLines = v->getNumSegments();
		for ( size_t i = 0; i < numLines; ++i )
		{
			const ViewBase::segmentT& s = v->getSegment( i );
			const AVector& pt1 = s->getPt1();
			const AVector& pt2 = s->getPt2();
			float x1 = scale * ( pt1( 0 ) / pt1( 2 ) ) + xOff;
			float y1 = scale * ( pt1( 1 ) / pt1( 2 ) ) + yOff;
			float x2 = scale * ( pt2( 0 ) / pt2( 2 ) ) + xOff;
			float y2 = scale * ( pt2( 1 ) / pt2( 2 ) ) + yOff;
			float nx = y2 - y1;
			float ny = -( x2 - x1 );
			float len = sqrt( nx * nx + ny * ny );
			nx = ( nx / len ) * 0.5;
			ny = ( ny / len ) * 0.5;

			if(!QGLHelper::isMultisamplingAvailable()) {
				QGLHelper::aaGlLineWidth(1.0);
				QGLHelper::aaGlBeginLines();
				QGLHelper::aaGlVertex2(x1, y1);
				QGLHelper::aaGlVertex2(x2, y2);
				QGLHelper::aaGlEndLines();
				continue;
			}
			glBegin( GL_QUADS );
			glColor3f( 1.0, 0.0, 0.0 );
			glVertex3f( x1 + nx, y1 + ny, 0.0f ); // Top Left
			glVertex3f( x1 - nx, y1 - ny, 0.0f );// Bottom Left
			glColor3f( 0.0, 1.0, 0.0 );
			glVertex3f( x2 - nx, y2 - ny, 0.0f );// Bottom Right
			glVertex3f( x2 + nx, y2 + ny, 0.0f );// Top Right
			glEnd();
		}
		glEndList();
	}
}

void FeaturesLayer::paint() const
{
	assert( 0 != feature );

	glPushMatrix();

	glTranslated( position.x(), position.y(), 0.2 );
	glColor3i( 0, 0, 0 );

	boost::shared_ptr<GLEnabler> enabler;
	if(!QGLHelper::isMultisamplingAvailable()) {
		glDisable(GL_DEPTH_TEST);
		enabler.reset(new GLEnabler(GL_BLEND));
		glBlendFunc(GL_SRC_ALPHA,  GL_ONE_MINUS_SRC_ALPHA);
		glColor4d(0.0, 0.0, 0.0, 1.0);
	}
	if ( listMode )
	{
		assert( 0 != featList && 0 != lineList );
		if ( featuresVisible )
			glCallList( featList );
		if ( linesVisible )
			glCallList( lineList );
		glPopMatrix();
		return;
	}

	if ( featuresVisible )
	{
		size_t cnt = offsets.size();
		for ( size_t i = 0; i < cnt; ++i )
		{
			glTranslatef( offsets[i].x, offsets[i].y, 0.0 );
			glCallList( feature );
		}
	}
	glPopMatrix();
}

GLuint FeaturesLayer::glMakeFeatureList() const
{
	static const float cross_sect = 7.0;
	static const float thickness = 1.0;

	GLuint list = glGenLists( 1 );
	glNewList( list, GL_COMPILE );
	if(!QGLHelper::isMultisamplingAvailable()) {
		QGLHelper::aaGlLineWidth(thickness);
		QGLHelper::aaGlBeginLines();
		QGLHelper::aaGlVertex2(-cross_sect * 0.5 , 0);
		QGLHelper::aaGlVertex2(cross_sect * 0.5, 0);
		QGLHelper::aaGlVertex2(0, -cross_sect * 0.5);
		QGLHelper::aaGlVertex2(0, cross_sect * 0.5);
		QGLHelper::aaGlEndLines();
	} else {
		glDrawRect( -( thickness / 2.0f ), -( cross_sect / 2.0f ), thickness, cross_sect );
		glDrawRect( -( cross_sect / 2.0f ), -( thickness / 2.0f ), ( cross_sect - thickness ) / 2.0f, thickness );
		glDrawRect( ( thickness / 2.0f ), -( thickness / 2.0f ), ( cross_sect - thickness ) / 2.0f, thickness );
	}
	glEndList();
	return list;
}

void FeaturesLayer::glDrawRect( float x, float y, float width, float height ) const
{
	glBegin( GL_QUADS );
	glVertex3f( x, y, 0.0f ); // Top Left
	glVertex3f( x + width, y, 0.0f );// Top Right
	glVertex3f( x + width, y + height, 0.0f );// Bottom Right
	glVertex3f( x, y + height, 0.0f );// Bottom Left
	glEnd();
}
