

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <QtCore/QDir>
#include <QtGui/QMouseEvent>

#include "qglhelper.h"

#include <adrvisual/glvisualizerdots.h>
#include <adrvisual/glnviewrenderer.h>

#include "constants.h"
#include "glwidget.h"
#include "groupboxgl.h"
#include "checkbuttongl.h"
#include "pushbuttongl.h"
#include "exclusiveoptionsgroup.h"
#include "slidergl.h"
#include "imagegl.h"

#include <boost/assign/list_inserter.hpp>

#include <map>
#include <string>

using std::map;
using std::string;

const int timerIntervall = 20;
enum Action { RotateCamera, RotateObject, Zoom, Pan, Pick };

GLWidget::GLWidget( const QGLFormat & format, vis_type vis, QWidget* parent )
		: QGLWidget( format, parent ),
		minSize( 100, 70 ),
		mouse_sensitivity( 1.0f ),
		wheel_pos( 0 ),

		vis_t( vis ),
		_visualizer( NULL ),
		config_file(),
		listener( 0 ),

		timer( this ),
		buttons( Qt::NoButton ),
		numKeysPressed( 0 ),
		rotateUp( false ),
		rotateDown( false ),
		rotateLeft( false ),
		rotateRight( false ),
		slideForward( false ),
		slideLeft( false ),
		slideBack( false ),
		slideRight( false ),
		glInitialized( false ),
		cameraAction( false ),
		reportAccu( false ) ,
		m_focused ( NULL ) ,
		m_animate_checkbox( NULL )
{
	setAttribute( Qt::WA_NoSystemBackground, true );
	setFocusPolicy( Qt::WheelFocus );
	connect( &timer, SIGNAL( timeout() ), this, SLOT( moveModel() ) );
	setMouseTracking( true );

	switch ( vis_t )
	{
		case vis_dots:
			_visualizer = new GLVisualizerDots;
			break;
		case vis_delaunay:
			_visualizer = new GLNViewRenderer<imgTypeT>;
			break;
		default:
			_visualizer = new GLVisualizerDots;
			vis_t = vis_dots;
	}
	setMinimumSize(300, 300);
}

void GLWidget::initializeGL()
{
	glClearColor(0.5, 0.5, 0.5, 0.0);
	glClearDepth( 1.0 );
	glEnable( GL_TEXTURE_2D );

	QGLHelper::prepareAATextures(this);

	m_navigate_options_box.reset( new GroupBoxGL );
	m_navigate_options_group.reset( new ExclusiveOptionsGroup );

	CheckButtonGL * checkb = new CheckButtonGL(tr("Rotate camera"));
	m_navigate_options_box->addControl(checkb);
	m_navigate_options_group->addOption(checkb, "Rotate camera");

	checkb = new CheckButtonGL(tr("Pan"));
	m_navigate_options_box->addControl(checkb);
	m_navigate_options_group->addOption(checkb, "Pan");

	checkb = new CheckButtonGL(tr("Zoom"));
	m_navigate_options_box->addControl(checkb);
	m_navigate_options_group->addOption(checkb, "Zoom");

	checkb = new CheckButtonGL(tr("Rotate object"));
	m_navigate_options_box->addControl(checkb);
	m_navigate_options_group->addOption(checkb, "Rotate object");

	m_navigate_options_box->initialize(this);
	m_navigate_options_group->activateOption("Rotate object");
	connect(m_navigate_options_group.get(), SIGNAL(sigOptionActivated(const std::string&)),
			this, SLOT(slotOptionActivated()));


	m_actions_box.reset( new GroupBoxGL );
	PushButtonGL * stButton = new PushButtonGL(tr("Standard"));
	connect(stButton, SIGNAL(clicked(ButtonGL*)), this, SLOT(slotSetStandardView()));
	m_actions_box->addControl(stButton);
	m_animate_checkbox = new CheckButtonGL(tr("Animate"));
	connect(m_animate_checkbox, SIGNAL(toggled(ButtonGL*,bool)), this, SLOT(slotToggleAnimation()));
	m_actions_box->addControl(m_animate_checkbox);

	m_actions_box->initialize(this);

	m_dots_slider.reset( new SliderGL(SliderGL::Vertical, 0, 1000) );
	m_dots_slider->initialize(this);
	connect(m_dots_slider.get(), SIGNAL(valueChanged(int)), this, SLOT(dotSliderMoved(int)));

	m_depth_slider.reset( new SliderGL(SliderGL::Vertical, 0, 1000) );
	m_depth_slider->setValue(1000);
	m_depth_slider->initialize(this);
	connect(m_depth_slider.get(), SIGNAL(valueChanged(int)), this, SLOT(depthSliderMoved(int)));

	m_image_dots.reset( new ImageGL(":/images/points.png") );
	m_image_dots->initialize(this);

	m_image_depth.reset( new ImageGL(":/images/triangle.png") );
	m_image_depth->initialize(this);

	initVisualizer();

	dotSliderMoved(0);
	depthSliderMoved(1000);

	glInitialized = true;
}

void GLWidget::slotOptionActivated()
{
	updateGL();
}

void GLWidget::initVisualizer()
{
	makeCurrent();
	_visualizer->init();
	_visualizer->focus();

	int displayflags = _visualizer->settings();
	displayflags |= DISPLAY_MESH | DISPLAY_TEXTURE | DISPLAY_LOCAL_FRAME;
	_visualizer->settings( displayflags );

	static const unsigned char indigo[4] = { 0x25, 0x0f, 0x64, 0xff };
	_visualizer->bgcolor( indigo );

	static const float eye[3] = {-32.0f,-32.0f,-32.0f};
	static const float aim[3] = {0.0f,0.0f,0.0f};
	static const float up[3] = {0.0f,0.0f,1.0f};
	_visualizer->lookAt( eye, aim, up );
}

void GLWidget::slotToggleShowCams()
{
	makeCurrent();
	int displayflags = _visualizer->settings();
	if ( displayflags & DISPLAY_CAMERAS ) displayflags &= ~DISPLAY_CAMERAS;
	else displayflags |= DISPLAY_CAMERAS;
	_visualizer->settings( displayflags );
	if ( glInitialized ) updateGL();
}

void GLWidget::slotSetStandardView()
{
	emit standardViewRequested();
}

void GLWidget::slotCamsOFF()
{
	makeCurrent();
	int displayflags = _visualizer->settings();
	if ( displayflags & DISPLAY_CAMERAS ) displayflags &= ~DISPLAY_CAMERAS;
	else return;
	_visualizer->settings( displayflags );
	if ( glInitialized ) updateGL();
}

void GLWidget::resizeGL( int w, int h )
{
	if ( h <= 0 ) h = 1;
	if ( w <= 0 ) w = 1;
	_visualizer->resize( w, h );
	_visualizer->perspective( 45.0f );
	resizeControls(w, h);
}

void GLWidget::resizeControls( int w, int h)
{
	double m_actions_width = m_actions_box->width();
	m_actions_box->setPosition(QPointF(double(w) - m_actions_width, 0));

	const int margin = 8;
	double X = margin;
	double Y = m_navigate_options_box->getSize().height() + margin;

	double diff = 0.5 * ( m_dots_slider->getSize().width() - m_image_dots->getSize().width() );
	m_image_dots->setPosition(QPointF(X + diff, Y));
	Y += m_image_dots->getSize().height();
	double L = height() - (Y + margin);

	m_dots_slider->setPosition(QPointF(X, Y));
	m_dots_slider->setLength(L);

	X += m_dots_slider->getSize().width() + margin + margin / 2;
	Y -= m_image_dots->getSize().height();
	m_image_depth->setPosition(QPointF(X + diff, Y));

	Y += m_image_dots->getSize().height();
	m_depth_slider->setPosition(QPointF(X, Y));
	m_depth_slider->setLength(L);
}

void GLWidget::paintGL()
{
	if ( !glInitialized ) return;

	QGLHelper::setCurrentContext(this);

	if ( _visualizer )
	{

		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		_visualizer->focus();
		_visualizer->paint();
	}
	paintControls();
}

void GLWidget::paintControls()
{
	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_LIGHTING);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	glOrtho( 0, width(), height(), 0, -1, 1 );

	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();

	m_navigate_options_box->paint();

	m_actions_box->paint();

	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND );
	float color[4] = { 1.0, 1.0, 1.0, 1.0 };
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, color);

	m_image_dots->paint();

	m_dots_slider->paint();

	m_image_depth->paint();

	m_depth_slider->paint();

	glFlush();

	glPopMatrix(); // GL_MODELVIEW

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib(); // GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT
}

void GLWidget::nViewSelected( NView *nv, ViewBase *v, const bool reset )
{
	makeCurrent();
	if ( !_visualizer ) return;
	_visualizer->attach( nv );
	_visualizer->init();
	if(v)
	{
		printf("=== GLWidget selects view\n");
		_visualizer->select(v, reset);
	}
	_visualizer->focus();
	const float y_axis[3] = {0.0f, 1.0f, 0.0f};
	_visualizer->setTurnTableSpine( y_axis );
	if ( glInitialized )
		updateGL();
}

void GLWidget::mouseMoveEvent( QMouseEvent* e )
{
	makeCurrent();

	if(e->buttons() == Qt::LeftButton) {
		double x = e->x(), y = e->y();
		if(m_focused != NULL) {
			m_focused->moved(x, y);
			return;
		}
	}

	const float dist_to_angle = 0.005f;
	const float dist_to_slide = 0.25f;
	const float tilt_gap = 0.1f;
	const float slide_factor = 0.1;

	QPoint mouse_new_pos = e->pos();
	QPoint mouse_move = mouse_new_pos - mouse_pos;
	mouse_pos = mouse_new_pos;

	mouse_move *= mouse_sensitivity;

	const float x_axis[3] = {1.0f, 0.0f, 0.0f};
//	const float y_axis[3] = {0.0f, 1.0f, 0.0f};
	float delta[3] = {0.0f, 0.0f, 0.0f};

	static map<string, Action> options;
	if(options.empty()) {
		boost::assign::insert(options)
				("Rotate camera", RotateCamera)
				("Rotate object", RotateObject)
				("Zoom", Zoom) ("Pan", Pan);
	}
	string option = m_navigate_options_group->activeOption();
	map<string, Action>::const_iterator it = options.find(option);

	Action cur_action = Pick;
	switch ( buttons )
	{
		case Qt::LeftButton:
		{
			if(it != options.end()) {
				cur_action = it->second;
			} else
				cur_action = RotateObject;
		}
		break;
		case Qt::RightButton:
			cur_action = RotateCamera;
		break;
		case Qt::MidButton:
			cur_action = Zoom;
		break;
		default: break; // Pick
	}

	switch ( cur_action )
	{
		case RotateCamera:
			_visualizer->turn( -mouse_move.x() * dist_to_angle );
			_visualizer->tilt( -mouse_move.y() * dist_to_angle,tilt_gap );
			break;
		case RotateObject:
		{
			float dx = mouse_move.x() * dist_to_angle;
			float dy = mouse_move.y() * dist_to_angle;
//			_visualizer->rotate( y_axis, dx ,false );
			_visualizer->turnTable( dx, false );
			maccu.rotY( dx, reportAccu );
			_visualizer->rotate( x_axis, dy,false );
			maccu.rotX( dy, reportAccu );
			break;
		}
		case Zoom:
		{
			float ds = mouse_move.y() * dist_to_slide;
			delta[2] = ds;
			maccu.slide( ds, reportAccu );
			_visualizer->slide( delta );
			break;
		}
		case Pan:
		{
			if(mouse_move.x() == 0 && mouse_move.y() == 0)
				break;
			if(mouse_move.x() < 0) {
				delta[0] = -mouse_move.x() * slide_factor;
			} else {
				delta[0] = -mouse_move.x() * slide_factor;
			}
			delta[1] = mouse_move.y() * slide_factor;
			_visualizer->slide( delta );
		}
		break;
		case Pick:
		default:
			_visualizer->pick( mouse_new_pos.x(), mouse_new_pos.y() );
			break;
	}
	if ( glInitialized ) updateGL();
}

void GLWidget::mousePressEvent ( QMouseEvent * e )
{
	buttons = e->button();
	if(e->button() == Qt::LeftButton) {
		double x = e->x(), y = e->y();
		if(m_navigate_options_box->rect().contains(x, y)) {
			m_navigate_options_box->pressed(x, y);
			m_focused = m_navigate_options_box.get();
		}
		else if(m_actions_box->rect().contains(x, y)) {
			m_actions_box->pressed(x, y);
			m_focused = m_actions_box.get();
		}
		else if(m_dots_slider->rect().contains(x, y)) {
			m_dots_slider->pressed(x, y);
			m_focused = m_dots_slider.get();
		}
		else if(m_depth_slider->rect().contains(x, y)) {
			m_depth_slider->pressed(x, y);
			m_focused = m_depth_slider.get();
		}
	}
}

void GLWidget::mouseReleaseEvent ( QMouseEvent * e )
{
	buttons = Qt::NoButton;
	if(e->button() == Qt::LeftButton) {
		double x = e->x(), y = e->y();
		if(m_focused != NULL) {
			m_focused->released(x, y);
			m_focused = NULL;
		}
	}
}

void GLWidget::wheelEvent ( QWheelEvent * e )
{
	double delta = -( double )e->delta() * 0.05;
	int sign = delta < 0 ? -1 : 1;
	delta = delta * delta * sign;
	if( e->orientation() == Qt::Vertical ) {
		double x = e->x(), y = e->y();
		if(m_dots_slider->rect().contains(x, y)) {
			m_dots_slider->move( delta );
			return;
		}
		else if(m_depth_slider->rect().contains(x, y)) {
			m_depth_slider->move( delta );
			return;
		}
	}
	else
	{
		e->ignore();
		return;
	}
	// disabled so far
	// picking becomes unstable when object is scaled
	//
	// because the intersection test does not know the size
	// of scaled screen points in the object space
	return;

	makeCurrent();
	const int wheel_quant = 24;			// should be a fraction of 120
	const float min_scale = 1.0f;
	const float max_scale = 16.0f;

	wheel_pos += e->delta() / wheel_quant;

	if ( wheel_pos<0 ) wheel_pos = 0;
	if ( wheel_pos > 100 ) wheel_pos = 100;
	float s = min_scale + ( ( max_scale-min_scale ) / 100.0f ) * wheel_pos;
	float factor[3] = {s,s,s};
	_visualizer->scale( factor,true );

	if ( glInitialized ) updateGL();
}

void GLWidget::updateKeyActions( QKeyEvent *e, bool set )
{
	makeCurrent();
	if ( e->isAutoRepeat() ) return;

	bool known = false;
	switch ( e->key() )
	{
		case Qt::Key_Up:
			rotateUp = set;
			known = true;
			break;
		case Qt::Key_Down:
			rotateDown = set;
			known = true;
			break;
		case Qt::Key_Left:
			rotateLeft = set;
			known = true;
			break;
		case Qt::Key_Right:
			rotateRight = set;
			known = true;
			break;
		case Qt::Key_W:
			slideForward = set;
			known = true;
			break;
		case Qt::Key_A:
			slideLeft = set;
			known = true;
			break;
		case Qt::Key_S:
			slideBack = set;
			known = true;
			break;
		case Qt::Key_D:
			slideRight = set;
			known = true;
			break;
		default:
			e->ignore();
			break;
	}

	if ( known )
	{
		int inc = set ? 1 : -1;
		numKeysPressed += inc;
		if ( numKeysPressed < 0 ) numKeysPressed = 0;
		e->accept();
	}

	moveModel();
}

void GLWidget::keyPressEvent( QKeyEvent *e )
{
	makeCurrent();

	if ( e->key() == Qt::Key_Delete )
	{
		emit deletePoint3D( true );
	}

	updateKeyActions( e, true );

	if ( numKeysPressed > 0 && !timer.isActive() )
		timer.start( timerIntervall );

}

void GLWidget::slotToggleAnimation()
{
	cameraAction = !cameraAction;
	if ( cameraAction )
	{
		GLRenderer *rendy = dynamic_cast<GLRenderer *>( _visualizer );
		if ( rendy )
		{
			string filename = QDir::homePath().toStdString() + "/.guide.cam";
			//tell the tourist camera fov, clip planes, etc.
			tourist.setCamera( rendy->getCamera() );
			tourist.reset();
			tourist.readScript( filename.c_str() );
			timer.start( timerIntervall );
		}
	}
	else
	{
		DBGOUT( DBGNORMAL, " --- Stopping Animation ---\n" );
		timer.stop();
		numKeysPressed = 0;
		updateGL();
	}
}

void GLWidget::dotSliderMoved( int i )
{
	vis_type vis_t = visualizerType();
	GLVisualizerDots* vis = 0;

	switch ( vis_t )
	{
		case vis_dots:
			vis = dynamic_cast<GLVisualizerDots*>( visualizer() );
			if ( vis ) vis->setDotSizePercentage( ( float )i/10 );
			break;
		case vis_delaunay:
			setPointSize( i );
			break;
		default:
			printf( "invalid renderer\n" );
			exit( -1 );
	}
	if ( glInitialized )
		updateGL();
}

void GLWidget::depthSliderMoved( int i )
{
	vis_type vis_t = visualizerType();
	GLVisualizerDots* vis1 = 0;

	switch ( vis_t )
	{
		case vis_dots:
			vis1 = dynamic_cast<GLVisualizerDots*>( visualizer() );
			if ( vis1 ) vis1->setDistancePercentage( ( float )i/10 );
			break;
		case vis_delaunay:
			setDeviation( i );
			break;
		default:
			printf( "invalid renderer\n" );
			exit( -1 );
	}
	if ( glInitialized )
		updateGL();
}

void GLWidget::setPointSize(int ticks)
{
	GLNViewRenderer<imgTypeT>* vis = 0;
	vis = dynamic_cast<GLNViewRenderer<imgTypeT>*>( visualizer() );
	if ( !vis ) return;

	static unsigned char penclr[] = { 0xff, 0xff, 0xff, 0xff };
	vis->penTicks( penclr, ticks, 1000 );
}

void GLWidget::setDeviation( int ticks )
{
	GLNViewRenderer<imgTypeT>* vis = 0;
	vis = dynamic_cast<GLNViewRenderer<imgTypeT>*>( visualizer() );
	if ( !vis ) return;
	vis->deviationTicks( ticks, 1000 );
}

void GLWidget::keyReleaseEvent( QKeyEvent *e )
{
	makeCurrent();
	const float y_axis[3] = {0.0f, 1.0f, 0.0f};
	//simple hotkeys
	switch ( e->key() )
	{
			/**
				this starts the animation
			*/
		case Qt::Key_O:
		{
			m_animate_checkbox->setState(!cameraAction);
			slotToggleAnimation();
			e->accept();
		}
		break;
		/**
			this prints the current cam and world trafos
		*/
		case Qt::Key_K:
		{
			GLRenderer *rendy = dynamic_cast<GLRenderer *>( _visualizer );
			if ( rendy )
			{
				tourist.printCamera( rendy->getCamera() );
				if ( rendy->getEntity() )
				{
					tourist.printWorldTrafo( rendy->getEntity()->world );
					tourist.printTurnTableSpine( rendy->getEntity()->getObjUp() );
				}
			}
		}
		break;
		/**
			this toggles the printing of the animation accus
		*/
		case Qt::Key_M:
			reportAccu = ! reportAccu;
			break;
			/**
				this resets the animation accus
			*/
		case Qt::Key_N:
			maccu.reset();
			break;
		case Qt::Key_J:
			_visualizer->setTurnTableSpine( y_axis );
			break;
		case Qt::Key_Delete:
			emit deletePoint3D( false );
			break;
		default:
			if ( !cameraAction ) updateKeyActions( e, false );
			break;
	}

	if ( numKeysPressed <= 0 && timer.isActive() && !cameraAction )
	{
		timer.stop();
		numKeysPressed = 0;
	}
}

void GLWidget::moveModel()
{
	makeCurrent();

	const float x_axis[3] = {1.0f, 0.0f, 0.0f};
//	const float y_axis[3] = {0.0f, 1.0f, 0.0f};

	float delta[3] = {0.0f,0.0f,0.0f};
	const float rstep = 0.05;
	const float sstep = 0.5;

	if ( cameraAction )
	{
		GLRenderer *rendy = dynamic_cast<GLRenderer *>( _visualizer );
		bool hasNext = false;
		if ( rendy )
		{
			hasNext = tourist.step( rendy );
		}
		if ( !hasNext )
		{
			cameraAction = false;
			DBGOUT( DBGNORMAL, " --- Stopping Animation ---\n" );
			timer.stop();
		}
	}
	else
	{
		if ( rotateUp )
		{
			_visualizer->rotate( x_axis, rstep,false );
			maccu.rotX( rstep, reportAccu );
		}
		if ( rotateDown )
		{
			_visualizer->rotate( x_axis, -rstep,false );
			maccu.rotX( -rstep, reportAccu );
		}
		if ( rotateLeft )
		{
//			_visualizer->rotate( y_axis, rstep,false );
			_visualizer->turnTable( rstep, false );
			maccu.rotY( rstep, reportAccu );
		}
		if ( rotateRight )
		{
//			_visualizer->rotate( y_axis, -rstep,false );
			_visualizer->turnTable( -rstep, false );
			maccu.rotY( -rstep, reportAccu );
		}
		if ( slideForward )
		{
			delta[2] = -sstep;
			_visualizer->slide( delta );
			maccu.slide( -sstep, reportAccu );
		}
		if ( slideLeft )
		{
			delta[0] = -sstep;
			_visualizer->slide( delta );
		}
		if ( slideBack )
		{
			delta[2] = sstep;
			_visualizer->slide( delta );
			maccu.slide( sstep, reportAccu );
		}
		if ( slideRight )
		{
			delta[0] = sstep;
			_visualizer->slide( delta );
		}
	}

	if ( glInitialized ) updateGL();
}

void GLWidget::setListener( SelectListener * li )
{
	if ( listener == li ) return;
	makeCurrent();
	listener = li;
	_visualizer->addListener( listener );
}

void GLWidget::set_GL_Config( const std::string & cfg )
{
	makeCurrent();
	config_file = cfg;
	if ( vis_delaunay == vis_t )
	{
		GLNViewRenderer<imgTypeT>* vis = dynamic_cast<GLNViewRenderer<imgTypeT>*>( _visualizer );
		if ( vis ) vis->loadConfig( config_file.c_str() );
	}
}

void GLWidget::nextVisualizer()
{
	makeCurrent();

	if ( vis_dots == vis_t ) vis_t = vis_delaunay;
	else vis_t = vis_dots;

	_visualizer->unfocus();
	NView* old_nview = _visualizer->nview();
	delete _visualizer;

	switch ( vis_t )
	{
		case vis_dots:
			_visualizer = new GLVisualizerDots;
			break;
		case vis_delaunay:
			_visualizer = new GLNViewRenderer<imgTypeT>;
			break;
		default:
			assert(false);
			_visualizer = new GLVisualizerDots;
	}
	_visualizer->addListener( listener );

	initVisualizer();
	set_GL_Config( config_file );
	resizeGL( width(), height() );
	if ( old_nview ) nViewSelected( old_nview );
	_visualizer->focus();
}

const QString & GLWidget::visualizerName() const
{
	static const QString dot( "Dots" );
	static const QString del( "Delaunay" );
	static const QString err( "Error!" );
	if ( vis_dots == vis_t ) return dot;
	else if ( vis_delaunay == vis_t ) return del;
	return err;
}
