

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include <QHeaderView>
#include <QMenu>
#include <QFileDialog>
#include <QTextStream>
#include <QSettings>

#include <vector>
#include <string>

#include "configconstants.h"
#include "configparser.h"

#include "selector.h"
#include "combodelegate.h"
#include "optionchooser.h"

Selector::Selector( QWidget * parent )
		: QWidget( parent )
{
	setupUi( this );
	init();
	makeConnections();
}

Selector::Selector( const QString & fname, QWidget * parent )
		: QWidget( parent )
{
	setupUi( this );
	init();
	makeConnections();
	load( fname );
}

Selector::~ Selector()
{
	QSettings settings( "ADORE2", "Selector" );
	settings.setValue( "lastDir", lastDir );
}

void Selector::init()
{
	std::vector< std::string > cc = batch::Parser::getCommands();
	int cnt = cc.size();
	for ( int i = 0; i < cnt; ++i )
		commands.append( QString::fromStdString( cc[i] ) );

	commandMenu = new QMenu;
	for ( int i = 0; i < cnt; ++i )
		commandMenu->addAction( commands[i] );
	addButton->setPopupMode( QToolButton::InstantPopup );
	addButton->setMenu( commandMenu );

	processingList->header()->setResizeMode( QHeaderView::Interactive );
	QStringList header;
	header << "Command" << "Algorithm" << "Options";
	model.setHorizontalHeaderLabels( header );
	processingList->setModel( &model );

	combo = new ComboDelegate;
	processingList->setItemDelegateForColumn( 1, combo );
	processingList->setEditTriggers( QAbstractItemView::AllEditTriggers );
	processingList->header()->setResizeMode( QHeaderView::ResizeToContents );

	QSettings settings( "ADORE2", "Selector" );
	lastDir = settings.value( "lastDir", QDir::homePath() ).toString();
}

void Selector::makeConnections()
{
	connect( commandMenu, SIGNAL( triggered( QAction* ) ), this, SLOT( slotAdd( QAction* ) ) );
	connect( removeButton, SIGNAL( pressed() ), this, SLOT( slotRemove() ) );
	connect( upButton, SIGNAL( pressed() ), this, SLOT( slotUp() ) );
	connect( downButton, SIGNAL( pressed() ), this, SLOT( slotDown() ) );
	connect( openButton, SIGNAL( pressed() ), this, SLOT( slotOpen() ) );
	connect( saveButton, SIGNAL( pressed() ), this, SLOT( slotSave() ) );
	connect( processingList, SIGNAL( doubleClicked( const QModelIndex& ) ),
	         this, SLOT( slotEditOptions( const QModelIndex& ) ) );
	connect( combo, SIGNAL( sigCommitData( int ) ), this, SLOT( slotSyncToCombo( int ) ) );

	connect( goButton, SIGNAL( clicked() ), parent(), SLOT( to3D() ) );
	connect( go2Button, SIGNAL( clicked() ), parent(), SLOT( to3D_2() ) );
	connect( newButton, SIGNAL( clicked() ), this, SLOT( slotNew() ) );
}

void Selector::slotNew()
{
	model.clear();
	QStringList header;
	header << "Command" << "Algorithm" << "Options";
	model.setHorizontalHeaderLabels( header );
}

void Selector::slotAdd( QAction* action )
{
	QList< QStandardItem* > row;
	QString text( action->text() );
	std::string cmd = text.toStdString();

	/// commando column ///
	QStandardItem* item = new QStandardItem( text );
	item->setCheckable( true );
	item->setEditable( false );
	item->setCheckState( Qt::Checked );
	row.append( item );

	/// algorithm column ///
	row.append( fillAlgItem( cmd ) );

	/// options column ///
	row.append( fillOptItem( cmd ) );

	model.appendRow( row );
}

QStandardItem * Selector::fillAlgItem( const std::string & cmd ) const
{
	std::vector< std::string > alg = batch::Parser::getAlgorithms( cmd );
	QList<QString> algList;
	int cnt = alg.size();
	for ( int i = 0; i < cnt; ++i )
		algList.append( QString::fromStdString( alg[i] ) );
	QStandardItem* item;
	if ( cnt )
	{
		QByteArray bytes;
		QDataStream stream( &bytes, QIODevice::WriteOnly ); stream << algList;
		item = new QStandardItem( algList[0] );
		item->setData( bytes, ComboDelegate::data_idx );
		item->setData( 0, ComboDelegate::idx_idx );
	}
	else
	{
		item = new QStandardItem;
		item->setData( -1, ComboDelegate::idx_idx );
		item->setEnabled( false );
	}
	return item;
}

QStandardItem * Selector::fillOptItem( const std::string & cmd ) const
{
	QStandardItem* item =
	        new QStandardItem( QIcon( ":/images/configure.png" ), "Configure" );
	item->setEditable( false );

	std::vector< std::string > alg = batch::Parser::getAlgorithms( cmd );
	int cnt = alg.size();
	for ( int i = 0; i < cnt; ++i )
	{
		QString opts( QString::fromStdString( batch::Parser::getOptions( cmd, alg[i] ) ) );
		item->setData( opts, item_opts_idx + i );
	}
	if ( !cnt )
	{
		QString opts( QString::fromStdString( batch::Parser::getOptions( cmd ) ) );
		item->setData( opts, item_opts_idx );
	}
	return item;
}

void Selector::slotRemove()
{
	QItemSelectionModel* selmodel = processingList->selectionModel();
	if ( !selmodel ) return;
	QModelIndexList list( selmodel->selectedRows() );
	if ( list.isEmpty() ) return;

	int row = list[0].row();
	QList< QStandardItem* > items( model.takeRow( row ) );
	for ( int i = 0; i < items.size(); ++i )
		delete items[i];
	selmodel->reset();
}

void Selector::slotUp()
{
	QItemSelectionModel* selmodel = processingList->selectionModel();
	if ( !selmodel ) return;
	QModelIndexList list( selmodel->selectedRows() );
	if ( list.isEmpty() ) return;

	int row = list[0].row();
	if ( 0 == row ) return;
	QList< QStandardItem* > items( model.takeRow( row ) );
	model.insertRow( row - 1, items );

	selmodel->reset();
	QModelIndex idx( items[0]->index() );
	selmodel->select( idx, QItemSelectionModel::SelectCurrent | QItemSelectionModel::Rows );
}

void Selector::slotDown()
{
	QItemSelectionModel* selmodel = processingList->selectionModel();
	if ( !selmodel ) return;
	QModelIndexList list( selmodel->selectedRows() );
	if ( list.isEmpty() ) return;

	int row = list[0].row();
	if ( row >= model.rowCount() - 1 ) return;
	QList< QStandardItem* > items( model.takeRow( row ) );
	model.insertRow( row + 1, items );

	selmodel->reset();
	QModelIndex idx( items[0]->index() );
	selmodel->select( idx, QItemSelectionModel::SelectCurrent | QItemSelectionModel::Rows );
}

void Selector::slotOpen()
{
	QString fn( QFileDialog::getOpenFileName( this, tr( "Open script" ), lastDir, tr( "NView script (*.txt)" ) ) );
	if ( fn.isEmpty() ) return;

	load( fn );
	lastDir = QFileInfo( fn ).absolutePath();
}

void Selector::load( const QString & fname )
{
	QFile file( fname );
	if ( !file.open( QFile::ReadOnly | QFile::Text ) )
		return;

	QTextStream in( &file );
	QString buffer;
	QApplication::setOverrideCursor( Qt::WaitCursor );
	buffer = in.readAll();

	batch::_cfg_ config;
	batch::Parser parser;
	parser.makeConform( buffer.toStdString(), config );
	QApplication::restoreOverrideCursor();

	model.clear();
	QStringList header;
	header << "Command" << "Algorithm" << "Options";
	model.setHorizontalHeaderLabels( header );

	int cnt = config.cmds.size();
	for ( int i = 0; i < cnt; ++i )
	{
		std::string cmd( config.cmds[i].name );

		QList< QStandardItem* > row;
		QStandardItem* item = new QStandardItem( QString::fromStdString( cmd ) );
		item->setCheckable( true );
		item->setEditable( false );
		item->setCheckState( Qt::Checked );
		row.append( item );

		item = fillAlgItem( cmd );
		std::string alg;
		bool has_alg = config.cmds[i].algs.empty() ? false : true;
		int aidx = -1;
		if ( has_alg )
		{
			alg = config.cmds[i].algs[0].name;
			aidx = batch::Parser::algIndex( cmd, alg );
			item->setText( QString::fromStdString( alg ) );
			item->setData( aidx, ComboDelegate::idx_idx );
		}
		else
		{
			item->setData( -1, ComboDelegate::idx_idx );
			item->setEnabled( false );
		}
		row.append( item );

		item = fillOptItem( cmd );

		std::string def_options;
		if ( has_alg ) def_options = batch::Parser::getOptions( cmd, alg );
		else def_options = batch::Parser::getOptions( cmd );
		bool has_options = def_options.empty() ? false : true;
		if ( has_options )
		{
			std::vector < batch::_opt_ > def_opts;
			parser.parseOptions( def_options, def_opts );
			std::vector < batch::_opt_ > options;
			if ( has_alg ) options = config.cmds[i].algs[0].opts;
			else options = config.cmds[i].opts;

			int opt_cnt = def_opts.size();
			for ( int j = 0; j < opt_cnt; ++j )
			{
				std::string opt_name( def_opts[j].name );
				std::vector < batch::_opt_ >::iterator ii = options.begin();
				std::vector < batch::_opt_ >::iterator end = options.end();
				for ( ; ii !=end; ++ii )
				{
					batch::_opt_& ooo = *ii;
					if ( opt_name == ooo.name )
					{
						if ( def_opts[j].type == ooo.type )
						{
							int def_val_cnt = def_opts[j].values.size();
							int val_cnt = ooo.values.size();
							if ( def_val_cnt == val_cnt )
							{
								for ( int l = 0; l < val_cnt; ++l )
									def_opts[j].values[l] = ooo.values[l];
							}
						}
						options.erase( ii );
						break;
					}
				}
			}
			if ( opt_cnt )
			{
				QString opt_string;
				for ( int j = 0; j < opt_cnt; ++j )
				{
					opt_string += "<option>";
					opt_string += "<type>" + QString::fromStdString( def_opts[j].type ) + "</type>";
					opt_string += "<name>" + QString::fromStdString( def_opts[j].name ) + "</name>";
					int val_cnt = def_opts[j].values.size();
					for ( int k = 0; k < val_cnt; ++k )
						opt_string += "<value>" + QString::fromStdString( def_opts[j].values[k] ) + "</value>";
					opt_string += "</option>\n";
				}
				if ( has_alg ) item->setData( opt_string, item_opts_idx + aidx );
				else item->setData( opt_string, item_opts_idx );
			}
		}
		else item->setEnabled( false );
		row.append( item );
		model.appendRow( row );
	}
}

void Selector::slotSave()
{
	QString progr( makeScript() );
	if ( progr.isEmpty() ) return;

	static const QString ext( ".txt" );
	QString fn( QFileDialog::getSaveFileName( this, tr( "Save script" ), lastDir, tr( "NView script (*%1)" ).arg( ext ) ) );
	if ( fn.isEmpty() ) return;

	if ( !fn.endsWith( ext ) ) fn += ext;
	QFile file( fn );
	if ( !file.open( QFile::WriteOnly | QFile::Text ) )
		return ;

	QTextStream out( &file );
	QApplication::setOverrideCursor( Qt::WaitCursor );
	out << progr;
	QApplication::restoreOverrideCursor();
}

void Selector::slotSyncToCombo( int idx )
{
	QItemSelectionModel* selmodel = processingList->selectionModel();
	if ( !selmodel ) return;
	QModelIndexList list( selmodel->selectedRows() );
	if ( list.isEmpty() ) return;

	int lastidx = list[0].sibling( list[0].row(), 1 ).data( ComboDelegate::idx_idx ).toInt(); //algorithm index from 2nd column
	if ( idx == lastidx ) return;

	QVariant var( list[0].sibling( list[0].row(), 1 ).data( ComboDelegate::data_idx ) );
	QByteArray bytes( var.toByteArray() );
	QDataStream stream( bytes );
	QList<QString> strings; stream >> strings;

	model.setData( list[0].sibling( list[0].row(), 1 ), strings[idx], ComboDelegate::text_idx );
	model.setData( list[0].sibling( list[0].row(), 1 ), idx, ComboDelegate::idx_idx );
}

void Selector::slotEditOptions( const QModelIndex & index )
{
	if ( 2 != index.column() ) return;

	QString cmd = index.sibling( index.row(), 0 ).data( ComboDelegate::text_idx ).toString(); //command from 1st column
	QString alg = index.sibling( index.row(), 1 ).data( ComboDelegate::text_idx ).toString(); //algorithm from 2nd column
	int idx = index.sibling( index.row(), 1 ).data( ComboDelegate::idx_idx ).toInt(); //algorithm index from 2nd column

	QString title( alg );
	if ( alg.isEmpty() ) title = cmd;

	QString options;
	if ( -1 == idx )
		options = index.data( item_opts_idx ).toString();
	else
		options = index.data( item_opts_idx + idx ).toString();
	if ( options.isEmpty() ) return;

	OptionChooser dlg( options );
	QPoint position( mapToGlobal( pos() ) );
	dlg.move( dlg.mapFromGlobal( position ) );
	dlg.setWindowTitle( title + " Options" );
	if ( QDialog::Rejected == dlg.exec() ) return;
	if ( -1 == idx ) idx = 0;
	model.setData( index, dlg.options(), item_opts_idx + idx );
}

QString Selector::makeScript() const
{
	QString progr;
	int cnt = model.rowCount();
	for ( int i = 0; i < cnt; ++i )
	{
		QStandardItem* item = model.item( i, 0 );
		if ( Qt::Checked != item->checkState() ) continue;
		progr += "<command>";
		progr += "<name>" + item->text() + "</name>\n";
		bool has_alg = false;
		item = model.item( i, 1 );
		QString alg( item->text() );
		if ( !alg.isEmpty() )
		{
			has_alg = true;
			progr +=  "<algorithm> <name>" + alg + "</name>\n";
		}
		int lastidx = item->data( ComboDelegate::idx_idx ).toInt(); //algorithm index from 2nd column
		item = model.item( i, 2 );
		QString opts;
		if ( -1 < lastidx )
			opts = item->data( item_opts_idx + lastidx ).toString();
		else
			opts = item->data( item_opts_idx ).toString();

		if ( !opts.isEmpty() ) progr += opts;
		if ( has_alg ) progr += "</algorithm>\n";
		progr += "</command>\n";
	}
	return progr;
}

